/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Standard Actions */

int
mops_movob_actcb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 mops_point_object *point = NULL;
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0;
 double dx = 0, dy = 0, dz = 0;
 double v1[3] = {0}, v2[3]={0};
 float euler[3] = {0};
 GLdouble mm[16];
 static GLdouble m[16] = {0};
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 char fname[] = "move_object";

  /* parse args */
  if(argc == 5)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);

	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;

	    /* get real direction of current level coordinate system */
	    glMatrixMode (GL_MODELVIEW);
	    glPushMatrix();
	    if(view->type != MOPS_VTTRIM)
	      if(mops_current_level&&mops_current_level->next)
		mops_get_all_isrtransform(mops_current_level->next);
	    glGetDoublev(GL_MODELVIEW_MATRIX, m);
	    glPopMatrix();

	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winx winy|-winxy winx winy]");
      return TCL_OK;
    }


  /* calc dx, dy, dz */
  dx = -(oldwinx - winx) * view->conv_x;
  dy = (oldwiny - winy) * view->conv_y;

  oldwinx = winx;
  oldwiny = winy;

   if((view->type == MOPS_VTFRONT) || (view->type == MOPS_VTTRIM))
    {
      v2[0]=dx;
      v2[1]=dy;
      v2[2]=0.0;
      MOPS_APTRAN3(v1,v2,m)
      dx = v1[0];
      dy = v1[1];
      dz = v1[2];
    }


  /* Side or Top view? */
  if(view->type == MOPS_VTSIDE)
    {
      v2[0]=0.0;
      v2[1]=dy;
      v2[2]=-dx;
      MOPS_APTRAN3(v1,v2,m)
      dx = v1[0];
      dy = v1[1];
      dz = v1[2];
    }
  if(view->type == MOPS_VTTOP)
    {
      v2[0]=dx;
      v2[1]=0.0;
      v2[2]=-dy;
      MOPS_APTRAN3(v1,v2,m)
      dx = v1[0];
      dy = v1[1];
      dz = v1[2];
    }

 
  /* move the objects / selected points */

  while(sel)
    {
      o = sel->object;
      if(!o)
	return TCL_OK;

      if(sel->object->selp)
	{
	  glMatrixMode(GL_MODELVIEW);
	  glPushMatrix();
	   glScaled (1.0/o->scalx, 1.0/o->scaly, 1.0/o->scalz);

	   mops_quat_to_euler(o->quat,euler);
	   glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
	   glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);

	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  v2[0]=dx;
	  v2[1]=dy;
	  v2[2]=dz;
	  MOPS_APTRAN3(v1,v2,mm)
	  dx = v1[0];
	  dy = v1[1];
	  dz = v1[2];

	  point = sel->object->selp;
	  while(point)
	    {
	      point->point[0]+=dx;
	      point->point[1]+=dy;
	      point->point[2]+=dz;

	      point = point->next;
	    }

	  o->modified = MOPS_TRUE;

	}
      else
      {
	o = sel->object;
	if(o)
	  {
	    o->movx += dx; 
	    o->movy += dy; 
	    o->movz += dz; 
	  }
      }
      sel = sel->next;
    }

  if(!mops_prefs.lazynotify)
    mops_status = mops_notify_parent();

  if((dx != 0.0) || (dy != 0.0) || (dz != 0.0))
    display_cb(togl);

 return TCL_OK;
}


int
mops_rotob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double height = Togl_Height(togl);
 double winx = 0.0, winy = 0.0, tpoint[4] = {0};
 double angle = 0.0;
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 float quat[4];
 float v1[2], v2[2];
 float alpha, beta;
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "rotate_object";
 
  /* parse args */
  if(argc == 5)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);

	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }

      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);

	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winx winy|-winxy winx winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode (GL_MODELVIEW);

  /* rotate the object(s) */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
 	   mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
    
	  owiny = height-owiny;
	    
	  v1[0] = oldwinx-owinx;
	  v1[1] = oldwiny-owiny;

	  if((v1[0]==0.0)&&(v1[1]==0.0))
	    break;
	  alpha = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	  if(v1[1]<0.0)
	    alpha = 360.0-alpha;
	     
	  v2[0] = winx-owinx;
	  v2[1] = winy-owiny;
	  if((v2[0]==0.0)&&(v2[1]==0.0))
	    break;
	  beta = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	  if(v2[1]<0.0)
	    beta = 360.0-beta;

	  angle = beta - alpha;

	  if(o->selp)
	    {
	      point = o->selp;

	      glPushMatrix();
	      glLoadIdentity();

	      switch(view->type)
		{
		case MOPS_VTSIDE:
		  /* rotate about x */
		  glRotatef(-angle,1.0,0.0,0.0);
		  break;
		case MOPS_VTFRONT:
		case MOPS_VTTRIM:
		  /* rotate about z */
		  glRotated(-angle,0.0,0.0,1.0);
		  break;
		case MOPS_VTTOP:
		  /* rotate about y */
		  glRotatef(-angle,0.0,1.0,0.0);
		  break;
		default:
		  break;
		} /* switch */

	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);

	      glPopMatrix();
	      
	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;

	    }
	  else
	    switch(view->type)
	      {
	      case MOPS_VTSIDE:
		/* rotate about x */

		o->rotx += angle;
		if(o->rotx >= 360.0) o->rotx -= 360.0;
		if(o->rotx <= -360.0) o->rotx += 360.0;

		mops_axis_to_quat(xaxis,MOPS_D2R(angle),quat);
		mops_add_quat(quat,o->quat,o->quat);
		break;
	      case MOPS_VTFRONT:
	      case MOPS_VTTRIM:
		/* rotate about z */

		o->rotz += angle;
		if(o->rotz >= 360.0) o->rotz -= 360.0;
		if(o->rotz <= -360.0) o->rotz += 360.0;

		mops_axis_to_quat(zaxis,MOPS_D2R(angle),quat);
		mops_add_quat(quat,o->quat,o->quat);
		break;
	      case MOPS_VTTOP:
		/* rotate about y */

		o->roty -= angle;
		if(o->roty >= 360.0) o->roty -= 360.0;
		if(o->roty <= -360.0) o->roty += 360.0;

		mops_axis_to_quat(yaxis,MOPS_D2R(angle),quat);
		mops_add_quat(quat,o->quat,o->quat);
		break;
	      default:
		break;
	      } /* switch */

	} /* if */
      
      sel = sel->next;
    } /* while */

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

  oldwinx = winx;
  oldwiny = winy;

 return TCL_OK;
}


int
mops_rotoba_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double height = Togl_Height(togl);
 double winx = 0.0, winy = 0.0, ax=0.0, ay=0.0, ay2=0.0, axo=0.0, ayo=0.0;
 double dummy;
 double angle = 0.0, tpoint[4] = {0};
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 double dx,dy,dax,day,zangle,yangle,xangle;
 float quat[4];
 float v1[3] = {0}, v2[3] = {0}, v3[3];
 float alpha, beta;
 GLfloat m[4][4];
 GLdouble mp[16], mm[16], owinx = 0.0, owiny = 0.0, owinz = 0.0;
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "rotate_object_about";

  /* parse args */
  if(argc == 7)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	Tcl_GetDouble(interp, argv[5], &ax);
	Tcl_GetDouble(interp, argv[6], &ay);

	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	    mops_snap_togrid(togl,&ax,&ay);
	  }
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    Tcl_GetDouble(interp, argv[5], &ax);
	    Tcl_GetDouble(interp, argv[6], &ay);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
		mops_snap_togrid(togl,&ax,&ay);
	      }
	    oldwinx = winx;
	    oldwiny = winy;

	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winx winy|-winxy winx winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode (GL_MODELVIEW);

  ay2 = ay;

  /* rotate the object(s) */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  ay = ay2;
	  glPushMatrix();
	  if(mops_current_level->next)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   /*mops_build_rotmatrix(m, o->quat);
	     glMultMatrixf(&m[0][0]);*/
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  owiny = height-owiny;

	  dx = owinx;
	  dy = owiny;
	  dax = dx - ax;
	  day = dy - ay;

	  owinx -= (owinx - ax);
	  owiny -= (owiny - ay);

	  v1[0] = oldwinx-owinx;
	  v1[1] = oldwiny-owiny;

	  if((v1[0]==0.0)&&(v1[1]==0.0))
	    break;
	  alpha = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	  if(v1[1]<0.0)
	    alpha = 360.0-alpha;
	     
	  v2[0] = winx-owinx;
	  v2[1] = winy-owiny;
	  if((v2[0]==0.0)&&(v2[1]==0.0))
	    break;
	  beta = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	  if(v2[1]<0.0)
	    beta = 360.0-beta;

	  angle = beta - alpha;

	  switch(view->type)
	    {
	    case MOPS_VTSIDE:
	      /* rotate about x */

	      /* get old rotation about X */
	      glPushMatrix();
	      mops_build_rotmatrix(m, o->quat);
	      glMultMatrixf(&m[0][0]);
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      MOPS_APTRAN3(v3,zaxis,mm);
	      v2[0] = v3[2];
	      v2[1] = v3[1];
	      xangle = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v2[1]<0.0)
		xangle = 360.0-xangle;

	      glPushMatrix();
	      if(mops_current_level->next)
		mops_get_all_transform(mops_current_level->next);
	      glTranslated(o->movx, o->movy, o->movz);
	      
	      glRotatef(xangle, 1.0, 0.0, 0.0);

	      /*	      glScaled (o->scalx, o->scaly, o->scalz);*/
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();
	      /*
	      gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	      */
	      ay = height-ay;
	      
	      gluUnProject(ax,ay,owinz,mm,mp,vp,&dummy,&ayo,&axo);

	      glPushMatrix();
	      glLoadIdentity();

	      glRotatef(xangle, 1.0, 0.0, 0.0);

	      glTranslated(0.0, -ayo, -axo);
	      glRotated(-angle,1.0,0.0,0.0);
	      glTranslated(0.0, ayo, axo);

	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      v1[0] = 0.0;
	      v1[1] = 0.0;
	      v1[2] = 0.0;

	      MOPS_APTRAN3(v2,v1,mm);

	      break;
	    case MOPS_VTFRONT:
	    case MOPS_VTTRIM:
	      /* rotate about z */

	      /* get old rotation about Z */
	      glPushMatrix();
	      mops_build_rotmatrix(m, o->quat);
	      glMultMatrixf(&m[0][0]);
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      MOPS_APTRAN3(v3,xaxis,mm);
	      v2[0] = v3[0];
	      v2[1] = v3[1];
	      zangle = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v2[1]<0.0)
		zangle = 360.0-zangle;

	      glPushMatrix();
	      if(mops_current_level->next)
		mops_get_all_transform(mops_current_level->next);
	      glTranslated(o->movx, o->movy, o->movz);
	      
	      glRotatef(zangle, 0.0, 0.0, 1.0);

	      /*	      glScaled (o->scalx, o->scaly, o->scalz);*/
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();
	      /*
	      gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	      */
	      ay = height-ay;
	      
	      gluUnProject(ax,ay,owinz,mm,mp,vp,&axo,&ayo,&dummy);

	      glPushMatrix();
	      glLoadIdentity();

	      glRotatef(zangle, 0.0, 0.0, 1.0);

	      glTranslated(-axo, -ayo, 0.0);
	      glRotated(-angle,0.0,0.0,1.0);
	      glTranslated(axo, ayo, 0.0);

	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      v1[0] = 0.0;
	      v1[1] = 0.0;

	      MOPS_APTRAN3(v2,v1,mm);

	      break;
	    case MOPS_VTTOP:
	      /* rotate about y */

	      /* get old rotation about Y */

	      glPushMatrix();
	      mops_build_rotmatrix(m, o->quat);
	      glMultMatrixf(&m[0][0]);
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      MOPS_APTRAN3(v3,xaxis,mm);
	      v2[0] = v3[0];
	      v2[1] = v3[2];
	      yangle = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v2[1]<0.0)
		yangle = 360.0-yangle;

	      glPushMatrix();
	      if(mops_current_level->next)
		mops_get_all_transform(mops_current_level->next);
	      glTranslated(o->movx, o->movy, o->movz);
	      
	      glRotatef(yangle, 0.0, 1.0, 0.0);

	      /*	      glScaled (o->scalx, o->scaly, o->scalz);*/
	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();
	      /*
	      gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	      */
	      ay = height-ay;
	      
	      gluUnProject(ax,ay,owinz,mm,mp,vp,&axo,&dummy,&ayo);

	      glPushMatrix();
	      glLoadIdentity();

	      glRotatef(yangle, 0.0, 1.0, 0.0);

	      glTranslated(-axo, 0.0, -ayo);
	      glRotated(-angle,0.0,1.0,0.0);
	      glTranslated(axo, 0.0, ayo);

	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      v1[0] = 0.0;
	      v1[1] = 0.0;
	      v1[2] = 0.0;

	      MOPS_APTRAN3(v2,v1,mm);

	      break;
	    default:
	      break;
	    } /* switch */


	  if(o->selp)
	    {
	      point = o->selp;

	      glPushMatrix();
	      glLoadIdentity();
	      switch(view->type)
		{
		case MOPS_VTSIDE:
		  /* rotate about x */
		  glRotated(-angle,1.0,0.0,0.0);
		  glTranslated(0.0,-v2[1],-v2[2]);
		  
		  break;
		case MOPS_VTFRONT:
		case MOPS_VTTRIM:
		  /* rotate about z */
		  glRotated(-angle,0.0,0.0,1.0);
		  glTranslated(-v2[0],-v2[1],0.0);

		  break;
		case MOPS_VTTOP:
		  /* rotate about y */
		  glRotated(-angle,0.0,1.0,0.0);
		  glTranslated(-v2[0],0.0,-v2[2]);
	      
		  break;
		default:
		  break;
		} /* switch */

	      glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;
	      
	    }
	  else
	    {
	      switch(view->type)
		{
		case MOPS_VTSIDE:
		  /* rotate about x */

	      
		  o->movz -= v2[2];
		  o->movy -= v2[1];
	
		  o->rotx += angle;
		  if(o->rotx >= 360.0) o->rotx -= 360.0;
		  if(o->rotx <= -360.0) o->rotx += 360.0;

		  mops_axis_to_quat(xaxis,MOPS_D2R(angle),quat);
		  mops_add_quat(quat,o->quat,o->quat);
		  break;
		case MOPS_VTFRONT:
		case MOPS_VTTRIM:
		  /* rotate about z */

		  o->movx -= v2[0];
		  o->movy -= v2[1];
	      
		  o->rotz += angle;
		  if(o->rotz >= 360.0) o->rotz -= 360.0;
		  if(o->rotz <= -360.0) o->rotz += 360.0;

		  mops_axis_to_quat(zaxis,MOPS_D2R(angle),quat);
		  mops_add_quat(quat,o->quat,o->quat);

		  break;
		case MOPS_VTTOP:
		  /* rotate about y */
	    
		  o->movx -= v2[0];
		  o->movz -= v2[2];

		  o->roty -= angle;
		  if(o->roty >= 360.0) o->roty -= 360.0;
		  if(o->roty <= -360.0) o->roty += 360.0;

		  mops_axis_to_quat(yaxis,MOPS_D2R(angle),quat);
		  mops_add_quat(quat,o->quat,o->quat);
		  break;
		default:
		  break;
		} /* switch */

	    } /* else */

	} /* if */
      
      sel = sel->next;
    } /* while */

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

  oldwinx = winx;
  oldwiny = winy;

 return TCL_OK;
}


int
mops_sc1DXob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscalx = 1.0;
 double tpoint[4]={0}, t1, t2, v1[2], v2[2];
 GLdouble vx[3], alpha, beta, gamma;
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DX_object";

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	

      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the object(s) / selected points */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	  gluProject(1.0,0.0,0.0,mm,mp,vp,&vx[0],&vx[1],&vx[2]);

	  owiny = height - owiny;
	  vx[1] = height - vx[1];

	  vx[0] -= owinx;
	  vx[1] -= owiny;

	  if((vx[0]!=0.0)||(vx[1]!=0.0))
	    {

	      alpha = MOPS_R2D(acos(vx[0]/MOPS_V2LEN(vx)));
	      if(vx[1]>0.0)
		alpha = 360.0-alpha;

	      v1[0] = (oldwinx-owinx);
	      v1[1] = (oldwiny-owiny);

	      v2[0] = (winx-owinx);
	      v2[1] = (winy-owiny);
	      
	      beta = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	      if(v1[1]<0.0)
		beta = 360.0-beta;

	      gamma = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v1[1]<0.0)
		gamma = 360.0-gamma;

	      t1 = cos(MOPS_D2R(alpha)+MOPS_D2R(beta))*MOPS_V2LEN(v1);
	      t2 = cos(MOPS_D2R(alpha)+MOPS_D2R(gamma))*MOPS_V2LEN(v2);

	      if(t1)
		dscalx = t2/t1;
	      else
		dscalx = 1.0;
	    }
	  else
	    dscalx = 1.0;
	      
	  if(o->selp)
	    {
	      point = o->selp;
	      
	      glPushMatrix();
	       glLoadIdentity();

	       mops_get_all_isrtransform(mops_current_level->next);

	       glScaled(dscalx,1.0,1.0);

	       mops_get_all_srtransform(mops_current_level->next);

	       glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;	      

	    }
	  else
	    {
	    
	      o->scalx *= dscalx;

	    }


	}
      
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;
  
  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


int
mops_sc1DYob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscaly = 1.0;
 double tpoint[4]={0}, t1, t2, v1[2], v2[2];
 GLdouble vy[3], alpha, beta, gamma;
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DY_object";

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the object(s) / selected points */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	  gluProject(0.0,1.0,0.0,mm,mp,vp,&vy[0],&vy[1],&vy[2]);

	  vy[1] = height - vy[1];
	  owiny = height - owiny;
	 
	  vy[0] -= owinx;
	  vy[1] -= owiny;

	  if((vy[0]!=0.0)||(vy[1]!=0.0))
	    {
	      alpha = MOPS_R2D(acos(vy[0]/MOPS_V2LEN(vy)));
	      if(vy[1]>0.0)
		alpha = 360.0-alpha;

	      v1[0] = (oldwinx-owinx);
	      v1[1] = (oldwiny-owiny);

	      v2[0] = (winx-owinx);
	      v2[1] = (winy-owiny);
	      
	      beta = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	      if(v1[1]<0.0)
		beta = 360.0-beta;

	      gamma = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v1[1]<0.0)
		gamma = 360.0-gamma;

	      t1 = cos(MOPS_D2R(alpha)+MOPS_D2R(beta))*MOPS_V2LEN(v1);
	      t2 = cos(MOPS_D2R(alpha)+MOPS_D2R(gamma))*MOPS_V2LEN(v2);

	      if(t1)
		dscaly = t2/t1;
	      else
		dscaly = 1.0;
	    }
	  else
	    dscaly = 1.0;
	      
	  if(o->selp)
	    {
	      point = o->selp;
	      
	      glPushMatrix();
	       glLoadIdentity();

	       mops_get_all_isrtransform(mops_current_level->next);

	       glScaled(1.0,dscaly,1.0);

	       mops_get_all_srtransform(mops_current_level->next);

	       glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;	

	    }
	  else
	    {
	      o->scaly *= dscaly;
	    }
	}
      
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


int
mops_sc1DZob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscalz = 1.0;
 double tpoint[4]={0}, t1, t2, v1[2], v2[2];
 GLdouble vz[3], alpha, beta, gamma;
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DZ_object";

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the object(s) / selected points */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);
	  gluProject(0.0,0.0,1.0,mm,mp,vp,&vz[0],&vz[1],&vz[2]);

	  vz[1] = height - vz[1];
	  owiny = height - owiny;
	 
	  vz[0] -= owinx;
	  vz[1] -= owiny;

	  if((vz[0]!=0.0)||(vz[1]!=0.0))
	    {
	      alpha = MOPS_R2D(acos(vz[0]/MOPS_V2LEN(vz)));
	      if(vz[1]>0.0)
		alpha = 360.0-alpha;

	      v1[0] = (oldwinx-owinx);
	      v1[1] = (oldwiny-owiny);

	      v2[0] = (winx-owinx);
	      v2[1] = (winy-owiny);
	      
	      beta = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	      if(v1[1]<0.0)
		beta = 360.0-beta;

	      gamma = MOPS_R2D(acos(v2[0]/MOPS_V2LEN(v2)));
	      if(v1[1]<0.0)
		gamma = 360.0-gamma;

	      t1 = cos(MOPS_D2R(alpha)+MOPS_D2R(beta))*MOPS_V2LEN(v1);
	      t2 = cos(MOPS_D2R(alpha)+MOPS_D2R(gamma))*MOPS_V2LEN(v2);

	      if(t1)
		dscalz = t2/t1;
	      else
		dscalz = 1.0;
	    }
	  else
	    dscalz = 1.0;
	      
	  if(o->selp)
	    {
	      point = o->selp;
	      
	      glPushMatrix();
	       glLoadIdentity();

	       mops_get_all_isrtransform(mops_current_level->next);

	       glScaled(1.0,1.0,dscalz);

	       mops_get_all_srtransform(mops_current_level->next);

	       glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;

	    }
	  else
	    {
	      o->scalz *= dscalz;
	    }
	}
      
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


/* selected points only */
int
mops_sc1DXworld_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscalx = 0.0;
 double tpoint[4]={0};
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DX_world";
 float euler[3] = {0};

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the selected points */
  while(sel)
    {
      o = sel->object;
      if(o->selp)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  if(view->type == MOPS_VTSIDE)
	    {
	      dscalx = 1.0;
	    }
	  else
	    {
	      if(oldwinx-owinx)
		{
		  dscalx = (winx-owinx)/(oldwinx-owinx);
		}
	      else
		{
		  dscalx = 1.0;
		}
	    }
	 
	  point = o->selp;
	      
	  glPushMatrix();
	   glLoadIdentity();

	   mops_get_all_isrtransform(mops_current_level->next);
	   
	   mops_quat_to_euler(o->quat,euler);
	   glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
	   glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);
	   glScaled(dscalx,1.0,1.0);
	   glRotatef(MOPS_R2D(-euler[2]), 1.0, 0.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[0]), 0.0, 0.0, 1.0);
	   mops_get_all_srtransform(mops_current_level->next);

	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	  o->modified = MOPS_TRUE;
	  
	}
     
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}

/* selected points only */
int
mops_sc1DYworld_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscaly = 1.0;
 double tpoint[4]={0};
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DY_world";
 float euler[3] = {0};

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);


  /* scale the selected points */
  while(sel)
    {
      o = sel->object;
      if(o->selp)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  owiny = height - owiny;
	  if(view->type == MOPS_VTTOP)
	    {
	      dscaly = 1.0;
	    }
	  else
	    {
	      if(oldwiny-owiny)
		{
		  dscaly = (winy-owiny)/(oldwiny-owiny);
		}
	      else
		{
		  dscaly = 1.0;
		}
	    }
	 
	  point = o->selp;
	      
	  glPushMatrix();
	   glLoadIdentity();

	   mops_get_all_isrtransform(mops_current_level->next);

	   mops_quat_to_euler(o->quat,euler);
	   glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
	   glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);
	   glScaled(1.0,dscaly,1.0);
	   glRotatef(MOPS_R2D(-euler[2]), 1.0, 0.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[0]), 0.0, 0.0, 1.0);
	   mops_get_all_srtransform(mops_current_level->next);

	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	  o->modified = MOPS_TRUE;

	}

      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


/* selected points only */
int
mops_sc1DZworld_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscalz=1.0;
 double tpoint[4]={0};
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale1DZ_world";
 float euler[3] = {0};

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);


  /* scale the selected points */
  while(sel)
    {
      o = sel->object;
      if(o->selp)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  owiny = height - owiny;
	  if(view->type == MOPS_VTFRONT || view->type == MOPS_VTTRIM)
	    {
	      dscalz = 1.0;
	    }
	  else
	    {
	      if(view->type == MOPS_VTTOP)
		{
		  if(oldwiny-owiny)
		    dscalz = (winy-owiny)/(oldwiny-owiny);
		  else
		    dscalz = 1.0;
		}
	      else
		{
		  if(view->type == MOPS_VTSIDE)
		    {
		      if(oldwinx-owinx)
			{
			  dscalz = (winx-owinx)/(oldwinx-owinx);
			}
		      else
			{
			  dscalz = 1.0;
			}
		    }
		}
	    }

	  point = o->selp;
	      
	  glPushMatrix();
	   glLoadIdentity();

	   mops_get_all_isrtransform(mops_current_level->next);

	   mops_quat_to_euler(o->quat,euler);
	   glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
	   glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);
	   glScaled(1.0,1.0,dscalz);
	   glRotatef(MOPS_R2D(-euler[2]), 1.0, 0.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[1]), 0.0, 1.0, 0.0);
	   glRotatef(MOPS_R2D(-euler[0]), 0.0, 0.0, 1.0);
	   mops_get_all_srtransform(mops_current_level->next);
	   
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	  o->modified = MOPS_TRUE;
	  
	}

      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


int
mops_sc3Dob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscal = 1.0;
 double tpoint[4]={0}, t1, t2, v1[2], v2[2];
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale3D_object";

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the object(s) / selected points */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  owiny = height - owiny;
	 
	  v1[0] = (oldwinx-owinx);
	  v1[1] = (oldwiny-owiny);

	  v2[0] = (winx-owinx);
	  v2[1] = (winy-owiny);
	      
	  t1 = MOPS_V2LEN(v1);
	  t2 = MOPS_V2LEN(v2);

	  if(t1)
	    dscal = t2/t1;
	  else
	    dscal = 1.0;
	      
	  if(o->selp)
	    {
	      point = o->selp;
	      
	      glPushMatrix();
	       glLoadIdentity();

	       mops_get_all_isrtransform(mops_current_level->next);

	       glScaled(dscal,dscal,dscal);

	       mops_get_all_srtransform(mops_current_level->next);

	       glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;

	    }
	  else
	    {
	      o->scalx *= dscal;
	      o->scaly *= dscal;
	      o->scalz *= dscal;
	    }
	}
      
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


int
mops_sc2Dob_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 double height = Togl_Height(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0, dscal = 1.0;
 double tpoint[4]={0}, t1, t2, v1[2], v2[2];
 GLdouble mp[16], mm[16], owinx, owiny, owinz;
 GLfloat m[4][4];
 GLint vp[4];
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 char fname[] = "scale2D_object";

  /* parse args */
  if(argc >= 4)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }	
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;
	  }

    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start $winx $winy|-winxy $winx $winy]");
      return TCL_OK;
    }

  if((oldwinx == winx) && (oldwiny == winy))
    return TCL_OK;

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_PROJECTION_MATRIX, mp);
  glMatrixMode(GL_MODELVIEW);

  /* scale the object(s) / selected points */
  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glPushMatrix();
	  if(view->type != MOPS_VTTRIM)
	    mops_get_all_transform(mops_current_level->next);
	   glTranslated(o->movx, o->movy, o->movz);
	   mops_build_rotmatrix(m, o->quat);
	   glMultMatrixf(&m[0][0]);
	   glScaled (o->scalx, o->scaly, o->scalz);
	   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();
	  gluProject(0.0,0.0,0.0,mm,mp,vp,&owinx,&owiny,&owinz);

	  owiny = height - owiny;
	 
	  v1[0] = (oldwinx-owinx);
	  v1[1] = (oldwiny-owiny);

	  v2[0] = (winx-owinx);
	  v2[1] = (winy-owiny);
	      
	  t1 = MOPS_V2LEN(v1);
	  t2 = MOPS_V2LEN(v2);

	  if(t1)
	    dscal = t2/t1;
	  else
	    dscal = 1.0;
	      
	  if(o->selp)
	    {
	      point = o->selp;
	      
	      glPushMatrix();
	       glLoadIdentity();

	       mops_get_all_isrtransform(mops_current_level->next);

	       switch(view->type)
		 {
		 case MOPS_VTFRONT:
		 case MOPS_VTTRIM:
		   glScaled(dscal,dscal,1.0);
		   break;
		 case MOPS_VTSIDE:
		   glScaled(1.0,dscal,dscal);
		   break;
		 case MOPS_VTTOP:
		   glScaled(dscal,1.0,dscal);
		   break;
		 default:
		   break;
		 }

	       mops_get_all_srtransform(mops_current_level->next);

	       glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	      glPopMatrix();

	      while(point)
		{
		  MOPS_APTRAN3(tpoint,point->point,mm);
		  memcpy(point->point,tpoint,3*sizeof(double));

		  point = point->next;
		}

	      o->modified = MOPS_TRUE;

	    }
	  else
	    {
	       switch(view->type)
		 {
		 case MOPS_VTFRONT:
		 case MOPS_VTTRIM:
		   o->scalx *= dscal;
		   o->scaly *= dscal;
		   break;
		 case MOPS_VTSIDE:
		   o->scaly *= dscal;
		   o->scalz *= dscal;
		   break;
		 case MOPS_VTTOP:
		   o->scalx *= dscal;
		   o->scalz *= dscal;
		   break;
		 default:
		   break;
		 }

	    }
	}
      
      sel = sel->next;
    }

  oldwinx = winx;
  oldwiny = winy;

  if(!mops_prefs.lazynotify)
    mops_notify_parent();

  display_cb(togl);

 return TCL_OK;
}


/* 
   the next functions are the interface to the transformations
   from the console
 */

int
mops_movob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
  double dx = 0, dy = 0, dz = 0; 
  mops_list_object *sel = mops_current_selection;
  mops_object *o = NULL;
  char fname[] = "move_object";

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  o->movx += dx; 
	  o->movy += dy; 
	  o->movz += dz; 
	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_movsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char *argv[])
{
 double dx = 0, dy = 0, dz = 0; 
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 GLdouble mm[16];
 char fname[] = "move_sel";
 double tpoint[4] = {0};

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glMatrixMode (GL_MODELVIEW);
	  glPushMatrix();
	  glTranslatef(dx,dy,dz);
	  glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  point = o->selp;
	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_scalob_tcmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char *argv[])
{
 double dx = 0, dy = 0, dz = 0; 
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 char fname[] = "scale_object";

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  if(dx == 0.0)
    dx = 1.0;
  if(dy == 0.0)
    dy = 1.0;
  if(dz == 0.0)
    dz = 1.0;

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  o->scalx *= dx; 
	  o->scaly *= dy; 
	  o->scalz *= dz; 
	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_scalsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char *argv[])
{
 double dx = 0, dy = 0, dz = 0; 
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 GLdouble mm[16];
 char fname[] = "scal_sel";
 double tpoint[4] = {0};

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  if(dx == 0.0)
    dx = 1.0;
  if(dy == 0.0)
    dy = 1.0;
  if(dz == 0.0)
    dz = 1.0;

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glMatrixMode (GL_MODELVIEW);
	  glPushMatrix();
	  glScalef(dx,dy,dz);
	  glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  point = o->selp;
	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}

int
mops_rotob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 double dx = 0, dy = 0, dz = 0; 
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 float quat[4];
 char fname[] = "rot_object";

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,"%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  while(sel)
    {
      o = sel->object;

      if(o)
	{
	  o->rotx += dx;
	  if(dx != 0.0)
	    {
	      mops_axis_to_quat(xaxis,dx*MOPS_PI/180.0,quat);
	      mops_add_quat(quat,o->quat,o->quat);
	    }
	  o->roty += dy; 
	  if(dy != 0.0)
	    {
	      mops_axis_to_quat(yaxis,dy*MOPS_PI/180.0,quat);
	      mops_add_quat(quat,o->quat,o->quat);
	    }
	  o->rotz += dz; 
	  if(dz != 0.0)
	    {
	      mops_axis_to_quat(zaxis,dz*MOPS_PI/180.0,quat);
	      mops_add_quat(quat,o->quat,o->quat);
	    }
	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_rotsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char *argv[])
{
 double dx = 0, dy = 0, dz = 0; 
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_point_object *point = NULL;
 GLdouble mm[16];
 char fname[] = "rot_sel";
 double tpoint[4] = {0};

  if(argc != 4)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "%dx %dy %dz");
      return TCL_OK;
    }

  Tcl_GetDouble(interp, argv[1], &dx);
  Tcl_GetDouble(interp, argv[2], &dy);
  Tcl_GetDouble(interp, argv[3], &dz);

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  glMatrixMode (GL_MODELVIEW);
	  glPushMatrix();
	  glRotatef(dx,1.0,0.0,0.0);
	  glRotatef(dy,0.0,1.0,0.0);
	  glRotatef(dz,0.0,0.0,1.0);
	  glGetDoublev(GL_MODELVIEW_MATRIX, mm);
	  glPopMatrix();

	  point = o->selp;
	  while(point)
	    {
	      MOPS_APTRAN3(tpoint,point->point,mm);
	      memcpy(point->point,tpoint,3*sizeof(double));

	      point = point->next;
	    }

	  o->modified = MOPS_TRUE;

	}
      
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_moview_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0;
 static double olddx = 0.0, olddy = 0.0;
 double dxw = 0.0, dyw = 0.0;
 double t[3] = {0}, t2[2] = {0};
 GLdouble mm[16]= {0}, rotx = 0.0, roty = 0.0;
 char fname[] = "move_view";

  /* parse args */
  if(argc == 5)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);
	    oldwinx = winx;
	    oldwiny = winy;
	    olddx = 0.0;
	    olddy = 0.0;
	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winx winy|-winxy winx winy]");
      return TCL_OK;
    }

  dxw = (oldwinx - winx) * view->conv_x;
  dyw = -(oldwiny - winy) * view->conv_y;
  oldwinx = winx;
  oldwiny = winy;

  if(view->type == MOPS_VTPERSP)
    {
     dxw *= (1.0/view->zoom)*4;
     dyw *= (1.0/view->zoom)*4;
    }

  t[0] = view->from_x - view->to_x;
  t[1] = view->from_y - view->to_y;
  t[2] = view->from_z - view->to_z;
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
   glLoadIdentity();
   if(!t[0] && !t[2])
     {
       glTranslated(dxw,0.0,-dyw);
     }
   else
     {
       roty = MOPS_R2D(atan2(t[0],t[2]));

       t2[0]=t[0];
       t2[1]=t[2];
       if(t[0]&&t[2])
	 rotx = MOPS_R2D(atan2(t[1],MOPS_V2LEN(t2)));

       glRotated(roty,0.0,1.0,0.0);
       glRotated(-rotx,1.0,0.0,0.0);
       if(view->up_y>0.0)
	 glTranslated(dxw,dyw,0.0);
       else
	 glTranslated(-dxw,-dyw,0.0);
       glRotated(rotx,1.0,0.0,0.0);
       glRotated(-roty,0.0,1.0,0.0);
     }
   glGetDoublev(GL_MODELVIEW_MATRIX, mm);
  glPopMatrix();

  t[0]=view->from_x;
  t[1]=view->from_y;
  t[2]=view->from_z;
  mops_apply_transform3(t,mm);
  view->from_x = t[0];
  view->from_y = t[1];
  view->from_z = t[2];
  t[0]=view->to_x;
  t[1]=view->to_y;
  t[2]=view->to_z;
  mops_apply_transform3(t,mm);
  view->to_x = t[0];
  view->to_y = t[1];
  view->to_z = t[2];

  reshape_cb(togl);
  display_cb(togl);

 return TCL_OK;
}


int
mops_zoomview_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwiny = 0.0;
 double winy = 0.0;
 double dy = 1.0;
 char fname[] = "zoom_view";

  /* parse args */
  if(argc == 4)
    {
      if(!strcmp(argv[2],"-winy"))
      {
	Tcl_GetDouble(interp, argv[3], &winy);
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winy);
	    oldwiny = winy;
	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winy|-winy winy]");
      return TCL_OK;
    }

  if(oldwiny == winy)
    return TCL_OK;

  dy = 1.1;
  if((oldwiny - winy)>0)
    dy = 0.9;

  /* zoom the view */
  view->zoom *= dy;

  oldwiny = winy;

  if(dy != 1.0)
    {
      reshape_cb(togl);
      display_cb(togl);
    }

 return TCL_OK;
}


int
mops_movezview_actcb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwiny = 0.0;
 double winy = 0.0;
 static double olddy = 0.0;
 double dyw = 0.0;
 double v[3] = {0};
 char fname[] = "movez_view";

  /* parse args */
  if(argc == 4)
    {
      if(!strcmp(argv[2],"-winy"))
      {
	Tcl_GetDouble(interp, argv[3], &winy);
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winy);
	    oldwiny = winy;
	    olddy = 0.0;
	  }
    }
  else
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "[-start winy|-winy winy]");
      return TCL_OK;
    }

  dyw = -(oldwiny - winy) * view->conv_x * 2.0;
  oldwiny = winy;

  /* move the view along its current z axis */
  v[0] = view->from_x - view->to_x;
  v[1] = view->from_y - view->to_y;
  v[2] = view->from_z - view->to_z;

  MOPS_V3SCAL(v,dyw)

  view->from_x += v[0];
  view->to_x += v[0];
  view->from_y += v[1];
  view->to_y += v[1];
  view->from_z += v[2];
  view->to_z += v[2];    

  oldwiny = winy;

  if(dyw != 0.0)
    {
      reshape_cb(togl);
      display_cb(togl);
    }

 return TCL_OK;
}
