/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Box */

int
mops_crtbox_tcmd(Tcl_Interp * interp,
		 int argc, char *argv[], mops_object *o)
{
 mops_box_object *box = NULL;
 char fname[] = "crtbox";


 if(!(box = calloc(1,sizeof(mops_box_object))))
   {
     mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
     return TCL_ERROR;
   }

 box->width  = 1.0;
 box->length = 1.0;
 box->height = 1.0;
   
 o->type = MOPS_OTBOX;
 o->object = box;

 return TCL_OK;
}


int
mops_free_box(mops_box_object *box)
{
  if(!box)
    return MOPS_OK;

  free(box);

 return MOPS_OK;
}


int
mops_copy_box(mops_box_object *source,
	      mops_box_object **dest)
{
 int mops_status = MOPS_OK;
 mops_box_object *box = NULL;

 if(!(box = calloc(1,sizeof(mops_box_object))))
   return MOPS_OUTOFMEM; 
 memcpy(box, source, sizeof(mops_box_object)); 
 *dest = box;

 return mops_status;
}


int
mops_draw_box_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_box_object *box = NULL;
 GLdouble wh = 0.0, hh = 0.0, lh = 0.0;

  if(!o)
    return MOPS_OK;

  box = o->object;
  if(!box)
    return MOPS_OK;    

  wh = (GLdouble)(box->width*0.5);
  lh = (GLdouble)(box->length*0.5);
  hh = (GLdouble)(box->height*0.5);

  /* draw */
  glBegin(GL_LINE_LOOP);
  glVertex3d(wh,hh,lh);
  glVertex3d(wh,-hh,lh);
  glVertex3d(-wh,-hh,lh);
  glVertex3d(-wh,hh,lh);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3d(wh,hh,-lh);
  glVertex3d(wh,-hh,-lh);
  glVertex3d(-wh,-hh,-lh);
  glVertex3d(-wh,hh,-lh);
  glEnd();
  glBegin(GL_LINES);
  glVertex3d(wh,hh,lh);
  glVertex3d(wh,hh,-lh);
  glVertex3d(wh,-hh,lh);
  glVertex3d(wh,-hh,-lh);
  glVertex3d(-wh,-hh,lh);
  glVertex3d(-wh,-hh,-lh);
  glVertex3d(-wh,hh,lh);
  glVertex3d(-wh,hh,-lh);
  glEnd();

 return mops_status;
}


int
mops_draw_box(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_box_object *box = NULL;
 GLdouble wh = 0.0, hh = 0.0, lh = 0.0;

  if(!o)
    return MOPS_OK;

  box = o->object;
  if(!box)
    return MOPS_OK;

  wh = (GLdouble)(box->width/2.0);
  lh = (GLdouble)(box->length/2.0);
  hh = (GLdouble)(box->height/2.0);

  glBegin(GL_QUADS);
  glNormal3d(0.0,0.0,1.0);
  glVertex3d(wh,hh,lh);
  glVertex3d(-wh,hh,lh);
  glVertex3d(-wh,-hh,lh);
  glVertex3d(wh,-hh,lh);

  glNormal3d(0.0,-1.0,0.0);
  glVertex3d(wh,hh,lh);
  glVertex3d(-wh,hh,lh);
  glVertex3d(-wh,hh,-lh);
  glVertex3d(wh,hh,-lh);

  glNormal3d(1.0,0.0,0.0);
  glVertex3d(wh,hh,lh);
  glVertex3d(wh,-hh,lh);
  glVertex3d(wh,-hh,-lh);
  glVertex3d(wh,hh,-lh);

  glNormal3d(0.0,0.0,1.0);
  glVertex3d(wh,hh,-lh);
  glVertex3d(-wh,hh,-lh);
  glVertex3d(-wh,-hh,-lh);
  glVertex3d(wh,-hh,-lh);

  glNormal3d(0.0,-1.0,0.0);
  glVertex3d(wh,-hh,lh);
  glVertex3d(-wh,-hh,lh);
  glVertex3d(-wh,-hh,-lh);
  glVertex3d(wh,-hh,-lh);

  glNormal3d(1.0,0.0,0.0);
  glVertex3d(-wh,hh,lh);
  glVertex3d(-wh,-hh,lh);
  glVertex3d(-wh,-hh,-lh);
  glVertex3d(-wh,hh,-lh);
  
  glEnd();


 return mops_status;
}


int
mops_wrib_box(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_box_object *box = NULL;
  RtPoint rect[4];

  if(!o)
   return MOPS_OK;

  box = (mops_box_object*)(o->object);

  RiSolidBegin(RI_PRIMITIVE);
  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  RiScale((RtFloat)box->width,(RtFloat)box->height,(RtFloat)box->length);

  rect[0][0] = 0.5;
  rect[0][1] = 0.5;
  rect[0][2] = 0.5;

  rect[1][0] = 0.5;
  rect[1][1] = 0.5;
  rect[1][2] = -0.5;

  rect[2][0] = 0.5;
  rect[2][1] = -0.5;
  rect[2][2] = 0.5;

  rect[3][0] = 0.5;
  rect[3][1] = -0.5;
  rect[3][2] = -0.5;

  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
  RiRotate(90.0, 0.0, 0.0, 1.0);
  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
  RiRotate(90.0, 0.0, 0.0, 1.0);
  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
  RiRotate(90.0, 0.0, 0.0, 1.0);
  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
  RiRotate(90.0, 0.0, 1.0, 0.0);
  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
  RiRotate(180.0, 0.0, 1.0, 0.0);
  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);

  RiTransformEnd();
  RiAttributeEnd(); 
  RiSolidEnd();

 return mops_status;
}


int
mops_get_box_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);


  Tcl_SetVar(interp,"propertyList","BoxAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  
  Tcl_SetVar(interp,"propertyProcList","buildBoxAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getBoxAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  
  Tcl_SetVar(interp,"propertySetProcList","setBoxAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
 
 return mops_status;
}

/* Tcl -> C! */
int
mops_set_boxattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="BoxAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 mops_box_object *box = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;
  
  toa = Tcl_NewStringObj(n1,-1);

  box = (mops_box_object *)o->object;

  ton = Tcl_NewStringObj("Width",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &box->width);

  Tcl_SetStringObj(ton,"Length",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &box->length);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &box->height);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

/* C -> Tcl! */
int
mops_get_boxattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="BoxAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 mops_box_object *box = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  toa = Tcl_NewStringObj(n1,-1);

  box = (mops_box_object *)o->object;

  ton = Tcl_NewStringObj("Width",-1);
  to = Tcl_NewDoubleObj(box->width);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Length",-1);
  to = Tcl_NewDoubleObj(box->length);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_NewDoubleObj(box->height);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}
