/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Clipboard & Selection */

/* The Selection */

int
mops_free_selection()
{
 mops_list_object *sel = mops_current_selection, *seln = NULL;

  while(sel)
    {
      seln = sel->next;
      sel->object->selected = 0;
      free(sel);
      sel = seln;
    }

  mops_current_selection = NULL;
  mops_last_selection = NULL;

  if(mops_point_edit_coords)
    free(mops_point_edit_coords);

  mops_point_edit_coords = NULL;

 return MOPS_OK;
}


int
mops_add_selection(mops_object *o)
{
 mops_list_object *new_sel = NULL;

  if(!(new_sel = calloc(1,sizeof(mops_list_object))))
    return MOPS_OUTOFMEM;

  new_sel->object = o;
  o->selected = 1;

  if(mops_current_selection)
    mops_last_selection->next = new_sel;
  else
    mops_current_selection = new_sel; 

  mops_last_selection = new_sel;

 return MOPS_OK;
}


/*
 * argv[] contains the split list of selected elements
 * of the hierarchy listbox
 *
 */
int
mops_set_select_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 int i = 0, j = 1, argvi = 0, start = 1;
 mops_object *o = mops_current_level->object;

  if(mops_current_selection)
    mops_free_selection();

  /* nothing selected ? */
  if(argc == 1)
    {
      /* no editable properties */
      Tcl_SetVar(interp,"propertyList","",TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
      return TCL_OK; 
    }

  if(!o)
    return TCL_OK;
  else
    {

      while(o)
	{
	  o->selected = MOPS_FALSE;
	  o = o->next;
	}
      o = mops_current_level->object;

    }

  if(o != mops_root)
    {
      /* ".." selected? */
      if(atoi(argv[1]) == 0)
	{
	  Tcl_SetVar(interp,"propertyList","",TCL_GLOBAL_ONLY |
		     TCL_LEAVE_ERR_MSG);
	  start++;
	}
    }
  else
    {
      /* "Scene" selected? */
      if(atoi(argv[1]) == 0)
	{
	  start++; 
	  mops_status = mops_add_selection(o);
	  if(argc == 2)
	    mops_status = mops_get_root_properties(interp);
	  /* prevent the rest of MOPS from destroying the root */
	  mops_free_selection();

	}

      if(o->next)
	o = o->next;
    }


  for(i=start;i<argc;i++)
    {
      argvi = atoi(argv[i]);
      while(j != argvi)
	{
	  j++;

	  if(o)
	    o = o->next;	

	  if(!o)
	    break;

	  /* no reset of o for next iteration, because we believe
	     that the arguments are ordered! */
	}

      /* found a selected object: add to the list */
      if(o)
	mops_status = mops_add_selection(o);

    }

  /* if a single object has been selected we ask
     the object to build a list of editable properties
     to display in the appropriate listbox */
  if(mops_current_selection)
    {
      if((!mops_current_selection->next) && (argc == 2))
	{
	  o = mops_current_selection->object;
	  switch (o->type)
	    {
	    case MOPS_OTLEVEL:
	      mops_status = mops_get_level_properties(interp);
	      break;
	    case MOPS_OTCUSTOM:
	      mops_status = mops_get_custom_properties(interp, o);
	      break;
	    case MOPS_OTNURBCURVE:
	      mops_status = mops_get_nurbcurve_properties(interp);
	      break;
	    case MOPS_OTNURBPATCH:
	      mops_status = mops_get_nurbpatch_properties(interp);
	      break;
	    case MOPS_OTLIGHT:
	      mops_status = mops_get_light_properties(interp, o);
	      break;
	    case MOPS_OTSPHERE:
	    case MOPS_OTCYLINDER:
	    case MOPS_OTDISK:
	    case MOPS_OTCONE:
	    case MOPS_OTTORUS:
	    case MOPS_OTHYPERBOLOID:
	    case MOPS_OTPARABOLOID:
	      mops_status = mops_get_quadric_properties(interp, o->type);
	      break;
	    case MOPS_OTBOX:
	      mops_status = mops_get_box_properties(interp);
	      break;
	    case MOPS_OTVIEW:
	      mops_status = mops_get_view_properties(interp, o);
	      break;
	    case MOPS_OTINSTANCE:
	      mops_status = mops_get_instance_properties(interp);
	      break;
	    }
	}
      else
	{ /* multiple-selected-objects -> no property editing (yet?) */
	  Tcl_UnsetVar(interp,"propertyList",TCL_LEAVE_ERR_MSG |
		       TCL_GLOBAL_ONLY);
	  Tcl_SetVar(interp,"propertyList","",TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
	}
    }

 return TCL_OK;
}


/* The Clipboard */


int
mops_delob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL, *clevel = NULL;
 mops_list_object *lev = mops_current_level;
 char fname[] = "del_object";

 clevel = lev->object;

 /* objects may get deleted */
 if(sel)
   mops_clear_undo_info();

  while(sel)
    {
      o = sel->object;
      if(o)
	{
	  if(o->refcount)
	    {
	      mops_error(interp,MOPS_REFCOUNT,fname,NULL);
	      return TCL_OK;
	    }
	  if(o->down)
	  if(mops_check_refcount(o->down))
	    {
	      mops_error(interp,MOPS_REFCOUNT,fname,NULL);
	      return TCL_OK;
	    }
	  mops_status = mops_unlink_object(o);
	  mops_status = mops_delete_object(o);
	}
      sel = sel->next;
    }

  if(mops_current_selection)
    mops_status = mops_free_selection();

  mops_notify_parent();

 return TCL_OK;
}


int
mops_cutob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_object *clip = mops_object_clipboard;
 char fname[] = "cut_object";

  if(sel)
    {
      /* objects may get deleted */
      mops_clear_undo_info();

      /* clear old clipboard content */
      while(clip)
	{
	  o = clip;
	  clip = clip->next;
	  mops_status = mops_delete_object(o);	  
	}
      mops_object_clipboard = NULL;

      /* cut first object */
      o = sel->object;
      if(o)
	{
	  if(o->refcount)
	    {
	      mops_error(interp,MOPS_REFCOUNT,fname,NULL);
	      return TCL_OK;
	    }
	  if(o->down)
	  if(mops_check_refcount(o->down))
	    {
	      mops_error(interp,MOPS_REFCOUNT,fname,NULL);
	      return TCL_OK;
	    }

	  mops_status = mops_unlink_object(o);
	  mops_object_clipboard = o;

	  /* single object selected ? */
	  if(!sel->next)
	    o->next = NULL;

	  sel = sel->next;

	  /* cut the rest */
	  while(sel)
	    {
	      if(mops_check_refcount(sel->object))
		{
		  mops_error(interp,MOPS_REFCOUNT,fname,NULL);
		  if(o)
		    o->next = NULL;
		  return TCL_OK;
		}

	      mops_status = mops_unlink_object(sel->object);
	      o->next = sel->object;
	      o = sel->object;

	      sel = sel->next;
	    }

	  /* terminate clipboard hierarchy */
	  o->next = NULL;
	}

    }

  mops_status = mops_free_selection();
  
  mops_notify_parent();

 return TCL_OK;
}


int
mops_pasob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_object *clip = mops_object_clipboard;
 mops_object *ins = NULL;
 char fname[] = "paste_object";

  if(clip)
    {
      o = clip;
      while(o)
	{

	  mops_status = mops_copy_object(o,&ins);
	  mops_status = mops_link_new_object(ins);
	  if(mops_status)
	    {
	      /* oops, couldn't insert object */
	      mops_error(interp, mops_status, fname, NULL);
	      mops_status = mops_delete_object(ins);
	      return TCL_ERROR;
	    }
	    
	  o = o->next;
	}

    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_copob_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 int mops_status;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_object *clip = mops_object_clipboard;
 char fname[] = "copy_object";

  if(!sel)
    return TCL_OK;

  /* clear old clip */
  while(clip)
    {
      o = clip;
      clip = clip->next;
      mops_status = mops_delete_object(o);	  
    }

  o = sel->object;
  mops_status = mops_copy_object(o, &mops_object_clipboard);

  MOPS_CRERROR

  clip = mops_object_clipboard;

  sel = sel->next;

  while(sel)
    {
      o = sel->object;

      mops_status = mops_copy_object(o, &(clip->next));

      MOPS_CRERROR

      clip = clip->next;
      sel = sel->next;
    }

  mops_status = mops_free_selection();
    
 return TCL_OK;
}
