/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Custom Objects */

Tcl_HashTable mops_custom_id_ht;
static Tcl_HashTable mops_custom_names_ht;
static Tcl_HashTable mops_custom_createcb_ht;
static Tcl_HashTable mops_custom_drawscb_ht;
static Tcl_HashTable mops_custom_drawhcb_ht;
static Tcl_HashTable mops_custom_drawwcb_ht;
static Tcl_HashTable mops_custom_deletecb_ht;
static Tcl_HashTable mops_custom_copycb_ht;
static Tcl_HashTable mops_custom_getpropcb_ht;
static Tcl_HashTable mops_custom_getpointcb_ht;
static Tcl_HashTable mops_custom_wribcb_ht;
static Tcl_HashTable mops_custom_readcb_ht;
static Tcl_HashTable mops_custom_writecb_ht;
static Tcl_HashTable mops_custom_notifycb_ht;
static Tcl_HashTable mops_custom_provides_ht;
static Tcl_HashTable mops_custom_bbccb_ht;

static char *typecounter = NULL;


/*************************************/
/* initialize custom hashtables      */
/*************************************/
int
mops_init_custom_ht()
{
 int mops_status = MOPS_OK;

  Tcl_InitHashTable(&mops_custom_id_ht, TCL_STRING_KEYS);
  Tcl_InitHashTable(&mops_custom_names_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_createcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_drawscb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_drawhcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_drawwcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_deletecb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_copycb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_getpropcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_getpointcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_wribcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_readcb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_writecb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_notifycb_ht, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&mops_custom_provides_ht, TCL_STRING_KEYS);
  Tcl_InitHashTable(&mops_custom_bbccb_ht, TCL_ONE_WORD_KEYS);

 return mops_status;
}


/*************************************/
/* register a new custom object type 
   type_name, create, delete, copy
   may not, all other callbacks may
   be NULL                           */
/*************************************/
int
mops_register_custom_objecttype(char *type_name,
				mops_create_custom_cb *customcreate_cb,
				mops_draw_custom_cb *customdraws_cb,
				mops_draw_custom_cb *customdrawh_cb,
				mops_draw_custom_cb *customdraww_cb,
				mops_delete_custom_cb *customdelete_cb,
				mops_copy_custom_cb *customcopy_cb,
				mops_getprop_custom_cb *customgetprop_cb,
				mops_getpoint_custom_cb *customgetpoint_cb,
				mops_wrib_custom_cb *customwrib_cb,
				mops_read_custom_cb *customread_cb,
				mops_write_custom_cb *customwrite_cb,
				mops_notify_custom_cb *customnotify_cb)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 Tcl_HashEntry *entry = NULL;

  /* check, if type is already registered */
  if((entry = Tcl_FindHashEntry(&mops_custom_id_ht, type_name)))
     return MOPS_TYPEEXISTS;

  typecounter++;
  entry = Tcl_CreateHashEntry(&mops_custom_id_ht, type_name, &new_item);
  Tcl_SetHashValue(entry, typecounter);

  /* register name */
  entry = Tcl_CreateHashEntry(&mops_custom_names_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, type_name);

  /* register create callback */
  entry = Tcl_CreateHashEntry(&mops_custom_createcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customcreate_cb);

  /* register draw surface callback */
  entry = Tcl_CreateHashEntry(&mops_custom_drawscb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customdraws_cb);

  /* register draw-handles callback */
  entry = Tcl_CreateHashEntry(&mops_custom_drawhcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customdrawh_cb);

  /* register draw-wire callback */
  entry = Tcl_CreateHashEntry(&mops_custom_drawwcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customdraww_cb);

  /* register delete callback */
  entry = Tcl_CreateHashEntry(&mops_custom_deletecb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customdelete_cb);

  /* register copy callback */
  entry = Tcl_CreateHashEntry(&mops_custom_copycb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customcopy_cb);

  /* register getprop callback */
  entry = Tcl_CreateHashEntry(&mops_custom_getpropcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customgetprop_cb);

  /* register getpoint callback */
  entry = Tcl_CreateHashEntry(&mops_custom_getpointcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customgetpoint_cb);

  /* register wrib callback */
  entry = Tcl_CreateHashEntry(&mops_custom_wribcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customwrib_cb);

  /* register read callback */
  entry = Tcl_CreateHashEntry(&mops_custom_readcb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customread_cb);

  /* register write callback */
  entry = Tcl_CreateHashEntry(&mops_custom_writecb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customwrite_cb);

  /* register notify callback */
  entry = Tcl_CreateHashEntry(&mops_custom_notifycb_ht, typecounter,
			      &new_item);
  Tcl_SetHashValue(entry, customnotify_cb);


 return mops_status;
}


/*************************************/
/* create a custom object            */
/*************************************/
int
mops_custom_create_tcmd(Tcl_Interp * interp,
			int argc, char *argv[], mops_object *o)
{
 int mops_status = MOPS_OK;
 char *customid;
 mops_custom_object *custom = NULL;
 mops_create_custom_cb *customcrtfunc = NULL;
 Tcl_HashEntry *entry;
 char fname[] = "custom_create";

  if(argc < 3)
    {
      mops_error(interp, MOPS_ARGS, fname, "name");
      return TCL_ERROR;
    }

  o->type = MOPS_OTCUSTOM;

  if(!(custom = calloc(1, sizeof(mops_object))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  o->object = custom;

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, argv[2]);
  if (entry != NULL)
    {
      customid = Tcl_GetHashValue(entry);
    }
  else
    {
      free(custom);
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_ERROR;
    }

  /* get create callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_createcb_ht, customid);
  if (entry != NULL)
    {
      customcrtfunc = (mops_create_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    {
      free(custom);
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_ERROR;
    }

  if(customcrtfunc)
    mops_status = customcrtfunc(interp, argc, argv, &(custom->object));

  if(mops_status)
    {
      free(custom);
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  custom->type = customid;

  entry = Tcl_FindHashEntry(&mops_custom_provides_ht, argv[2]);
  if (entry != NULL)
    {
      custom->provides = (unsigned int)Tcl_GetHashValue(entry);
    }

  o->modified = MOPS_TRUE;

 return TCL_OK;
}


/*************************************/
/* get all loaded custom objects     */
/*************************************/
int
mops_get_custom_objects_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 Tcl_HashEntry *entry;
 char *i = NULL;
 char *name = NULL;
 char fname[] = "get_custom_objects";

  if(argc < 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "varname");
      return TCL_OK;
    }

  if(!typecounter)
    return TCL_OK;
    
  for(i = typecounter; i; i--)
    {
      entry = Tcl_FindHashEntry(&mops_custom_names_ht, i);
      if(entry != NULL)
	{
	  name = (char *)Tcl_GetHashValue(entry);
	  if(name)
	    Tcl_SetVar(interp, argv[1], name, TCL_APPEND_VALUE |
		       TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	}
    }

 return TCL_OK;
}


/*************************************/
/* get name of a custom object type  */
/*************************************/
int
mops_get_custom_objectname(mops_object *o, char **result)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get name from id */
  entry = Tcl_FindHashEntry(&mops_custom_names_ht, custom->type);
  if (entry != NULL)
    *result = (char *) Tcl_GetHashValue(entry);
  else
    return MOPS_HASH;

 return mops_status;
}


/*************************************/
/* draw wireframe of a custom object */
/*************************************/
int
mops_draw_custom_w(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 mops_object *i = NULL;
 mops_custom_object *custom = NULL;
 mops_draw_custom_cb *customdrawwfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      i = o;
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_drawwcb_ht, custom->type);
  if (entry != NULL)
    {
      customdrawwfunc = (mops_draw_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customdrawwfunc)
    {
      if(instance)
	{
	  mops_status = customdrawwfunc(togl, i);
	}
      else
	{
	  mops_status = customdrawwfunc(togl, o);
	}
    }

 return mops_status;
}


/*****************************************/
/* draw surface/curve of a custom object */
/*****************************************/
int
mops_draw_custom_s(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 mops_object *i = NULL;
 mops_custom_object *custom = NULL;
 mops_draw_custom_cb *customdrawsfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      i = o;
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_drawscb_ht, custom->type);
  if (entry != NULL)
    {
      customdrawsfunc = (mops_draw_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    {
      return MOPS_HASH;
    }

  if(customdrawsfunc)
    {
      if(instance)
	{
	  mops_status = customdrawsfunc(togl, i);
	}
      else
	{
	  mops_status = customdrawsfunc(togl, o);
	}
    }

 return mops_status;
}


/***********************************/
/* draw handles of a custom object */
/***********************************/
int
mops_draw_custom_h(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_draw_custom_cb *customdrawhfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_drawhcb_ht, custom->type);
  if (entry != NULL)
    {
      customdrawhfunc = (mops_draw_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customdrawhfunc)
    mops_status = customdrawhfunc(togl, o);

 return mops_status;
}


/***************************************************/
/* get point (for pointediting) of a custom object */
/***************************************************/
int
mops_get_custom_point(mops_object *o, double objX, double objY, double objZ)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_getpoint_custom_cb *customgetpointfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_getpointcb_ht, custom->type);
  if (entry != NULL)
    {
      customgetpointfunc = (mops_getpoint_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customgetpointfunc)
    mops_status = customgetpointfunc(o, objX, objY, objZ);

 return mops_status;
}


/*************************************/
/* get properties of a custom object */
/*************************************/
int
mops_get_custom_properties(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_getprop_custom_cb *customgetpropertiesfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_getpropcb_ht, custom->type);
  if (entry != NULL)
    {
      customgetpropertiesfunc = (mops_getprop_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customgetpropertiesfunc)
    mops_status = customgetpropertiesfunc(interp);

 return mops_status;
}


/*************************************/
/* free a custom object              */
/*************************************/
int
mops_free_custom(mops_custom_object *c)
{
 int mops_status = MOPS_OK;
 mops_delete_custom_cb *customdeletefunc = NULL;
 Tcl_HashEntry *entry;

  if(!c)
    return MOPS_OK;
 
  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_deletecb_ht, c->type);
  if (entry != NULL)
    {
      customdeletefunc = (mops_delete_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customdeletefunc)
    mops_status = customdeletefunc(c->object);

  if(!mops_status)
    free(c); /* XXXX rather free() it anyway? */

 return mops_status;
}


/*************************************/
/* copy a custom object              */
/*************************************/
int
mops_copy_custom(mops_custom_object *source,
		 mops_custom_object **dest)
{
 int mops_status = MOPS_OK;
 mops_copy_custom_cb *customcopyfunc = NULL;
 Tcl_HashEntry *entry;
 mops_custom_object *new = NULL;

  if(!source)
    return MOPS_OK;
 
  /* copy generic custom object */
  if(!(new = calloc(1,sizeof(mops_custom_object))))
    return MOPS_OUTOFMEM;
  
  *dest = new;

  memcpy(new, source, sizeof(mops_custom_object));
  /* danger! links point to original hierarchy */
  new->object = NULL;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_copycb_ht, source->type);
  if (entry != NULL)
    {
      customcopyfunc = (mops_copy_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    {
      free(new);
      *dest = NULL;
      return MOPS_HASH;
    }

  if(customcopyfunc)
    mops_status = customcopyfunc(source->object, &(new->object));

  if(mops_status)
    {
      free(new);
      *dest = NULL;
      return mops_status;
    }

 return mops_status;
}


/*************************************/
/* export a custom object to RIB     */
/*************************************/
int
mops_wrib_custom(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_wrib_custom_cb *customwribfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_wribcb_ht, custom->type);
  if (entry != NULL)
    {
      customwribfunc = (mops_wrib_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customwribfunc)
    mops_status = customwribfunc(o);

 return mops_status;
}

int 
mops_read_custom(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 char *customname = NULL;
 char *customid = NULL;
 Tcl_HashEntry *entry;
 mops_read_custom_cb *customreadfunc = NULL;
 mops_custom_object *custom = NULL;
 char *name = NULL;

  if(!(custom = calloc(1, sizeof(mops_custom_object))))
    { return MOPS_OUTOFMEM; }

  o->object = custom;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &name);
  o->name = name;

  mops_read_string(fileptr, &customname);

  /* get id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, customname);
  if (entry != NULL)
    {
      customid = Tcl_GetHashValue(entry);
    }
  else
    {
      free(custom); free(customname);
      return MOPS_HASH;
    }

  custom->type = customid;

  /* get cb from id */
  entry = Tcl_FindHashEntry(&mops_custom_readcb_ht, customid);
  if (entry != NULL)
    {
      customreadfunc = (mops_read_custom_cb *) Tcl_GetHashValue(entry);
    }
  else
    {
      free(custom); free(customname);
      return MOPS_HASH;
    }

  if(customreadfunc)
    mops_status = customreadfunc(fileptr, custom);

  if(mops_status)
    {
      free(custom); free(customname);
      return mops_status;
    }

  entry = Tcl_FindHashEntry(&mops_custom_provides_ht, customname);
  if (entry != NULL)
    {
      custom->provides = (unsigned int)Tcl_GetHashValue(entry);
    }


  free(customname);

 return mops_status;
}


int 
mops_write_custom(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_write_custom_cb *customwritefunc = NULL;
 Tcl_HashEntry *entry;
 char *customname = NULL;

  if(!o)
    return MOPS_OK;

  custom = (mops_custom_object*)(o->object);
 
  if(!custom)
    return MOPS_OK;

  fprintf(fileptr, "%d\n", MOPS_OTCUSTOM);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

  /* get name from id */
  entry = Tcl_FindHashEntry(&mops_custom_names_ht, custom->type);
  if (entry != NULL)
    {
      customname = (char *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;
  
  if(customname)
    fprintf(fileptr,"%s\n", customname);
  
  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_writecb_ht, custom->type);
  if (entry != NULL)
    {
      customwritefunc = (mops_write_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;
  
  if(customwritefunc)
    mops_status = customwritefunc(fileptr, custom);

 return mops_status;
}


int
mops_notify_custom(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_notify_custom_cb *customnotifyfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_notifycb_ht, custom->type);
  if (entry != NULL)
    {
      customnotifyfunc = (mops_notify_custom_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(customnotifyfunc)
    mops_status = customnotifyfunc(o);

 return mops_status;
}


int
mops_register_provider(Tcl_Interp *interp, char *type,
		       unsigned int target_type,
		       mops_provider_cb *cb)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 Tcl_HashEntry *entry = NULL;
 char  fname[] = "register_provider";

  /* check, if provider for type is already registered */
  if((entry = Tcl_FindHashEntry(&mops_provider_ht, type)))
    {
      mops_error(interp, MOPS_ERROR, fname, "provider already registered");
      return TCL_OK;
    }

  entry = Tcl_CreateHashEntry(&mops_provider_ht, type, &new_item);
  Tcl_SetHashValue(entry, cb);

  if((entry = Tcl_FindHashEntry(&mops_custom_provides_ht, type)))
    {
      mops_error(interp, MOPS_ERROR, fname, "provider already registered");
      return TCL_OK;
    }

  entry = Tcl_CreateHashEntry(&mops_custom_provides_ht, type, &new_item);
  Tcl_SetHashValue(entry, target_type);

 return mops_status;
}


int
mops_provide_object(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 char *type = NULL;
 Tcl_HashEntry *entry = NULL;
 mops_provider_cb *cb = NULL;

 mops_status = mops_get_objectname(o, MOPS_TRUE, &type);
 if(mops_status || (!type))
   return mops_status;

   
 if(!(entry = Tcl_FindHashEntry(&mops_provider_ht, type)))
   {
     free(type);
     return MOPS_HASH;
   }

 if(entry)
   cb = (mops_provider_cb *)Tcl_GetHashValue(entry);

 if(cb)
   {
     mops_status = cb(o, target);
   }

  free(type);

 return mops_status;
}


int
mops_register_bbccb(Tcl_Interp *interp, char *type, mops_bbc_cb *cb)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 Tcl_HashEntry *entry = NULL;
 char  fname[] = "register_bbccb";

  /* check, if bbc-callback for type is already registered */
  if((entry = Tcl_FindHashEntry(&mops_custom_bbccb_ht, type)))
    {
      mops_error(interp, MOPS_ERROR, fname, "bbc callback already registered");
      return TCL_OK;
    }

  entry = Tcl_CreateHashEntry(&mops_custom_bbccb_ht, type, &new_item);
  Tcl_SetHashValue(entry, cb);


 return mops_status;
}


int
mops_bb_custom(mops_object *o, double *bb)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_bbc_cb *custombbcfunc = NULL;
 Tcl_HashEntry *entry;

  if(!o)
    return MOPS_OK;

  custom = o->object;

  if(!custom)
    return MOPS_OK;

  /* get callback from id */
  entry = Tcl_FindHashEntry(&mops_custom_bbccb_ht, custom->type);
  if (entry != NULL)
    {
      custombbcfunc = (mops_bbc_cb *)Tcl_GetHashValue(entry);
    }
  else
    return MOPS_HASH;

  if(custombbcfunc)
    mops_status = custombbcfunc(o, bb);

 return mops_status;
}
