/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Error */

/* Global Variables */
int mops_errno;

void
mops_error(Tcl_Interp *interp, int code, char *where, char *what)
{
  Tcl_DString ds;
  char gentxterr[] = "Error: ", gentxtwarn[] = "Warning: ";
  char warncmd[] = "puts stdout \"", errcmd[] = "puts stderr \"";

  mops_errno = code;

  if(code == MOPS_OK)
    return;

  Tcl_DStringInit(&ds);

  if(code==MOPS_WARNING)
    Tcl_DStringAppend(&ds, warncmd, -1);
  else
    Tcl_DStringAppend(&ds, errcmd, -1);

  if(where)
    {
      Tcl_DStringAppend(&ds, where, -1);
      Tcl_DStringAppend(&ds, ": ", -1);
    }

  switch(code)
    {
    case MOPS_WARNING:
      Tcl_DStringAppend(&ds, gentxtwarn, -1);
      break;
    case MOPS_OUTOFMEM:
      Tcl_DStringAppend(&ds, MOPS_OUTOFMEMSTR, -1);
      break;
    case MOPS_OPENFILE:
      Tcl_DStringAppend(&ds, MOPS_OPENFILESTR, -1);
      break;
    case MOPS_CLOSEFILE:
      Tcl_DStringAppend(&ds, MOPS_CLOSEFILESTR, -1);
      break;
    case MOPS_NOTAMOPS:
      Tcl_DStringAppend(&ds, MOPS_NOTAMOPSSTR, -1);
      break;
    case MOPS_NOSEL:
      Tcl_DStringAppend(&ds, MOPS_NOSELSTR, -1);
      break;
    case MOPS_ARGS:
      Tcl_DStringAppend(&ds, MOPS_ARGSSTR, -1);
      break;
    case MOPS_HASH:
      Tcl_DStringAppend(&ds, MOPS_HASHSTR, -1);
      break;
    case MOPS_HIER:
      Tcl_DStringAppend(&ds, MOPS_HIERSTR, -1);
      break;
    case MOPS_TYPEEXISTS:
      Tcl_DStringAppend(&ds, MOPS_TYPEEXISTSSTR, -1);
      break;
    case MOPS_REFCOUNT:
      Tcl_DStringAppend(&ds, MOPS_REFCOUNTSTR, -1);
      break;
    case MOPS_ERROR:
      Tcl_DStringAppend(&ds, gentxterr, -1);
      break;
    default:
      Tcl_DStringAppend(&ds, gentxterr, -1);
      break;
    }

  if(what)
    Tcl_DStringAppend(&ds, what, -1);

  Tcl_DStringAppend(&ds, "\"", -1);

  Tcl_Eval(interp, Tcl_DStringValue(&ds));

  Tcl_DStringFree(&ds);

  Tcl_SetVar(interp, "mops_error", "1", TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);


 return;
}
