/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Instances */

int
mops_create_oid_ht(mops_object *o)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 int found = MOPS_FALSE;
 Tcl_HashEntry *entry = NULL;
 mops_tag_object *tag = NULL;

 if(!o)
   {
     Tcl_DeleteHashTable(&mops_oidptr_ht);
     Tcl_InitHashTable(&mops_oidptr_ht, TCL_STRING_KEYS);
     return MOPS_OK;
   }

  while(o)
    {
      if(o->refcount)
	{
	  
	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{
		  if((entry = Tcl_FindHashEntry(&mops_oidptr_ht, tag->val)))
		    return MOPS_ERROR; /* OID already registered? */

		  entry = Tcl_CreateHashEntry(&mops_oidptr_ht,
					      tag->val, &new_item);
		  Tcl_SetHashValue(entry, (char*)o); 

		  found = MOPS_TRUE;
		}

	      tag = tag->next;
	    }

	  if(!found)
	    return MOPS_ERROR; /* this should never happen */

	}

      if(o->down)
	mops_status = mops_create_oid_ht(o->down);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}

int
mops_create_readoid_ht(mops_object *o)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 int found = MOPS_FALSE;
 Tcl_HashEntry *entry = NULL;
 mops_tag_object *tag = NULL;

 if(!o)
   {
     Tcl_DeleteHashTable(&mops_oidptr_ht);
     Tcl_InitHashTable(&mops_oidptr_ht, TCL_STRING_KEYS);
     return MOPS_OK;
   }

  while(o)
    {
      if(o->type != MOPS_OTINSTANCE)
	{
	  
	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{
		  if((entry = Tcl_FindHashEntry(&mops_oidptr_ht, tag->val)))
		    return MOPS_ERROR; /* OID already registered? */

		  entry = Tcl_CreateHashEntry(&mops_oidptr_ht,
					      tag->val, &new_item);
		  Tcl_SetHashValue(entry, (char*)o); 

		  found = MOPS_TRUE;
		}

	      tag = tag->next;
	    }


	}

      if(o->down)
	mops_status = mops_create_readoid_ht(o->down);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}

int
mops_connect_instances(mops_object *o)
{
 int mops_status = MOPS_OK;
 int found = MOPS_FALSE;
 Tcl_HashEntry *entry = NULL;
 mops_tag_object *tag = NULL;
 mops_object *target = NULL;

 if(!o)
   return MOPS_OK;

  while(o)
    {
      if((o->type == MOPS_OTINSTANCE) && (!o->object))
	{
	  
	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{
		  if(!(entry = Tcl_FindHashEntry(&mops_oidptr_ht, tag->val)))
		    return MOPS_ERROR; /* OID not registered? */

		  target = (mops_object *)Tcl_GetHashValue(entry);
		  o->object = target;
		  if(target)
		    target->refcount++;
		  found = MOPS_TRUE;
		}

	      tag = tag->next;
	    }

	  if(!found)
	    return MOPS_ERROR; /* this should never happen */

	}

      if(o->down)
	mops_status = mops_connect_instances(o->down);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}


int
mops_create_oid(char **dest)
{
 static unsigned int counter = 0;

 if(!dest)
   {
     counter=0;
     return MOPS_OK;
   }
 else
   counter++;

 /*
   TeaMan, 24.3.1999 on #AmigaGER (IRCNet):
   "32 chars sind genug"
   "es geht ja wohl um Praxisanwendung und nicht
    um akademische Betrachtungen" ;)
 */

 if(!(*dest = calloc(32, sizeof(char))))
   return MOPS_OUTOFMEM;

 sprintf(*dest,"%u",counter);

 return MOPS_OK;
}

int
mops_create_orig_ids(mops_object *o)
{
 int mops_status = MOPS_OK;
 char *tname = NULL, *tval = NULL;
 int found = MOPS_FALSE;
 mops_tag_object *tag = NULL, *newtag = NULL;

  while(o)
    {
      if(o->refcount)
	{
	  tname = NULL;
	  tval = NULL;

	  mops_status = mops_create_oid(&tval);

	  if(mops_status)
	    return mops_status;

	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{
		  free(tag->val);
		  tag->val = tval;
		  found = MOPS_TRUE;
		}
	      tag = tag->next;
	    }

	  if(!found)
	    {
	      if(!(tname = calloc(3,sizeof(char))))
		return MOPS_OUTOFMEM;

	      strcpy(tname,"OI");
	      if(!(newtag = calloc(1,sizeof(mops_tag_object))))
		return MOPS_OUTOFMEM;

	      newtag->name = tname;
	      newtag->val = tval;
	      newtag->type = mops_oi_tagtype;
	      newtag->next = o->tags;
	      o->tags = newtag;
	    }

	}

      if(o->down)
	mops_status = mops_create_orig_ids(o->down);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}

int
mops_create_instance_ids(mops_object *o)
{
 int mops_status = MOPS_OK;
 int found = MOPS_FALSE;
 mops_tag_object *tag = NULL, *origtag = NULL; 
 mops_object *orig = NULL;

  while(o)
    {
      if(o->type == MOPS_OTINSTANCE)
	{
	  orig = (mops_object *) o->object;

	  found = MOPS_FALSE;
	  tag = orig->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{

		  origtag = tag;
		  found = MOPS_TRUE;
		}
	      tag = tag->next;
	    }

	  if(!found)
	    return MOPS_ERROR; /* This should never happen! */

	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{ 
		  /* copy val from origtag to oitag of instance object */
		  free(tag->val);
		  if(!(tag->val=calloc(strlen(origtag->val)+1,sizeof(char))))
		    return MOPS_OUTOFMEM;
		  strcpy(tag->val, origtag->val);

		  found = MOPS_TRUE;
		}
	      tag = tag->next;
	    }

	  if(!found)
	    {
	      /* need to create a new tag object */
	      mops_copy_tag(origtag,&tag);
	      tag->next = o->tags;
	      o->tags = tag;
	    }

	}

      if(o->down)
	mops_status = mops_create_instance_ids(o->down);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}


int
mops_check_refcount(mops_object *o)
{
  
  while(o)
    {
      if(o->refcount)
	return MOPS_TRUE;

      if(o->down)
	if(mops_check_refcount(o->down))
	  return MOPS_TRUE;

      o = o->next;
    }

  return MOPS_FALSE;
}

int
mops_crtinstance_tcmd(Tcl_Interp * interp,
		      int argc, char *argv[], mops_object *o)
{
 mops_list_object *sel = mops_current_selection;
 mops_object *target = NULL;
 char fname[] = "crtinstance";

 if(!sel)
   {
     mops_error(interp,MOPS_NOSEL,fname,NULL);
     return TCL_ERROR;
   }

 if(!(target=sel->object))
   {
     mops_error(interp,MOPS_ERROR,fname,NULL);
     return TCL_ERROR;
   }

 if(target->type == MOPS_OTINSTANCE || target->type == MOPS_OTVIEW ||
    target->type == MOPS_OTROOT)
   {
     mops_error(interp,MOPS_ERROR,fname,
		"cannot create instance of this type of object");
     return TCL_ERROR;
   }

 target->refcount++;
 o->object = target;
 o->type = MOPS_OTINSTANCE;
 mops_copy_attributes(target,o);

 return TCL_OK;
}


int
mops_get_instance_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  Tcl_UnsetVar(interp,"propertyList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyGetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertySetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyDataList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);

  /* add transformation property */
  Tcl_SetVar(interp,"propertyList","Transformations", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildTransformationGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getTransfProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setTransfProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","transfPropData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}


int
mops_wrib_instance(mops_object *o, char *file)
{
 int mops_status = MOPS_OK;
 char *iafilename = NULL;
 double movx, movy, movz;
 double rotx, roty, rotz;
 double scalx, scaly, scalz;
 float quat[4];
 int found = MOPS_FALSE;
 mops_object *orig = NULL;
 mops_tag_object *tag = NULL;

 if(!mops_prefs.resolveinstances)
  {
    tag = o->tags;
    while(tag && !found)
      {
	if(tag->type == mops_oi_tagtype)
	  {
	    RiTransformBegin();
	    mops_status = mops_wrib_trafos(o);
	    if(!file)
	      RiReadArchive(tag->val,RI_NULL,RI_NULL);
	    else
	      {
		if(!(iafilename = calloc(1,
			    strlen(tag->val)+strlen(file)+2)))
		  return MOPS_OUTOFMEM;
		strcpy(iafilename,file);
		iafilename[strlen(file)] = '-';
		strcpy(&(iafilename[strlen(file)+1]),tag->val);
		RiReadArchive(iafilename,RI_NULL,RI_NULL);
		free(iafilename);

	      }

	    RiTransformEnd();
	    found = MOPS_TRUE;
	  }
	tag = tag->next;
      }

    if(!found)
      return MOPS_ERROR; /* This should never happen! */

    return MOPS_OK; /* XXXX early exit! */
  }

 orig = (mops_object *)o->object;

 if(!orig)
   return MOPS_ERROR;

 movx = orig->movx;
 movy = orig->movy;
 movz = orig->movz;

 rotx = orig->rotx;
 roty = orig->roty;
 rotz = orig->rotz;

 scalx = orig->scalx;
 scaly = orig->scaly;
 scalz = orig->scalz;

 quat[0] = orig->quat[0];
 quat[1] = orig->quat[1];
 quat[2] = orig->quat[2];
 quat[3] = orig->quat[3];

 orig->movx = o->movx;
 orig->movy = o->movy;
 orig->movz = o->movz;

 orig->rotx = o->rotx;
 orig->roty = o->roty;
 orig->rotz = o->rotz;

 orig->scalx = o->scalx;
 orig->scaly = o->scaly;
 orig->scalz = o->scalz;

 orig->quat[0] = o->quat[0];
 orig->quat[1] = o->quat[1];
 orig->quat[2] = o->quat[2];
 orig->quat[3] = o->quat[3];

 mops_status = mops_wrib_object(orig,file);

 if(orig->type == MOPS_OTLIGHT)
   mops_status = mops_wrib_lights(o, file, MOPS_FALSE);

 orig->movx = movx;
 orig->movy = movy;
 orig->movz = movz;

 orig->rotx = rotx;
 orig->roty = roty;
 orig->rotz = rotz;

 orig->scalx = scalx;
 orig->scaly = scaly;
 orig->scalz = scalz;

 orig->quat[0] = quat[0];
 orig->quat[1] = quat[1];
 orig->quat[2] = quat[2];
 orig->quat[3] = quat[3];

 return mops_status;
}


int
mops_resolve_instance(Tcl_Interp *interp, mops_object *i)
{
 int mops_status = MOPS_OK;
 char *iname = NULL;
 double movx, movy, movz;
 double rotx, roty, rotz;
 double scalx, scaly, scalz;
 float quat[4];
 mops_object *orig = NULL, *temp = NULL;
 mops_object *inext = NULL;

 movx = i->movx;
 movy = i->movy;
 movz = i->movz;

 rotx = i->rotx;
 roty = i->roty;
 rotz = i->rotz;

 scalx = i->scalx;
 scaly = i->scaly;
 scalz = i->scalz;

 quat[0] = i->quat[0];
 quat[1] = i->quat[1];
 quat[2] = i->quat[2];
 quat[3] = i->quat[3];

 if(i->type != MOPS_OTINSTANCE)
   return MOPS_ERROR;

 orig = (mops_object *)i->object;
 if(!orig)
   return MOPS_ERROR;

 inext = i->next;
 iname = i->name;

 mops_status = mops_copy_object(orig, &temp);
 if(mops_status)
   return mops_status;

 memcpy(i, temp, sizeof(mops_object));
 if(i->name) free(i->name);
 i->name = iname;

 i->next = inext;

 i->movx = movx;
 i->movy = movy;
 i->movz = movz;

 i->rotx = rotx;
 i->roty = roty;
 i->rotz = rotz;

 i->scalx = scalx;
 i->scaly = scaly;
 i->scalz = scalz;

 i->quat[0] = quat[0];
 i->quat[1] = quat[1];
 i->quat[2] = quat[2];
 i->quat[3] = quat[3];

 orig->refcount--;

 return mops_status;
}


int
mops_resolve_instance_tcmd(ClientData clientData, Tcl_Interp * interp,
			   int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 char fname[] = "resolveIn";

 if(!sel)
   {
     mops_error(interp,MOPS_NOSEL,fname,NULL);
     return TCL_OK;
   }

 if(!(o = sel->object))
   {
     mops_error(interp,MOPS_ERROR,fname,NULL);
     return TCL_OK;
   }

 if(!o->type == MOPS_OTINSTANCE)
   {
     mops_error(interp, MOPS_ERROR, fname, "Object is not of type Instance!");
     return TCL_OK;
   }

 mops_status = mops_resolve_instance(interp, o);

 if(mops_status)
   {
     mops_error(interp,mops_status,fname,NULL);
     return TCL_OK;
   }

 return TCL_OK;
}
