/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"
#include <limits.h>

/* Lights */

/********************************************/
/* mops_draw_light:
   draws a lightsource                      */
/********************************************/
int
mops_draw_light(struct Togl *togl, mops_object *o)
{
 mops_shader *shader = NULL;
 mops_shader_arg *sarg = NULL;
 mops_light_object *light = NULL;
 GLdouble fromx = 0.0, fromy = 0.0, fromz = 0.0;
 GLdouble tox = 0.0, toy = 0.0, toz = 0.0;
 GLdouble w;
 GLdouble mvm[16], pm[16], win1x, win1y, win1z, win2x, win2y, win2z;
 GLdouble p1x, p1y, p2x, p2y, alpha;
 GLint vp[4];
 double radius = 0.0;
 int has_from = 0, has_to = 0,i,a;
 double c1[24] = {0};

 w = (GLdouble)(sqrt(2.0)*0.5);

  if(!o)
    return MOPS_OK;

  light = (mops_light_object *)o->object;

  switch(light->type)
    {
    case MOPS_LITPOINT:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      break;
    case MOPS_LITSPOT:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      has_to = MOPS_TRUE;
      tox = 0.0;
      toy = 0.0;
      toz = -10.0;
      break;
    case MOPS_LITDIST:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      has_to = MOPS_TRUE;
      tox = 0.0;
      toy = 0.0;
      toz = -10.0;
      break;
    case MOPS_LITCUSTOM:
      shader = o->shaders;
      while(shader)
	{
	  if(shader->type == MOPS_STLIGHT)
	    {
	      sarg = shader->arg;
	      while(sarg)
		{
		  if(!strcmp(sarg->name,"from"))
		    {
		      has_from = MOPS_TRUE;
		      fromx = sarg->val.point[0];
		      fromy = sarg->val.point[1];
		      fromz = sarg->val.point[2];
		    }
		  if(!strcmp(sarg->name,"to"))
		    {
		      has_to = MOPS_TRUE;
		      tox = sarg->val.point[0];
		      toy = sarg->val.point[1];
		      toz = sarg->val.point[2];

		    }
		  sarg = sarg->next;	      
		}/* while */
	    } /* if */
	  shader = shader->next;
	} /* while */
      break;

    } /* switch */

  if(has_from)
    {
      glBegin(GL_LINES);
      glVertex3d(fromx-(1.0/o->scalx),fromy,fromz);
      glVertex3d(fromx+(1.0/o->scalx),fromy,fromz);
      glVertex3d(fromx,fromy-(1.0/o->scaly),fromz);
      glVertex3d(fromx,fromy+(1.0/o->scaly),fromz);
      glVertex3d(fromx,fromy,fromz-(1.0/o->scalz));
      glVertex3d(fromx,fromy,fromz+(1.0/o->scalz));
      glEnd();

      if(has_to)
	{
	  glBegin(GL_LINES);
	  glVertex3d(fromx,fromy,fromz);
	  glVertex3d(tox,toy,toz);
	  glEnd();


	  /* draw direction pointer */
	  glGetDoublev(GL_MODELVIEW_MATRIX, mvm);
	  glGetDoublev(GL_PROJECTION_MATRIX, pm);
	  glGetIntegerv(GL_VIEWPORT, vp);
	  
	  gluProject((GLdouble)tox, (GLdouble)toy,
		     (GLdouble)toz, mvm, pm, vp,
		     &win1x, &win1y, &win1z);


	  gluProject((GLdouble)fromx, (GLdouble)fromy,
		     (GLdouble)fromz, mvm, pm, vp,
		     &win2x, &win2y, &win2z);

	  p1x = win1x+MOPS_POINTER;
	  p1y = win1y+MOPS_POINTER;

	  p2x = win1x+MOPS_POINTER;
	  p2y = win1y-MOPS_POINTER;

	  if(win2x-win1x != 0.0)
	    {
	      alpha = MOPS_R2D(atan((win2y-win1y)/(win2x-win1x)));

	      if(win2x<win1x)
		alpha = 180+alpha;
	    }
	  else
	    if(win2y>win1y)
	      alpha = 90;
	    else
	      alpha = -90;

	  glMatrixMode(GL_PROJECTION);
	  glPushMatrix();
	   glLoadIdentity();
	   glOrtho(0, Togl_Width(togl), 0, Togl_Height(togl), -100.0, 100.0);

	   glMatrixMode(GL_MODELVIEW);
	   glPushMatrix();
	    glLoadIdentity();
	    glTranslated(win1x,win1y,0.0);
	    glRotated(alpha,0.0,0.0,1.0);
	    glTranslated(-win1x,-win1y,0.0);
	    glBegin(GL_LINE_STRIP);
	     glVertex3d(p1x,p1y,win1z);
	     glVertex3d(win1x,win1y,win1z);
	     glVertex3d(p2x,p2y,win1z);
	    glEnd();
	  
	   glPopMatrix();
	   glMatrixMode(GL_PROJECTION);
	  glPopMatrix();
	  glMatrixMode(GL_MODELVIEW);

	}
    }

  if(light->type == MOPS_LITSPOT)
    {
      radius = tan(light->cone_angle)*10.0;

      c1[1] = radius;
      c1[3] = w*radius;
      c1[4] = w*radius;
      c1[6] = radius;
      c1[9] = w*radius;
      c1[10] = -w*radius;
      c1[13] = -radius;
      c1[15] = -w*radius;
      c1[16] = -w*radius;
      c1[18] = -radius;
      c1[21] = -w*radius;
      c1[22] = w*radius;

      a = 0;
      glBegin(GL_LINES);
       for(i=0;i<4;i++)
	 {
	   glVertex3d(fromx,fromy,fromz);
	   glVertex3d(c1[a],c1[a+1],-10.0);
	   a+=6;
	 }
      glEnd();
      a = 0;
      glBegin(GL_LINE_LOOP);
       for(i=0;i<8;i++)
	 {
	   glVertex3d(c1[a],c1[a+1],-10.0);
	   a+=3;
	 }
      glEnd();
    }

 return MOPS_OK;
}


/********************************************/
/* mops_draw_light_h:
   draws only handles of a lightsource      */
/********************************************/
int
mops_draw_light_h(struct Togl *togl, mops_object *o)
{

 return MOPS_OK;
}


/********************************************/
/* mops_create_light:
   create a lightsource                     */
/********************************************/
int
mops_create_light(mops_light_object **lightptr)
{
 int mops_status = MOPS_OK;
 mops_light_object *newlight = NULL;

  if(!(newlight = calloc(1, sizeof(mops_light_object))))
    return MOPS_OUTOFMEM;

  newlight->type = MOPS_LITCUSTOM;  
  newlight->on = MOPS_TRUE;
  newlight->samples = 1;

  newlight->colr = 255;
  newlight->colg = 255;
  newlight->colb = 255;
  newlight->intensity = 1.0;

  newlight->cone_angle = MOPS_D2R(30.0);
  newlight->cone_delta_angle = MOPS_D2R(5.0);
  newlight->beam_distribution = 1.0;

  *lightptr = newlight;

 return mops_status;
}


/********************************************/
/* mops_free_light:
   free a lightsource                       */
/********************************************/
int
mops_free_light(mops_light_object *light)
{

  if(!light)
    return MOPS_OK;

  free(light);

 return MOPS_OK;
}


/********************************************/
/* mops_copy_light:
   copy a lightsource                       */
/********************************************/
int
mops_copy_light(mops_light_object *source,
		mops_light_object **dest)
{
 int mops_status = MOPS_OK;
 mops_light_object *newo = NULL;

  if(!(newo = calloc(1,sizeof(mops_light_object))))
    return MOPS_OUTOFMEM; 

  memcpy(newo, source, sizeof(mops_light_object)); 
  /* danger! links point to original hierachy */

  *dest = newo;

 return mops_status;
}


/********************************************/
/*mops_get_light_point:
  single point editing for lights of type:
  distant, point, spot                      */
/********************************************/
int
mops_get_light_point(mops_object *o, double objX, double objY,
		     double objZ)
{
 int mops_status = MOPS_OK;

  if(mops_point_edit_coords)
    free(mops_point_edit_coords);

  mops_point_edit_coords = NULL;

 return mops_status;
}


/********************************************/
/*mops_get_light_properties:
                                            */
/********************************************/
int
mops_get_light_properties(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_light_object *light = NULL;

  if(!o)
    return TCL_OK;

  light = (mops_light_object *)o->object;

  Tcl_UnsetVar(interp,"propertyList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyGetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertySetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyDataList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);

  /* add transformation property */
  Tcl_SetVar(interp,"propertyList","Transformations", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildTransformationGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getTransfProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setTransfProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","transfPropData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* Light-Shader */
  if(light->type == MOPS_LITCUSTOM)
    {
      Tcl_SetVar(interp,"propertyList","LightShader", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildlightShaderGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getSProp light lightShaderData",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setSProp light lightShaderData",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyDataList","lightShaderData",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }

  /* Light-Attrib */
  Tcl_SetVar(interp,"propertyList","LightAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  switch(light->type)
    {
    case MOPS_LITCUSTOM:
      Tcl_SetVar(interp,"propertyProcList","buildclightAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_LITPOINT:
      Tcl_SetVar(interp,"propertyProcList","buildplightAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_LITDIST:
      Tcl_SetVar(interp,"propertyProcList","builddlightAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_LITSPOT:
      Tcl_SetVar(interp,"propertyProcList","buildslightAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    }

  Tcl_SetVar(interp,"propertyGetProcList","getLAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setLAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","lattrPropData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);


 return mops_status;
}


/* C -> Tcl! */
int
mops_get_lattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="lattrPropData";
 mops_light_object *light = NULL;
 Tcl_Obj *to = NULL, *ton = NULL, *toa = NULL;

  if(!sel)
    return TCL_OK;

  o = sel->object; 
  light = (mops_light_object *) o->object;

  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("Type",-1);
  to = Tcl_NewIntObj(light->type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"IsOn",-1);
  to = Tcl_NewIntObj(light->on);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Shadows",-1);
  to = Tcl_NewIntObj(light->shadows);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Samples",-1);
  to = Tcl_NewIntObj(light->samples);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Color_R",-1);
  to = Tcl_NewIntObj(light->colr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_G",-1);
  to = Tcl_NewIntObj(light->colg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_B",-1);
  to = Tcl_NewIntObj(light->colb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_A",-1);
  to = Tcl_NewIntObj(light->cola);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Intensity",-1);
  to = Tcl_NewDoubleObj(light->intensity);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ConeAngle",-1);
  to = Tcl_NewDoubleObj(MOPS_R2D(light->cone_angle));
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ConeDAngle",-1);
  to = Tcl_NewDoubleObj(MOPS_R2D(light->cone_delta_angle));
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"BeamDistrib",-1);
  to = Tcl_NewDoubleObj(light->beam_distribution);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_lattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_light_object *light = NULL;
 double dtemp = 0.0;
 char *n1="lattrPropData";
 Tcl_Obj *to = NULL, *ton = NULL, *toa = NULL;

  if(!sel)
    return TCL_OK;

  o = sel->object; 
  light = (mops_light_object *) o->object;

  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("Type",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->type));

  Tcl_SetStringObj(ton,"IsOn",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->on));

  Tcl_SetStringObj(ton,"Shadows",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->shadows));

  Tcl_SetStringObj(ton,"Samples",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->samples));

  Tcl_SetStringObj(ton,"Intensity",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(light->intensity));

  Tcl_SetStringObj(ton,"Color_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->colr));
  o->colr = light->colr;
  Tcl_SetStringObj(ton,"Color_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->colg));
  o->colg = light->colg;
  Tcl_SetStringObj(ton,"Color_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->colb));
  o->colb = light->colb;
  Tcl_SetStringObj(ton,"Color_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(light->cola));
  o->cola = light->cola;

  Tcl_SetStringObj(ton,"ConeAngle",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(dtemp));

  light->cone_angle = MOPS_D2R(dtemp);

  Tcl_SetStringObj(ton,"ConeDAngle",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(dtemp));

  light->cone_delta_angle = MOPS_D2R(dtemp);

  Tcl_SetStringObj(ton,"BeamDistrib",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(light->beam_distribution));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/********************************************/
/*mops_crtlight_tcmd:
                                            */
/********************************************/
int
mops_crtlight_tcmd(Tcl_Interp * interp,
		   int argc, char *argv[], mops_object *o)
{
 int mops_status;
 char fname[] = "create_light";

  o->type = MOPS_OTLIGHT;

  /* we create the light-object */
  mops_status = mops_create_light((mops_light_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

 return TCL_OK;
}


int
mops_check_lights(mops_object *o)
{
 int  i = 0;
 mops_light_object *light = NULL;

  if(!o)
    return 0;

  while(o)
    {
      if(o->down)
	i = mops_check_lights(o->down);

      if(o->type == MOPS_OTLIGHT)
	{
	  light = (mops_light_object *)o->object;
	  if(light->on)
	    return 1;
	}
      o = o->next;
    }

 return i;
}

int
mops_wrib_light_instance(mops_object *o, char *file)
{
 int mops_status = MOPS_OK;
 char *iafilename = NULL;
 double movx, movy, movz;
 double rotx, roty, rotz;
 double scalx, scaly, scalz;
 float quat[4];
 int found = MOPS_FALSE;
 mops_object *orig = NULL;
 mops_tag_object *tag = NULL;

 orig = (mops_object *)o->object;

 if(!orig)
   return MOPS_ERROR;

 if(orig->type != MOPS_OTLIGHT)
   return MOPS_OK;

 if(!mops_prefs.resolveinstances)
  {
    tag = o->tags;
    while(tag && !found)
      {
	if(tag->type == mops_oi_tagtype)
	  {
	    RiTransformBegin();
	    mops_status = mops_wrib_trafos(o);
	    if(!file)
	      RiReadArchive(tag->val,RI_NULL,RI_NULL);
	    else
	      {
		if(!(iafilename = calloc(1,
			    strlen(tag->val)+strlen(file)+2)))
		  return MOPS_OUTOFMEM;
		strcpy(iafilename,file);
		iafilename[strlen(file)] = '-';
		strcpy(&(iafilename[strlen(file)+1]),tag->val);
		RiReadArchive(iafilename,RI_NULL,RI_NULL);
		free(iafilename);

	      }

	    RiTransformEnd();
	    found = MOPS_TRUE;
	  }
	tag = tag->next;
      }

    if(!found)
      return MOPS_ERROR; /* This should never happen! */

    return MOPS_OK; /* XXXX early exit! */
  }

 orig = (mops_object *)o->object;

 if(!orig)
   return MOPS_ERROR;

 movx = orig->movx;
 movy = orig->movy;
 movz = orig->movz;

 rotx = orig->rotx;
 roty = orig->roty;
 rotz = orig->rotz;

 scalx = orig->scalx;
 scaly = orig->scaly;
 scalz = orig->scalz;

 quat[0] = orig->quat[0];
 quat[1] = orig->quat[1];
 quat[2] = orig->quat[2];
 quat[3] = orig->quat[3];

 orig->movx = o->movx;
 orig->movy = o->movy;
 orig->movz = o->movz;

 orig->rotx = o->rotx;
 orig->roty = o->roty;
 orig->rotz = o->rotz;

 orig->scalx = o->scalx;
 orig->scaly = o->scaly;
 orig->scalz = o->scalz;

 orig->quat[0] = o->quat[0];
 orig->quat[1] = o->quat[1];
 orig->quat[2] = o->quat[2];
 orig->quat[3] = o->quat[3];

 mops_status = mops_wrib_object(orig,file);

 if(orig->type == MOPS_OTLIGHT)
   mops_status = mops_wrib_lights(o, file, MOPS_FALSE);

 orig->movx = movx;
 orig->movy = movy;
 orig->movz = movz;

 orig->rotx = rotx;
 orig->roty = roty;
 orig->rotz = rotz;

 orig->scalx = scalx;
 orig->scaly = scaly;
 orig->scalz = scalz;

 orig->quat[0] = quat[0];
 orig->quat[1] = quat[1];
 orig->quat[2] = quat[2];
 orig->quat[3] = quat[3];

 return mops_status;
}

int
mops_wrib_lights(mops_object *o, char *file, int wrib_archive)
{
 int mops_status = MOPS_OK;
 mops_light_object *light = NULL;
 char *onstr = "on", *offstr = "off";
 RtColor color = {0};
 RtPoint from, to;
 RtFloat intensity, coneangle, conedeltaangle, beamdistribution;
 RtFloat rim[4][4];

  if(!o)
    return mops_status;

  while(o)
    {
      if(o->type == MOPS_OTLEVEL)
	{
	  RiTransformBegin();
	  if(o->movx != 0.0 || o->movy != 0.0 || o->movz != 0.0)
	    RiTranslate((RtFloat)o->movx, (RtFloat)o->movy, (RtFloat)o->movz);

	  if(o->quat[0] != 0.0 || o->quat[1] != 0.0 || o->quat[2] != 0.0 ||
	     o->quat[3] != 1.0)
	    {
	      mops_build_rotmatrix(rim, o->quat);
	      RiConcatTransform(rim);
	    }

	  if(o->scalx != 1.0 || o->scaly != 1.0 || o->scalz != 1.0)
	    RiScale((RtFloat)o->scalx, (RtFloat)o->scaly, (RtFloat)o->scalz);

	  mops_wrib_lights(o->down, file, wrib_archive);
	  RiTransformEnd();
	}

      if(!wrib_archive)
	if(o->type == MOPS_OTINSTANCE)
	  mops_wrib_light_instance(o,file);

      if(o->type == MOPS_OTLIGHT)
	{
	  light = (mops_light_object *)o->object;

	  RiTransformBegin();
	  if(!wrib_archive)
	   {
	    if(o->movx != 0.0 || o->movy != 0.0 || o->movz != 0.0)
	     RiTranslate((RtFloat)o->movx, (RtFloat)o->movy, (RtFloat)o->movz);

	    if(o->quat[0] != 0.0 || o->quat[1] != 0.0 || o->quat[2] != 0.0 ||
	       o->quat[3] != 1.0)
	      {
		mops_build_rotmatrix(rim, o->quat);
		RiConcatTransform(rim);
	      }

	    if(o->scalx != 1.0 || o->scaly != 1.0 || o->scalz != 1.0)
	      RiScale((RtFloat)o->scalx, (RtFloat)o->scaly, (RtFloat)o->scalz);
	   }
	  from[0] = (RtFloat)0.0;
	  from[1] = (RtFloat)0.0;
	  from[2] = (RtFloat)0.0;

	  to[0] = (RtFloat)0.0;
	  to[1] = (RtFloat)0.0;
	  to[2] = (RtFloat)-10.0;

	  intensity = (RtFloat) light->intensity;
	  color[0] = (RtFloat)light->colr/255.0;
	  color[1] = (RtFloat)light->colg/255.0;
	  color[2] = (RtFloat)light->colb/255.0;

	  if(light->shadows)
	    RiAttribute("light","shadows",(RtPointer)&onstr, RI_NULL);
	  else
	    RiAttribute("light","shadows",(RtPointer)&offstr, RI_NULL);

	  switch(light->type)
	    {
	    case MOPS_LITCUSTOM:
	      if(o->down)
		{ /* this is an AreaLight */
		  RiAttributeBegin();
		}

	      if(light->samples > 1)
		RiAttribute("light","nsamples",
			    (RtPointer)(&light->samples), RI_NULL);

	      mops_status = mops_wrib_shaders(o);

	      if(o->down)
		{ /* this is an AreaLight */
		  mops_status = mops_wrib_object(o->down, file);
		  RiAttributeEnd();
		}
	      break;
	    case MOPS_LITPOINT:
	      RiTranslate((RtFloat)o->movx, (RtFloat)o->movy,
			  (RtFloat)o->movz);
	      RiDeclare("intensity","float");
	      RiDeclare("lightcolor","color");
	      RiDeclare("from","point");

	      mops_light_handle = RiLightSource("pointlight",
			"intensity",(RtPointer)&intensity,
			"lightcolor",(RtPointer)&color,
			"from",(RtPointer)&from, RI_NULL);
	      break;
	    case MOPS_LITSPOT:
	      RiDeclare("intensity","float");
	      RiDeclare("lightcolor","color");
	      RiDeclare("from","point");
	      RiDeclare("to","point");
	      RiDeclare("coneangle","float");
	      RiDeclare("conedeltaangle","float");
	      RiDeclare("beamdistribution","float");

	      coneangle = (RtFloat) light->cone_angle;
	      conedeltaangle = (RtFloat) light->cone_delta_angle;
	      beamdistribution = (RtFloat) light->beam_distribution;
	      mops_light_handle = RiLightSource("spotlight",
			"intensity",(RtPointer)&intensity,
			"lightcolor",(RtPointer)&color,
			"from",(RtPointer)&from, 
			"to",(RtPointer)&to, 
			"coneangle",(RtPointer)&coneangle,
			"conedeltaangle",(RtPointer)&conedeltaangle,
			"beamdistribution",(RtPointer)&beamdistribution,
			RI_NULL);
	      break;
	    case MOPS_LITDIST:
	      RiDeclare("intensity","float");
	      RiDeclare("lightcolor","color");
	      RiDeclare("from","point");
	      RiDeclare("to","point");

	      mops_light_handle = RiLightSource("distantlight",
			"intensity",(RtPointer)&intensity,
			"lightcolor",(RtPointer)&color,
			"from",(RtPointer)&from, 
			"to",(RtPointer)&to, RI_NULL);
	      break;
	      
	    } /* switch */

	  if(light->on)
	    RiIlluminate(mops_light_handle,RI_TRUE);
	  else
	    RiIlluminate(mops_light_handle,RI_FALSE);
	  RiTransformEnd();
	}
      o = o->next;
    }

  return mops_status;
}


