/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Miscellaneous */

int
mops_add_surfshader_property(Tcl_Interp *interp)
{
  /* Surface-Shader */
  Tcl_SetVar(interp,"propertyList","SurfShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildsurfaceShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getSProp surface surfaceShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setSProp surface surfaceShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","surfaceShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return MOPS_OK;
}


int
mops_add_dispshader_property(Tcl_Interp *interp)
{
  /* Displacement-Shader */
  Tcl_SetVar(interp,"propertyList","DispShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","builddisplacementShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList",
	     "getSProp displacement displacementShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList",
	     "setSProp displacement displacementShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","displacementShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return MOPS_OK;
}

int
mops_add_intshader_property(Tcl_Interp *interp)
{
  /* Interior-Shader */
  Tcl_SetVar(interp,"propertyList","IntShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildvolumeShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList",
	     "getSProp volume volumeShaderData interior",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList",
	     "setSProp volume volumeShaderData interior",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","volumeShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return MOPS_OK;
}

int
mops_add_extshader_property(Tcl_Interp *interp)
{
  /* Exterior-Shader */
  Tcl_SetVar(interp,"propertyList","ExtShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildvolumeShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList",
	     "getSProp volume volumeShaderData exterior",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList",
	     "setSProp volume volumeShaderData exterior",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","volumeShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return MOPS_OK;
}


/* mops_default_attributes:
   add default attribute settings to
   mops_object *o */
int
mops_default_attributes(mops_object *o)
{
  if(!o)
    return MOPS_OK;

  o->quat[3]=1.0;

  o->scalx = 1.0;
  o->scaly = 1.0;
  o->scalz = 1.0;

  o->colr = 0.8*255.0;
  o->colg = 0.75*255.0;
  o->colb = 0.65*255.0;
  o->cola = 255;

  o->opr = 255.0;
  o->opg = 255.0;
  o->opb = 255.0;
  o->opa = 255;

  o->shading_rate = 1.0;

  /* full visibility */
  o->editor = 1;
  o->camera = 1;
  o->reflection = 1;
  o->shadow = 1;

  /* no radiosity attributes: */
  o->avr = -1;
  o->emr = -1;
  o->spr = -1;

  o->patch_size = -1.0;
  o->elem_size = -1.0;
  o->min_size = -1.0;

 return MOPS_OK;
}

/* mops_copy_attributes:  
   mops_object *o */
int
mops_copy_attributes(mops_object *src, mops_object *dst)
{

  if((!src)||(!dst))
    return MOPS_OK;

  dst->movx = src->movx;
  dst->movy = src->movy;
  dst->movz = src->movz;

  dst->rotx = src->rotx;
  dst->roty = src->roty;
  dst->rotz = src->rotz;

  memcpy(dst->quat, src->quat, 4*sizeof(float));

  dst->scalx = src->scalx;
  dst->scaly = src->scaly;
  dst->scalz = src->scalz;

  dst->colr = src->colr;
  dst->colg = src->colg;
  dst->colb = src->colb;
  dst->cola = src->cola;

  dst->opr = src->opr;
  dst->opg = src->opg;
  dst->opb = src->opb;
  dst->opa = src->opa;

  dst->avr = src->avr;
  dst->avg = src->avg;
  dst->avb = src->avb;
  dst->ava = src->ava;

  dst->emr = src->emr;
  dst->emg = src->emg;
  dst->emb = src->emb;
  dst->ema = src->ema;

  dst->spr = src->spr;
  dst->spg = src->spg;
  dst->spb = src->spb;
  dst->spa = src->spa;

  dst->patch_size = src->patch_size;
  dst->elem_size = src->elem_size;
  dst->min_size = src->min_size;

  dst->zonal = src->zonal;
  dst->shading_rate = src->shading_rate;
  dst->cast_shadows = src->cast_shadows;

  dst->true_displacement = src->true_displacement;
  dst->d_bound = src->d_bound;
  dst->d_bound_val = src->d_bound_val;
  dst->has_caustics = src->has_caustics;

  dst->editor = src->editor;
  dst->camera = src->camera;
  dst->reflection = src->reflection;
  dst->shadow = src->shadow;

  /*  dst-> = src->;*/

 return MOPS_OK;
}

/* mops_copy_transform:  
   copy transformation attributes from src to dst */
int
mops_copy_transform(mops_object *src, mops_object *dst)
{

  if((!src)||(!dst))
    return MOPS_OK;

  dst->movx = src->movx;
  dst->movy = src->movy;
  dst->movz = src->movz;

  dst->rotx = src->rotx;
  dst->roty = src->roty;
  dst->rotz = src->rotz;

  dst->scalx = src->scalx;
  dst->scaly = src->scaly;
  dst->scalz = src->scalz;

  memcpy(dst->quat, src->quat, 4*sizeof(float));

 return MOPS_OK;
}


/* mops_get_version_tcmd:
   get all version information from the current OpenGL/GLU and Mops */
int
mops_get_version_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[])
{
  char *glver, *glven, *glren, *gluver, *gluext;

  glver = (char *)glGetString(GL_VERSION);
  Tcl_SetVar(interp,"gl_ver", glver, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

  glven = (char *)glGetString(GL_VENDOR);
  Tcl_SetVar(interp,"gl_ven", glven, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

  glren = (char *)glGetString(GL_RENDERER);
  Tcl_SetVar(interp,"gl_ren", glren, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

  gluver = (char *)gluGetString(GLU_VERSION);
  Tcl_SetVar(interp,"glu_ver", gluver, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

  gluext = (char *)gluGetString(GLU_EXTENSIONS);
  Tcl_SetVar(interp,"glu_ext", gluext, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

  Tcl_SetVar(interp,"mops_version", MOPS_VERSIONSTR, TCL_GLOBAL_ONLY |
	     TCL_LEAVE_ERR_MSG);

 return TCL_OK;
}



int
mops_register_crt(Tcl_Interp *interp, char *name, mops_createcb *func)
{
 int new_item;
 Tcl_HashEntry *entry;

 entry = Tcl_CreateHashEntry(&mops_createcb_ht, name, &new_item);
 Tcl_SetHashValue(entry, func);

 return TCL_OK;
}


int
mops_crtob_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *new = NULL;
 mops_createcb *crtfunc = NULL;
 int result = TCL_OK;
 Tcl_HashEntry *entry;
 char fname[] = "create_object";

  /* parse args */
  if(argc<2)
    {
      mops_error(interp, MOPS_ARGS, fname, "type [typeparameters]");
      return TCL_OK;
    }

  entry = Tcl_FindHashEntry(&mops_createcb_ht, argv[1]);
  if (entry != NULL) {
    crtfunc = (mops_createcb *)Tcl_GetHashValue(entry);
  } else {
    mops_error(interp, MOPS_HASH, fname, NULL);
    return TCL_OK;
  }

  if(!(new = calloc(1,sizeof(mops_object))))
    {
      mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
      return TCL_OK;
    }

  mops_status = mops_default_attributes(new);

  if(crtfunc)
    if((result = crtfunc(interp, argc, argv, new)) != TCL_OK)
      {
	mops_error(interp,MOPS_ERROR,fname,"create function failed");
	free(new);
	return TCL_OK;
      }

  mops_status = mops_link_new_object(new);
  if(mops_status)
    {
      mops_delete_object(new);
      mops_error(interp, mops_status, fname, NULL);
      return TCL_OK;
    }

  mops_notify_parent();

 return TCL_OK;
}


void
mops_get_all_transform(mops_list_object *lo)
{
 mops_object *o = NULL;
 GLfloat m[4][4];

  if(!lo)
    return;

  o = lo->object;

  if(o->type == MOPS_OTNURBPATCH)
    return;

  if(lo->next)
   {
     mops_get_all_transform(lo->next->next);
   }

  if((o->type != MOPS_OTROOT) && o->down)
    {
      glTranslated(o->movx, o->movy, o->movz);
      mops_build_rotmatrix(m, o->quat);
      glMultMatrixf(&m[0][0]);
      glScaled (o->scalx, o->scaly, o->scalz);
    }

 return;
}


void
mops_get_all_itransform(mops_list_object *lo)
{
  mops_object *o = NULL;
  float euler[3] = {0};

  if(!lo)
    return;

  o = lo->object;

  if(o->type == MOPS_OTNURBPATCH)
    return;


  if((o->type != MOPS_OTROOT) && o->down)
    {
      glScaled (1/o->scalx, 1/o->scaly, 1/o->scalz);

      mops_quat_to_euler(o->quat,euler);
      glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
      glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
      glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);

      glTranslated(-o->movx, -o->movy, -o->movz);
    }

  if(lo->next)
   {
     mops_get_all_itransform(lo->next->next);
   }

 return;
}

void
mops_get_all_isrtransform(mops_list_object *lo)
{
  mops_object *o = NULL;
  float euler[3] = {0};

  if(!lo)
    return;

  o = lo->object;

  if(o->type == MOPS_OTNURBPATCH)
    return;

  if((o->type != MOPS_OTROOT) && o->down)
    {
      glScaled (1/o->scalx, 1/o->scaly, 1/o->scalz);

      mops_quat_to_euler(o->quat,euler);

      glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
      glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
      glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);

    }

  if(lo->next)
   {
     mops_get_all_isrtransform(lo->next->next);
   }

 return;
}


void
mops_get_all_srtransform(mops_list_object *lo)
{
  mops_object *o = NULL;
  float euler[3] = {0};

  if(!lo)
    return;

  o = lo->object;

  if(o->type == MOPS_OTNURBPATCH)
    return;

  if(lo->next)
   {
     mops_get_all_srtransform(lo->next->next);
   }

  if((o->type != MOPS_OTROOT) && o->down)
    {
      glScaled (o->scalx, o->scaly, o->scalz);
      mops_quat_to_euler(o->quat,euler);
      glRotatef(MOPS_R2D(-euler[2]), 1.0, 0.0, 0.0);
      glRotatef(MOPS_R2D(-euler[1]), 0.0, 1.0, 0.0);
      glRotatef(MOPS_R2D(-euler[0]), 0.0, 0.0, 1.0);
    }

 return;
}


void
mops_get_all_irtransform(mops_list_object *lo)
{
  mops_object *o = NULL;
  float euler[3] = {0};

  if(!lo)
    return;

  o = lo->object;

  if(o->type == MOPS_OTNURBPATCH)
    return;


  if((o->type != MOPS_OTROOT) && o->down)
    {
      mops_quat_to_euler(o->quat,euler);
      glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
      glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
      glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);
    }

  if(lo->next)
   {
     mops_get_all_irtransform(lo->next->next);
   }

 return;
}


void
mops_apply_transform3(double *c, double *m)
{
  double result[3] = {0};

  result[0] = m[0]*c[0] + m[4]*c[1] + m[8]*c[2]  + m[12]*1.0;
  result[1] = m[1]*c[0] + m[5]*c[1] + m[9]*c[2]  + m[13]*1.0;
  result[2] = m[2]*c[0] + m[6]*c[1] + m[10]*c[2] + m[14]*1.0;

  memcpy(c, result, (size_t)(3*sizeof(double)));

 return;
}

void
mops_apply_transform4(double *c, double *m)
{
  double result[4] = {0};

  result[0] = m[0]*c[0] + m[4]*c[1]  + m[8]*c[2]  + m[12]*c[3];
  result[1] = m[1]*c[0] + m[5]*c[1]  + m[9]*c[2]  + m[13]*c[3];
  result[2] = m[2]*c[0] + m[6]*c[1]  + m[10]*c[2] + m[14]*c[3];
  result[3] = m[3]*c[0] + m[7]*c[1]  + m[11]*c[2] + m[15]*c[3];

  memcpy(c, result, (size_t)(4*sizeof(double)));

 return;
}

int
mops_gettmp_tcmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char *argv[])
{
 char *tmpname = NULL;
 Tcl_Obj *to = NULL, *ton = NULL;
 char fname[] = "get_tmpname";
 
  if(argc<3)
    {
      mops_error(interp, MOPS_ARGS, fname, "tmpdir varname");
      return TCL_OK;
    }

#ifdef MOPSHAVETEMPNAM
  if(!(tmpname = tempnam(argv[1],"mops")))
     {
       mops_error(interp,MOPS_ERROR,fname,"cannot create tempname");
       return TCL_OK;
     }
#else
  if(!(tmpname = tmpnam(argv[1])))
     {
       mops_error(interp,MOPS_ERROR,fname,"cannot create tempname");
       return TCL_OK;
     }
#endif

  ton = Tcl_NewStringObj(argv[2],-1);
  to = Tcl_NewStringObj(tmpname,-1);
  
  Tcl_ObjSetVar2(interp,ton,NULL,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  free(tmpname);

 return TCL_OK;
}


int
mops_notify_parent(void)
{
 int mops_status = MOPS_OK;
 mops_list_object *lev = mops_current_level;
 mops_object *o = NULL;

 while(lev)
  {
    if(lev->next)
      if(lev->next->object)
	o = lev->next->object;

    if(o)
      {
	switch (o->type)
	  {
	  case MOPS_OTCUSTOM:
	    mops_status = mops_notify_custom(o);
	    break;
	  default:
	    break;
	  }
	
      }

    lev = lev->next;
    if(lev)
      lev = lev->next;
  }

 return mops_status;
}


int
mops_gettype_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char *name = NULL;
 mops_list_object *sel = mops_current_selection;
 char fname[] = "getType";

 if(!sel)
   return TCL_OK;

 if(!sel->object)
   return TCL_OK;

 if(argc<2)
   {
     mops_error(interp,MOPS_ARGS,fname,"varname");
     return TCL_OK;
   }

  mops_status = mops_get_objectname(sel->object, MOPS_TRUE, &name);
  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_OK;
    }

  Tcl_SetVar(interp,argv[1], name, TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  free(name);

 return TCL_OK;
}

int
mops_haschild_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;

 if(!sel)
   return TCL_OK;

 if(!sel->object)
   return TCL_OK;

 if(sel->object->down)
   Tcl_SetResult(interp,"1",TCL_VOLATILE);
 else
   Tcl_SetResult(interp,"0",TCL_VOLATILE);

 return TCL_OK;
}



int
mops_delegtran_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *parent=NULL, *child=NULL;
 double scalx,scaly,scalz,movx,movy,movz;
 float euler[3] = {0};
 float quat[4];
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};

 if(!sel)
   return TCL_OK;

 while(sel)
   {
     if(!sel->object)
       return TCL_OK;

     parent = sel->object;

     if(!sel->object->down)
       return TCL_OK;

     movx = parent->movx;
     movy = parent->movy;
     movz = parent->movz;

     scalx = parent->scalx;
     scaly = parent->scaly;
     scalz = parent->scalz;

     mops_quat_to_euler(parent->quat,euler);

     child = sel->object->down;
     while(child)
       {
	 child->movx += movx;
	 child->movy += movy;
	 child->movz += movz;

	 child->scalx *= scalx;
	 child->scaly *= scaly;
	 child->scalz *= scalz;

	 child->rotx -= MOPS_R2D(euler[2]);
	 child->roty -= MOPS_R2D(euler[1]);
	 child->rotz -= MOPS_R2D(euler[0]);

	 mops_axis_to_quat(xaxis,euler[2],quat);
	 mops_add_quat(quat,child->quat,child->quat);	
	 mops_axis_to_quat(yaxis,euler[1],quat);
	 mops_add_quat(quat,child->quat,child->quat);	
	 mops_axis_to_quat(zaxis,euler[0],quat);
	 mops_add_quat(quat,child->quat,child->quat);	 

	 child = child->next;
       }

     parent->movx=0.0;
     parent->movy=0.0;
     parent->movz=0.0;

     parent->scalx=1.0;
     parent->scaly=1.0;
     parent->scalz=1.0;

     parent->rotx = 0.0;
     parent->roty = 0.0;
     parent->rotz = 0.0;

     parent->quat[0]=0.0;
     parent->quat[1]=0.0;
     parent->quat[2]=0.0;
     parent->quat[3]=1.0;

     sel = sel->next;
   }

 return TCL_OK;
}



void
mops_clear_point_selection(mops_object *o)
{
  mops_point_object *next = NULL, *p = NULL;

  if(!o)
    return;

  p = o->selp;

  while(p)
    {
      next = p->next;
      free(p);
      p = next;
    }

  o->selp = NULL;

 return;
}


int
mops_show_tcmd(ClientData clientData, Tcl_Interp * interp,
	       int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;

 if(!sel)
   return TCL_OK;

 while(sel)
   {
     if(!sel->object)
       return TCL_OK;
     o = sel->object;

     o->editor = MOPS_TRUE;

     sel = sel->next;
   }

 return TCL_OK;
}


int
mops_hide_tcmd(ClientData clientData, Tcl_Interp * interp,
	       int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;

 if(!sel)
   return TCL_OK;

 while(sel)
   {
     if(!sel->object)
       return TCL_OK;
     o = sel->object;

     o->editor = MOPS_FALSE;

     sel = sel->next;
   }

 return TCL_OK;
}


int
mops_convert_object_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char *type = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 Tcl_HashEntry *entry = NULL;
 mops_converter_cb *cb;
 char fname[] = "convert_object";

 if(!sel)
   {
     mops_error(interp, MOPS_NOSEL, fname, NULL);
     return TCL_OK;
   }

 while(sel)
   {
     o = sel->object;

     type = NULL;
     mops_status = mops_get_objectname(o, MOPS_TRUE, &type);
     if(mops_status || (!type))
       {
	 mops_error(interp, MOPS_ERROR, fname, NULL);
	 return TCL_OK;
       }
   
     if(!(entry = Tcl_FindHashEntry(&mops_converter_ht, type)))
       {
	 mops_error(interp, MOPS_ERROR, fname,
		    "no converter registered for this type");
	 free(type);
	 return TCL_OK;
       }
     
     if(entry)
       cb = (mops_converter_cb *)Tcl_GetHashValue(entry);

     if(cb)
       {
	 mops_status = cb(interp, o);
	 if(mops_status)
	   {
	     mops_error(interp, MOPS_ERROR, fname, "converter failed");
	     free(type); return TCL_OK;
	   }
       }

     free(type);

     sel = sel->next;
   }
 return TCL_OK;
}


int
mops_register_converter(Tcl_Interp *interp, char *type, mops_converter_cb *cb)
{
 int mops_status = MOPS_OK;
 int new_item = 0;
 Tcl_HashEntry *entry = NULL;
 char  fname[] = "register_converter";

  /* check, if converter for type is already registered */
  if((entry = Tcl_FindHashEntry(&mops_converter_ht, type)))
    {
      mops_error(interp, MOPS_ERROR, fname, "converter already registered");
      return TCL_OK;
    }

  entry = Tcl_CreateHashEntry(&mops_converter_ht, type, &new_item);
  Tcl_SetHashValue(entry, cb);

 return mops_status;
}


int
mops_snap_togrid(struct Togl *togl, double *winx, double *winy)
{
 mops_view_info *view = Togl_GetClientData(togl);
 int width = Togl_Width (togl);
 int height = Togl_Height (togl);
 double refx = 0.0, refy = 0.0;
 double gdx = 0.0, gdy = 0.0;
 double gridx = view->grid/view->conv_x,
   gridy = view->grid/view->conv_y, gh = view->grid/view->conv_x/2.0;

  switch (view->type)
    {
    case MOPS_VTFRONT:
    case MOPS_VTTRIM:
      refx = (width/2.0)-(view->from_x/view->conv_x);
      refy = (height/2.0)+(view->from_y/view->conv_y);
      break;
    case MOPS_VTSIDE:
      refx = (width/2.0)-(view->from_z/view->conv_x);
      refy = (height/2.0)+(view->from_y/view->conv_y);
      break;
    case MOPS_VTTOP:
      refx = (width/2.0)-(view->from_x/view->conv_x);
      refy = (height/2.0)-(view->from_z/view->conv_y);
      break;
    }
  gdx = fmod(*winx-refx, gridx);
  if(*winx>refx)
    {
      if (gdx >= gh)
	*winx += (gridx-gdx);
      else
	*winx -= gdx;
    }
  else
    {
      if (gdx <= -gh)
	*winx -= (gridx+gdx);
      else
	*winx -= gdx;
    }

  gdy = fmod(*winy-refy, gridy);

  if(*winy>refy)
    {
      if (gdy >= gh)
	*winy += (gridy-gdy);
      else
	*winy -= gdy;
    }
  else
    {
      if (gdy <= -gh)
	*winy -= (gridy+gdy);
      else
	*winy -= gdy;
    }


 return MOPS_OK;
}


int
mops_get_object(mops_object *o, unsigned int type, mops_object **dst)
{
 int mops_status = MOPS_OK;
 mops_object *orig = NULL;
 mops_object *resolved = NULL;
 mops_object *new = NULL;
 mops_custom_object *custom = NULL;

 /* is it an instance? */
 if(o->type == MOPS_OTINSTANCE)
   {
     orig = (mops_object *)o->object;

     /* is it an instance of an object of the correct type? */
     if(orig->type == type)
       {
	 /* yes, we just have to copy the instance and resolve it */
	 mops_status = mops_copy_object(o, &new);
	 mops_status = mops_resolve_instance(NULL, new);
	 *dst = new;
       }
     else
       {
	 /* is this original of the instance perhaps providing what we want? */
	 if(orig->type == MOPS_OTCUSTOM)
	   {
	     custom = (mops_custom_object *)orig->object;
	   }
	 else
	   {
	     *dst = NULL;
	     return mops_status; /* XXXX early exit */
	   }
	 if(custom->provides == type)
	   {
	     /* it does, fine */
	     mops_status = mops_copy_object(o, &resolved);
	     /* we need to resolve here, to get the transformation
		attributes from the right object */
	     mops_status = mops_resolve_instance(NULL, resolved);
	     mops_status = mops_provide_object(resolved, (void*)&new);
	     /* provided objects are private memory of the
		providing object! */
	     mops_status = mops_copy_object(new, dst);
	     mops_status = mops_delete_object(resolved);
	   }
	 else
	   {
	     /* no, it does not */
	     *dst = NULL;
	   }
       }
   }
 else
   {/* it is not an instance, but perhaps providing what we want */
     if(o->type == MOPS_OTCUSTOM)
       {
	 custom = (mops_custom_object *)o->object;
       }
     else
       {
	 *dst = NULL;
	 return mops_status; /* XXXX early exit */
       }

     if(custom->provides == type)
       {
	 /* it does, fine */
	 mops_status = mops_provide_object(o, (void*)&new);
	 /* provided objects are private memory of the
	    providing object! */
	 mops_status = mops_copy_object(new, dst);
       }
     else
       {
	 /* no, it does not */
	 *dst = NULL;
       }
   }

 return mops_status;
}


int
mops_force_notification(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_object *od = NULL;

  od = o->down;
  while(od)
    {
      mops_status = mops_force_notification(od);
      od = od->next;
    }

  /* call the notification callback */
  /* luckily, no core type supports notification */
  if(o->type == MOPS_OTCUSTOM)
    mops_status = mops_notify_custom(o);

  /* XXXX ignore mops_status? */

 return MOPS_OK;
}


int
mops_forcenot_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;

  if(!sel)
    {
      o = mops_root->next;
      while(o)
	{
	  mops_status = mops_force_notification(o);
	  o = o->next;
	}
    }
  else
    {
      while(sel)
	{
	  mops_status = mops_force_notification(sel->object);
	  sel = sel->next;
	}

      mops_notify_parent();

    }

 return TCL_OK;
}



