#
# The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
#
# The Mops is copyrighted (C) 1998-2000 by Randolf Schultz
# (rschultz@informatik.uni-rostock.de)
#
# See the file License for details.
#
# mopstcl.tcl [15. Feb 2000]
#
# The main Mops Tcl code

# Some global Variables and Arrays
set filename ""
set curlevel ""
set mops_tempfiles ""
set mops_views ""
set currentView ""
set prop_sema 0
set oview_sema 0
set view1Geomvalid 0
# Preferences
array set prefs {
    BorderWidth 0
    AutoResize 1
    PickEpsilon 0.2
    UpdateTrafo 1
    SaveViews 1
    ReadRiOpt 1
    OnError 0
    tmpdir "/tmp"
    RIBFile "Scene"
    Image "RIB"
    ResInstances 0
    CheckLights 1
    QRender "rgl -rd 4 %s"
    Render "rendrib -d 4 %s"
    DailyTips {
{Always click on drawn pixels, when picking vertices.}
{There is a console-menu on <Mouse-3> (the right mouse button).}        
{Use <Shift+TAB> to move the input focus away from the console.}
{Documentation is available in the doc subdirectory of the distribution.}
{You may drag the borderline above the console to adjust it's height.}
{Picking does not work right? Adjust PickEpsilon in Scene-Prefs!}
{The NURB circles and spheres are not to be tampered with!}
{New! Documentation is available in the doc subdirectory!}
{Tired of that "ever-resizing" main window? See the Scene-Prefs.}
{Copy a single parameter of a property by doubleclicking on its name,
then press <Ctrl+Shift+i> (Copy Marked Prop).}
{All views appear as childs of the Scene-object!}
{Rather want a traditional 3(4)-view interface? Simply save a
scene containing just your views and load that on startup!}
{Have I mentioned the documentation lately?}
}
    cm ""
    pasteProp 0
    thisPropGUIStatic 0
    Shaders ""
    KeepNTmpFiles 5
    idrGeom ""
    toolBoxGeom ""
    mainGeom ""
    view1Geom ""
    zoomWheel 1
    RIStandard 0
    MOPSLIB "/usr/local/lib/mops"
    balloon 1500
    LazyNotify 0
    toolBoxList {transforms solids crnurb points curves curves2 custom camera misc}
    mopsdoc ""
    docdirs {}
    docview {netscape -remote openFile($mopsdoc)}
    docview2 {netscape $mopsdoc &}
    scripts {}
    twmcompat 0
}

# g lobal p latform dependant settings
array set mops_gp {
 mopsrc "~/.mopsrc"
 separator ":"
}

if { $tcl_platform(platform) == "windows" } {
    set mops_gp(separator) ";"
    set mops_gp(mopsrc) "~/mopsrc"
    set prefs(tmpdir) "$env(TEMP)"
    set prefs(PickEpsilon) 0.0
}

# if envvar MOPSRC is set, use it
if { [string length [array names env MOPSRC]] != 0 } {
    set mops_gp(mopsrc) $env(MOPSRC)
} else {
    if { $tcl_platform(platform) == "windows" } {
# if envvar HOME is not set (Win95?), use TEMP
	if { [string length [array names env HOME]] == 0 } {
	    set mops_gp(mopsrc) "$env(TEMP)/mopsrc"
	}
    }
}

# standard shortcuts for main window
array set mops_main_sc {
    Quit "Ctrl-q"
    New  "Ctrl-N"
    Replace "Ctrl-l"
    Insert "Ctrl-i"
    SaveAs "Ctrl-S"
    Save "Ctrl-s"
    ExportRIB "Ctrl-E"

    Apply "Ctrl-A"

    Copy "Ctrl-c"
    Cut "Ctrl-x"
    Paste "Ctrl-v"

    CopyP "Ctrl-C"
    ICopyP "Ctrl-I"
    PasteP "Ctrl-V"

    Name "Ctrl-n"
    Swap "Ctrl-w"
    Undo "Ctrl-u"

    Help "F1"
    Hide "Ctrl-z"
    SwCon "Shift-Tab"
}

# standard shortcuts for a view window
array set mops_view_sc {
    Close "Ctrl-w"
    QRender "Ctrl-r"
    Render "Ctrl-R"
    Redraw "Ctrl-d"
    Undo "Ctrl-u"

    Front "Ctrl-f"
    Side "Ctrl-s"
    Top "Ctrl-t"
    Persp "Ctrl-p"
    Trim "Ctrl-T"
    Halve "Ctrl-H"
    Double "Ctrl-D"
    
    RotL "Left"
    RotR "Right"
    RotU "Up"
    RotD "Down"

    ZoomI "KP_Add"
    ZoomO "KP_Subtract"
    
    Wire "Ctrl-W"
    Surf "Ctrl-S"
    Grid "Ctrl-G"
    SGrid "Ctrl-g"
    Light "Ctrl-L"
    Auto "Ctrl-A"

    ZoomTO "Ctrl-o"
}
# standard shortcuts for object actions
array set mops_oa_sc {
    MoveO "m"
    RotO "r"
    RotA "a"
    ScalX "x"
    ScalY "y"
    ScalZ "z"
    Scal2 "s"
    Scal3 "S"
    Stretch "Alt-s"

    Edit "e"
    EditD "p"
    WeightE "w"
    Select "t"
    DeSelect "D"
    InsertP "i"
    DelP "d"
    SplitC "c"
    FindU "u"

    ScalXW "X"
    ScalYW "Y"
    ScalZW "Z"
}
# standard shortcuts for view actions
array set mops_va_sc {
    MoveV "v"
    RotV "R"
    ZoomV "o"
    Break "Escape"
    MoveVButton "3"
    ZoomVButton "2"
}

array set addViewBindings {

}

array set editPointDarray {
    x 0.0
    y 0.0
    z 0.0
    w 0.0
}

# Shaders
set shaderTypeList [list surface displacement imager light volume transformation]
set newshaderindex ""

foreach i $shaderTypeList {
    eval "array set ${i}ShaderData \{ dummy 0 \}"
    eval "array set ${i}shaders \{ shadernames \"\" shaderpaths \"\" \}"
}

# Version information
set gl_ext ""
set gl_ver ""
set gl_ren ""
set gl_ven ""
set glu_ver ""
set glu_ext ""

# Properties
set stdpropertyList [list Transformations Attributes Tags]
set propertyList ""

set stdpropertyProcList [list buildTransformationGUI buildAttributesGUI buildTagsGUI]
set propertyProcList ""


set stdpropertyGetProcList [list getTransfProp getAttrProp getTagsProp]
set propertyGetProcList ""


set stdpropertySetProcList [list setTransfProp setAttrProp setTagsProp]
set propertySetProcList ""

set stdpropertyDataList [list transfPropData attrPropData tagPropData]
set propertyDataList ""

array set resetProp { dummy 0 }
array set copyProp { dummy 0 }
array set copyPropOmit { dummy 0 }

# keep last entered value in:
set cbspnum 4
set nclen 4
set npwidth 4
set npheight 4
set splitu 0.5
set elevd 1
set insknu 0.5
set insknr 1

proc uMS {} {updateMopsSelect}
proc sL {} {selectLast}
proc rAV {} {redrawAllViews}

if { $tcl_platform(platform) == "windows" } {
image create photo emptyimg
emptyimg blank
emptyimg configure -width 6 -height 6
}

# Procedures

# Shaders

proc scanAllShaders {} {
    global env prefs shaderTypeList MOPSNOSLCARGS mops_error

    puts stdout "Scanning for shaders..."

    if { $MOPSNOSLCARGS == 1 } {
	puts stdout "libslcargs not available..."
	return;
    }

    foreach i [list surface displacement imager light volume transformation] {
	eval global ${i}shaders
	eval "set ${i}shaders\(shadernames\) \"\" "
	eval "set ${i}shaders\(shaderpaths\) \"\" "
    }

    set temp ""
    if { $prefs(Shaders) == "" } {
	puts stderr "Shaders are not configured."
	puts stderr "Trying to load initial configuration from env(SHADERS)."
	if {[string length [array names env SHADERS]] == 0} then {
	    puts stderr "Environment variable SHADERS not set."
	    puts stderr "Please configure Shaders in Scene/Prefs."
	    return;
	} else {
	    set prefs(Shaders) $env(SHADERS)
	}
    } else { set env(SHADERS) $prefs(Shaders) }

    set spathstr $prefs(Shaders)

    global mops_gp
    regsub -all $mops_gp(separator) $spathstr " " spathstr


    foreach p $spathstr {
	    append temp "[glob -nocomplain $p/*.slc] "
	}

    set allshaders ""
    foreach s $temp {
	# silently omit unreadable shaders
	if {[file readable $s]} {
	    lappend allshaders $s
	}
    }

    foreach s $allshaders {
	regsub -all "/|\.slc" $s " " dummy
	set dummy [lindex $dummy end]
	# XXXX shader-file-names may not contain more than 2 "."s!
	set dummy [file rootname $dummy]

	set shaderarguments ""
	
	set mops_error 0
	scanShader $dummy shaderarguments
	if { $mops_error } {
	   break;
	}
	set shadertype [lindex $shaderarguments 1]

	set shadernamelistname ${shadertype}shaders(shadernames)
	eval set shadernamelist \$$shadernamelistname
	set shaderpathlistname ${shadertype}shaders(shaderpaths)

	if {[string first " $dummy " $shadernamelist ] == -1} {
	    lappend $shadernamelistname $dummy
	    lappend $shaderpathlistname $s
	}

    }

    # sort all lists
    foreach i [list surface displacement imager light volume transformation] {
	set shadernamelistname ${i}shaders(shadernames)
	eval set list \$$shadernamelistname
	set $shadernamelistname [lsort $list]
    }
    # luniq (remove multiple entries)
    foreach i [list surface displacement imager light volume transformation] {
	set shadernamelistname ${i}shaders(shadernames)
	eval set list \$$shadernamelistname
	set j [llength $list]
	set k 0
	set $shadernamelistname ""
	while { $j > 0 } {
	    if { [lindex $list $k] != [lindex $list [expr $k + 1]] } {
		lappend $shadernamelistname [lindex $list $k]
	    }
	    incr j -1
	    incr k
	}
    }

    set n [llength $allshaders]
    if { $n > 0} {
	set out [format "%d shaders found." $n]
	puts stdout $out
    } else {
	puts stderr "No shaders found."
	puts stderr "Please check the setting of Scene/Prefs/Shaders."
    }

return;
}


#############################################################
# this one is tricky, we let the user choose a shader of type 
# type, and then we build a new buildShaderGUI-proc
# and invoke it
proc selectShader { win type } {
upvar #0 ${type}ShaderData sArgArray

global env prefs ${type}shaders MOPSNOSLCARGS mops_error

if { $MOPSNOSLCARGS == 1 } {
    puts stdout "libslcargs not available..."
    return;
}

eval "set shaders \$${type}shaders(shadernames)"
eval "set shaderpaths \$${type}shaders(shaderpaths)"

# request a new shader
set w .setShaderw
catch {destroy $w}
toplevel $w
wm title $w "Mops - Set Shader"
wm iconname $w "Mops"
wm transient $w .

set f [frame $w.f1]
listbox $f.lb -width 20 -height 10 -selectmode single\
	-exportselection 0 -yscrollcommand {.setShaderw.f1.fsc.sc set}
eval [subst "$f.lb insert 0 $shaders"]
pack $f.lb -in $f -side left -fill both -expand yes
set f [frame $f.fsc]
scrollbar $f.sc -command {.setShaderw.f1.lb yview} -takefocus 0
pack $f.sc -in $f -side right -fill y -expand yes
set f $w.f1
pack $f.fsc -in $f -side right -fill y
pack $f -in $w -side top -fill both -expand yes

set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command {
    global newshaderindex
    set newshaderindex [.setShaderw.f1.lb curselection]
    grab release .setShaderw
    focus .
    destroy .setShaderw
}

button $f.bca -text "Cancel" -width 5 -command "\
    global newshaderindex;\
    set newshaderindex \"\";\
    destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x

centerWindowp $w
grab $w
focus $f.bok
tkwait window $w

# now we have the new shader
global newshaderindex
if {$newshaderindex == ""} {return;}

set shadername [lindex $shaders $newshaderindex]

set shaderarguments ""
set mops_error 0
scanShader $shadername shaderarguments
if { $mops_error } {
    puts stderr "Oops, could not scan shader!"
    break;
}

# paranoia?
if { [lindex $shaderarguments 1] != $type } { 
    if { [tk_messageBox -icon warning -type okcancel -message "An error\
	    occured, indicating that the database of shaders (built at\
	    program startup) might have become invalid. Select \"Ok\"\
	    to rebuild database or \"Cancel\" to stop the current\
	    operation."] == "ok" } { 
	scanAllShaders; selectShader $w $type; return; } else { return; }
    }

# we can now build the new proc
# the first thing is always the select-another-shader-button
# or we are stuck in the new GUI afterwards
set newGUIproc "proc build${type}ShaderGUI { w } {"
append newGUIproc "global ${type}ShaderData;"
append newGUIproc "global prefs;"
append newGUIproc "set prefs(thisPropGUIStatic) 1;"
append newGUIproc "addCommandEntry \$w c1 \"Set new shader.\" \"selectShader \$w $type\"; "

append newGUIproc "addCommandEntry \$w c2 \"Delete shader.\" \"delSelectedShader\";"

# current shadername
append newGUIproc "addTextEntry \$w e1 $shadername; "

catch {unset $sArgArray}
set sArgArray(Name) $shadername
# set sArgArray(File) $shaderpath
set sArgArray(ArgNames) ""
set sArgArray(ArgTypes) ""

set i [lindex $shaderarguments 2]

set j 3
while { $i > 0 } {

    set argname [lindex $shaderarguments $j]
    incr j
    set argtype [lindex $shaderarguments $j]
    incr j
    # XXXX jump over arrays!
    if { [lindex $shaderarguments $j] == 0 } {

    incr j

    lappend sArgArray(ArgNames) $argname
    lappend sArgArray(ArgTypes) $argtype

    switch $argtype {
     color {
      set sArgArray(${argname}_R) [expr int([lindex $shaderarguments $j]*255)]
      incr j
      set sArgArray(${argname}_G) [expr int([lindex $shaderarguments $j]*255)]
      incr j
      set sArgArray(${argname}_B) [expr int([lindex $shaderarguments $j]*255)]
      incr j
      set sArgArray(${argname}_A) 0

      append newGUIproc "addColorEntry \$w ${type}ShaderData $argname; "
     }
     point {
      set sArgArray(${argname}_X) [lindex $shaderarguments $j]
      incr j
      set sArgArray(${argname}_Y) [lindex $shaderarguments $j]
      incr j
      set sArgArray(${argname}_Z) [lindex $shaderarguments $j]
      incr j
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_X;"
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_Y;"
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_Z;"
     }
     vector {
      set sArgArray(${argname}_X) [lindex $shaderarguments $j]
      incr j
      set sArgArray(${argname}_Y) [lindex $shaderarguments $j]
      incr j
      set sArgArray(${argname}_Z) [lindex $shaderarguments $j]
      incr j
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_X;"
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_Y;"
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname}_Z;"
     }
     float {
      set sArgArray(${argname}) [lindex $shaderarguments $j]
      incr j
      append newGUIproc "addParamEntry \$w ${type}ShaderData ${argname};"
     }
     string {
      set sArgArray(${argname}) [lindex $shaderarguments $j]
      incr j
      append newGUIproc "addFileEntry \$w ${type}ShaderData ${argname};"
     }
     default {puts stdout "Oops: Ignoring unknown argument type: $argtype!";}
    }
} else {
incr j
    switch $argtype {
     color {
      incr j 3
#      incr j
#      incr j

     }
     point {
      incr j 3
#      incr j
#      incr j
     }
     vector {
      incr j 3
#      incr j
#      incr j
     }
     float {
      incr j
     }
     string {
      incr j
     }
     default {puts stdout "Oops: Ignoring unknown argument type: $argtype!";}
    }
}
incr i -1
}

append newGUIproc "}"

# make new proc
eval $newGUIproc

# fire up the new GUI
set index [.fu.fMain.fProp.flb.li curselection]
applyPropGUIChanges $index
updateMopsPList

return;
}


# we need the 'build a buildShaderGUIproc'-functionality also
# when reading shader-data from the C-context
# ${type}shaderData has to be filled appropriately!
# XXXX doppelmoppel
proc buildNewSBuildGUIProcproc { type } {
upvar #0 ${type}ShaderData sArgArray

set newGUIproc "proc build${type}ShaderGUI { w } {"
append newGUIproc "global ${type}ShaderData;"
append newGUIproc "global prefs;"
append newGUIproc "set prefs(thisPropGUIStatic) 1;"
append newGUIproc "addCommandEntry \$w c1 \"Set new shader.\" \"selectShader \$w $type\"; "

append newGUIproc "addCommandEntry \$w c2 \"Delete shader.\" \"delSelectedShader\";"

set shadername $sArgArray(Name)

if { $shadername != "" } {
    # current shadername
    append newGUIproc "addTextEntry \$w e1 $shadername; "

    set argnames $sArgArray(ArgNames)
    set argtypes $sArgArray(ArgTypes)
    set index 0
    foreach argname $argnames {
	set argtype [lindex $argtypes $index]

	switch $argtype {
	    color {
		append newGUIproc "addColorEntry \$w ${type}ShaderData\
			$argname; "
	    }
	    point {
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_X;"
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_Y;"
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_Z;"
	    }
	    vector {
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_X;"
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_Y;"
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname}_Z;"
	    }
	    float {
		append newGUIproc "addParamEntry \$w ${type}ShaderData\
			${argname};"
	    }
	    string {
		append newGUIproc "addFileEntry \$w ${type}ShaderData\
			${argname};"  
	    }
	}
	incr index
    }


}

append newGUIproc "}"

# make new proc
eval $newGUIproc
update

return;
}

proc delSelectedShader { } {

set index [.fu.fMain.fProp.flb.li curselection]
set name [.fu.fMain.fProp.flb.li get $index]

switch $name {
 "SurfShader"  { set type "surface" }
 "DispShader"  { set type "displacement" }
 "LightShader"  { set type "light" }
 "IntShader"  { set type "interior" }
 "ExtShader"  { set type "exterior" }
 "AtmShader"  { set type "atmosphere" }
 "ImgShader"  { set type "imager" }
 default { return }
}

delShader $type

updateMopsPList

}


# Misc

proc saveArchive { } {


}

###################################
# moveOrResize:
#
proc moveOrResize { w newgeom } {

    if { $newgeom == "" } { return }

#    update idletasks
    set oldgeom [wm geom $w]
    regexp {([0-9]+)?x?([0-9]+)?(\+|\-)?([0-9]+)?(\+|\-)?([0-9]+)?} $oldgeom blurb width height blurb2 x blurb3 y

    regexp {([0-9]+)?x?([0-9]+)?(\+|\-)?([0-9]+)?(\+|\-)?([0-9]+)?} $newgeom blurb nw nh blurb2 nx blurb3 ny

    if { $nw == "" } {
	if { $nx != "" } { set x $nx }
	if { $ny != "" } { set y $ny }
	if { [expr abs($x)] >= [winfo screenwidth $w] } {
	    set x 0
	}
	if { [expr abs($y)] >= [winfo screenheight $w] } {
	    set y 0
	}

	set newgeom ""
	if { $x >= 0 } { append newgeom "+$x" } else { append newgeom "-$x" }
	if { $y >= 0 } { append newgeom "+$y" } else { append newgeom "-$y" }
	wm geom $w $newgeom
    } else {
	if { $nx == "" } {

	    if { $nw != "" } { set width $nw } 
	    if { $nh != "" } { set height $nh } 

	    set newgeom "${width}x${height}"

	    wm geom $w $newgeom
	} else {
	    if { $nx != "" } { set x $nx }
	    if { $ny != "" } { set y $ny }
	    if { $nw != "" } { set width $nw } 
	    if { $nh != "" } { set height $nh } 
	    if { [expr abs($x)] >= [winfo screenwidth $w] } {
		set x 0
	    }
	    if { [expr abs($y)] >= [winfo screenheight $w] } {
		set y 0
	    }

	    set newgeom "${width}x${height}"
	    if { $x >= 0 } { 
		append newgeom "+$x" 
	    } else { 
		append newgeom "-$x" 
	    }
	    if { $y >= 0 } { 
		append newgeom "+$y"
	    } else {
		append newgeom "-$y"
	    }
	    wm geom $w $newgeom
	}
    }

    update idletasks

 return;
}

###################################
# savePrefsp:
#
proc savePrefsp { } {
global mops_gp
set mopsrc $mops_gp(mopsrc)

if { [file exists $mopsrc ] } {
    if { ! [file writable $mopsrc ] } {
	puts stderr "$mopsrc is not writable! No preferences saved!"
	return;
    }
 
    file copy -force -- $mopsrc ${mopsrc}.bak

    set oldfile [open ${mopsrc}.bak]
    set newfile [open $mopsrc w]
    set done 0
    set line ""
# copy old mopsrc to new until end or magic cookie
    while {! $done } {
	gets $oldfile line
	puts $newfile $line
	if { [eof $oldfile] } {
	    set done 1
	    puts $newfile "# saved preferences follow, do not modify or delete this line!"
	} else {
	if { [string first "# saved preferences follow," $line ] >= 0 } {
	    set done 1
	}
    }
    }
} else {
    if { $mopsrc == "" } { set mopsrc "~/.mopsrc" }
    set newfile [open $mopsrc w]
    puts $newfile "# saved preferences follow, do not modify or delete this line!"
}

global prefs

# get main geometry
global tcl_platform
if { ($tcl_platform(platform) != "windows") &&\
	($prefs(twmcompat) != 1) } {
    set x [winfo rootx .]
    set y [winfo rooty .]
    set ng ""
    if { $x >= 0 } { append ng "+$x" } else { append ng "-$x" }
    if { $y >= 0 } { append ng "+$y" } else { append ng "-$y" }
    set prefs(mainGeom) $ng
} else {
    set prefs(mainGeom) [wm geom .]
}

# get toolbox geometry
if { [winfo exists .tb] } {
    if { ($tcl_platform(platform) != "windows") &&\
	    ($prefs(twmcompat) != 1) } {
	set x [winfo rootx .tb]
	set y [winfo rooty .tb]
	set ng ""
	if { $x >= 0 } { append ng "+$x" } else { append ng "-$x" }
	if { $y >= 0 } { append ng "+$y" } else { append ng "-$y" }
	set prefs(toolBoxGeom) $ng
    } else {
	set prefs(toolBoxGeom) [wm geom .tb]
    }
}

# get view1 geometry
global view1Geomvalid
if { $view1Geomvalid == 1 } {
    if { ($tcl_platform(platform) != "windows") &&\
	    ($prefs(twmcompat) != 1) } {
	set x [winfo rootx .view1]
	set y [winfo rooty .view1]
	set ng ""
	if { $x >= 0 } { append ng "+$x" } else { append ng "-$x" }
	if { $y >= 0 } { append ng "+$y" } else { append ng "-$y" }	
	set prefs(view1Geom) $ng
    } else {
	set prefs(view1Geom) [wm geom .view1]
    }
}

# actually write the prefs array
foreach pref [array names prefs ] {
    eval [subst "set val {{\$prefs($pref)}}"]
    puts $newfile "set prefs($pref) $val"
}

puts $newfile "return"

close $newfile
catch {close $oldfile}

puts "Done saving preferences to $mopsrc."

return;
}

###################################
# runToolp:
# execute a tool, asking for arguments
# argvars:    list of global varnames
# argstrings: labels for the argument entry fields
# command:    the command that calls the tool
# example:
# runToolp splitCurveu {"Split at u:"} "splitCurve %splitCurveu"
proc runToolp { argvars argstrings command } {

set w .rtw
catch {destroy $w}
toplevel $w
wm title $w "Mops"
wm iconname $w "Mops"
wm transient $w .

set f [frame $w.f1]
pack $f -in $w -side top -fill x


set okscript ""
append okscript "set command \"$command\";"

set index 0
foreach i $argvars {
    # create GUI
    set f2 [frame $f.f$index]

    label $f2.l -text [lindex $argstrings $index ] -width 12
    entry $f2.e -width 12
    global $i
    eval [subst "$f2.e insert @0 \$$i"]

    pack $f2.l $f2.e -in $f2 -padx 2 -pady 2 -side left
    pack $f2 -in $w.f1 -side top
    # and simultaneously create a script
    append okscript "global $i; set $i \[$f2.e get\];update;"
    append okscript "regsub -all \"%$i\" \$command \[$f2.e get\] command;"
    incr index
}

append okscript "eval \"\$command\"; redrawAllViews;"
append okscript "grab release .rtw; focus .; destroy .rtw"

set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command $okscript

button $f.bca -text "Cancel" -width 5 -command "destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x -expand yes

centerWindowp $w
grab $w
focus $f.bok
tkwait window $w

return;
}


##############################
# hideMops:
# iconify all windows
#
proc hideMops { } {
    global mops_views

    foreach i $mops_views {
	wm iconify $i
    }

    if { [winfo exists .tb] } { wm iconify .tb }
    if { [winfo exists .iw] } { wm iconify .iw }

    wm iconify .

 return;
}

##############################
# cleanTmpFilesp:
# 
#
proc cleanTmpFilesp { all } {
    global prefs mops_tempfiles

    if { $all == 1 } {
	puts "Removing all temporary files..."
	foreach i $mops_tempfiles {
	    cd [file dirname $i]
	    set files [glob [file tail $i]*]
	    foreach j $files { file delete $j }
	}
    } else {

	set new_tempfiles ""
	set count 0
	set todelete [expr [llength $mops_tempfiles]-$prefs(KeepNTmpFiles)]
	
	foreach i $mops_tempfiles {
	    if { $count < $todelete } {
#		puts "Removing some temporary files..."
		set oldcd [pwd]
		cd [file dirname $i]
		set files [glob [file tail $i]*]
		foreach j $files { file delete $j }
		cd $oldcd
	    } else {
		lappend new_tempfiles $i
	    }


	    incr count
	}
	set mops_tempfiles $new_tempfiles

    }

}


##############################
# centerWindowp:
# center window w on screen
#
proc centerWindowp { w } {

    wm withdraw $w
    update idletasks

    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
            - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
            - [winfo vrooty [winfo parent $w]]]

    moveOrResize $w "+${x}+${y}"

    wm deiconify $w

 return;
}


##############################
# updatePositionsp:
# get the positions of all views
# put them in the C context view_info
proc updatePositionsp { } {
 global mops_views tcl_platform

    update idletasks

    foreach w $mops_views {

	if { $tcl_platform(platform) != "windows" } {
	    set x [winfo rootx $w]
	    set y [winfo rooty $w]
	} else {
	    set oldgeom [wm geom [winfo toplevel $w]]
	    regexp {([0-9]+)?x?([0-9]+)?(\+|\-)?([0-9]+)?(\+|\-)?([0-9]+)?} $oldgeom blurb width height blurb2 x blurb3 y
	}

	$w.f3D.togl updatePosition $x $y
    }

 return;
}


##############################
# set title bar of window w (for mops views only)
# type = "" -> no change, action = "" -> no change
proc viewTitlep { w type action } {

    set w [winfo toplevel $w]

    set oldname [wm title $w] 

    catch {scan $oldname "%s - %s - %s" vname oldtype oldaction}

    scan $w ".view%d" number

    set newvname [format "View%d" $number]

    if { $type != "" } { set oldtype $type }
    if { $action != "" } { set oldaction $action }
    set newname [format "%s - %s - %s" $newvname $oldtype $oldaction]
    wm title $w $newname
    update idletasks

return;
}


# Properties

# the following procedures implement the
# standard Property-GUI-elements 

# toggleCopyOmit:
# arranges for all variables in varnames
# to be omitted by the next copy property action
proc toggleCopyOmit { label varnames } {
    upvar #0 copyPropOmit omita
    set oldlabel [$label cget -text] 

    if { [string first "!" $oldlabel ] == -1 } {
	set name "!$oldlabel"
	$label configure -text $name
	foreach i $varnames {
	    set omita($i) ""
	}
    } else {
	set name [string trimleft $oldlabel !]
	$label configure -text $name
	foreach i $varnames {
	    unset omita($i)
	}
    }
}

# 
#
#
proc updateParamEntry { w prop name op } {
    global $prop

    set f $w.f${name}
    
    set oldval [$f.e get]
    set newval $oldval

    if {$op == "*2"} {
	set newval [expr $oldval * 2]
    } elseif { $op == "/2"} {
	set newval [expr $oldval / 2]
    }

    set ${prop}(${name}) $newval

return;
}

#
#
#
proc addParamEntry { w prop name } {
global $name prefs

set bw $prefs(BorderWidth)
incr bw
set f [frame $w.f${name} -relief sunken -bd $bw]

label $f.l -width 12 -text ${name}:

bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

button $f.b1 -pady 0 -bd $bw -text "/2" -command "updateParamEntry $w $prop $name /2" -takefocus 0

entry $f.e -width 6 -textvariable ${prop}(${name}) -bd $bw

button $f.b2 -pady 0 -bd $bw -text "*2"  -command "updateParamEntry $w $prop $name *2" -takefocus 0
global tcl_platform
if { $tcl_platform(platform) == "windows" } {
    pack $f.l -in $f -side left
    pack $f.b1 -in $f -side left -pady 0 -fill x -expand yes
    pack $f.e -in $f -side left -pady 0 -fill y
    pack $f.b2 -in $f -side left -pady 0 -fill x -expand yes
} else {
    pack $f.l $f.b1 $f.e $f.b2 -in $f -side left -fill both
}
pack $f -in $w -side top -fill x
return;
}

#
#
#
proc updateColorEntry { w prop name button} {
    global $prop 

    set rname ${prop}(${name}_R)
    set gname ${prop}(${name}_G)
    set bname ${prop}(${name}_B)

    set newcolor [tk_chooseColor -initialcolor [$button cget -background]]   
    if { $newcolor != "" } {
	$button configure -background $newcolor

	scan $newcolor "#%2x%2x%2x" $rname $gname $bname
    }

return;
}

#
#
#
proc updateColorEntryFromE { w prop name button} {
 global $prop 

 set rname ${prop}(${name}_R)
 set gname ${prop}(${name}_G)
 set bname ${prop}(${name}_B)

 set red [subst "\$$rname"]
 set green [subst "\$$gname"]
 set blue [subst "\$$bname"]

    if { ($red <= 255) && ($red >= 0) && ($green <= 255) && ($green >= 0)\
	    && ($blue <= 255) && ($blue >= 0)} {

	set newcolor [format "#%02x%02x%02x" $red $green $blue]
	$button configure -background $newcolor
    }



return;
}

#
#
#
proc addColorEntry { w prop name } {
    global $prop prefs

    # first, we create a label on its own line
    set f [frame $w.fl${name}]
    
    label $f.l -text $name
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l {${name}_R\
	    ${name}_G ${name}_B ${name}_A}"
    pack $f.l -in $f
    pack $f -in $w -side top

    # now the color entries
    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    entry $f.er -width 4 -textvariable ${prop}(${name}_R) -bd $bw
    entry $f.eg -width 4 -textvariable ${prop}(${name}_G) -bd $bw
    entry $f.eb -width 4 -textvariable ${prop}(${name}_B) -bd $bw
    entry $f.ea -width 4 -textvariable ${prop}(${name}_A) -bd $bw

    bind $f.er <Return> "updateColorEntryFromE $w $prop $name $f.b1"
    bind $f.eg <Return> "updateColorEntryFromE $w $prop $name $f.b1"
    bind $f.eb <Return> "updateColorEntryFromE $w $prop $name $f.b1"
    bind $f.ea <Return> "updateColorEntryFromE $w $prop $name $f.b1"

    set red [subst \$${prop}(${name}_R)]
    set green [subst \$${prop}(${name}_G)]
    set blue [subst \$${prop}(${name}_B)]

    if { ($red <= 255) && ($red >= 0) && ($green <= 255) && ($green >= 0)\
	    && ($blue <= 255) && ($blue >= 0)} {
	set bcolor [format "#%02x%02x%02x" $red $green $blue]
    } else {
	set bcolor "#000000"
    }
    global tcl_platform

    if { $tcl_platform(platform) == "windows" } {
	button $f.b1 -pady 0 -background $bcolor\
		-command "updateColorEntry $w $prop $name $f.b1"\
		-bd $bw -width 3
    } else {
	button $f.b1 -pady 0 -background $bcolor\
		-command "updateColorEntry $w $prop $name $f.b1"\
		-bd $bw
    }

    pack $f.er $f.eg $f.eb $f.ea $f.b1 -in $f -fill both -expand yes -side left -padx 2
    pack $f -in $w -side top -fill x -expand yes

return;
}


#
#
#
proc addCheckEntry { w prop name } {
    global $prop prefs

    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    label $f.l -width 12 -text ${name}:
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

    global tcl_platform

    if { $tcl_platform(platform) == "windows" } {
	# damn windows
	checkbutton $f.cb -image emptyimg -variable ${prop}(${name})\
		-bd $bw -indicatoron 0 -selectcolor #b03060

	pack $f.l -in $f -side left
	pack $f.cb -in $f -side left -padx 50 -pady 2
    } else {
	checkbutton $f.cb -variable ${prop}(${name}) -bd $bw -pady 0
	pack $f.l -in $f -side left
	pack $f.cb -in $f -side left -fill x -expand yes
    }


    pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addMenuEntry { w prop name elist } {
    global $prop prefs tcl_platform

    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    label $f.l -width 12 -text ${name}:
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

    menubutton $f.mb -text Eimer -menu $f.mb.m -relief raised -bd $bw\
	    -padx 0 -pady 0

    if { $tcl_platform(platform) == "windows" } {
	$f.mb configure -pady 1
    }
    set m [menu $f.mb.m -tearoff 0]

    set val 0

    foreach i $elist {
	$m add command -label $i -command "global $prop
	  set ${prop}($name) $val
	  $f.mb configure -text $i"
	incr val
    }


    $m invoke [subst \$${prop}($name)]

    pack $f.l -in $f -side left -fill x
    pack $f.mb -in $f -side left -fill x -expand yes -pady 0
    pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addStringEntry { w prop name } {
    global $prop prefs

    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]
    
    label $f.l -width 12 -text ${name}:
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

    entry $f.e -textvariable ${prop}(${name}) -width 15 -bd $bw
    
    pack $f.l -in $f -side left -fill x
    pack $f.e -in $f -side left -fill both -expand yes
    pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addFileEntry { w prop name } {
    global $prop prefs
    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    label $f.l -width 10 -text ${name}:
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

    entry $f.e -textvariable ${prop}(${name}) -width 15 -bd $bw
    button $f.b -text "Set" -bd $bw -padx 0 -pady 0 -command "\
	global $prop;\
	set filen \[$f.e get\];\
	set filen \[tk_getOpenFile -parent .\
	-title \"Set File:\"];\
	if { \$filen != \"\" } {\
         $f.e delete 0 end;\
	 $f.e insert 0 \$filen;\
	 set ${prop}($name) \$filen
        }
	" 
	pack $f.l -in $f -side left -fill x
	pack $f.b -in $f -side right -fill x -expand yes
	pack $f.e -in $f -side left -fill both
	pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addMDirEntry { w prop name } {
    global $prop prefs
    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    label $f.l -width 10 -text ${name}:
    bind $f.l <Double-ButtonPress-1> "toggleCopyOmit $f.l $name"

    entry $f.e -textvariable ${prop}(${name}) -width 15 -bd $bw
    button $f.b -text "Add" -bd $bw -padx 0 -pady 0 -command "\
	global $prop;\
	set filen \[$f.e get\];\
	global mops_gp;\
	set sep \$mops_gp(separator);\
	set filen \[tk_getOpenFile\
	-title \"Select File:\"\];\
	if { \$filen != \"\" } {\
	 if { \$${prop}($name) != \"\" } {\
          set ${prop}($name) \$${prop}($name)\$sep\[file dirname \$filen\];\
	 } else {\
	  set ${prop}($name) \[file dirname \$filen\];\
         };\
        };\
	" 
	pack $f.l -in $f -side left -fill x
	pack $f.b -in $f -side right -fill x -expand yes
	pack $f.e -in $f -side left -fill both
	pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addCommandEntry { w name text command } {
    global prefs
    set bw $prefs(BorderWidth)
    incr bw
    set f [frame $w.f${name} -relief sunken -bd $bw]

    button $f.b -text $text -bd $bw -command $command -pady 0

    pack $f.b -in $f -side left -fill x -expand yes
    pack $f -in $w -side top -fill x

return;
}

#
#
#
proc addTextEntry { w name text } {

set f [frame $w.${name}]

label $f.l -text $text

pack $f.l -in $f
pack $f -in $w -side top
return;
}


# the following procedures implement all the
# property GUIs

#
#
#
proc buildRiOptGUI { w } {
    global prefs
    set prefs(thisPropGUIStatic) 1

    addTextEntry $w e0 "Image:"
    addParamEntry $w riopt Width
    addParamEntry $w riopt Height

    addTextEntry $w e1 "Antialias:"
    addParamEntry $w riopt Variance
    addParamEntry $w riopt Samples_X
    addParamEntry $w riopt Samples_Y
    addMenuEntry $w riopt FilterFunc [list Gaussian Triangle CatmullRom Box Sinc]
    addParamEntry $w riopt FilterWidth
    addParamEntry $w riopt FilterHeight

    addTextEntry $w e2 "Quantization:"
    addParamEntry $w riopt ExpGain
    addParamEntry $w riopt ExpGamma
    addParamEntry $w riopt RGBA_ONE
    addParamEntry $w riopt RGBA_MIN
    addParamEntry $w riopt RGBA_MAX
    addParamEntry $w riopt RGBA_Dither

    addTextEntry $w e3 "BMRT-Specific:"
    addTextEntry $w e4 "Rendering:"
    addParamEntry $w riopt MinSamples
    addParamEntry $w riopt MaxSamples
    addParamEntry $w riopt MaxRayLevel
    addParamEntry $w riopt ShadowBias
    addMenuEntry $w riopt PRManSpec [list RiStandard PRMan]

    addTextEntry $w e5 "Radiosity:"
    addParamEntry $w riopt RadSteps
    addParamEntry $w riopt PatchSamples

    addTextEntry $w e6 "Paths:"
    addStringEntry $w riopt Textures
    addStringEntry $w riopt Includes
    addStringEntry $w riopt Shaders

    addTextEntry $w e7 "Limits:"
    addParamEntry $w riopt TextureMem
    addParamEntry $w riopt GeomMem


 return;
}

#
#
#
proc buildPrefsGUI { w } {

    addTextEntry $w e0 "Shaders:"
    addMDirEntry $w prefs Shaders
    addCommandEntry $w c1 "Scan Shaders!" "scanAllShaders"

    addTextEntry $w e1 "GUI:"
    addMenuEntry $w prefs BorderWidth [list 1 2]
    addCheckEntry $w prefs AutoResize

    addTextEntry $w e6 "Modeling:"
    addParamEntry $w prefs PickEpsilon
    addCheckEntry $w prefs UpdateTrafo
    addCheckEntry $w prefs LazyNotify

    addTextEntry $w e7 "Errors:"
    addMenuEntry $w prefs OnError [list Stop Continue]

    addTextEntry $w e8 "Files:"
    addCheckEntry $w prefs SaveViews 
    addCheckEntry $w prefs ReadRiOpt

    addTextEntry $w e3 "RIB-Export:"
    addFileEntry $w prefs RIBFile
    addFileEntry $w prefs Image
    addCheckEntry $w prefs ResInstances
    addCheckEntry $w prefs CheckLights
    addCheckEntry $w prefs RIStandard

    addTextEntry $w e4 "Rendering:"
    addStringEntry $w prefs QRender
    addStringEntry $w prefs Render

return;
}

proc getPrefsPropp { } {

    getPrefsProp

 return;
}

proc setPrefsPropp { } {
 global prefs tkCon

    setPrefsProp

 return;
}


#
#
#
proc getCameraPropp { name } {
    global view currentView
    upvar #0 cameraData localcameraData

    .$name.f3D.togl makecurrent
    .$name.f3D.togl getViewConfig

    set localcameraData(From_X) $view(From_X)
    set localcameraData(From_Y) $view(From_Y)
    set localcameraData(From_Z) $view(From_Z)

    set localcameraData(To_X) $view(To_X)
    set localcameraData(To_Y) $view(To_Y)
    set localcameraData(To_Z) $view(To_Z)

    set localcameraData(Roll) $view(Roll)
    set localcameraData(Zoom) $view(Zoom)

    $currentView makecurrent
    $currentView getViewConfig


 return;
}


#
#
#
proc setCameraPropp { name } {
    global currentView cameraData view

    set arguments ""

    append arguments [subst "\
-fromx $cameraData(From_X) -fromy $cameraData(From_Y)\
-fromz $cameraData(From_Z)\
-tox $cameraData(To_X) -toy $cameraData(To_Y) -toz $cameraData(To_Z)\
-upx 0.0 -upy 1.0 -upz 0.0 -updroty 1\
-roll $cameraData(Roll) -zoom $cameraData(Zoom)"]


    eval ".${name}.f3D.togl makecurrent"
    eval ".${name}.f3D.togl setViewConfig $arguments"
    $currentView makecurrent

#XXXX reset actions!

 return;
}

#
#
#
proc getViewAttribPropp { name } {
    global view currentView
    upvar #0 viewAttribData localviewAttribData

    .$name.f3D.togl makecurrent
    .$name.f3D.togl getViewConfig

    set localviewAttribData(Type) $view(Type)
    set localviewAttribData(OrigType) $view(Type)
    set localviewAttribData(Width) $view(Width)
    set localviewAttribData(Height) $view(Height)
    set localviewAttribData(Grid) $view(Grid)
    set localviewAttribData(DrawWire) $view(DrawWire)
    set localviewAttribData(DrawSurf) $view(DrawSurf)
    set localviewAttribData(DrawGrid) $view(DrawGrid)
    set localviewAttribData(DrawSel) $view(DrawSel)

    set localviewAttribData(Lighting) $view(Lighting)
    set localviewAttribData(Redraw) $view(Redraw)

    $currentView makecurrent

 return;
}


#
#
#
proc setViewAttribPropp { name } {
    global currentView viewAttribData

    set arguments ""
    wm geometry .$name ""
    .$name.f3D.togl configure -width $viewAttribData(Width)\
	    -height $viewAttribData(Height)
    update

    if { $viewAttribData(Type) != $viewAttribData(OrigType)} {
	set arguments [subst "-type $viewAttribData(Type)"]
    }

append arguments [subst "\
-draww $viewAttribData(DrawWire) -draws $viewAttribData(DrawSurf)\
-drawg $viewAttribData(DrawGrid) -dsel $viewAttribData(DrawSel)\
-draw $viewAttribData(Redraw) -light $viewAttribData(Lighting)\
-grid $viewAttribData(Grid)"]

eval ".${name}.f3D.togl makecurrent"
eval ".${name}.f3D.togl setViewConfig $arguments"

 return;
}


#
#
#
proc buildCameraGUI { w } {
    global cameraData prefs

    addTextEntry $w e2 "From:"
    addParamEntry $w cameraData From_X
    addParamEntry $w cameraData From_Y
    addParamEntry $w cameraData From_Z

    addTextEntry $w e3 "To:"
    addParamEntry $w cameraData To_X
    addParamEntry $w cameraData To_Y
    addParamEntry $w cameraData To_Z

    addTextEntry $w e5 "Misc:"
    addParamEntry $w cameraData Roll
    addParamEntry $w cameraData Zoom

 return;
}


#
#
#
proc buildViewAttribGUI { w } {
    global viewAttribData prefs

    addMenuEntry $w viewAttribData Type [list Front Side Top Persp Trim]

    addTextEntry $w e4 "Window:"
    addParamEntry $w viewAttribData Width
    addParamEntry $w viewAttribData Height

    addTextEntry $w e5 "Grid:"
    addParamEntry $w viewAttribData Grid

    addTextEntry $w e6 "Draw what:"
    addCheckEntry $w viewAttribData DrawWire
    addCheckEntry $w viewAttribData DrawSurf
    addCheckEntry $w viewAttribData DrawGrid
    addCheckEntry $w viewAttribData DrawSel
    addTextEntry $w e7 "Misc:"
    addCheckEntry $w viewAttribData Lighting
    addCheckEntry $w viewAttribData Redraw

return;
}


#
#
#
proc buildTransformationGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e1 "Translation:"
addParamEntry $w transfPropData Translate_X
addParamEntry $w transfPropData Translate_Y
addParamEntry $w transfPropData Translate_Z

addTextEntry $w e2 "Rotation:"
addParamEntry $w transfPropData Rotate_X
addParamEntry $w transfPropData Rotate_Y
addParamEntry $w transfPropData Rotate_Z

addTextEntry $w e3 "Scale:"
addParamEntry $w transfPropData Scale_X
addParamEntry $w transfPropData Scale_Y
addParamEntry $w transfPropData Scale_Z

addTextEntry $w e4 "Quaternion:"
addParamEntry $w transfPropData Quat0
addParamEntry $w transfPropData Quat1
addParamEntry $w transfPropData Quat2
addParamEntry $w transfPropData Quat3

return;
}

#
#
#
proc buildAttributesGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addColorEntry $w attrPropData Color

addColorEntry $w attrPropData Opacity

addTextEntry $w e3 "Shading:"
addParamEntry $w attrPropData ShadingRate
addMenuEntry $w attrPropData Interpolation [list Constant Smooth]

addTextEntry $w e4 "Displacement:"
addCheckEntry $w attrPropData TrueDisp
addMenuEntry $w attrPropData BoundCoord [list Object Shader Camera]
addParamEntry $w attrPropData Bound

addTextEntry $w e5 "Misc:"
addMenuEntry $w attrPropData CastShadows [list Os None Opaque Shade]

addTextEntry $w e7 "Visibility:"
addCheckEntry $w attrPropData Editor
addCheckEntry $w attrPropData Camera
addCheckEntry $w attrPropData Reflection
addCheckEntry $w attrPropData Shadow

addTextEntry $w e6 "Radiosity:"
addColorEntry $w attrPropData AverageColor
addColorEntry $w attrPropData EmissionColor
addColorEntry $w attrPropData SpecularColor
addParamEntry $w attrPropData PatchSize
addParamEntry $w attrPropData ElemSize
addParamEntry $w attrPropData MinSize
addMenuEntry $w attrPropData Zonal [list DontSet None ZReceives ZShoots Full]
addCheckEntry $w attrPropData HasCaustics
return;
}

#
#
#
proc buildclightAttrGUI { w } {
addMenuEntry $w lattrPropData Type [list Custom Point Spot Distant]
addCheckEntry $w lattrPropData IsOn
addMenuEntry $w lattrPropData Shadows [list Off On]
addParamEntry $w lattrPropData Samples
return;
}

#
#
#
proc buildplightAttrGUI { w } {
addMenuEntry $w lattrPropData Type [list Custom Point Spot Distant]
addCheckEntry $w lattrPropData IsOn
addMenuEntry $w lattrPropData Shadows [list Off On]
addParamEntry $w lattrPropData Intensity
addColorEntry $w lattrPropData Color
return;
}

#
#
#
proc buildslightAttrGUI { w } {
addMenuEntry $w lattrPropData Type [list Custom Point Spot Distant]
addCheckEntry $w lattrPropData IsOn
addMenuEntry $w lattrPropData Shadows [list Off On]
addParamEntry $w lattrPropData Intensity
addColorEntry $w lattrPropData Color
addParamEntry $w lattrPropData ConeAngle
addParamEntry $w lattrPropData ConeDAngle
addParamEntry $w lattrPropData BeamDistrib
return;
}

#
#
#
proc builddlightAttrGUI { w } {
addMenuEntry $w lattrPropData Type [list Custom Point Spot Distant]
addCheckEntry $w lattrPropData IsOn
addMenuEntry $w lattrPropData Shadows [list Off On]
addParamEntry $w lattrPropData Intensity
addColorEntry $w lattrPropData Color
return;
}

#
# initial Surface-Shader-GUI
#
proc buildsurfaceShaderGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1
addCommandEntry $w c1 "Set new shader!" "selectShader $w surface"
return;
}
#
# initial Light-Shader-GUI
#
proc buildlightShaderGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1
addCommandEntry $w c1 "Set new shader!" "selectShader $w light"
return;
}
#
# initial Displacement-Shader-GUI
#
proc builddisplacementShaderGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1
addCommandEntry $w c1 "Set new shader!" "selectShader $w displacement"
return;
}
#
# initial Volume-Shader-GUI
#
proc buildvolumeShaderGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1
addCommandEntry $w c1 "Set new shader!" "selectShader $w volume"
return;
}

#
# initial Imager-Shader-GUI
#
proc buildimagerShaderGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1
addCommandEntry $w c1 "Set new shader!" "selectShader $w imager"
return;
}

#
#
#
proc buildNCAttrGUI { w } {
global prefs
#set prefs(thisPropGUIStatic) 1
addTextEntry $w e1 "Curve:"
addParamEntry $w ncattrPropData Length
addParamEntry $w ncattrPropData Order
addMenuEntry $w ncattrPropData Knot-Type [list Bezier B-Spline NURB Custom]
addStringEntry $w ncattrPropData Knots
addCheckEntry $w ncattrPropData Has-Weights
addCheckEntry $w ncattrPropData Closed
addTextEntry $w e2 "Handles:"
addParamEntry $w ncattrPropData Handlesize
addCheckEntry $w ncattrPropData W-Handles
addTextEntry $w e3 "GLU-Parameters:"
addParamEntry $w ncattrPropData Tolerance
return;
}

#
#
#
proc buildNPAttrGUI { w } {
global prefs
#set prefs(thisPropGUIStatic) 1
addTextEntry $w e1 "Patch:"
addParamEntry $w npattrPropData Width
addParamEntry $w npattrPropData Height
addParamEntry $w npattrPropData Order_U
addParamEntry $w npattrPropData Order_V
addMenuEntry $w npattrPropData Knot-Type_U [list Bezier B-Spline NURB Custom]
addStringEntry $w npattrPropData Knots_U
addMenuEntry $w npattrPropData Knot-Type_V [list Bezier B-Spline NURB Custom]
addStringEntry $w npattrPropData Knots_V
addCheckEntry $w npattrPropData Has-Weights
addTextEntry $w e2 "Handles:"
addParamEntry $w npattrPropData Handlesize
addCheckEntry $w npattrPropData W-Handles
addTextEntry $w e3 "GLU-Parameters:"
addParamEntry $w npattrPropData Tolerance
addMenuEntry $w npattrPropData Display [list Fill OutlinePoly OutlinePatch]
return;
}

#
#
#
proc buildBoxAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addParamEntry $w BoxAttrData Width
addParamEntry $w BoxAttrData Length
addParamEntry $w BoxAttrData Height

return;
}

#
#
#
proc buildSphereAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData Radius
addParamEntry $w QAttrData ZMin
addParamEntry $w QAttrData ZMax
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildConeAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData Radius
addParamEntry $w QAttrData Height
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildDiskAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addParamEntry $w QAttrData Radius
addParamEntry $w QAttrData Height
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildCylinderAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData Radius
addParamEntry $w QAttrData ZMin
addParamEntry $w QAttrData ZMax
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildHyperbAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData P1_X
addParamEntry $w QAttrData P1_Y
addParamEntry $w QAttrData P1_Z
addParamEntry $w QAttrData P2_X
addParamEntry $w QAttrData P2_Y
addParamEntry $w QAttrData P2_Z
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildParabAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData RMax
addParamEntry $w QAttrData ZMin
addParamEntry $w QAttrData ZMax
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc buildTorusAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addCheckEntry $w QAttrData Closed
addParamEntry $w QAttrData MajorRad
addParamEntry $w QAttrData MinorRad
addParamEntry $w QAttrData PhiMin
addParamEntry $w QAttrData PhiMax
addParamEntry $w QAttrData ThetaMax

return;
}

#
#
#
proc Revolve_buildRevolveAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e3 "Revolution:"
addParamEntry $w RevolveAttrData ThetaMax

addTextEntry $w e2 "Caps:"
addCheckEntry $w RevolveAttrData HasUpperCap
addCheckEntry $w RevolveAttrData HasLowerCap

addTextEntry $w e1 "GLU-Parameters:"
addParamEntry $w RevolveAttrData Tolerance
addMenuEntry $w  RevolveAttrData Display [list Fill OutlinePoly OutlinePatch]

return;
}

#
#
#
proc Extrude_buildExtrudeAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e4 "Extrusion:"
addParamEntry $w ExtrudeAttrData Height
addTextEntry $w e2 "Caps:"
addCheckEntry $w ExtrudeAttrData HasUpperCap
addCheckEntry $w ExtrudeAttrData HasLowerCap
addTextEntry $w e3 "Beveling:"
addCheckEntry $w ExtrudeAttrData UpperBevel
addCheckEntry $w ExtrudeAttrData LowerBevel
addMenuEntry $w  ExtrudeAttrData BevelType [list Round Linear Ridge]
addParamEntry $w ExtrudeAttrData BevelRadius
addTextEntry $w e1 "GLU-Parameters:"
addParamEntry $w ExtrudeAttrData Tolerance
addMenuEntry $w  ExtrudeAttrData Display [list Fill OutlinePoly OutlinePatch]

return;
}

#
#
#
proc Sweep_buildSweepAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e2 "Sweep:"
addParamEntry $w SweepAttrData Sections
addCheckEntry $w SweepAttrData Rotate
addCheckEntry $w SweepAttrData Interpolate
addTextEntry $w e1 "GLU-Parameters:"
addParamEntry $w SweepAttrData Tolerance
addMenuEntry $w  SweepAttrData Display [list Fill OutlinePoly OutlinePatch]

return;
}

#
#
#
proc ICurve_buildICurveAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e3 "Curve:"
addCheckEntry $w ICurveAttrData Closed

addTextEntry $w e2 "Interpolation:"
addMenuEntry $w  ICurveAttrData Mode [list C2-Cubic Global4D]
addParamEntry $w ICurveAttrData Order
addParamEntry $w ICurveAttrData IParam

addTextEntry $w e1 "GLU-Parameters:"
addParamEntry $w ICurveAttrData Tolerance


return;
}

#
#
#
proc Skin_buildSkinAttrGUI { w } {
global prefs
set prefs(thisPropGUIStatic) 1

addTextEntry $w e2 "Skin:"
addCheckEntry $w SkinAttrData Interpolate
addParamEntry $w SkinAttrData Order_U
addMenuEntry $w SkinAttrData Knot-Type_U [list Bezier B-Spline NURB]

addTextEntry $w e1 "GLU-Parameters:"
addParamEntry $w SkinAttrData Tolerance
addMenuEntry $w  SkinAttrData Display [list Fill OutlinePoly OutlinePatch]

return;
}

#
#
#
proc updateTagsData { } {
global tagPropData

 set tw .fu.fMain.fArg.flb.ca.w.fTags.text
 set args ""
 set end 0
 set index 1
 set tagPropData(ttypes) " "
 set tagPropData(tvalues) " "
 while { $end != 1 } {
     
   set line [$tw get $index.0 $index.0lineend]
   if { $line == "" } { set end 1 } else {

       if { [string first " " $line] != -1 } {
	   set type ""
	   set value ""
	   scan $line "%s %s" type value
	   
	   append tagPropData(ttypes) " "
	   append tagPropData(ttypes) $type

	   append tagPropData(tvalues) " "
	   append tagPropData(tvalues) $value
       }
   }

   incr index 1
 }

 return;
}
#
#
#
proc buildTagsGUI { w } {
global prefs tagPropData

set bw $prefs(BorderWidth)

set f [frame $w.fTags -relief sunken -bd $bw]

text $f.text -height 15 -width 40 -yscrollcommand "$f.scy set"
scrollbar $f.scy -takefocus 0 -command "$f.text yview"



$f.text delete 0.0 end

set index 0
foreach tag $tagPropData(ttypes) {
 set line [format "%s %s\n" $tag [lindex $tagPropData(tvalues) $index] ]
 $f.text insert [expr $index + 1].0 $line

 incr index
}

bind $f.text <Down> +updateTagsData
bind $f.text <Up> +updateTagsData
bind $f.text <Enter> +updateTagsData
bind $f.text <Leave> +updateTagsData

pack $f.scy -side right -fill y -expand yes
pack $f.text -fill both -expand yes
pack $f -in $w -side top -fill x -expand yes
return;
}

#
#
#
proc getTagsProp { } {
 global tagPropData

 set tagPropData(ttypes) ""
 set tagPropData(tvalues) ""

 getTags tagPropData(ttypes) tagPropData(tvalues)

 return;
}


#
#
#
proc setTagsProp { } {
 global tagPropData
 set args ""
 set index 0
 foreach tag $tagPropData(ttypes) {

  set value [lindex $tagPropData(tvalues) $index]

  append args " "
  append args $tag
  append args " "
  append args $value

  incr index 1
 }
 eval [subst "setTags $args"]

 return;
}

#
#
#
proc applyPropGUIChanges { index } {
    global propertySetProcList propertyGetProcList prefs
    eval [lindex $propertySetProcList $index]
    redrawAllViews
    update
    if { $prefs(thisPropGUIStatic) == 0 } {
	updateMopsPList
    } else {
	eval [lindex $propertyGetProcList $index]
    }

return;
}

#
#
#
proc undoPropGUIChanges { index } {
    global propertyDataList resetProp 
    global [lindex $propertyDataList $index]
    array set [lindex $propertyDataList $index] [array get resetProp]
    applyPropGUIChanges $index
return;
}

#
#
#
proc updatePropGUI { } {
    global propertyProcList propertyDataList\
	    propertyGetProcList propertyList prefs prop_sema
    upvar #0 resetProp localresetProp

    if { $prop_sema == 1 } {
     # there can only be one!
     return;
    }

    set prop_sema 1
    update

    set w .fu.fMain.fArg.flb.ca.w
    set index [.fu.fMain.fProp.flb.li curselection]

    if { $prefs(thisPropGUIStatic) == 1 } {

	# get values from object
	eval [lindex $propertyGetProcList $index]

	# save undo-information
	global [lindex $propertyDataList $index]
	catch {unset localresetProp}
	array set localresetProp [array get [lindex $propertyDataList $index]]

	.fu.fMain.fArg.fb.b1 configure -command "applyPropGUIChanges $index"
	.fu.fMain.fArg.fb.b2 configure -command "undoPropGUIChanges $index"
	set prop_sema 0
	update
	return;
    }

    # clear old GUI
    catch {.fu.fMain.fArg.flb.ca delete all}
    destroy $w
    update
    frame $w
    .fu.fMain.fArg.flb.ca create window 5 5 -anchor nw -window $w

    if { $index == "" } {
	set prop_sema 0
	update
	return;
    }
    if { $propertyList == "" } {
	set prop_sema 0
	update
	return;
    }

    # get values from object
    eval [lindex $propertyGetProcList $index]

    # create new GUI
    eval {[lindex $propertyProcList $index] $w}

    # realize new GUI
    update

    set f .fu.fMain.fArg.flb

    # resize canvas
    set width [expr [winfo reqwidth $w] + 10]
    set height [expr [winfo reqheight $w] + 10]
    .fu.fMain.fArg.flb.ca configure -width $width
    .fu.fMain.fArg.flb.ca configure -scrollregion [list 0 5 $width $height]
    update
    # resize main?
    if { $prefs(AutoResize) == 1 } {
	set newwidth [winfo reqwidth .fu]
	set newheight [winfo height .]

	set ng ${newwidth}x${newheight}
	set oldgeom [wm geom [winfo toplevel $w]]
	regexp {([0-9]+)?x?([0-9]+)?(\+|\-)?([0-9]+)?(\+|\-)?([0-9]+)?} $oldgeom blurb width height blurb2 x blurb3 y


	global tcl_platform prefs
	if { ($tcl_platform(platform) != "windows") &&\
		($prefs(twmcompat) != 1) } {
	    set x [winfo rootx .]
	    set y [winfo rooty .]

	    if { $x >= 0 } { append ng "+$x" } else { append ng "$x" }
	    if { $y >= 0 } { append ng "+$y" } else { append ng "$y" }
	}
	wm geometry . ""
	wm geometry . $ng
    }

    # save undo-information
    global [lindex $propertyDataList $index]
    catch {unset localresetProp}
    array set localresetProp [array get [lindex $propertyDataList $index]]

    .fu.fMain.fArg.fb.b1 configure -command "applyPropGUIChanges $index"
    .fu.fMain.fArg.fb.b2 configure -command "undoPropGUIChanges $index"

    .fu.fMain.fProp.flb.li selection clear 0 end
    .fu.fMain.fProp.flb.li selection set $index
    set prop_sema 0
    update
return;
}

#
#
#
proc copyPropp { } {
    global propertyDataList copyPropName copyPropOmit prefs
    upvar #0 copyProp localcopyProp

    set index [.fu.fMain.fProp.flb.li curselection]
    
    if { $index == "" } { return; }

    set copyPropName [.fu.fMain.fProp.flb.li get $index]

    catch {unset localcopyProp}

    set arrayname [lindex $propertyDataList $index]
    global $arrayname
    set avnames [array names $arrayname]

    foreach i $avnames {
        if { ! [info exists copyPropOmit($i)] } {
	    eval [subst "set localcopyProp\(\$i\) \{\$$arrayname\(\$i\)\}"]
	}   
    }

 return;
}

#
#
#
proc icopyPropp { } {
    global propertyDataList copyPropName copyPropOmit prefs
    upvar #0 copyProp localcopyProp

    set index [.fu.fMain.fProp.flb.li curselection]
    
    if { $index == "" } { return; }

    set copyPropName [.fu.fMain.fProp.flb.li get $index]

    catch {unset localcopyProp}

    array set localcopyProp { dummy 0 }

    set arrayname [lindex $propertyDataList $index]
    global $arrayname

    set avnames [array names $arrayname]
    foreach i $avnames {
        if { [info exists copyPropOmit($i)] } {
	    eval [subst "set localcopyProp\(\$i\) \{\$$arrayname\(\$i\)\}"]
	}   
    }

return;
}

#
#
#
proc pastePropp { } {
    global propertyDataList copyProp copyPropName copyPropOmit prefs
    global propertyGetProcList propertySetProcList
    upvar #0 resetProp localresetProp

    set selected [.fu.fMain.fHier.flb.li curselection]

    setSelect
    updateMopsSelect

    .fu.fMain.fProp.flb.li selection clear 0 end
    updateMopsPList

    if {$selected == ""} {return;}

    set prefs(pasteProp) 1

    foreach selobject $selected {
	.fu.fMain.fHier.flb.li selection clear 0 end
	.fu.fMain.fHier.flb.li selection set $selobject
	setSelect $selobject
	updateMopsPList
	set numprops [.fu.fMain.fProp.flb.li size ]
	set propindex 0
	while {$propindex != $numprops} {
	    set propname [.fu.fMain.fProp.flb.li get $propindex]
	    if { $propname == $copyPropName } {

		eval [lindex $propertyGetProcList $propindex]

		array set saveResetData [array get localresetProp]
		upvar #0 [lindex $propertyDataList $propindex] propArray
		
		set avnames [array names copyProp]
		foreach i $avnames {
			set propArray($i) $copyProp($i)
		}

		eval [lindex $propertySetProcList $propindex]

		array set localresetProp [array get saveResetData]
		update
		eval [lindex $propertyGetProcList $propindex]
	    }
	    incr propindex
	}
    }

    set prefs(pasteProp) 0

 return;
}

#
#
#
proc tipoftheDay { } {
    global prefs 

    set count [llength $prefs(DailyTips)]
    expr srand([clock seconds])

    set tip [lindex $prefs(DailyTips) [expr round(rand()*double($count-1))]]
    puts stdout $tip

 return;
}

#
#
#
proc aboutMops {} {
set w .wabout
# there can only be one
catch {destroy $w}
toplevel $w
wm title $w "About The Mops"
wm iconname $w "Mops"

frame $w.ftext
frame $w.fbutton

pack $w.ftext -in $w -side top -expand yes -fill both
pack $w.fbutton -in $w -side bottom

text $w.ftext.text -yscrollcommand "$w.ftext.sbar set" \
	-setgrid 1 -height 17 -width 65
scrollbar $w.ftext.sbar -takefocus 0 -command "$w.ftext.text yview"
pack $w.ftext.text -in $w.ftext -side left -fill both -expand yes
pack $w.ftext.sbar -in $w.ftext -side right -fill y

$w.ftext.text insert end \
{
The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler

The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
(rschultz@informatik.uni-rostock.de)
All Rights Reserved

See the file License for details.

For additional information, new versions etc.
see The Mops Home Page:
http://www.informatik.uni-rostock.de/~rschultz/mops/
http://wwwicg.informatik.uni-rostock.de/~rschultz/mops/
http://www.ebusinessrevolution.com/mops/

Version Information:
}

$w.ftext.text tag add "tag1" 1.0 8.end
$w.ftext.text tag add "tag2" 2.0 2.end
$w.ftext.text tag configure tag1 -justify center
$w.ftext.text tag configure tag2 -underline yes
global mops_version tcl_version tk_version gl_ren gl_ver gl_ven gl_ext glu_ver glu_ext tk_patchLevel tcl_patchLevel

$w.ftext.text insert end \
"
The Mops: $mops_version

Tcl
Version: $tcl_version
PatchLevel: $tcl_patchLevel

Tk
Version: $tk_version
PatchLevel: $tk_patchLevel

OpenGL 
Version:  $gl_ver
Renderer: $gl_ren
Vendor:   $gl_ven
"

foreach i $gl_ext {
$w.ftext.text insert end "Extension: $i
" 
}

$w.ftext.text insert end \
"
GLU
Version: $glu_ver
"

foreach i $glu_ext {
$w.ftext.text insert end "Extension: $i
" 
}

$w.ftext.text insert end \
"

The Mops is based on the following fine software:

o BMRT (Blue Moon Rendering Tools) by Larry Gritz
o The Togl widget by Benjamin Bederson, Brian Paul et. al.
o The tkMegaWidget set by Jeffrey Hobbs
o The 3DMF parser from Apple (Duet Development Corp.)
o Tcl/Tk, OpenGL, The RenderMan Interface (too many to list)

The Mops learned a lot from the code of:

o The Affine Toolkit by Thomas E. Burge
o GLUT by M. Kilgard
o The NURBS Book by L. Piegl and W. Tiller
o The NURBS Procedure Library by W.T. Hewitt and D. Yipp
o The NURBS++ library by P. Lavoie

OpenGL (R) is a registered trademark of Silicon Graphics, Inc.

The RenderMan (R) Interface Procedures and Protocol are:
Copyright 1988, 1989, Pixar
All Rights Reserved

RenderMan (R) is a registered trademark of Pixar

The Affine Libraries and Tools are
Copyright (c) 1995, 1996, 1997, 1998 Thomas E. Burge
All rights reserved.

Affine (R) is a registered trademark of Thomas E. Burge
"

$w.ftext.text tag add "tag3" end-14lines end

$w.ftext.text tag configure tag3 -justify center

$w.ftext.text configure -state disabled
bind $w <Next> "$w.ftext.text yview scroll 1 pages"
bind $w <Prior> "$w.ftext.text yview scroll -1 pages"

button $w.fbutton.b -text "Dismiss" -command "destroy $w"
pack $w.fbutton.b -in $w.fbutton

}


#
# Help!
# display the documentation
proc mopsHelpp {} {
global prefs
set mopsdoc $prefs(mopsdoc)
set choices { "/usr/doc/mops/html/mops.html" "/usr/local/mops/doc/html/mops.html" "/opt/mops/doc/html/mops.html" }
# search for the docs
if { $mopsdoc != ""} {
    if { ![file exists $mopsdoc] } {
	lappend choices $mopsdoc
	set mopsdoc ""
    }
}

if { $mopsdoc == ""} {
    set tmp [file dirname [info nameofexecutable]]
    append tmp /../doc/html/mops.html
    lappend choices $tmp
    if { $prefs(docdirs) != "" } { lappend choices $prefs(docdirs) }
    puts stdout "prefs(mopsdoc) is not set or not found. Looking in:"
    foreach choice $choices {
	puts -nonewline stdout "${choice}... "
	if  { [file exists $choice ] } {
	    set mopsdoc $choice
	    break
	}
	puts stderr "No."
    }
    if { $mopsdoc != "" } {
	puts stdout "Yes."
    }
}

if { $mopsdoc != "" } {
    regsub -all {\$mopsdoc} $prefs(docview) $mopsdoc command

    if {[string length "[info commands "[lindex "$command" 0]"]"]} {
	catch {eval $command}
    } else {
	catch {eval exec -- $command}
    }
    global errorCode
    if { $errorCode != "NONE" } {
	regsub -all {\$mopsdoc} $prefs(docview2) $mopsdoc command

	if {[string length "[info commands "[lindex "$command" 0]"]"]} {
	    eval $command
	} else {
	    eval exec -- $command
	}
    }

} else {
    puts stderr "Could not find the documentation, please set prefs(mopsdoc)."
}

return;
}

# replace control
proc repcont { string } {
    regsub -all "Ctrl" $string "Control" dummy
    return $dummy
}

##############################
# Setup Keybindings for Main
# 
#
proc setupKeyBindings { w } {
    global mops_main_sc

    if { [winfo exists .fl.con] == 1 } {
	bind .fl.con.console <[repcont $mops_main_sc(SwCon)]> {
	    focus [tk_focusNext %W] }

	bind .fl.con.console <ButtonPress-4> {
	.fl.con.console yview scroll -1 pages
	break
	}
	bind .fl.con.console <ButtonPress-5> {
	.fl.con.console yview scroll 1 pages
        break
	}

    }

    bind $w <[repcont $mops_main_sc(Quit)]> {.fu.fMenu.fil.m invoke 15}
    .fu.fMenu.fil.m entryconfigure 15 -accelerator $mops_main_sc(Quit)
    bind $w <[repcont $mops_main_sc(New)]> ".fu.fMenu.fil.m invoke 1"
    .fu.fMenu.fil.m entryconfigure 1 -accelerator $mops_main_sc(New)
    bind $w <[repcont $mops_main_sc(Replace)]> ".fu.fMenu.fil.m invoke 3"
    .fu.fMenu.fil.m entryconfigure 3 -accelerator $mops_main_sc(Replace)
    bind $w <[repcont $mops_main_sc(Insert)]> ".fu.fMenu.fil.m invoke 4"
    .fu.fMenu.fil.m entryconfigure 4 -accelerator $mops_main_sc(Insert)
    bind $w <[repcont $mops_main_sc(SaveAs)]> ".fu.fMenu.fil.m invoke 6"
    .fu.fMenu.fil.m entryconfigure 6 -accelerator $mops_main_sc(SaveAs)
    bind $w <[repcont $mops_main_sc(Save)]> ".fu.fMenu.fil.m invoke 7"
    .fu.fMenu.fil.m entryconfigure 7 -accelerator $mops_main_sc(Save)
    bind $w <[repcont $mops_main_sc(ExportRIB)]> ".fu.fMenu.fil.m invoke 9"
    .fu.fMenu.fil.m entryconfigure 9 -accelerator $mops_main_sc(ExportRIB)

    bind $w <[repcont $mops_main_sc(Apply)]> ".fu.fMain.fArg.fb.b1 invoke"

    bind $w <[repcont $mops_main_sc(Copy)]> ".fu.fMenu.ed.m invoke 1"
    .fu.fMenu.ed.m entryconfigure 1 -accelerator $mops_main_sc(Copy)
    bind $w <[repcont $mops_main_sc(Cut)]> ".fu.fMenu.ed.m invoke 2"
    .fu.fMenu.ed.m entryconfigure 2 -accelerator $mops_main_sc(Cut)
    bind $w <[repcont $mops_main_sc(Paste)]> ".fu.fMenu.ed.m invoke 3"
    .fu.fMenu.ed.m entryconfigure 3 -accelerator $mops_main_sc(Paste)
    bind $w <[repcont $mops_main_sc(CopyP)]> ".fu.fMenu.ed.m invoke 6"
    .fu.fMenu.ed.m entryconfigure 6 -accelerator $mops_main_sc(CopyP)
    bind $w <[repcont $mops_main_sc(ICopyP)]> ".fu.fMenu.ed.m invoke 7"
    .fu.fMenu.ed.m entryconfigure 7 -accelerator $mops_main_sc(ICopyP)
    bind $w <[repcont $mops_main_sc(PasteP)]> ".fu.fMenu.ed.m invoke 8"
    .fu.fMenu.ed.m entryconfigure 8 -accelerator $mops_main_sc(PasteP)
    bind $w <[repcont $mops_main_sc(Name)]> ".fu.fMenu.ed.m invoke 10"
    .fu.fMenu.ed.m entryconfigure 10 -accelerator $mops_main_sc(Name)
    bind $w <[repcont $mops_main_sc(Swap)]> ".fu.fMenu.ed.m invoke 12"
    .fu.fMenu.ed.m entryconfigure 12 -accelerator $mops_main_sc(Swap)
    bind $w <[repcont $mops_main_sc(Undo)]> ".fu.fMenu.ed.m invoke 14"
    .fu.fMenu.ed.m entryconfigure 14 -accelerator $mops_main_sc(Undo)

    bind $w <[repcont $mops_main_sc(Help)]> ".fu.fMenu.hlp.m invoke 1"
    .fu.fMenu.hlp.m entryconfigure 1 -accelerator $mops_main_sc(Help)

    bind $w <[repcont $mops_main_sc(Hide)]> "hideMops"
}

##############################
# Setup Keybindings for a 3D-View
#
#
proc setupKeyBindings3D { w } {
    global mops_view_sc

    bind $w <[repcont $mops_view_sc(Undo)]> ".fu.fMenu.ed.m invoke 14"

    bind $w <[repcont $mops_view_sc(QRender)]> "$w.fMenu.v.m invoke 1"
    $w.fMenu.v.m entryconfigure 1 -accelerator $mops_view_sc(QRender)
    bind $w <[repcont $mops_view_sc(Render)]> "$w.fMenu.v.m invoke 2"
    $w.fMenu.v.m entryconfigure 2 -accelerator $mops_view_sc(Render)
    bind $w <[repcont $mops_view_sc(Redraw)]> "$w.fMenu.v.m invoke 3"
    $w.fMenu.v.m entryconfigure 3 -accelerator $mops_view_sc(Redraw)
    bind $w <[repcont $mops_view_sc(Close)]> "closeMopsView $w"
    $w.fMenu.v.m entryconfigure 5 -accelerator $mops_view_sc(Close)

    bind $w <[repcont $mops_view_sc(Front)]> "$w.fMenu.c.m.tp invoke 1"
    $w.fMenu.c.m.tp entryconfigure 1 -accelerator $mops_view_sc(Front)
    bind $w <[repcont $mops_view_sc(Side)]> "$w.fMenu.c.m.tp invoke 2"
    $w.fMenu.c.m.tp entryconfigure 2 -accelerator $mops_view_sc(Side)
    bind $w <[repcont $mops_view_sc(Top)]> "$w.fMenu.c.m.tp invoke 3"
    $w.fMenu.c.m.tp entryconfigure 3 -accelerator $mops_view_sc(Top)
    bind $w <[repcont $mops_view_sc(Persp)]> "$w.fMenu.c.m.tp invoke 5"
    $w.fMenu.c.m.tp entryconfigure 5 -accelerator $mops_view_sc(Persp)
    bind $w <[repcont $mops_view_sc(Trim)]> "$w.fMenu.c.m.tp invoke 7"
    $w.fMenu.c.m.tp entryconfigure 7 -accelerator $mops_view_sc(Trim)

    bind $w <[repcont $mops_view_sc(Wire)]> "$w.fMenu.c.m.drw invoke 1"
    $w.fMenu.c.m.drw entryconfigure 1 -accelerator $mops_view_sc(Wire)
    bind $w <[repcont $mops_view_sc(Surf)]> "$w.fMenu.c.m.drw invoke 2"
    $w.fMenu.c.m.drw entryconfigure 2 -accelerator $mops_view_sc(Surf)
    bind $w <[repcont $mops_view_sc(Grid)]> "$w.fMenu.c.m.drw invoke 4"
    $w.fMenu.c.m.drw entryconfigure 4 -accelerator $mops_view_sc(Grid)

    bind $w <[repcont $mops_view_sc(SGrid)]> "$w.fMenu.c.m invoke 3"
    $w.fMenu.c.m entryconfigure 3 -accelerator $mops_view_sc(SGrid)
    bind $w <[repcont $mops_view_sc(Light)]> "$w.fMenu.c.m invoke 5"
    $w.fMenu.c.m entryconfigure 5 -accelerator $mops_view_sc(Light)
    bind $w <[repcont $mops_view_sc(Auto)]> "$w.fMenu.c.m invoke 6"
    $w.fMenu.c.m entryconfigure 6 -accelerator $mops_view_sc(Auto)

    bind $w <[repcont $mops_view_sc(Halve)]> "$w.fMenu.c.m invoke 8"
    $w.fMenu.c.m entryconfigure 8 -accelerator $mops_view_sc(Halve)
    bind $w <[repcont $mops_view_sc(Double)]> "$w.fMenu.c.m invoke 9"
    $w.fMenu.c.m entryconfigure 9 -accelerator $mops_view_sc(Double)

    $w.fMenu.c.m entryconfigure 11 -accelerator $mops_view_sc(ZoomTO)
    bind $w <[repcont $mops_view_sc(ZoomTO)]> {%W.f3D.togl zoomToOb}

    bind $w <[repcont $mops_view_sc(RotL)]>\
	    "$w.f3D.togl setViewConfig -drotx 5.0 -droty 0.0"
    bind $w <[repcont $mops_view_sc(RotR)]>\
	    "$w.f3D.togl setViewConfig -drotx -5.0 -droty 0.0"
    bind $w <[repcont $mops_view_sc(RotU)]>\
	    "$w.f3D.togl setViewConfig -droty -5.0 -drotx 0.0"
    bind $w <[repcont $mops_view_sc(RotD)]>\
	    "$w.f3D.togl setViewConfig -droty 5.0 -drotx 0.0"

    bind $w <[repcont $mops_view_sc(ZoomI)]> {
	%W.f3D.togl setViewConfig -zoomd 0.9
    }

    bind $w <[repcont $mops_view_sc(ZoomO)]> {
	%W.f3D.togl setViewConfig -zoomd 1.1
    }

    bind $w <Enter> {
	if { $oview_sema != 1 } {
	    set w [winfo toplevel %W]
	    $w.f3D.togl makecurrent
	    $w.f3D.togl getViewConfig
	    set currentView $w.f3D.togl
	    update
	}
    }

}


###############################
# the following bunch of procs
# finally sets up bindings
# for interactive modeling
proc setupWEditBinding { w } {

    bind $w <ButtonPress-1> {
	saveUndo
	$currentView startPointEdit %x %y
	$currentView editWac -start %x
    }

    bind $w <B1-Motion> {
	$currentView editWac -winx %x 
	$currentView render
    }

return;
}

proc setupPEditBinding { w plane } {
    set w $w.f3D.togl

    bind $w <ButtonPress-1> {
	saveUndo
	$currentView startPointEdit %x %y
	$currentView editPac -start %x %y
    }

    bind $w <B1-Motion> {
	$currentView editPac -winxy %x %y
	$currentView render
    }

}

proc setupPSelectBinding { w } {

    bind $w <ButtonPress-1> {
	set oldx %x
	set oldy %y
	$currentView selectPoints %x %y
    }

    bind $w <ButtonRelease-1> {
	if { ($oldx != %x) || ($oldy != %y)} {
	    $currentView selectPoints %x %y $oldx $oldy
	}
	$w.f3D.togl setViewConfig -clip $oldx $oldy %x %y 0
	redrawAllViews
	update
    }

    bind $w <B1-Motion> {
	$w.f3D.togl setViewConfig -clip $oldx $oldy %x %y 1
    }

}

proc actionMoveOb { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W movObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W movObac -winxy %x %y
	update
    }
}

proc actionMoveView { w } {
    bind $w <ButtonPress-1> {
	%W movVwac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W movVwac -winxy %x %y
	update
    }
}

proc actionZoomView { w } {
    bind $w <ButtonPress-1> {
	%W zomVwac -start %y
	update
    }

    bind $w <B1-Motion> {
	%W zomVwac -winy %y
	update
    }
}

proc actionMoveZView { w } {
    bind $w <ButtonPress-1> {
	%W movzVwac -start %y
	update
    }

    bind $w <B1-Motion> {
	%W movzVwac -winy %y
	update
    }
}

proc actionRotView { w } {
    bind $w <ButtonPress-1> {
	set oldx %x
	set oldy %y
	update
    }

    bind $w <B1-Motion> {
	%W setViewConfig -drotx [expr ($oldx - %x)] -droty [expr ($oldy - %y)]
	set oldx %x
	set oldy %y
	update
    }
}

proc actionRotOb { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W rotObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W rotObac -winxy %x %y
	update
    }


}

proc actionRotObabindp { w x y } {
    bind $w <ButtonPress-1> "%W rotObaac -start %x %y $x $y"
    bind $w <B1-Motion> "%W rotObaac -winxy %x %y $x $y"
}

proc actionRotOba { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	update
	actionRotObabindp %W %x %y
    }
    bind $w <ButtonRelease-1> ""
    bind $w <B1-Motion> ""
}

proc actionSc1DObx { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DXObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DXObac -winxy %x %y
	update
    }
}
proc actionSc1DOby { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DYObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DYObac -winxy %x %y
	update
    }
}
proc actionSc1DObz { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DZObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DZObac -winxy %x %y
	update
    }
}
proc actionSc1DWorldx { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DXWorldac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DXWorldac -winxy %x %y
	update
    }
}
proc actionSc1DWorldy { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DYWorldac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DYWorldac -winxy %x %y
	update
    }
}
proc actionSc1DWorldz { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DZWorldac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DZWorldac -winxy %x %y
	update
    }
}
proc actionSc3DOb { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc3DObac -start %x %y
	update
    }
    bind $w <B1-Motion> {
	%W sc3DObac -winxy %x %y
	update
    }
}
proc actionSc2DOb { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc2DObac -start %x %y
	update
    }
    bind $w <B1-Motion> {
	%W sc2DObac -winxy %x %y
	update
    }
}
proc actionStr2DObxy { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DXObac -start %x %y
	%W sc1DYObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DXObac -winxy %x %y
	%W sc1DYObac -winxy %x %y
	update
    }
}
proc actionStr2DObzy { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DZObac -start %x %y
	%W sc1DYObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DZObac -winxy %x %y
	%W sc1DYObac -winxy %x %y
	update
    }
}
proc actionStr2DObxz { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W sc1DXObac -start %x %y
	%W sc1DZObac -start %x %y
	update
    }

    bind $w <B1-Motion> {
	%W sc1DXObac -winxy %x %y
	%W sc1DZObac -winxy %x %y
	update
    }
}

proc actionEditPointD { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W pointEditD %x %y
	update
    }
    bind $w <B1-Motion> {}
 
}

proc actionInsertPoint { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W insertPoint %x %y
	update
    }
    bind $w <B1-Motion> {}
 
}

proc actionSplitCurve { w } {

    $w splitCurve -start

    bind $w <ButtonPress-1> {
	saveUndo
	%W splitCurve %x %y
	update
    }
    bind $w <B1-Motion> {}

    bind $w <ButtonRelease-1> {
	%W splitCurve -end
	bind %W <ButtonPress-1> ""
	bind %W <B1-Motion> ""
	viewTitlep [winfo toplevel %W] ""  "None"
	uMS
    }
 
}

proc actionfindU { w } {

    $w findU -start

    bind $w <ButtonPress-1> {
	saveUndo
	%W findU %x %y
	update
    }
    bind $w <B1-Motion> {}

    bind $w <ButtonRelease-1> {

	%W findU -end
	bind %W <ButtonPress-1> ""
	bind %W <B1-Motion> ""
	viewTitlep [winfo toplevel %W] ""  "None"
	global mops_views currentView
	foreach i $mops_views {
	    if { "${i}.f3D.togl" != "%W" } {
		${i}.f3D.togl makecurrent
		${i}.f3D.togl getViewConfig
		if { $view(Type) != 3 } {
		    ${i}.f3D.togl findU -end
		    bind ${i}.f3D.togl <ButtonPress-1> ""
		    bind ${i}.f3D.togl <B1-Motion> ""
		    viewTitlep $i ""  "None"
		}
	    }
	}
	$currentView makecurrent
	$currentView getViewConfig
    }
    
}

proc actionDeletePoint { w } {
    bind $w <ButtonPress-1> {
	saveUndo
	%W deletePoint %x %y
	update
    }
    bind $w <B1-Motion> {}
 
}

proc releaseBindp { w } {
    bind $w.f3D.togl <ButtonRelease-1> {
	if { $prefs(UpdateTrafo) == 1 } { getTransfProp }
	if { $prefs(LazyNotify) == 1 } { forceNot }
	redrawAllViews
    }
}

###################################
# Setup Action-Keys for a 3D-View
#
#
proc setupActionKeys { w viewtype } {
    global mops_view_sc mops_oa_sc mops_va_sc

    if { ($viewtype == "xy") || ($viewtype == "tr")} {
	bind $w <[repcont $mops_oa_sc(MoveO)]> {
	    actionMoveOb %W.f3D.togl
	    releaseBindp $w
	    viewTitlep %W "" MoveXY
	}
	bind $w <[repcont $mops_oa_sc(RotO)]> { 
	    actionRotOb %W.f3D.togl 
	    releaseBindp $w
	    viewTitlep %W "" RotateZ
	}
	bind $w <[repcont $mops_oa_sc(RotA)]> { 
	    actionRotOba %W.f3D.togl 
	    viewTitlep %W "" RotateAboutZ
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Edit)]> {
	    setupPEditBinding %W "xy"
	    viewTitlep %W "" EditXY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Scal2)]> {
	    actionSc2DOb %W.f3D.togl
	    viewTitlep %W "" ScaleXY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Stretch)]> {
	    actionStr2DObxy %W.f3D.togl
	    viewTitlep %W "" StretchXY
	    releaseBindp $w
	}
    }
    if { $viewtype == "yz" } {
	bind $w <[repcont $mops_oa_sc(MoveO)]> { 
	    actionMoveOb  %W.f3D.togl 
	    viewTitlep %W "" MoveYZ
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(RotO)]> { 
	    actionRotOb %W.f3D.togl 
	    viewTitlep %W "" RotateX
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(RotA)]> { 
	    actionRotOba %W.f3D.togl 
	    viewTitlep %W "" RotateAboutX
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Scal2)]> {
	    actionSc2DOb %W.f3D.togl
	    viewTitlep %W "" ScaleZY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Edit)]> {
	    setupPEditBinding %W "zy"
	    viewTitlep %W "" EditZY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Stretch)]> {
	    actionStr2DObzy %W.f3D.togl
	    viewTitlep %W "" StretchZY
	    releaseBindp $w
	}

    }
    if { $viewtype == "xz" } {
	bind $w <[repcont $mops_oa_sc(MoveO)]> {
	    actionMoveOb %W.f3D.togl 
	    viewTitlep %W "" MoveXZ
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(RotO)]> { 
	    actionRotOb %W.f3D.togl 
	    viewTitlep %W "" RotateY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(RotA)]> { 
	    actionRotOba %W.f3D.togl 
	    viewTitlep %W "" RotateAboutY
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Scal2)]> {
	    actionSc2DOb %W.f3D.togl
	    viewTitlep %W "" ScaleXZ
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Edit)]> {
	    setupPEditBinding %W "xz"
	    viewTitlep %W ""  EditXZ
	    releaseBindp $w
	}
	bind $w <[repcont $mops_oa_sc(Stretch)]> {
	    actionStr2DObxz %W.f3D.togl
	    viewTitlep %W "" StretchXZ
	    releaseBindp $w
	}
    }

bind $w <[repcont $mops_oa_sc(Scal3)]> {
    actionSc3DOb %W.f3D.togl
    viewTitlep %W ""  ScaleXYZ
    releaseBindp $w
}

bind $w <[repcont $mops_oa_sc(WeightE)]> {
    setupWEditBinding %W.f3D.togl
    viewTitlep %W ""  EditW
}
bind $w <[repcont $mops_oa_sc(Select)]> {
    setupPSelectBinding %W.f3D.togl
    viewTitlep %W ""  SelectPoints
}
bind $w <[repcont $mops_va_sc(MoveV)]> {
    actionMoveView %W.f3D.togl
    viewTitlep %W "" MoveView
}
bind $w <[repcont $mops_va_sc(ZoomV)]> {
    actionZoomView %W.f3D.togl
    viewTitlep %W "" ZoomView
}
bind $w <[repcont $mops_oa_sc(ScalX)]> {
    actionSc1DObx %W.f3D.togl
    viewTitlep %W "" ScaleX
    releaseBindp $w
}
bind $w <[repcont $mops_oa_sc(ScalY)]> {
    actionSc1DOby %W.f3D.togl
    releaseBindp $w
    viewTitlep %W "" ScaleY
}
bind $w <[repcont $mops_oa_sc(ScalZ)]> {
    actionSc1DObz %W.f3D.togl
    releaseBindp $w
    viewTitlep %W "" ScaleZ
}
bind $w <[repcont $mops_oa_sc(ScalXW)]> {
    actionSc1DWorldx %W.f3D.togl
    viewTitlep %W "" ScaleXWorld
}
bind $w <[repcont $mops_oa_sc(ScalYW)]> {
    actionSc1DWorldy %W.f3D.togl
    viewTitlep %W "" ScaleYWorld
}
bind $w <[repcont $mops_oa_sc(ScalZW)]> {
    actionSc1DWorldz %W.f3D.togl
    viewTitlep %W "" ScaleZWorld
}
bind $w <[repcont $mops_oa_sc(EditD)]> {
    actionEditPointD %W.f3D.togl
    viewTitlep %W "" EditPoint
}
bind $w <[repcont $mops_oa_sc(InsertP)]> {
    actionInsertPoint %W.f3D.togl
    viewTitlep %W "" InsertPoint
    releaseBindp $w
}
bind $w <[repcont $mops_oa_sc(DelP)]> {
    actionDeletePoint %W.f3D.togl
    viewTitlep %W "" DeletePoint
    releaseBindp $w
}
bind $w <[repcont $mops_oa_sc(SplitC)]> {
    actionSplitCurve %W.f3D.togl
    viewTitlep %W "" SplitCurve
    bind $w.f3D.togl <ButtonRelease-1> {
	setSelect; updateMopsSelect; updateMopsPList;
	redrawAllViews
    }
}
bind $w <[repcont $mops_oa_sc(FindU)]> {
    actionfindU %W.f3D.togl
    viewTitlep %W "" FindU
}
bind $w <[repcont $mops_va_sc(RotV)]> {
    actionRotView %W.f3D.togl
    viewTitlep %W "" RotateView
}

bind $w <[repcont $mops_oa_sc(DeSelect)]> {
    $currentView deselectPoints
    redrawAllViews
}

bind $w <[repcont $mops_va_sc(Break)]> {
    bind %W.f3D.togl <ButtonPress-1> ""
    bind %W.f3D.togl <B1-Motion> ""
    viewTitlep %W ""  "None"
}

bind $w <ButtonPress-$mops_va_sc(MoveVButton)> {
    %W movVwac -start %x %y
    update
}

bind $w <B$mops_va_sc(MoveVButton)-Motion> {
    %W movVwac -winxy %x %y
    update
}

bind $w <Shift-ButtonPress-$mops_va_sc(MoveVButton)> {
    %W movzVwac -start %y
    update
}

bind $w <Shift-B$mops_va_sc(MoveVButton)-Motion> {
    %W movzVwac -winy %y
    update
}


bind $w <ButtonPress-$mops_va_sc(ZoomVButton)> {
    %W zomVwac -start %y
    update
}

bind $w <B$mops_va_sc(ZoomVButton)-Motion> {
    %W zomVwac -winy %y
    update
}

global prefs
if {$prefs(zoomWheel) } {

    bind $w <ButtonPress-4> {
	%W setViewConfig -zoomd 0.5
	update
    }

    bind $w <ButtonPress-5> {
	%W setViewConfig -zoomd 2
	update
    }

}


if { $viewtype == "clear" } {
    global mops_oa_sc
    set actions [array names mops_oa_sc]
    foreach i $actions {
	eval "bind \$w \$mops_oa_sc(${i}) \"\""
    }
    bind $w.f3D.togl <B1-Motion> ""
}

}


#
#
#  
proc updateMopsPList { } {
    global propertyList stdpropertyList\
	    stdpropertyProcList propertyProcList\
	    stdpropertyGetProcList propertyGetProcList\
	    stdpropertySetProcList propertySetProcList\
	    stdpropertyDataList propertyDataList prefs

 set lb .fu.fMain.fProp.flb.li
 
 set cursel [.fu.fMain.fProp.flb.li curselection]

 set selobj [.fu.fMain.fHier.flb.li curselection]

 # delete all current entries
 $lb delete 0 end

 # get properties
 if { $selobj != "" } {
     set propertyList $stdpropertyList
     set propertyProcList $stdpropertyProcList
     set propertyGetProcList $stdpropertyGetProcList
     set propertySetProcList $stdpropertySetProcList
     set propertyDataList $stdpropertyDataList
     eval [subst "setSelect $selobj"]
     redrawAllViews
 }

 # insert properties
 if { $propertyList != "" } {

    eval [subst "$lb insert end $propertyList"]

 }

 if { $cursel != "" } {
     $lb selection set $cursel
     set prefs(thisPropGUIStatic) 0
     updatePropGUI
 } else {
 # clear old propGUI
 set w .fu.fMain.fArg.flb.ca.w
 catch {.fu.fMain.fArg.flb.ca delete all}
 destroy $w
 update
}
 return;
}


#
#
#
proc updateMopsSelect {} {
global curlevel propertyList

set lb .fu.fMain.fHier.flb.li

set selobj [$lb curselection]

update

# delete all current entries
$lb delete 0 end

# get current objects
getCLevel curlevel

eval [subst "$lb insert end $curlevel"]

if { $selobj != "" } {
    foreach i $selobj {
	$lb selection set $i
    }
}

#setSelect

#updateMopsPList

return;
}


#
#
#
proc selectNext { } {
    set lb .fu.fMain.fHier.flb.li
    set cs [lindex [$lb curselection] 0]
    incr cs
    if { $cs <= [$lb index end] } {
	$lb selection clear 0 end
	$lb selection set $cs
	setSelect $cs
    }
 return;
}

#
#
#
proc clearSelect { } {

    .fu.fMain.fHier.flb.li selection clear 0 end
    setSelect
    updateMopsSelect
    updateMopsPList
    
 return;
}

#
#
#
proc selectLast { } {
    set lb .fu.fMain.fHier.flb.li
    $lb selection clear 0 end
    setSelect
    $lb selection set end
    setSelect [$lb index end]
    updateMopsPList
    
 return;
}

#
#
#
proc forAll { recursive command } {
    set lb .fu.fMain.fHier.flb.li

    set selection [$lb curselection]
    if { $selection == "" } {
	$lb selection set 1 end
	set selection [$lb curselection]
    }
    $lb selection clear 0 end
    updateMopsSelect
    set i 0
    foreach sel $selection {
	cancelled
	incr i
	$lb selection clear 0 end
	$lb selection set $sel
	setSelect $sel
	updateMopsPList
	# go down?
	if { $recursive } {
	    if { [hasChild] == 1} {
		goDown $sel
		updateMopsSelect
		$lb selection clear 0 end
		setSelect
		forAll 1 $command
		goUp
		updateMopsSelect
		$lb selection set $sel
		setSelect $sel
		updateMopsPList
	    }
	}
	$lb selection clear 0 end
	$lb selection set $sel
	setSelect $sel
	updateMopsPList
	eval $command
    }

    updateMopsSelect

 return;
}

#
#
#
proc forAllT { type recursive command } {
    set lb .fu.fMain.fHier.flb.li

    set selection [$lb curselection]
    if { $selection == "" } {
	$lb selection set 1 end
	set selection [$lb curselection]
    }
    $lb selection clear 0 end
    updateMopsSelect
    set i 0
    foreach sel $selection {
	cancelled
	incr i
	$lb selection clear 0 end
	$lb selection set $sel
	setSelect $sel
	updateMopsPList
	# go down?
	if { $recursive } {
	    if { [hasChild] == 1} {
		goDown $sel
		updateMopsSelect
		$lb selection clear 0 end
		setSelect
		forAllT $type 1 $command
		goUp
		updateMopsSelect
		$lb selection set $sel
		setSelect $sel
		updateMopsPList
	    }
	}
	$lb selection clear 0 end
	$lb selection set $sel
	setSelect $sel
	updateMopsPList
	global otype
	set otype 0
	getType otype
	if { $otype == $type } {
	    eval $command
	}
    }

    updateMopsSelect

 return;
}

# Procedures for View-Windows

##############################
# setGridp:
# 
#
proc setGridp { w } {
set w .setGrid
catch {destroy $w}
toplevel $w
wm title $w "Mops - Set GridSize"
wm iconname $w "Mops"

set f [frame $w.f1]
pack $f -in $w -side top -fill x

set f [frame $f.fgr]

label $f.l -text "New Grid-Size:"
button $f.b1 -pady 0  -text "/2" -command {
    set val [.setGrid.f1.fgr.e get]
    .setGrid.f1.fgr.e delete 0 end
    .setGrid.f1.fgr.e insert 0 [expr ($val/2.0)] } -takefocus 0
    entry $f.e -width 8
button $f.b2 -pady 0  -text "*2" -command {
    set val [.setGrid.f1.fgr.e get]
    .setGrid.f1.fgr.e delete 0 end
    .setGrid.f1.fgr.e insert 0 [expr ($val*2.0)] } -takefocus 0

global view
$f.e insert @0 $view(Grid)

pack $f.l $f.b1 $f.e $f.b2 -in $f -padx 2 -pady 2 -side left
pack $f -in $w.f1 -side top
set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command {
    global currentView view
    $currentView setViewConfig -grid [.setGrid.f1.fgr.e get]
    $currentView getViewConfig
    $currentView render
    grab release .setGrid
    destroy .setGrid
}

button $f.bca -text "Cancel" -width 5 -command "grab release .setGrid; destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x

centerWindowp $w
grab $w
focus .setGrid.f1.fgr.e
tkwait window $w
global currentView
focus $currentView

return;
}


##############################
# closeMopsView:
# 
#
proc closeMopsView { w } {
  global mops_views currentView

  # delete view from list
  set temp ""
  set currentView ""
  foreach view $mops_views {
    if { $view != $w } {
	set temp [concat $temp $view]
	set currentView ${view}.f3D.togl
    }
  }

  set mops_views $temp

  catch {$currentView makecurrent}
  update

  # now delete the window
  destroy $w.f3D.togl
  destroy $w

  updateMopsSelect

  global MOPSIDR
  if { $MOPSIDR == 1 } {
    openIDR
  }

 return;
}


##############################
# openMopsView:
# 
#
proc openMopsView { w width height } {
global tcl_platform

catch {destroy $w}
toplevel $w

global mops_views mops_view_sc

scan $w ".%s" name
wm iconname $w $name

frame $w.fMenu -bd 2 -relief raised
frame $w.f3D

pack $w.fMenu -side top -fill x

menubutton $w.fMenu.v -text "View" -menu $w.fMenu.v.m
menu $w.fMenu.v.m
$w.fMenu.v.m add command\
-label "Quick Render"\
-command {\
global env prefs currentView mops_tempfiles

getTmpFile $prefs(tmpdir) tmpfile

 $currentView wrib -file $tmpfile -temp

 global tcl_platform
 if { $tcl_platform(platform) == "windows" } {
    # Windows sucks big time!
    regsub -all {\\} $tmpfile {\\\\\\\\\\\\\\\\} tmpfile
 }

 lappend mops_tempfiles $tmpfile

 set command "exec "

 regsub -all {%s} $prefs(QRender) $tmpfile command2

 append command $command2
 append command " &"

 eval [subst "$command"]

 update
 cleanTmpFilesp 0

}

$w.fMenu.v.m add command\
-label "Render"\
-command {\
 global env prefs currentView mops_tempfiles

 getTmpFile $prefs(tmpdir) tmpfile

 $currentView wrib -file $tmpfile -image ${tmpfile}.tiff -temp

 global tcl_platform
 if { $tcl_platform(platform) == "windows" } {
    # Windows sucks big time!
    regsub -all {\\} $tmpfile {\\\\\\\\\\\\\\\\} tmpfile
 }

 lappend mops_tempfiles $tmpfile

 set command "exec "

 regsub -all {%s} $prefs(Render) $tmpfile command2

 append command $command2
 append command " &"

 eval [subst "$command"]

 update

 cleanTmpFilesp 0

}

$w.fMenu.v.m add command -label "Redraw" -command {
    global currentView view
    set redraw $view(Redraw)
    $currentView setViewConfig -draw 1
    $currentView render
    $currentView setViewConfig -draw $redraw
}

$w.fMenu.v.m add separator

$w.fMenu.v.m add command -label "Close" -command "closeMopsView $w"

#####

menubutton $w.fMenu.c -text "Configure" -menu $w.fMenu.c.m
menu $w.fMenu.c.m
set m $w.fMenu.c.m.tp
$w.fMenu.c.m add cascade -label "Viewtype" -menu $m
menu $m
$m add command -label "Front" -command {
    global currentView view
    set togl $currentView
    set w [winfo toplevel $togl]
    $togl getViewConfig
    if { $view(Type) == 3 } {
	$togl setViewConfig -zoom [expr $view(Zoom)*12]
    }
    $togl setViewConfig -type 0 -fromx 0.0 -fromy 0.0 -fromz 10.0\
	    -tox 0.0 -toy 0.0 -toz 0.0 -upx 0.0 -upy 1.0 -upz 0.0

    update
    setupActionKeys $w clear
    setupActionKeys $w xy
    viewTitlep $w Front None
    $togl render
    updateMopsPList
}
$m add command -label "Side" -command {
    global currentView view
    set togl $currentView
    set w [winfo toplevel $togl]
    $togl getViewConfig
    if { $view(Type) == 3 } {
	$togl setViewConfig -zoom [expr $view(Zoom)*12]
    }
    $togl setViewConfig -type 1 -fromx 10.0 -fromy 0.0 -fromz 0.0\
	    -tox 0.0 -toy 0.0 -toz 0.0 -upx 0.0 -upy 1.0 -upz 0.0
    update
    setupActionKeys $w clear
    setupActionKeys $w yz
    viewTitlep $w Side None
    $togl render
    updateMopsPList
}
$m add command -label "Top" -command {
    global currentView view
    set togl $currentView
    set w [winfo toplevel $togl]
    $togl getViewConfig
    if { $view(Type) == 3 } {
	$togl setViewConfig -zoom [expr $view(Zoom)*12]
    }
    $togl setViewConfig -type 2 -fromx 0.0 -fromy 10.0 -fromz 0.0\
	    -tox 0.0 -toy 0.0 -toz 0.0 -upx 0.0 -upy 0.0 -upz -1.0
    update
    setupActionKeys $w clear
    setupActionKeys $w xz
    viewTitlep $w Top None
    $togl render
    updateMopsPList
}
$m add separator
$m add command -label "Perspective" -command {
    global currentView view
    set togl $currentView
    set w [winfo toplevel $togl]
    $togl getViewConfig
    if { $view(Type) != 3 } {
	$togl setViewConfig -zoom [expr $view(Zoom)/12]
    }
    $togl setViewConfig -type 3 -fromx 0.0 -fromy 0 -fromz 15.0\
	    -tox 0.0 -toy 0.0 -toz 0.0  -upx 0.0 -upy 1.0 -upz 0.0\
	    -drawg 1 -grid 1 -drotx 45 -droty -30
    update
    setupActionKeys $w clear
    viewTitlep $w Persp None
    $togl render
    updateMopsPList
}
$m add separator
$m add command -label "Trim" -command {
    global currentView
    set togl $currentView
    set w [winfo toplevel $togl]
    $togl setViewConfig -type 4 -fromx 0.0 -fromy 0.0 -fromz 10.0\
	    -tox 0.0 -toy 0.0 -toz 0.0 -upx 0.0 -upy 1.0 -upz 0.0\
	    -zoom 1.0
    update
    setupActionKeys $w clear
    setupActionKeys $w tr
    viewTitlep $w Trim None
    update
    $togl render
    updateMopsPList
}

set m $w.fMenu.c.m.drw
$w.fMenu.c.m add cascade -label "Draw what" -menu $m
menu $m
global view
$m add check -label "Wireframe" -variable view(DrawWire) -command {
    global currentView view
    $currentView setViewConfig -draww $view(DrawWire)
    updateMopsPList
}
$m add check -label "Curve/Surface" -variable view(DrawSurf) -command {
    global currentView view
    $currentView setViewConfig -draws $view(DrawSurf)
    updateMopsPList
}
$m add separator
$m add check -label "Draw Grid" -variable view(DrawGrid) -command {
    global currentView view
    $currentView setViewConfig -drawg $view(DrawGrid)
    updateMopsPList
}
set m $w.fMenu.c.m
$m add command -label "Set Gridsize" -command "\
setGridp $w; focus $w
"
$m add check -label "Draw Selection only" -variable view(DrawSel)\
	-command {
    global currentView view
    $currentView setViewConfig -dsel $view(DrawSel) 
    updateMopsPList
}
$m add check -label "Lighting" -variable view(Lighting)\
	-command {
    global currentView view
    $currentView setViewConfig -light $view(Lighting)
    updateMopsPList
}
$m add check -label "Automatic Redraw" -variable view(Redraw)\
	-command {
    global currentView view
    $currentView setViewConfig -draw $view(Redraw)
    updateMopsPList
}
$m add separator
$m add command -label "Half Size" -command {
    global currentView
    upvar #0 view localview

    set localview(Width) [expr ($localview(Width) / 2)]
    set localview(Height) [expr ($localview(Height) / 2)]
    set hview $currentView
    wm geometry [winfo toplevel $hview] ""
    $hview configure -width $localview(Width)\
	    -height $localview(Height)
    update
    $hview getViewConfig
    $hview setViewConfig
    $hview makecurrent
    $hview render
    update
    global $currentView
    $currentView makecurrent
    $currentView getViewConfig
}

$m add command -label "Double Size" -command {
    global currentView
    upvar #0 view localview

    set localview(Width) [expr ($localview(Width) * 2)]
    set localview(Height) [expr ($localview(Height) * 2)]
    set hview $currentView
    wm geometry [winfo toplevel $hview] ""
    $currentView configure -width $localview(Width)\
	    -height $localview(Height)
    update
    $hview getViewConfig
    $hview setViewConfig
    $hview makecurrent
    $hview render
    update
    global $currentView
    $currentView makecurrent
    $currentView getViewConfig
}

$m add separator

$m add command -label "Zoom to Object" -command { global currentView;
$currentView zoomToOb }

if { $tcl_platform(platform) == "windows" } {
$w.fMenu.v configure -pady 1
$w.fMenu.c configure -pady 1
}

pack $w.fMenu.v $w.fMenu.c -in $w.fMenu -side left

#####

togl $w.f3D.togl -rgba true -double true -depth true\
	-ident $name -width $width -height $height

set mops_views [concat $mops_views $w]

pack $w.f3D -in $w -fill both -expand yes
pack $w.f3D.togl -in $w.f3D -fill both -expand yes
update
global currentView
set currentView $w.f3D.togl

# default config (Front, Zoom 3, Draw Wireframe)
$w.fMenu.c.m.tp invoke 1

$w.f3D.togl setViewConfig -draww 1 -draw 1 -zoom 3

if { [winfo exist .iw] } {
  global MOPSIDR
  if { $MOPSIDR == 1 } {
    openIDR
  }
}

update

return;
}


##############################
# newViewp:
# create new 3D-View
#
proc newViewp { width height } {
 global mops_views oview_sema

 set oview_sema 1
 update

 if { [llength $mops_views] == 0 } {
  set vname .view1
 } else {
  set lastname [lindex $mops_views end]
  scan $lastname ".view%d" lastid
  set vname .view[expr $lastid + 1]
 }

 openMopsView $vname $width $height

 setupKeyBindings3D $vname

 update

 updateMopsSelect

 set oview_sema 0
 update

 return;
}


##############################
# redrawAllViews: 
# 
#
proc redrawAllViews {} {
 global mops_views currentView

    catch {set tmp $currentView}

    foreach view $mops_views {
	if { [winfo ismapped $view] } {
	    set view ${view}.f3D.togl
#	    $view makecurrent
	    $view render
	}
    }

    catch {$tmp makecurrent}
    catch {set currentView $tmp}

 return;
}


##############################
# closeAllViews: 
# 
#
proc closeAllViews {} {
 global mops_views currentView

    foreach view $mops_views {
	closeMopsView $view
    }

    set mops_views ""
    set currentView ""
    update
 return;
}



# File-Menu

##############################
# newScenep:
# clear Scene
#
proc newScenep { } {
 global filename

 set filename ""

 goTop
 updateMopsSelect
 newScene
 updateMopsSelect
 redrawAllViews

 return;
}


##############################
# replaceScenep:
# clear Scene, read 
#
proc replaceScenep { } {
global filename

if { $filename == "" } {
    set dirname [pwd]
} else {
    set dirname [file dirname $filename]
    if { $dirname == "." } { set dirname [pwd] }
}

set types {{"Mops Scene" ".mop"} {"All files" *}}
    
set newfilename [tk_getOpenFile -filetypes $types -parent .\
	-initialfile [file tail $filename] -initialdir $dirname\
	-title "Select file to load:"]

if { $newfilename != "" } {
    closeAllViews
    .fu.fMain.fHier.fsel.bnon invoke
    update
    set filename $newfilename
    replaceScene $filename
    updateMopsSelect
    redrawAllViews
    set windowfilename [file tail [file rootname $filename]]
    wm title . "The Mops - Main - $windowfilename"

}
}


##############################
# insertScenep:
#
#
proc insertScenep { } {
global filename

set types {{"Mops Scene" ".mop"} {"All files" *}}

if { $filename == "" } {
    set dirname [pwd]
} else {
    set dirname [file dirname $filename]
    if { $dirname == "." } { set dirname [pwd] }
}
set types {{"Mops Scene" ".mop"} {"All files" *}}

set ifilename [tk_getOpenFile -filetypes $types -parent .\
	-initialfile [file tail $filename] -initialdir $dirname\
	-title "Select file to load:"]

if { $ifilename != "" } {
    .fu.fMain.fHier.fsel.bnon invoke
    update
    insertScene $ifilename
    updateMopsSelect
    redrawAllViews
}
}


##############################
# saveScenep:
# 
#
proc saveScenep { ask } {
    global filename
    set tmp $filename

    if { ($ask == "ask") || ($filename == "") } {
	if { $filename == "" } {
	    set filename "unnamed.mop"
	    set dirname [pwd]
	} else {
	    set dirname [file dirname $filename]
	    if { $dirname == "." } { set dirname [pwd] }
	}
	set types {{"Mops Scene" ".mop"} {"All files" *}}
	set filename [tk_getSaveFile -filetypes $types -parent .\
		-initialfile [file tail $filename]\
		-initialdir $dirname -title "Select file to save to:"]
    }

if { $filename != "" } {
    # fix window positions
    updatePositionsp
    saveScene $filename
    set windowfilename [file tail [file rootname $filename]]
    wm title . "The Mops - Main - $windowfilename"
} else { set filename $tmp }

}


##############################
# exportRIBp:
# 
# 
proc exportRIBp { } {
set w .exportRIBw
catch {destroy $w}
toplevel $w
wm title $w "Mops - Export RIB"
wm iconname $w "Mops"
wm transient $w .

set f [frame $w.f1]
pack $f -in $w -side top -fill x

label $f.l -text "Select View:"
set f [frame $f.fview]
scrollbar $f.sc -command {.exportRIBw.f1.fview.lb yview} -takefocus 0
listbox $f.lb -width 15 -height 5 -selectmode single\
	-exportselection 0 -yscrollcommand {.exportRIBw.f1.fview.sc set}
global mops_views currentView view

set index 0
set firstselect 0
foreach i $mops_views {
    eval [subst "$i.f3D.togl getViewConfig"]
    if { $view(TypeStr) == "Persp" } { set firstselect $index }
    scan $i ".%s" name
    if { $view(TypeStr) != "Trim" } {
	$f.lb insert end "$name ($view(TypeStr))"
	incr index
    }
}

catch {eval [subst "$currentView getViewConfig"]}

$f.lb selection set $firstselect

pack $f.lb -in $f -side left -fill x -fill y
pack $f.sc -in $f -side right -fill y
pack $w.f1.l -in $w.f1 -side top
pack $f -in $w.f1 -side top -fill x -fill y


set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command { 
    global currentView view prefs filename
    set selection [.exportRIBw.f1.fview.lb curselection]
    if {$selection != ""} {
	set sname [.exportRIBw.f1.fview.lb get $selection]
	scan $sname "%s (%s)" name type
	if { $prefs(RIBFile) == "Ask" || $prefs(RIBFile) == "" } {
	    set filetypes {{"RIB" ".rib"} {"All files" *}}
	    set efilename [tk_getSaveFile -filetypes $filetypes -parent .\
		    -title "Export to file:"]
	} else {
	    if { $prefs(RIBFile) == "Scene" } {
		if { $filename != "" } {
		    set efilename [file rootname $filename].rib
		} else {
		    set efilename unnamed.rib
		}
	    } else {
		set efilename $prefs(RIBFile)
	    }
	}

	if { $prefs(Image) == "Ask"} {
	    set filetypes {{"TIFF" ".tiff"} {"TIF" ".tif"} {"All files" *}}
	    set imagename "[tk_getSaveFile -filetypes $filetypes -parent .\
		    -title "Render to file:"]"
	    if { $imagename == "" } { set efilename "" }
	} else {
	    if { $prefs(Image) == "RIB" } {
		set imagename [file rootname $efilename].tiff
	    } else {
		set imagename $prefs(Image)
	    }
	}

	if { $efilename != ""} {
	    if { $imagename != "" } {
		.$name.f3D.togl wrib -file $efilename -image $imagename
	    } else {
		.$name.f3D.togl wrib -file $efilename
	    }
	}
    }
    focus .
    grab release .exportRIBw
    destroy .exportRIBw
}

button $f.bca -text "Cancel" -width 5 -command "destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x

centerWindowp $w
grab $w
focus $f.bok
tkwait window $w

return;
}

# lc - load custom object:
# 
# 
proc lc { filename } {

    getCustomObjects customnames
    set num [llength $customnames]

    set sopath [file dirname $filename]
    if { "$sopath" != "" } {
	set oldcdir [pwd]
	cd $sopath
	load $filename
	cd $oldcdir
    } else {
	load $filename
    }
    update
    set customnames ""
    getCustomObjects customnames
    if { $num < [llength $customnames] } {

	set co [lindex $customnames 0]
	.fu.fMenu.cr.m.cus add command -label $co\
	    -command "crtOb custom $co; updateMopsSelect; selectLast; redrawAllViews"
    }
 return;
}


##############################
# loadCustomp:
# 
# 
proc loadCustomp { } {
    global prefs

    set filetypes {{"Custom Object" ".so"} {"Custom Object" ".dll"} {"All files" *}}

    set filename [tk_getOpenFile -filetypes $filetypes -parent .\
		    -title "Select Custom Object:"]

    if { $filename != "" } { 
	lc $filename
    }

 return;
}


# Tools Menu
##############################
# nameObp:

proc nameObp { } {
set w .nameOw
catch {destroy $w}
toplevel $w
wm title $w "Mops"
wm iconname $w "Mops"
wm transient $w .

set f [frame $w.f1]
pack $f -in $w -side top -fill x

set f [frame $f.fnm]

label $f.l -text "Name"
entry $f.e -width 12
set lb .fu.fMain.fHier.flb.li
set selected [$lb curselection]
if { $selected == "" } { destroy $w; return; }
set oldname [$lb get [lindex $selected 0]]
$f.e insert @0 $oldname

pack $f.l $f.e -in $f -padx 2 -pady 2 -side left
pack $f -in $w.f1 -side top
set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command { 
    setON [.nameOw.f1.fnm.e get]
    updateMopsSelect
    redrawAllViews
    grab release .nameOw
    focus .
    destroy .nameOw
}

button $f.bca -text "Cancel" -width 5 -command "destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x

centerWindowp $w
grab $w
focus $w.f1.fnm.e
tkwait window $w

return;
}


##############################
#
#
proc crtRevolvep { } {
    global mops_error
    set mops_error 0
    set lb .fu.fMain.fHier.flb.li
    set selected [$lb curselection]
    if { $selected == "" } { return }
    crtOb custom Revolve
    if { $mops_error } { return; }
    $lb selection clear 0 end
    eval [subst "$lb selection set [lindex $selected 0]"]
    eval [subst "setSelect [lindex $selected 0]"]
    cutOb
    updateMopsSelect
    goDown [expr [$lb index end] -1 ]
    pasOb
    goUp
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}

##############################
#
#
proc crtICurvep { } {
    global mops_error nclen
    set mops_error 0
    set lb .fu.fMain.fHier.flb.li
    set selected [$lb curselection]
    if { $selected == "" } { 
	crtOb ncurve -length $nclen; uMS; sL; set selected [$lb curselection]
    }
    crtOb custom ICurve
    if { $mops_error } { return; }
    $lb selection clear 0 end
    eval [subst "$lb selection set [lindex $selected 0]"]
    eval [subst "setSelect [lindex $selected 0]"]
    cutOb
    updateMopsSelect
    goDown [expr [$lb index end] -1 ]
    pasOb
    goUp
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}


##############################
#
#
proc crtExtrudep { } {
    global mops_error
    set mops_error 0
    set lb .fu.fMain.fHier.flb.li
    set selected [$lb curselection]
    if { $selected == "" } { return }
    crtOb custom Extrude
    if { $mops_error } { return; }
    $lb selection clear 0 end
    foreach i $selected {
	eval [subst "$lb selection set $i"]
    }
    eval [subst "setSelect $selected"]
    cutOb
    updateMopsSelect
    goDown [expr [$lb index end] -1 ]
    pasOb
    goUp
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}


##############################
#
#
proc crtSweepp { } {
    global mops_error
    set mops_error 0
    set lb .fu.fMain.fHier.flb.li
    set selected [$lb curselection]
    if { $selected == "" } { return }
    crtOb custom Sweep 4
    if { $mops_error } { return; }
    $lb selection clear 0 end
    eval [subst "$lb selection set [lindex $selected 0] [lindex $selected 1]"]
    eval [subst "setSelect [lindex $selected 0] [lindex $selected 1]"]
    cutOb
    updateMopsSelect
    goDown [expr [$lb index end] -1 ]
    pasOb
    goUp
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}
##############################
#
#
proc crtSkinp { } {
    global mops_error
    set mops_error 0
    set lb .fu.fMain.fHier.flb.li
    set selected [$lb curselection]
    if { $selected == "" } { return }
    crtOb custom Skin
    if { $mops_error } { return; }
    $lb selection clear 0 end
    foreach i $selected {
	eval [subst "$lb selection set $i"]
    }
    eval [subst "setSelect $selected"]
    cutOb
    updateMopsSelect
    goDown [expr [$lb index end] -1 ]
    pasOb
    goUp
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}

##############################
# newLightp:
# create a new Light
proc newLightp { } {

    crtOb light
    updateMopsSelect
    selectLast
    redrawAllViews

return;
}

##############################
# editPointDp:
# directly edit coordinates
# 
proc editPointDp { } {
upvar #0 editPointDarray array

set w .editPointDw
catch {destroy $w}
toplevel $w
wm title $w "Mops - Edit Points"
wm iconname $w "Mops"

set f [frame $w.f1]
pack $f -in $w -side top -fill x

set f [frame $f.fx]

label $f.l -text "X" -width 4
entry $f.e -width 8
$f.e insert @0 $array(x)

pack $f.l $f.e -in $f -padx 2 -pady 2 -side left -fill x -expand yes
pack $f -in $w.f1 -side top  -fill x


set f $w.f1
set f [frame $f.fy]

label $f.l -text "Y" -width 4
entry $f.e -width 8
$f.e insert @0 $array(y)

pack $f.l $f.e -in $f -padx 2 -pady 2 -side left -fill x -expand yes
pack $f -in $w.f1 -side top  -fill x


set f $w.f1
set f [frame $f.fz]

label $f.l -text "Z" -width 4
entry $f.e -width 8
$f.e insert @0 $array(z)

pack $f.l $f.e -in $f -padx 2 -pady 2 -side left -fill x -expand yes
pack $f -in $w.f1 -side top  -fill x

set f $w.f1
set f [frame $f.fw]

label $f.l -text "W" -width 4
entry $f.e -width 8
$f.e insert @0 $array(w)

pack $f.l $f.e -in $f -padx 2 -pady 2 -side left -fill x -expand yes
pack $f -in $w.f1 -side top  -fill x

set f [frame $w.f2]
button $f.bok -text "Ok" -width 5 -command { 
    upvar #0 editPointDarray array
    
    set array(x) [.editPointDw.f1.fx.e get]
    set array(y) [.editPointDw.f1.fy.e get]
    set array(z) [.editPointDw.f1.fz.e get]
    set array(w) [.editPointDw.f1.fw.e get]

    grab release .editPointDw
    destroy .editPointDw
}

button $f.bca -text "Cancel" -width 5 -command "destroy $w"

pack $f.bok $f.bca -in $f -side left -fill x -expand yes
pack $f -in $w -side bottom -fill x

centerWindowp $w
grab $w
focus $f.bok
tkwait window $w

return;
}

######################
# balloon
# original code by Richard Suchenwirth
proc balloon {w help} {
    global balloonwin balloontime prefs
    set balloonwin ""
    set balloontime $prefs(balloon)

# 27.07.1999, #AmigaGER (IRCNet):
# <coto>Naja, tcl/tk ist eben geil :->
# ( realizing that I implemented the browse-balloons-feature in < 5 min ) 

    bind $w <Enter> "
     if { \$balloonwin == \"\" } {
      after \$balloontime {set balloonwin %W ; balloon:show %W [list $help] };
     }"

    bind $w <Leave> "
	if { \$balloonwin != \"\" } {
	    set balloontime 100;
	    set balloonwin \"\";
	}
	destroy %W.balloon;\
	after cancel {set balloonwin %W ; balloon:show %W [list $help] };
	after 500 [list set balloontime \$prefs(balloon)]"
}
proc balloon:show {w arg} {
    if {[eval winfo containing  [winfo pointerxy .]]!=$w} {return}
    set top $w.balloon
    catch {destroy $top}
    toplevel $top -bd 1 -bg black
    wm overrideredirect $top 1
    pack [message $top.txt -aspect 10000 -bg lightyellow -text $arg] 

    set wmx [winfo rootx $w]
    # don't move offscreen
    set rw [winfo reqwidth $top.txt]
    if { [expr $wmx + $rw] > [winfo screenwidth $top] } {
	set wmx [expr $wmx - (([expr $wmx + $rw])-[winfo screenwidth $top])]
    }

    set wmy [expr [winfo rooty $w]+[winfo height $w]]
    wm geometry $top \
	    [winfo reqwidth $top.txt]x[winfo reqheight $top.txt]+$wmx+$wmy
    raise $top
}

######################
# The Toolbox
proc openToolBoxp { } {
    global prefs toolBoxList mops_oa_sc mops_va_sc

    set w .tb
    catch {destroy $w}

    set bp "$prefs(MOPSLIB)"

    global MOPSWRAPPED
    if { $MOPSWRAPPED == 0 } {
	puts stdout "Looking for icon images:"
	set choices $bp
	lappend choices [file dirname [info nameofexecutable]]/../lib/
	lappend choices [file dirname [info nameofexecutable]]/../buttons/
	lappend choices [file dirname [info nameofexecutable]]/lib/
	lappend choices [file dirname [info nameofexecutable]]/buttons/

	set bp ""
	foreach choice $choices {
	    puts -nonewline stdout "${choice}... "
	    if  { [file exists $choice/box.gif ] } {
		set bp $choice
		break
	    }
	    puts stderr "No."
	}
	if { $bp != "" } {
	    puts stdout "Yes."
	} else {
	    puts stderr "Could not find icon images!"
	    puts stderr "Please set the environment variable MOPSLIB."
	    return;
	}
    }

    toplevel $w -visual truecolor
    wm title $w "Tools"
    wm iconname $w "Tools"
    wm group .

    if { $MOPSWRAPPED == 1 } {
	set files { box sphere cylinder cone disk torus hyperb parab ncurve ncircle nrect npatch move rot rota scal2d scal3d scal1dx scal1dy scal1dz edit editw editd tag insert delete revert split view rotv movev zoomv revolve extrude sweep skin icurve show convert notify findu refine concat clamp }

	foreach i $files {
	    set f [wrap::open lib/${i}.gif]
	    set hello [read $f]
	    close $f
	    image create photo ${i}img -format GIF -data $hello
	}
    

    } else {

    image create photo boximg -format GIF\
	    -file "$bp/box.gif"
    image create photo sphereimg -format GIF\
	    -file "$bp/sphere.gif"
    image create photo cylinderimg -format GIF\
	    -file "$bp/cylinder.gif"
    image create photo coneimg -format GIF\
	    -file "$bp/cone.gif"

    image create photo diskimg -format GIF\
	    -file "$bp/disk.gif"
    image create photo torusimg -format GIF\
	    -file "$bp/torus.gif"
    image create photo hyperbimg -format GIF\
	    -file "$bp/hyperb.gif"
    image create photo parabimg -format GIF\
	    -file "$bp/parab.gif"

    image create photo ncurveimg -format GIF\
	    -file "$bp/ncurve.gif"
    image create photo ncircleimg -format GIF\
	    -file "$bp/ncircle.gif"
    image create photo nrectimg -format GIF\
	    -file "$bp/nrect.gif"
    image create photo npatchimg -format GIF\
	    -file "$bp/npatch.gif"

    image create photo moveimg -format GIF\
	    -file "$bp/move.gif"
    image create photo rotimg -format GIF\
	    -file "$bp/rot.gif"
    image create photo rotaimg -format GIF\
	    -file "$bp/rota.gif"
    image create photo scal2dimg -format GIF\
	    -file "$bp/scal2d.gif"
    image create photo scal3dimg -format GIF\
	    -file "$bp/scal3d.gif"
    image create photo scal1dximg -format GIF\
	    -file "$bp/scal1dx.gif"
    image create photo scal1dyimg -format GIF\
	    -file "$bp/scal1dy.gif"
    image create photo scal1dzimg -format GIF\
	    -file "$bp/scal1dz.gif"

    image create photo editimg -format GIF\
	    -file "$bp/edit.gif"
    image create photo editwimg -format GIF\
	    -file "$bp/editw.gif"
    image create photo editdimg -format GIF\
	    -file "$bp/editd.gif"
    image create photo tagimg -format GIF\
	    -file "$bp/tag.gif"

    image create photo insertimg -format GIF\
	    -file "$bp/insert.gif"
    image create photo deleteimg -format GIF\
	    -file "$bp/delete.gif"
    image create photo revertimg -format GIF\
	    -file "$bp/revert.gif"
    image create photo splitimg -format GIF\
	    -file "$bp/split.gif"

    image create photo viewimg -format GIF\
	    -file "$bp/view.gif"
    image create photo rotvimg -format GIF\
	    -file "$bp/rotv.gif"
    image create photo movevimg -format GIF\
	    -file "$bp/movev.gif"
    image create photo zoomvimg -format GIF\
	    -file "$bp/zoomv.gif"

    image create photo revolveimg -format GIF\
	    -file "$bp/revolve.gif"
    image create photo extrudeimg -format GIF\
	    -file "$bp/extrude.gif"
    image create photo sweepimg -format GIF\
	    -file "$bp/sweep.gif"
    image create photo skinimg -format GIF\
	    -file "$bp/skin.gif"
    image create photo icurveimg -format GIF\
	    -file "$bp/icurve.gif"

    image create photo showimg -format GIF\
	    -file "$bp/show.gif"
    image create photo convertimg -format GIF\
	    -file "$bp/convert.gif"
    image create photo notifyimg -format GIF\
	    -file "$bp/notify.gif"


    image create photo finduimg -format GIF\
	    -file "$bp/findu.gif"
    image create photo refineimg -format GIF\
	    -file "$bp/refine.gif"
    image create photo concatimg -format GIF\
	    -file "$bp/concat.gif"
    image create photo clampimg -format GIF\
	    -file "$bp/clamp.gif"

    }

    #########
    global prefs
    foreach i $prefs(toolBoxList) {

     if { $i == "solids" } {
	 set f [frame $w.fsol1 ]

	 button $f.bb -image boximg -padx 0 -pady 0 -command {
	   crtOb box; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bb "create Box"
	 pack $f.bb -in $f -side left
	 button $f.bs -image sphereimg -padx 0 -pady 0 -command {
	   crtOb quadric sphere; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bs "create Sphere"
	 pack $f.bs -in $f -side left
	 button $f.bcy -image cylinderimg -padx 0 -pady 0 -command {
	   crtOb quadric cylinder; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bcy "create Cylinder"
	 pack $f.bcy -in $f -side left
	 button $f.bco -image coneimg -padx 0 -pady 0 -command {
	   crtOb quadric cone; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bco "create Cone"
	 pack $f.bco -in $f -side left

	 pack $f -in $w -side top
	 #########
	 set f [frame $w.fsol2 ]
	 button $f.bd -image diskimg -padx 0 -pady 0 -command {
	     crtOb quadric disk; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bd "create Disk"
	 pack $f.bd -in $f -side left
	 button $f.bto -image torusimg -padx 0 -pady 0 -command {
	     crtOb quadric torus; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bto "create Torus"
	 pack $f.bto -in $f -side left
	 button $f.bhy -image hyperbimg -padx 0 -pady 0 -command {
	crtOb quadric hyperboloid; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bhy "create Hyperboloid"
	 pack $f.bhy -in $f -side left

	 button $f.bpar -image parabimg -padx 0 -pady 0 -command {
	 crtOb quadric paraboloid; updateMopsSelect; selectLast; redrawAllViews
	 }
	 balloon $f.bpar "create Paraboloid"
	 pack $f.bpar -in $f -side left

	 pack $f -in $w -side top -anchor w
	 #########
	 set f [frame $w.fbarsol -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes
	 #########

     }

     if { $i == "crnurb" } {
	 set f [frame $w.fcrn ]

	 button $f.bnc -image ncurveimg -padx 0 -pady 0 -command {
	   global nclen; crtOb ncurve -length $nclen;
	     uMS; sL; redrawAllViews
	 }
	 balloon $f.bnc "create NURB curve\nShift: create ICurve"
	 pack $f.bnc -in $f -side left
	 bind $f.bnc <Shift-ButtonPress-1> {
	     crtICurvep
	     break
	 }
	 ######
	 button $f.bnci -image ncircleimg -padx 0 -pady 0 -command {
	   crtOb ncircle; uMS; sL; redrawAllViews
	 }
	 bind $f.bnci <Shift-ButtonPress-1> {
	     global  cbspnum 
	     crtOb cbsp $cbspnum; uMS; sL; redrawAllViews
	     break
	 }
	 balloon $f.bnci "create NURB circle\nShift: create closed BSpline"
	 pack $f.bnci -in $f -side left
	 ######
	 button $f.bnrec -image nrectimg -padx 0 -pady 0 -command {
	   crtOb trimrect; uMS; sL; redrawAllViews
	 }
	 balloon $f.bnrec "create Rectangle"
	 pack $f.bnrec -in $f -side left
	 button $f.bnp -image npatchimg -padx 0 -pady 0 -command {
	     global npwidth npheight
	     crtOb npatch -width $npwidth -height $npheight;
	     uMS; sL; redrawAllViews
	 }
	 balloon $f.bnp "create NURB patch"
	 pack $f.bnp -in $f -side left


	 pack $f -in $w -side top -anchor w
	 #########
	 set f [frame $w.fbarcrnc -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes
	 #########

     }


     if { $i == "transforms" } {
	 set f [frame $w.ftran1 ]

	 button $f.bm -image moveimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     actionMoveOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" MoveXY
		 }
		 if { $view(Type) == 1 } {
		     actionMoveOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" MoveZY
		 }
		 if { $view(Type) == 2 } {
		     actionMoveOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" MoveXZ
		 }
		 if { $view(Type) != 3 } {
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bm "move\n\[ $mops_oa_sc(MoveO) \]"
	 pack $f.bm -in $f -side left
	 ######
	 button $f.br -image rotimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     actionRotOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateZ
		 }
		 if { $view(Type) == 1 } {
		     actionRotOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateX
		 }
		 if { $view(Type) == 2 } {
		     actionRotOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateY
		 }
		 if { $view(Type) != 3 } {
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.br "rotate (around origin)\n\[ $mops_oa_sc(RotO) \]"
	 pack $f.br -in $f -side left
	 ######
	 button $f.bra -image rotaimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     actionRotOba ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateAboutZ
		 }
		 if { $view(Type) == 1 } {
		     actionRotOba ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateAboutX
		 }
		 if { $view(Type) == 2 } {
		     actionRotOba ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" RotateAboutY
		 }

	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bra "rotate (around point)\n\[ $mops_oa_sc(RotA) \]"
	 pack $f.bra -in $f -side left
	 ######
	 button $f.bs3 -image scal3dimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     actionSc3DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" Scale
		 }
		 if { $view(Type) == 1 } {
		     actionSc3DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" Scale
		 }
		 if { $view(Type) == 2 } {
		     actionSc3DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" Scale
		 }
		 if { $view(Type) != 3 } {
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bs3 "scale 3D\n\[ $mops_oa_sc(Scal3) \]"
	 pack $f.bs3 -in $f -side left
	 ######
	 pack $f -in $w -side top
	 ######
	 ######
	 set f [frame $w.ftran2 ]

	 button $f.bs2 -image scal2dimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     actionSc2DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleXY
		 }
		 if { $view(Type) == 1 } {
		     actionSc2DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleYZ
		 }
		 if { $view(Type) == 2 } {
		     actionSc2DOb ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleXZ
		 }
		 if { $view(Type) != 3 } {
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }
	     
	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bs2 "scale 2D\n\[ $mops_oa_sc(Scal2) \]"
	 pack $f.bs2 -in $f -side left
	 ######
	 button $f.bs1x -image scal1dximg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) != 3) } {
		     actionSc1DObx ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleX
		     
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bs1x "scale X\n\[ $mops_oa_sc(ScalX) \]"
	 pack $f.bs1x -in $f -side left
	 ######
	 button $f.bs1y -image scal1dyimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) != 3) } {
		     actionSc1DOby ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleY
		     
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bs1y "scale Y\n\[ $mops_oa_sc(ScalY) \]"
	 pack $f.bs1y -in $f -side left
	 ######
	 button $f.bs1z -image scal1dzimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) != 3) } {
		     actionSc1DObz ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" ScaleZ
		     
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bs1z "scale Z\n\[ $mops_oa_sc(ScalZ) \]"
	 pack $f.bs1z -in $f -side left

	 pack $f -in $w -side top -anchor w

	 set f [frame $w.fbartran -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes

     }

     ##################

     if { $i == "points" } {

	 set f [frame $w.fp]
	 
	 button $f.bed -image editimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { ($view(Type) == 0) || ($view(Type) == 4) } {
		     setupPEditBinding $i "xy" 
		     viewTitlep ${i}.f3D.togl "" EditXY
		 }
		 if { $view(Type) == 1 } {
		     setupPEditBinding $i "zy" 
		     viewTitlep ${i}.f3D.togl "" EditZY
		 }
		 if { $view(Type) == 2 } {
		     setupPEditBinding $i "xz" 
		     viewTitlep ${i}.f3D.togl "" EditXZ
		 }
		 if { $view(Type) != 3 } {
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bed "edit points\n\[ $mops_oa_sc(Edit) \]"
	 pack $f.bed -in $f -side left
	 ######
	 button $f.bedw -image editwimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { $view(Type) != 3 } {
		     setupWEditBinding ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" EditW
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bedw "edit weights\n\[ $mops_oa_sc(WeightE) \]"
	 pack $f.bedw -in $f -side left
	 ######
	 button $f.bedd -image editdimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 
		 if { $view(Type) != 3 } {
		     actionEditPointD ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" EditDirect
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	  
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 pack $f.bedd -in $f -side left
	 balloon $f.bedd "directly edit coordinates\n\[ $mops_oa_sc(EditD) \]"
	 ######
	 button $f.sel -image tagimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 
		 if { $view(Type) != 3 } {
		     setupPSelectBinding ${i}.f3D.togl
		     viewTitlep $i ""  SelectPoints
		     bind ${i}.f3D.togl <ButtonRelease-1> "+redrawAllViews"
		 }
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 bind $f.sel <Shift-ButtonPress-1> {
	     global  currentView 
	     $currentView deselectPoints
	     redrawAllViews
	     break;
	 }
	 balloon $f.sel "select (tag) points\n\[ $mops_oa_sc(Select) \]\nShift: deselect points\n\[ $mops_oa_sc(DeSelect) \]"
	 pack $f.sel -in $f -side left

	 pack $f -in $w -side top -anchor w

	 set f [frame $w.fbarpoint -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes

     }

     ##################
     if { $i == "curves" } {
	 set f [frame $w.fnc]

	 button $f.bin -image insertimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { $view(Type) != 3 } {
		     actionInsertPoint ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" InsertPoint
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
	   
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bin "insert point\n\[ $mops_oa_sc(InsertP) \]"
	 pack $f.bin -in $f -side left
	 ######
	 button $f.bde -image deleteimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { $view(Type) != 3 } {
		     actionDeletePoint ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" DeletePoint
		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
		 
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bde "delete point\n\[ $mops_oa_sc(DelP) \]"
	 pack $f.bde -in $f -side left
	 ######
	 button $f.br -image revertimg -padx 0 -pady 0 -command {
	     revertNC; updateMopsSelect; redrawAllViews
	 }
	 balloon $f.br "revert curve"
	 pack $f.br -in $f -side left
	 ######
	 button $f.bsp -image splitimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { $view(Type) != 3 } {
		     actionSplitCurve ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" SplitCurve
#		     bind ${i}.f3D.togl <ButtonRelease-1> { redrawAllViews }
		 }
		 
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bsp "split curve\n\[ $mops_oa_sc(SplitC) \]"
	 pack $f.bsp -in $f -side left
	 ######
	 pack $f -in $w -side top -anchor w

	 set f [frame $w.fbar4 -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes

     }

     ##################
     if { $i == "curves2" } {
	 set f [frame $w.fnc2]

	 button $f.bfu -image finduimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig
		 if { $view(Type) != 3 } {
		     actionfindU ${i}.f3D.togl
		     viewTitlep ${i}.f3D.togl "" FindU
		 }
	   
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bfu "find u\n\[ $mops_oa_sc(FindU) \]"
	 pack $f.bfu -in $f -side left
	 ######
	 button $f.brf -image refineimg -padx 0 -pady 0 -command {
	     refineNC;redrawAllViews
	 }
	 balloon $f.brf "refine"
	 pack $f.brf -in $f -side left
	 ######
	 button $f.bcl -image clampimg -padx 0 -pady 0 -command {
	     clampNC; updateMopsSelect; redrawAllViews
	 }
	 balloon $f.bcl "clamp"
	 pack $f.bcl -in $f -side left
	 ######
	 button $f.bcon -image concatimg -padx 0 -pady 0 -command {
	     crtOb concatnc; updateMopsSelect; redrawAllViews
	 }
	 balloon $f.bcon "concat"
	 pack $f.bcon -in $f -side left
	 ######
	 pack $f -in $w -side top -anchor w

	 set f [frame $w.fcurves2 -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes

     }


     ##################
     if { $i == "camera" } {
	 set f [frame $w.fcam]

	 ######
	 button $f.bv -image viewimg -padx 0 -pady 0 -command {
	     newViewp 400 300
	 }
	 balloon $f.bv "new view"
	 pack $f.bv -in $f -side left
	 ######
	 button $f.brot -image rotvimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig

		 actionRotView ${i}.f3D.togl
		 viewTitlep ${i}.f3D.togl "" RotateView
		 bind ${i}.f3D.togl <ButtonRelease-1> ""
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.brot "rotate view\n\[ $mops_va_sc(RotV) \]"
	 pack $f.brot -in $f -side left
	 ######
	 button $f.bmov -image movevimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig

		 actionMoveView ${i}.f3D.togl
		 viewTitlep ${i}.f3D.togl "" MoveView
		 bind ${i}.f3D.togl <ButtonRelease-1> ""
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bmov\
"move view\n\[ $mops_va_sc(MoveV) \] or Mouse-$mops_va_sc(MoveVButton)"
	 pack $f.bmov -in $f -side left
	 #####
	 button $f.bzom -image zoomvimg -padx 0 -pady 0 -command {
	     global mops_views currentView view
	     foreach i $mops_views {
		 ${i}.f3D.togl makecurrent
		 ${i}.f3D.togl getViewConfig

		 actionZoomView ${i}.f3D.togl
		 viewTitlep ${i}.f3D.togl "" ZoomView
		 bind ${i}.f3D.togl <ButtonRelease-1> ""
	     }

	     $currentView makecurrent
	     $currentView getViewConfig
	 }
	 balloon $f.bzom\
"zoom view\n\[ $mops_va_sc(ZoomV) \] or Mouse-$mops_va_sc(ZoomVButton)"
	 pack $f.bzom -in $f -side left
	 #####
	 pack $f -in $w -side top -anchor w

	 set f [frame $w.fbar5 -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes
     }

     if { $i == "custom" } {
	 set f [frame $w.fcust ]

	 button $f.brev -image revolveimg -padx 0 -pady 0 -command {
	   crtRevolvep
	 }
	 balloon $f.brev "revolve"
	 pack $f.brev -in $f -side left
	 #####
	 button $f.bex -image extrudeimg -padx 0 -pady 0 -command {
	   crtExtrudep
	 }
	 balloon $f.bex "extrude"
	 pack $f.bex -in $f -side left
	 #####
	 button $f.bswp -image sweepimg -padx 0 -pady 0 -command {
	   crtSweepp
	 }
	 balloon $f.bswp "sweep"
	 pack $f.bswp -in $f -side left
	 #####
	 button $f.bsk -image skinimg -padx 0 -pady 0 -command {
	     crtSkinp
	 }
	 balloon $f.bsk "skin"
	 pack $f.bsk -in $f -side left
	 #####
	 pack $f -in $w -side top -anchor w
	 #########
	 set f [frame $w.fbarcus -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes
	 #########

     }

     if { $i == "misc" } {
	 set f [frame $w.fmisc ]

	 button $f.bhide -image showimg -padx 0 -pady 0 -command {
	   hide; redrawAllViews
	 }
	 bind $f.bhide <Shift-ButtonPress-1> {
	     show; redrawAllViews; break;
	 }
	 balloon $f.bhide "hide\nShift: show"
	 pack $f.bhide -in $f -side left
	 #####
	 button $f.bconv -image convertimg -padx 0 -pady 0 -command {
	   convOb;uMS;clearSelect;redrawAllViews
	 }
	 balloon $f.bconv "convert"
	 pack $f.bconv -in $f -side left
	 #####
	 button $f.bnot -image notifyimg -padx 0 -pady 0 -command {
	   forceNot; redrawAllViews
	 }
	 balloon $f.bnot "force notification"
	 pack $f.bnot -in $f -side left

	 #####
	 pack $f -in $w -side top -anchor w
	 #########
	 set f [frame $w.fbarmisc -relief raised -bd 2 -height 3]
	 pack $f -in $w -side top -fill x -expand yes
	 #########

     }


 }
 # foreach


 global prefs
 update
 if { $prefs(toolBoxGeom) != "" } {
     moveOrResize .tb $prefs(toolBoxGeom)
 }
 return;
}

proc openIDR { } {
 global prefs mops_views

  set w .iw
  catch {destroy $w}
  toplevel $w
  wm title $w "IDR"
  wm iconname $w "IDR"
  wm group .

  set f [frame $w.fv -relief sunken -borderwidth 2]
  label $f.l -text "View:"
  menubutton $f.mb -text NoView! -menu $f.mb.m -relief raised
  set m [menu $f.mb.m -tearoff 0]
  # select view
  foreach i $mops_views {
      scan $i ".view%d" index
      $m add command -label View$index -command "global idr_view
      set idr_view $i
      $f.mb configure -text View$index"      
  }

  if { $mops_views != "" } {
      $m invoke 0
  }
  pack $f.l -in $f -side top
  pack $f.mb -in $f -side top
  pack $f -in $w -side top -fill x

  # select Phase
  set f [frame $w.fp -relief sunken -borderwidth 2]
  label $f.l -text "Phase:"
  menubutton $f.mb -text Modeling -menu $f.mb.m -relief raised
  set m [menu $f.mb.m -tearoff 0]
  $m add command -label Automatic -command "global idr_phase
  set idr_phase 0
  $f.mb configure -text Automatic"

  $m add command -label Modeling -command "global idr_phase
  set idr_phase 1
  $f.mb configure -text Modeling"

  $m add command -label Texturing -command "global idr_phase
  set idr_phase 2
  $f.mb configure -text Texturing"

  $m add command -label Lighting -command "global idr_phase
  set idr_phase 3
  $f.mb configure -text Lighting"

  $m add command -label Final -command "global idr_phase
  set idr_phase 4
  $f.mb configure -text Final"


  $m invoke 0
  pack $f.l -in $f -side top
  pack $f.mb -in $f -side top
  pack $f -in $w -side top -fill x

  # select Importance mode
  set f [frame $w.fi -relief sunken -borderwidth 2]
  label $f.l -text "Importance:"
  checkbutton $f.cbs -text "from selection" -textvariable idr_selection
  checkbutton $f.cbr -text "from RoI" -textvariable idr_roi
  checkbutton $f.cbo -text "per object" -textvariable idr_definedobject
  checkbutton $f.cba -text "automatic" -textvariable idr_automatic
  pack $f.l -in $f -side top
  pack $f.cbs -in $f -side top -anchor w
  pack $f.cbr -in $f -side top -anchor w
  pack $f.cbo -in $f -side top -anchor w
  pack $f.cba -in $f -side top -anchor w

  pack $f -in $w -side top -fill x

  set f [frame $w.fr]
  button $f.br -text "Render!" -command {
      global idr_view

      $idr_view.f3D.togl idrRender
  }
  pack $f.br -in $f 

  pack $f -in $w -side bottom
  if { $prefs(idrGeom) != "" } {
      moveOrResize $w $prefs(idrGeom)
  }
}



# The Main Window
######################################
######################################
######################################
set w .
wm title $w "The Mops - Main"
wm iconname $w "Mops"
# the upper frame
frame .fu
# the lower frame (contains the console)
frame .fl

# The Main Menu
frame .fu.fMenu -bd 2 -relief raised
pack .fu.fMenu -side top -fill x

# File
menubutton .fu.fMenu.fil -text "File" -menu .fu.fMenu.fil.m

menu .fu.fMenu.fil.m
set m .fu.fMenu.fil.m
$m add command -label "New"\
	-command "newScenep; setSelect; updateMopsSelect; updateMopsPList;"
$m add separator
$m add command -label "Replace" -command "replaceScenep"
$m add command -label "Insert" -command "insertScenep"
$m add separator
$m add command -label "Save as" -command "saveScenep ask"
$m add command -label "Save" -command "saveScenep dontask"
$m add separator
$m add command -label "Export RIB" -command "exportRIBp"
$m add separator
$m add command -label "Load Custom" -command "loadCustomp"
$m add separator
$m add command -label "Save Prefs" -command "savePrefsp"
$m add separator
$m add command -label "Exit!" -command {
    
    cleanTmpFilesp 1
    
    puts "Good Bye!"
    update
    exit
}
pack .fu.fMenu.fil -in .fu.fMenu -side left

# Edit
menubutton .fu.fMenu.ed -text "Edit" -menu .fu.fMenu.ed.m
menu .fu.fMenu.ed.m
set m .fu.fMenu.ed.m
$m add command -label "Copy" -command copOb
$m add command -label "Cut" -command "cutOb; clearSelect; redrawAllViews"
$m add command -label "Paste" -command "pasOb;\
	clearSelect; redrawAllViews"
$m add command -label "Delete" -command "delOb;\
	clearSelect; redrawAllViews"
$m add separator
$m add command -label "Copy Property" -command copyPropp
$m add command -label "Copy Marked Prop" -command icopyPropp
$m add command -label "Paste Property" -command pastePropp
$m add separator
$m add command -label "Name" -command "nameObp; updateMopsSelect"
$m add separator
$m add command -label "Swap" -command "swapOb; clearSelect; "
$m add separator
$m add command -label "Undo" -command "undo; redrawAllViews"
pack .fu.fMenu.ed -in .fu.fMenu -side left

# Create
menubutton .fu.fMenu.cr -text "Create" -menu .fu.fMenu.cr.m
menu .fu.fMenu.cr.m
set m .fu.fMenu.cr.m
$m add command -label "NURBCurve" -command {
    runToolp nclen "Length:" "crtOb ncurve -length %nclen;uMS;sL"
}
$m add command -label "NURBPatch" -command {
    runToolp {npwidth npheight} {"Width:" "Height:"}\
	    "crtOb npatch -width %npwidth -height %npheight;uMS;sL"
}
$m add separator
$m add cascade -menu $m.sl -label "Solid"
menu $m.sl
$m.sl add command -label "Box"\
-command "crtOb box; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Sphere"\
-command "crtOb quadric sphere; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Disk"\
-command "crtOb quadric disk; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Cone"\
-command "crtOb quadric cone; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Cylinder"\
-command "crtOb quadric cylinder; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Torus"\
-command "crtOb quadric torus; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Hyperboloid"\
-command "crtOb quadric hyperboloid; updateMopsSelect; selectLast; redrawAllViews"
$m.sl add command -label "Paraboloid"\
-command "crtOb quadric paraboloid; updateMopsSelect; selectLast; redrawAllViews"

$m add cascade -menu $m.csg -label "Level"
menu $m.csg
$m.csg add command -label "Union"\
-command "crtOb level union; updateMopsSelect; selectLast; redrawAllViews"
$m.csg add command -label "Intersection"\
-command "crtOb level int; updateMopsSelect; selectLast; redrawAllViews"
$m.csg add command -label "Difference"\
-command "crtOb level diff; updateMopsSelect; selectLast; redrawAllViews"
$m.csg add separator
$m.csg add command -label "Primitive"\
-command "crtOb level prim; updateMopsSelect; selectLast; redrawAllViews"
$m.csg add command -label "Level"\
-command "crtOb level level; updateMopsSelect; selectLast; redrawAllViews"
$m.csg add command -label "TrimLoop"\
-command "crtOb level tloop; updateMopsSelect; selectLast; redrawAllViews"
$m add command -label "Light" -command "newLightp"
$m add separator
$m add cascade -menu $m.cus -label "Custom Object"
menu $m.cus
#$m.cus add command -label "Revolve"\
# -command "crtOb custom Revolve; updateMopsSelect; redrawAllViews"
#$m.cus add command -label "Extrude"\
# -command "crtOb custom Extrude; updateMopsSelect; redrawAllViews"
#$m.cus add command -label "Sweep"\
# -command "crtOb custom Sweep; updateMopsSelect; redrawAllViews"
$m add separator
$m add command -label "Instance" \
	-command "crtOb instance; updateMopsSelect; selectLast; redrawAllViews"
$m add separator
$m add command -label "View" \
	-command "newViewp 400 300"
pack .fu.fMenu.cr -in .fu.fMenu -side left

# Tools
menubutton .fu.fMenu.tool -text "Tools" -menu .fu.fMenu.tool.m

set m [menu .fu.fMenu.tool.m]

$m add cascade -menu $m.nc -label "Create"
menu $m.nc
$m.nc add command -label "NURBCircle" -command "crtOb ncircle; updateMopsSelect; selectLast; redrawAllViews"
$m.nc add command -label "NURBSphere" -command "crtOb nsphere; updateMopsSelect; selectLast; redrawAllViews"
$m.nc add command -label "NURBSphere2" -command "crtNSphere2; updateMopsSelect; selectLast; redrawAllViews"
$m.nc add separator
$m.nc add command -label "TrimRect" -command "crtOb trimrect; updateMopsSelect; selectLast; redrawAllViews"
$m.nc add separator
$m.nc add command -label "ClosedBSpline" -command {
    runToolp cbspnum "Length:" "crtOb cbsp %cbspnum;uMS;sL"
}

$m.nc add separator
$m.nc add command -label "ICurve" -command "crtICurvep"
$m add separator
$m add cascade -menu $m.nct -label "NURBCurve"
menu $m.nct
$m.nct add command -label "Revolve" -command "crtRevolvep;"
$m.nct add command -label "Extrude" -command "crtExtrudep;"
$m.nct add command -label "Sweep" -command "crtSweepp;"
$m.nct add command -label "Skin" -command "crtSkinp;"
$m.nct add separator
$m.nct add command -label "Revert" -command "revertNC; updateMopsSelect; redrawAllViews"
$m.nct add command -label "Concat" -command "crtOb concatnc; updateMopsSelect; redrawAllViews"
$m.nct add command -label "Split" -command {
    runToolp splitu {"Split at:"} "splitNC %splitu;uMS;sL" }
$m.nct add command -label "Refine" -command { refineNC;uMS;sL }
$m.nct add command -label "Clamp" -command { clampNC;uMS;sL }
$m.nct add command -label "Elevate" -command {
    runToolp elevd {"Elevate by:"} "elevateNC %elevd;uMS;sL" }
$m.nct add command -label "Insert Knot" -command {
    runToolp [list insknu insknr] [list "Insert knot at:" "Insert times:"] "insKn %insknu %insknr;uMS;sL" }

$m.nct add separator
$m.nct add command -label "Collapse Points" -command { coll; rAV; }
$m.nct add command -label "Explode Points" -command { expl; rAV; }

$m add separator
$m add cascade -menu $m.npt -label "NURBPatch"
menu $m.npt
$m.npt add command -label "Split to Curves (u)" -command "splitNPatch u;updateMopsSelect;clearSelect;redrawAllViews"
$m.npt add command -label "Split to Curves (v)" -command "splitNPatch v;updateMopsSelect;clearSelect;redrawAllViews"
$m.npt add command -label "Build from Curves" -command "buildNPatch;updateMopsSelect;clearSelect;redrawAllViews"
$m add separator
$m add command -label "Hide" -command "hide; redrawAllViews"
$m add command -label "Show" -command "show; redrawAllViews"
$m add separator
$m add command -label "Convert" -command "convOb;updateMopsSelect;clearSelect;redrawAllViews"
$m add separator
$m add command -label "Force Notification" -command "forceNot; redrawAllViews"


pack .fu.fMenu.tool -in .fu.fMenu -side left

# Custom
menubutton .fu.fMenu.cust -text "Custom" -menu .fu.fMenu.cust.m
set prefs(cm) [menu .fu.fMenu.cust.m]
pack .fu.fMenu.cust -in .fu.fMenu -side left

# Help
menubutton .fu.fMenu.hlp -text "Help" -menu .fu.fMenu.hlp.m

menu .fu.fMenu.hlp.m
.fu.fMenu.hlp.m add command -label "Help" -command "mopsHelpp"
.fu.fMenu.hlp.m add command -label "About" -command "aboutMops"
pack .fu.fMenu.hlp -in .fu.fMenu -side right

if { $tcl_platform(platform) == "windows" } {
.fu.fMenu.fil configure -pady 1
.fu.fMenu.ed configure -pady 1
.fu.fMenu.cr configure -pady 1
.fu.fMenu.tool configure -pady 1
.fu.fMenu.cust configure -pady 1
.fu.fMenu.hlp configure -pady 1
}
#############################################################

frame .fu.fMain

# The Object-Hierarchy Listbox
frame .fu.fMain.fHier -bd 1 -relief sunken
set f .fu.fMain.fHier
frame $f.flb
set f $f.flb
label $f.la -text "Hierarchy:"
listbox $f.li -width 16 -selectmode extended\
	-yscrollcommand {.fu.fMain.fHier.flb.s set} -exportselection 0
bind $f.li <Double-ButtonPress-1> {
set list [.fu.fMain.fHier.flb.li curselection]
eval [subst "goDown $list"]
setSelect; updateMopsSelect; updateMopsPList;
}
bind $f.li <Return> {
set list [.fu.fMain.fHier.flb.li curselection]
eval [subst "goDown $list"]
setSelect; updateMopsSelect; updateMopsPList;
}

bind $f.li <ButtonRelease-1> { after 200 updateMopsPList }
bind $f.li <KeyRelease-space> { updateMopsPList }

bind $f.li <ButtonPress-4> {
	.fu.fMain.fHier.flb.li yview scroll -1 pages
        break
}
bind $f.li <ButtonPress-5> {
	.fu.fMain.fHier.flb.li yview scroll 1 pages
        break
}

scrollbar $f.s -command {.fu.fMain.fHier.flb.li yview} -takefocus 0
pack $f.la -in $f -side top
pack $f.li -in $f -side left -fill both -expand yes
pack $f.s -in $f -side right -fill both -expand yes

# selection buttons
frame .fu.fMain.fHier.fsel
set f .fu.fMain.fHier.fsel
button $f.bnon -text "None" -width 5 -padx 0 -pady 0 -command {
.fu.fMain.fHier.flb.li selection clear 0 end
setSelect
redrawAllViews
updateMopsSelect
updateMopsPList
global prop_sema
set prop_sema 0
} -takefocus 0
button $f.ball -text "All"  -width 5 -padx 0 -pady 0 -command {
.fu.fMain.fHier.flb.li selection set 0 end
updateMopsPList
} -takefocus 0
button $f.binv -text "Inv" -width 5 -padx 0 -pady 0 -command {
set selected [.fu.fMain.fHier.flb.li curselection]
.fu.fMain.fHier.flb.li selection set 0 end
foreach element $selected {
.fu.fMain.fHier.flb.li selection clear $element
}

updateMopsPList
} -takefocus 0
pack $f.ball $f.bnon $f.binv -in $f -side left -fill x

# navigation buttons
frame .fu.fMain.fHier.fnavi
set f .fu.fMain.fHier.fnavi 
button $f.bup -text "Up" -padx 0 -width 5 -pady 0 -command {
goUp
setSelect; updateMopsSelect; updateMopsPList;
} -takefocus 0
button $f.btop -text "Top" -padx 0 -width 5 -pady 0  -command {
goTop
setSelect; updateMopsSelect; updateMopsPList;
} -takefocus 0
button $f.bdwn -text "Down" -padx 0 -pady 0 -width 5 -command {
set list [.fu.fMain.fHier.flb.li curselection]
eval [subst "goDown $list"]
setSelect; updateMopsSelect; updateMopsPList;
} -takefocus 0
pack $f.bup $f.btop $f.bdwn -in $f -side left
pack .fu.fMain.fHier.fnavi -in .fu.fMain.fHier -side bottom
pack .fu.fMain.fHier.fsel -in .fu.fMain.fHier -side bottom
pack .fu.fMain.fHier.flb -in .fu.fMain.fHier -side top -fill both -expand yes
pack .fu.fMain.fHier -in .fu.fMain -side left -fill both -expand no

########################
# The Properties Listbox
frame .fu.fMain.fProp  -bd 1 -relief sunken
set f .fu.fMain.fProp
frame $f.flb
set f $f.flb
label $f.la -text "Properties:"
listbox $f.li -width 16 -height 12 -yscrollcommand {.fu.fMain.fProp.flb.s set}\
	-selectmode single -exportselection 0
bind $f.li <ButtonRelease-1> {
    global copyPropOmit
    updateMopsPList
    unset copyPropOmit
    array set copyPropOmit { dummy 0 }
}

bind $f.li <KeyRelease-space> updatePropGUI

bind $f.li <ButtonPress-4> {
	.fu.fMain.fProp.flb.li yview scroll -1 pages
        break
}
bind $f.li <ButtonPress-5> {
	.fu.fMain.fProp.flb.li yview scroll 1 pages
        break
}


scrollbar $f.s -command {.fu.fMain.fProp.flb.li yview} -takefocus 0
pack $f.la -in $f -side top
pack $f.li $f.s -in $f -side left -fill y -expand yes
pack .fu.fMain.fProp.flb -in .fu.fMain.fProp -side top -fill y -expand yes
pack .fu.fMain.fProp -in .fu.fMain -side left -fill y

######################
# The Arguments Canvas
frame .fu.fMain.fArg -bd 1 -relief sunken
set f .fu.fMain.fArg
label $f.la -text "Arguments:"
pack $f.la -in $f -side top

set f [frame $f.fb]
button $f.b1 -text "Apply" -padx 10 -pady 0 -command "update" -takefocus 0
button $f.b2 -text "Reset" -padx 10 -pady 0 -command "update" -takefocus 0
pack $f.b1 $f.b2 -in $f -side left -fill x -expand yes 
pack $f -in .fu.fMain.fArg -side bottom -fill x

set f .fu.fMain.fArg
scrollbar $f.s -command {.fu.fMain.fArg.flb.ca yview} -takefocus 0
pack $f.s -in $f -side right -fill y 

set f .fu.fMain.fArg
frame $f.flb -bd 2 -relief sunken
set f $f.flb
canvas $f.ca -width 6.5c -yscrollcommand {.fu.fMain.fArg.s set}

set propWindow [frame $f.ca.w]
$f.ca create window 5 5 -anchor nw -window $propWindow

bind . <ButtonPress-4> {
	.fu.fMain.fArg.flb.ca yview scroll -1 pages
}
bind . <ButtonPress-5> {
	.fu.fMain.fArg.flb.ca yview scroll 1 pages
}

pack $f.ca -in $f -side left -fill both -expand yes

pack .fu.fMain.fArg.flb -in .fu.fMain.fArg -side top -fill both -expand yes
pack .fu.fMain.fArg -in .fu.fMain -side left -fill both -expand yes 

pack .fu.fMain -in .fu -fill both -expand yes
pack .fu -in . -fill both -expand 1


# load widget.tcl pane.tcl and console.tcl, if we are wrapped
if { $MOPSWRAPPED == 1 } {
wrap::source "widget.tcl"
wrap::source "pane.tcl"
wrap::source "console.tcl"
}

#
##########
# load widget.tcl, if we are not standalone and not wrapped
set oldcd [pwd] 
set scdir [info script]
if {$scdir != ""} { cd [file dirname $scdir] }
if { $MOPSSTANDALONE == 0 } {
    if { [file exists "widget.tcl"] } {
	source "widget.tcl"
    } else {
	destroy .fl
	puts stderr "The file widget.tcl, which is needed for the"
	puts stderr "Mops-Console could not be found."
	puts stderr "Please install it where mopssh is."
	puts stderr "Currently, The Mops cannot run without widget.tcl!"
	puts stderr "The Mops is bailing out..."
	exit;
    }
}

# load console.tcl, if we are not standalone and not wrapped
if { $MOPSSTANDALONE == 0 } {
    if { [file exists "console.tcl"] } {
	source "console.tcl"
	# The Console
	pack .fl -in . -side bottom -fill both
	frame .fl.dummy
	console .fl.con -showmenu 0 -height 5 -width 60
	pack .fl.con -in .fl -expand 1 -fill both
	pack .fl -in . -expand 1 -fill both
	bind .fl.con.console <Shift-Tab> { focus [tk_focusNext %W] }
    } else {
	destroy .fl
	tk_messageBox -icon warning -type ok -message "\
		The file console.tcl, which is needed for the Mops-Console\
		could not be found. Please install it where mopssh is."
    }
} else {
	# The Console
	pack .fl -in . -side bottom -fill both
	frame .fl.dummy
	console .fl.con -showmenu 0 -height 5 -width 60
	pack .fl.con -in .fl -expand 1 -fill both
	pack .fl -in . -expand 1 -fill both
        bind .fl.con.console <Shift-Tab> { focus [tk_focusNext %W] }
}
update

# load pane.tcl, if we are not standalone and not wrapped
# and the console is available
if { [winfo exists .fl.con] == 1 } {
    if { $MOPSSTANDALONE == 0 } {
	if { [file exists "pane.tcl"] } {
	    source "pane.tcl"
	    set vheight [winfo rooty .fl]
	    pane .fu .fl -orient vertical
	    pane_constrain . .__h1 .fu .fl height y 1
	    pane_motion $vheight . .__h1 height y 1
	} else {
	    tk_messageBox -icon warning -type ok -message "\
		    The file pane.tcl, which is needed to make the\
		    Mops-Console a paned window could not be found.\
		    Please install it where mopssh is."
	}
    } else {
	set vheight [winfo rooty .fl]
	pane .fu .fl -orient vertical
	pane_constrain . .__h1 .fu .fl height y 1
	pane_motion $vheight . .__h1 height y 1
    }
}
cd $oldcd


##########
# Ignition

# clear console
if { [winfo exists .fl.con] == 1 } { .fl.con clear }

setPrefsProp

puts stdout "Mops-Startup-Sequence initiated."

updateMopsSelect

puts stdout "Loading CustomObjects..."
set oldcd [pwd] 
set scdir [info script]
#catch {lc csphere.so}
catch {lc mfio.so}
cd $oldcd

# the hunt for the red mopsrc
puts stdout "Sourcing $mops_gp(mopsrc)..."
if { [file exists $mops_gp(mopsrc)] } {
    if { [source $mops_gp(mopsrc)] != "" } {
	puts stderr "Oops? mopsrc returned something?"
    }
    setPrefsProp
} else {

    puts stdout "Not found. Please install the sample mopsrc from the distribution,"
    puts stdout "or save preferences."

}

scanAllShaders

# env var MOPSLIB overrides preference setting
if {[string length [array names env MOPSLIB]] != 0} then {
    array set prefs "MOPSLIB $env(MOPSLIB)"
}

puts stdout "Processing arguments..."
set i 0
set j 0
while { $i < $argc } {
    set arg [lindex $argv $i]
    if { [file extension $arg] == ".mop" } {
	set newfilename $arg
	if { $j == 0 } {
	    closeAllViews
	    set filename $newfilename
	    replaceScene $filename
	    set windowfilename [file tail [file rootname $newfilename]]
	    wm title . "The Mops - Main - $windowfilename"
	    set j 1
	} else {
	    set ifilename $newfilename
	    insertScene $ifilename
	}
	updateMopsSelect
    }
 incr i
}

redrawAllViews
puts stdout "Establishing key bindings..."
setupKeyBindings .


# End
###########

if { [winfo exists .fl.con] == 1 } {
    set .fl.con(-prompt) {[file tail [pwd]]>}
    Console:prompt .fl.con
}

# if there is no view present, we open the first one
if { $mops_views == "" } { 
    global view1Geomvalid
    set view1Geomvalid 1
    openMopsView .view1 400 300
    set currentView .view1.f3D.togl

    # make our view #1 a Front-View
    .view1.fMenu.c.m.tp invoke 1

    setupKeyBindings3D .view1
    setupActionKeys .view1 "xy"

    .view1.f3D.togl getViewConfig

    if { $prefs(view1Geom) != "" } {
	moveOrResize .view1 $prefs(view1Geom)
    }
}

# open tool window
openToolBoxp

# open idr commander
if { $MOPSIDR == 1 } {
    openIDR
}

set gl_ext [.view1.f3D.togl configure -extensions]
getVersion

if { $prefs(mainGeom) != "" } {
    moveOrResize . $prefs(mainGeom)
}

# another windows layout work-around:
if { $tcl_platform(platform) == "windows" } {
    .fu.fMain.fHier.flb.s configure\
	    -width [expr [.fu.fMain.fProp.flb.s cget -width]/2]   
}

# allow canceling of user Tcl-scripts in the console
bind .fl.con.console <Control-KeyPress-C> {set cancelled 1}
bind .fl.con.console <Control-KeyRelease-C> {set cancelled 0}

set cancelled 0
proc cancelled {} {
    global cancelled
    # give the user a chance to press Ctrl+C
    update
    if { $cancelled == 1 } {
	error "Cancelled!"
    }
}

rename while _while
proc while {condition body} {
    uplevel [list _while $condition "cancelled;$body"]
}

rename for _for
proc for {start test next body} {
    uplevel [list _for $start $test $next "cancelled;$body"]
}

if { $prefs(scripts) != "" } {
    puts stdout "Running some user defined scripts..."
    foreach script $prefs(scripts) {
	if { [file exists $script] } {
	    puts stdout ${script}; source $script
	} else {
	    puts -nonewline stderr "Specified file \""
	    puts -nonewline stderr $script
	    puts stderr "\" does not exist!"
	}
    }
}


puts stdout "Mops-Startup-Sequence finished."
puts stdout "The Tip of the Day is:"
tipoftheDay

# QED
