/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Standard Properties */

/* C -> Tcl! */
int
mops_get_transfprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="transfPropData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Translate_X",-1);
  to = Tcl_NewDoubleObj(o->movx);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Translate_Y",-1);
  to = Tcl_NewDoubleObj(o->movy);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Translate_Z",-1);
  to = Tcl_NewDoubleObj(o->movz);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton, "Quat0",-1);
  to = Tcl_NewDoubleObj((double)o->quat[0]);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Quat1",-1);
  to = Tcl_NewDoubleObj((double)o->quat[1]);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Quat2",-1);
  to = Tcl_NewDoubleObj((double)o->quat[2]);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Quat3",-1);
  to = Tcl_NewDoubleObj((double)o->quat[3]);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton, "Rotate_X",-1);
  to = Tcl_NewDoubleObj(o->rotx);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Rotate_Y",-1);
  to = Tcl_NewDoubleObj(o->roty);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Rotate_Z",-1);
  to = Tcl_NewDoubleObj(o->rotz);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Scale_X",-1);
  to = Tcl_NewDoubleObj(o->scalx);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Scale_Y",-1);
  to = Tcl_NewDoubleObj(o->scaly);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton, "Scale_Z",-1);
  to = Tcl_NewDoubleObj(o->scalz);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_transfprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="transfPropData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 double dtemp;
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 float quat[4], drot;
 int pasteProp = 0;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 

  toa = Tcl_NewStringObj("prefs",-1);
  ton = Tcl_NewStringObj("pasteProp",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &pasteProp);

  Tcl_SetStringObj(toa, n1,-1);
  Tcl_SetStringObj(ton, "Translate_X",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(o->movx));
  Tcl_SetStringObj(ton, "Translate_Y",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(o->movy));
  Tcl_SetStringObj(ton, "Translate_Z",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(o->movz));

  Tcl_SetStringObj(ton, "Quat0",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  o->quat[0] = dtemp;
  Tcl_SetStringObj(ton, "Quat1",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  o->quat[1] = dtemp;
  Tcl_SetStringObj(ton, "Quat2",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  o->quat[2] = dtemp;
  Tcl_SetStringObj(ton, "Quat3",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  o->quat[3] = dtemp;

  Tcl_SetStringObj(ton, "Rotate_X",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(o->rotx != dtemp)
    {
      if(!pasteProp)
	{
	  drot = (float)(o->rotx - dtemp);
	  mops_axis_to_quat(xaxis,drot*MOPS_PI/180.0,quat);
	  mops_add_quat(quat,o->quat,o->quat);
	}
      o->rotx = dtemp;
    }
  Tcl_SetStringObj(ton, "Rotate_Y",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(o->roty != dtemp)
    {
      if(!pasteProp)
	{     
	  drot = (float)(o->roty - dtemp);
	  mops_axis_to_quat(yaxis,drot*MOPS_PI/180.0,quat);
	  mops_add_quat(quat,o->quat,o->quat);
	}
      o->roty = dtemp;
    }
  Tcl_SetStringObj(ton, "Rotate_Z",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(o->rotz != dtemp)
    {
      if(!pasteProp)
	{
	  drot = (float)(o->rotz - dtemp);
	  mops_axis_to_quat(zaxis,drot*MOPS_PI/180.0,quat);
	  mops_add_quat(quat,o->quat,o->quat);
	}
      o->rotz = dtemp;
    }

  if(o->rotx == 0.0 && o->roty == 0.0 && o->rotz == 0.0)
    {o->quat[0]=0.0; o->quat[1]=0.0; o->quat[2]=0.0; o->quat[3]=1.0;}

  Tcl_SetStringObj(ton, "Scale_X",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(dtemp != 0.0) o->scalx = dtemp;
  Tcl_SetStringObj(ton, "Scale_Y",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(dtemp != 0.0) o->scaly = dtemp;
  Tcl_SetStringObj(ton, "Scale_Z",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  if(dtemp != 0.0) o->scalz = dtemp;

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  mops_notify_parent();

 return TCL_OK;
}

/* Standard Attributes */

/* C -> Tcl! */
int
mops_get_attrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="attrPropData";
 Tcl_Obj *to = NULL, *ton = NULL, *toa = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 


  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Color_R",-1);
  to = Tcl_NewIntObj(o->colr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_G",-1);
  to = Tcl_NewIntObj(o->colg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_B",-1);
  to = Tcl_NewIntObj(o->colb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Color_A",-1);
  to = Tcl_NewIntObj(o->cola);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Opacity_R",-1);
  to = Tcl_NewIntObj(o->opr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Opacity_G",-1);
  to = Tcl_NewIntObj(o->opg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Opacity_B",-1);
  to = Tcl_NewIntObj(o->opb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Opacity_A",-1);
  to = Tcl_NewIntObj(o->opa);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);


  Tcl_SetStringObj(ton,"ShadingRate",-1);
  to = Tcl_NewDoubleObj(o->shading_rate);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Interpolation",-1);
  to = Tcl_NewIntObj(o->shading_interpolation);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"TrueDisp",-1);
  to = Tcl_NewIntObj(o->true_displacement);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"BoundCoord",-1);
  to = Tcl_NewIntObj(o->d_bound);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Bound",-1);
  to = Tcl_NewDoubleObj(o->d_bound_val);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"CastShadows",-1);
  to = Tcl_NewIntObj(o->cast_shadows);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Editor",-1);
  to = Tcl_NewIntObj(o->editor);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Camera",-1);
  to = Tcl_NewIntObj(o->camera);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Reflection",-1);
  to = Tcl_NewIntObj(o->reflection);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Shadow",-1);
  to = Tcl_NewIntObj(o->shadow);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);


  Tcl_SetStringObj(ton,"AverageColor_R",-1);
  to = Tcl_NewIntObj(o->avr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"AverageColor_G",-1);
  to = Tcl_NewIntObj(o->avg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"AverageColor_B",-1);
  to = Tcl_NewIntObj(o->avb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"AverageColor_A",-1);
  to = Tcl_NewIntObj(o->ava);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"EmissionColor_R",-1);
  to = Tcl_NewIntObj(o->emr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"EmissionColor_G",-1);
  to = Tcl_NewIntObj(o->emg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"EmissionColor_B",-1);
  to = Tcl_NewIntObj(o->emb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"EmissionColor_A",-1);
  to = Tcl_NewIntObj(o->ema);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"SpecularColor_R",-1);
  to = Tcl_NewIntObj(o->spr);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"SpecularColor_G",-1);
  to = Tcl_NewIntObj(o->spg);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"SpecularColor_B",-1);
  to = Tcl_NewIntObj(o->spb);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"SpecularColor_A",-1);
  to = Tcl_NewIntObj(o->spa);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"PatchSize",-1);
  to = Tcl_NewDoubleObj(o->patch_size);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ElemSize",-1);
  to = Tcl_NewDoubleObj(o->elem_size);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"MinSize",-1);
  to = Tcl_NewDoubleObj(o->min_size);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Zonal",-1);
  to = Tcl_NewIntObj(o->zonal);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"HasCaustics",-1);
  to = Tcl_NewIntObj(o->has_caustics);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_attrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="attrPropData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Color_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->colr));
  Tcl_SetStringObj(ton,"Color_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->colg));
  Tcl_SetStringObj(ton,"Color_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->colb));
  Tcl_SetStringObj(ton,"Color_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->cola));

  Tcl_SetStringObj(ton,"Opacity_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->opr));
  Tcl_SetStringObj(ton,"Opacity_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->opg));
  Tcl_SetStringObj(ton,"Opacity_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->opb));
  Tcl_SetStringObj(ton,"Opacity_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->opa));

  Tcl_SetStringObj(ton,"ShadingRate",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(o->shading_rate));

  Tcl_SetStringObj(ton,"Interpolation",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->shading_interpolation)); 

  Tcl_SetStringObj(ton,"TrueDisp",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->true_displacement));

  Tcl_SetStringObj(ton,"BoundCoord",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->d_bound));

  Tcl_SetStringObj(ton,"Bound",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(o->d_bound_val));

  Tcl_SetStringObj(ton,"CastShadows",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->cast_shadows)); 

  Tcl_SetStringObj(ton,"Editor",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->editor)); 

  Tcl_SetStringObj(ton,"Camera",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->camera)); 

  Tcl_SetStringObj(ton,"Reflection",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->reflection)); 

  Tcl_SetStringObj(ton,"Shadow",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->shadow)); 


  Tcl_SetStringObj(ton,"AverageColor_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->avr));
  Tcl_SetStringObj(ton,"AverageColor_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->avg));
  Tcl_SetStringObj(ton,"AverageColor_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->avb));
  Tcl_SetStringObj(ton,"AverageColor_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->ava));

  Tcl_SetStringObj(ton,"EmissionColor_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->emr));
  Tcl_SetStringObj(ton,"EmissionColor_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->emg));
  Tcl_SetStringObj(ton,"EmissionColor_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->emb));
  Tcl_SetStringObj(ton,"EmissionColor_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->ema));

  Tcl_SetStringObj(ton,"SpecularColor_R",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->spr));
  Tcl_SetStringObj(ton,"SpecularColor_G",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->spg));
  Tcl_SetStringObj(ton,"SpecularColor_B",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->spb));
  Tcl_SetStringObj(ton,"SpecularColor_A",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->spa));

  Tcl_SetStringObj(ton,"PatchSize",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(o->patch_size));

  Tcl_SetStringObj(ton,"ElemSize",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(o->elem_size));

  Tcl_SetStringObj(ton,"MinSize",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp, to, &(o->min_size));

  Tcl_SetStringObj(ton,"Zonal",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->zonal));

  Tcl_SetStringObj(ton,"HasCaustics",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp, to, &(o->has_caustics));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  mops_notify_parent();

 return TCL_OK;
}


/* Scene-Preferences */

/* C -> Tcl! */
int
mops_get_prefsprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 char *n1="prefs";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("PickEpsilon",-1);
  to = Tcl_NewDoubleObj(mops_prefs.pick_epsilon);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"OnError",-1);
  to = Tcl_NewIntObj(mops_prefs.onerror);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"SaveViews",-1);
  to = Tcl_NewIntObj(mops_prefs.saveviews);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ReadRiOpt",-1);
  to = Tcl_NewIntObj(mops_prefs.readriopt);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"RIBFile",-1);
  to = Tcl_NewStringObj(mops_prefs.ribfile,-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Image",-1);
  to = Tcl_NewStringObj(mops_prefs.image,-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ResInstances",-1);
  to = Tcl_NewIntObj(mops_prefs.resolveinstances);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"CheckLights",-1);
  to = Tcl_NewIntObj(mops_prefs.checklights);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"RIStandard",-1);
  to = Tcl_NewIntObj(mops_prefs.ristandard);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"LazyNotify",-1);
  to = Tcl_NewIntObj(mops_prefs.lazynotify);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);


  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_prefsprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 char *n1="prefs";
 int itemp = 0;
 double dtemp = 0.0;
 char *string = NULL;
 int stringlen;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 char fname[] = "set_prefsprop";
 
  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("PickEpsilon",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_prefs.pick_epsilon = dtemp;

  Tcl_SetStringObj(ton,"OnError",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.onerror = (char)itemp;

  Tcl_SetStringObj(ton,"SaveViews",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.saveviews = itemp;

  Tcl_SetStringObj(ton,"ReadRiOpt",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.readriopt = itemp;

  Tcl_SetStringObj(ton,"RIBFile",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  string = Tcl_GetStringFromObj(to, &stringlen);
  if(mops_prefs.ribfile)
    { free(mops_prefs.ribfile); mops_prefs.ribfile = NULL; }
  if(!(mops_prefs.ribfile = calloc(stringlen+1,sizeof(char))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }
  strcpy(mops_prefs.ribfile, string);

  Tcl_SetStringObj(ton,"Image",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  string = Tcl_GetStringFromObj(to, &stringlen);
  if(mops_prefs.ribfile)
    { free(mops_prefs.image); mops_prefs.image = NULL; }
  if(!(mops_prefs.image = calloc(stringlen+1,sizeof(char))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }
  strcpy(mops_prefs.image, string);

  Tcl_SetStringObj(ton,"ResInstances",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.resolveinstances = itemp;

  Tcl_SetStringObj(ton,"CheckLights",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.checklights = itemp;

  Tcl_SetStringObj(ton,"RIStandard",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.ristandard = itemp;

  Tcl_SetStringObj(ton,"LazyNotify",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_prefs.lazynotify = itemp;

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

/* RiOptions */

/* C -> Tcl! */
int
mops_get_rioptprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 char *n1="riopt";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("Variance",-1);
  to = Tcl_NewDoubleObj(mops_riopt.Variance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Width",-1);
  to = Tcl_NewIntObj(mops_riopt.width);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_NewIntObj(mops_riopt.height);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Samples_X",-1);
  to = Tcl_NewIntObj(mops_riopt.Samples_X);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Samples_Y",-1);
  to = Tcl_NewIntObj(mops_riopt.Samples_Y);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"FilterFunc",-1);
  to = Tcl_NewIntObj(mops_riopt.FilterFunc);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"FilterWidth",-1);
  to = Tcl_NewIntObj(mops_riopt.FilterWidth);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"FilterHeight",-1);
  to = Tcl_NewIntObj(mops_riopt.FilterHeight);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"ExpGain",-1);
  to = Tcl_NewDoubleObj(mops_riopt.ExpGain);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"ExpGamma",-1);
  to = Tcl_NewDoubleObj(mops_riopt.ExpGamma);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"RGBA_ONE",-1);
  to = Tcl_NewIntObj(mops_riopt.RGBA_ONE);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"RGBA_MIN",-1);
  to = Tcl_NewIntObj(mops_riopt.RGBA_MIN);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"RGBA_MAX",-1);
  to = Tcl_NewIntObj(mops_riopt.RGBA_MAX);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"RGBA_Dither",-1);
  to = Tcl_NewDoubleObj(mops_riopt.RGBA_Dither);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"MinSamples",-1);
  to = Tcl_NewIntObj(mops_riopt.MinSamples);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"MaxSamples",-1);
  to = Tcl_NewIntObj(mops_riopt.MaxSamples);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"MaxRayLevel",-1);
  to = Tcl_NewIntObj(mops_riopt.MaxRayLevel);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"ShadowBias",-1);
  to = Tcl_NewDoubleObj(mops_riopt.ShadowBias);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"PRManSpec",-1);
  to = Tcl_NewIntObj(mops_riopt.PRManSpec);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"RadSteps",-1);
  to = Tcl_NewIntObj(mops_riopt.RadSteps);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"PatchSamples",-1);
  to = Tcl_NewIntObj(mops_riopt.PatchSamples);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Textures",-1);
  to = Tcl_NewStringObj(mops_riopt.textures,-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Includes",-1);
  to = Tcl_NewStringObj(mops_riopt.includes,-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"Shaders",-1);
  to = Tcl_NewStringObj(mops_riopt.shaders,-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"TextureMem",-1);
  to = Tcl_NewIntObj(mops_riopt.texturemem);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"GeomMem",-1);
  to = Tcl_NewIntObj(mops_riopt.geommem);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_rioptprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 char *n1="riopt";
 int itemp = 0; double dtemp = 0.0;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 char *result;
 char fname[] = "set_riopt";

  toa = Tcl_NewStringObj(n1,-1);
 
  ton = Tcl_NewStringObj("Variance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_riopt.Variance = dtemp;
  Tcl_SetStringObj(ton,"Width",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.width = itemp;
  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.height = itemp;
  Tcl_SetStringObj(ton,"Samples_X",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.Samples_X = itemp;
  Tcl_SetStringObj(ton,"Samples_Y",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.Samples_Y = itemp;
  Tcl_SetStringObj(ton,"FilterFunc",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.FilterFunc = (char)itemp;

  Tcl_SetStringObj(ton,"FilterWidth",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.FilterWidth = itemp;
  Tcl_SetStringObj(ton,"FilterHeight",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.FilterHeight = itemp;
  Tcl_SetStringObj(ton,"ExpGain",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_riopt.ExpGain = dtemp;
  Tcl_SetStringObj(ton,"ExpGamma",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_riopt.ExpGamma = dtemp;
  Tcl_SetStringObj(ton,"RGBA_ONE",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.RGBA_ONE = itemp;
  Tcl_SetStringObj(ton,"RGBA_MIN",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.RGBA_MIN = itemp;
  Tcl_SetStringObj(ton,"RGBA_MAX",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.RGBA_MAX = itemp;
  Tcl_SetStringObj(ton,"RGBA_Dither",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_riopt.RGBA_Dither = dtemp;

  Tcl_SetStringObj(ton,"MinSamples",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.MinSamples = itemp;
  Tcl_SetStringObj(ton,"MaxSamples",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.MaxSamples = itemp;
  Tcl_SetStringObj(ton,"MaxRayLevel",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.MaxRayLevel = itemp;
  Tcl_SetStringObj(ton,"ShadowBias",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &dtemp);
  mops_riopt.ShadowBias = dtemp;
  Tcl_SetStringObj(ton,"PRManSpec",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.PRManSpec = itemp;

  Tcl_SetStringObj(ton,"RadSteps",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.RadSteps = itemp;
  Tcl_SetStringObj(ton,"PatchSamples",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.PatchSamples = itemp;

  result = Tcl_GetVar2(interp, n1, "Textures",
		       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  if(mops_riopt.textures)
    {
      free(mops_riopt.textures);
      mops_riopt.textures = NULL;
    }

  if(!(mops_riopt.textures = calloc(strlen(result)+1, sizeof(char))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }
  strcpy(mops_riopt.textures, result);

  result = Tcl_GetVar2(interp, n1, "Includes",
		       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  if(mops_riopt.includes)
    {
      free(mops_riopt.includes);
      mops_riopt.includes = NULL;
    }

  if(!(mops_riopt.includes = calloc(strlen(result)+1, sizeof(char))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }
  strcpy(mops_riopt.includes, result);

  result = Tcl_GetVar2(interp, n1, "Shaders",
		       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  if(mops_riopt.shaders)
    {
      free(mops_riopt.shaders);
      mops_riopt.shaders = NULL;
    }

  if(!(mops_riopt.shaders = calloc(strlen(result)+1, sizeof(char))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }
  strcpy(mops_riopt.shaders, result);


  Tcl_SetStringObj(ton,"TextureMem",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.texturemem = itemp;

  Tcl_SetStringObj(ton,"GeomMem",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &itemp);
  mops_riopt.geommem = itemp;

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

/**********************/
/**********************/
/**********************/

/* list object specific properties */

int
mops_get_level_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

 return mops_status;
}


int
mops_get_root_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
  
  Tcl_UnsetVar(interp,"propertyList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyList","Prefs", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_UnsetVar(interp,"propertyProcList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildPrefsGUI", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyGetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyGetProcList","getPrefsPropp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_UnsetVar(interp,"propertySetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertySetProcList","setPrefsPropp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_UnsetVar(interp,"propertyDataList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyDataList","prefs",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* RiOptions */
  Tcl_SetVar(interp,"propertyList","RiOptions", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildRiOptGUI", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyGetProcList","getRiOptProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertySetProcList","setRiOptProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyDataList","riopt",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* Atmosphere-Shader */
  Tcl_SetVar(interp,"propertyList","AtmShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildvolumeShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList",
	     "getSProp volume volumeShaderData atmosphere",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList",
	     "setSProp volume volumeShaderData atmosphere",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","volumeShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* Imager-Shader */
  Tcl_SetVar(interp,"propertyList","ImgShader", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildimagerShaderGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList",
	     "getSProp imager imagerShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList",
	     "setSProp imager imagerShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","imagerShaderData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}


int
mops_get_view_properties(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_view_info *v = NULL;
 Tcl_DString ds;

  if(o)
    if(o->object)
      v = (mops_view_info *)o->object;
 
  Tcl_UnsetVar(interp,"propertyList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyList","Camera", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_UnsetVar(interp,"propertyProcList", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildCameraGUI", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_UnsetVar(interp,"propertyGetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_DStringInit(&ds);
  Tcl_DStringAppend(&ds, "getCameraPropp ", -1);
  Tcl_DStringAppend(&ds, Togl_Ident(v->togl), -1);
  Tcl_SetVar(interp,"propertyGetProcList", Tcl_DStringValue(&ds),
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_DStringFree(&ds);

  Tcl_UnsetVar(interp,"propertySetProcList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_DStringInit(&ds);
  Tcl_DStringAppend(&ds, "setCameraPropp ", -1);
  Tcl_DStringAppend(&ds, Togl_Ident(v->togl), -1);
  Tcl_SetVar(interp,"propertySetProcList", Tcl_DStringValue(&ds),
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_DStringFree(&ds);

  Tcl_UnsetVar(interp,"propertyDataList", TCL_LEAVE_ERR_MSG |
	       TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyDataList","cameraData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* ViewAttrib */
  Tcl_SetVar(interp,"propertyList","ViewAttrib", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"propertyProcList","buildViewAttribGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_DStringInit(&ds);
  Tcl_DStringAppend(&ds, "getViewAttribPropp ", -1);
  Tcl_DStringAppend(&ds, Togl_Ident(v->togl), -1);
  Tcl_SetVar(interp,"propertyGetProcList", Tcl_DStringValue(&ds),
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_DStringFree(&ds);

  Tcl_DStringInit(&ds);
  Tcl_DStringAppend(&ds, "setViewAttribPropp ", -1);
  Tcl_DStringAppend(&ds, Togl_Ident(v->togl), -1);
  Tcl_SetVar(interp,"propertySetProcList", Tcl_DStringValue(&ds),
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_DStringFree(&ds);

  Tcl_SetVar(interp,"propertyDataList","viewAttribData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}


