/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Quadrics */

int
mops_crtquad_tcmd(Tcl_Interp * interp,
		  int argc, char *argv[], mops_object *o)
{
 mops_sphere_object *sphere = NULL;
 mops_cone_object *cone = NULL;
 mops_disk_object *disk = NULL;
 mops_cylinder_object *cylinder = NULL;
 mops_hyperboloid_object *hyperboloid = NULL;
 mops_paraboloid_object *paraboloid = NULL;
 mops_torus_object *torus = NULL;
 char fname[] = "crtquad";

 if(argc<3)
   {
     mops_error(interp,MOPS_ARGS,fname,"type");
     return TCL_ERROR;
   }

 if(!strcmp(argv[2],"sphere"))
   {
     if(!(sphere = calloc(1,sizeof(mops_sphere_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     sphere->closed = MOPS_TRUE;
     sphere->radius = 1.0;
     sphere->zmin = -1.0;
     sphere->zmax = 1.0;
     sphere->thetamax = 360.0;
     o->type = MOPS_OTSPHERE;
     o->object = sphere;
   } else
 if(!strcmp(argv[2],"cone"))
   {
     if(!(cone = calloc(1,sizeof(mops_cone_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     cone->closed = MOPS_TRUE;
     cone->radius = 1.0;
     cone->height = 1.0;
     cone->thetamax = 360.0;
     o->type = MOPS_OTCONE;
     o->object = cone;
   } else
 if(!strcmp(argv[2],"cylinder"))
   {
     if(!(cylinder = calloc(1,sizeof(mops_cylinder_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     cylinder->closed = MOPS_TRUE;
     cylinder->radius = 1.0;
     cylinder->zmin = -1.0;
     cylinder->zmax = 1.0;
     cylinder->thetamax = 360.0;
     o->type = MOPS_OTCYLINDER;
     o->object = cylinder;
   } else
 if(!strcmp(argv[2],"disk"))
   {
     if(!(disk = calloc(1,sizeof(mops_disk_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     disk->height = 0.0;
     disk->radius = 1.0;
     disk->thetamax = 360.0;
     o->type = MOPS_OTDISK;
     o->object = disk;
   } else
 if(!strcmp(argv[2],"torus"))
   {
     if(!(torus = calloc(1,sizeof(mops_torus_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     torus->closed = MOPS_TRUE;
     torus->majorrad = 0.75;
     torus->minorrad = 0.25;
     torus->phimin = 0.0;
     torus->phimax = 360.0;
     torus->thetamax = 360.0;
     o->type = MOPS_OTTORUS;
     o->object = torus;
   } else
 if(!strcmp(argv[2],"hyperboloid"))
   {
     if(!(hyperboloid = calloc(1,sizeof(mops_hyperboloid_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     hyperboloid->closed = MOPS_TRUE;
     hyperboloid->p1x = 0.0; hyperboloid->p1y = 1.0; hyperboloid->p1z = -0.5;
     hyperboloid->p2x = 1.0; hyperboloid->p2y = 0.0; hyperboloid->p2z = 0.5;
     hyperboloid->thetamax = 360.0;
     o->type = MOPS_OTHYPERBOLOID;
     o->object = hyperboloid;
   } else
 if(!strcmp(argv[2],"paraboloid"))
   {
     if(!(paraboloid = calloc(1,sizeof(mops_paraboloid_object))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_ERROR;
       }
     paraboloid->closed = MOPS_TRUE;
     paraboloid->rmax = 1.0;
     paraboloid->zmin = 0.0;
     paraboloid->zmax = 1.0;
     paraboloid->thetamax = 360.0;
     o->type = MOPS_OTPARABOLOID;
     o->object = paraboloid;
   }

 return TCL_OK;
}


int
mops_free_quadric(void *quadric, int type)
{

  if(!quadric)
    return MOPS_OK;

  free(quadric);

 return MOPS_OK;
}


int
mops_copy_quadric(void *source,
		  void **dest, int type)
{
 int mops_status = MOPS_OK;
 mops_sphere_object *sphere = NULL;
 mops_cone_object *cone = NULL;
 mops_disk_object *disk = NULL;
 mops_cylinder_object *cylinder = NULL;
 mops_hyperboloid_object *hyperboloid = NULL;
 mops_paraboloid_object *paraboloid = NULL;
 mops_torus_object *torus = NULL;

 switch(type)
   {
   case MOPS_OTSPHERE:
     if(!(sphere = calloc(1,sizeof(mops_sphere_object))))
       return MOPS_OUTOFMEM; 
     memcpy(sphere, source, sizeof(mops_sphere_object)); 
     *dest = sphere;

     break;
   case MOPS_OTCONE:
     if(!(cone = calloc(1,sizeof(mops_cone_object))))
       return MOPS_OUTOFMEM; 
     memcpy(cone, source, sizeof(mops_cone_object)); 
     *dest = cone;
     break;
   case MOPS_OTCYLINDER:
     if(!(cylinder = calloc(1,sizeof(mops_cylinder_object))))
       return MOPS_OUTOFMEM; 
     memcpy(cylinder, source, sizeof(mops_cylinder_object)); 
     *dest = cylinder;
     break;
   case MOPS_OTDISK:
     if(!(disk = calloc(1,sizeof(mops_disk_object))))
       return MOPS_OUTOFMEM; 
     memcpy(disk, source, sizeof(mops_disk_object)); 
     *dest = disk;
     break;
   case MOPS_OTTORUS:
     if(!(torus = calloc(1,sizeof(mops_torus_object))))
       return MOPS_OUTOFMEM; 
     memcpy(torus, source, sizeof(mops_torus_object)); 
     *dest = torus;
     break;
   case MOPS_OTHYPERBOLOID:
     if(!(hyperboloid = calloc(1,sizeof(mops_hyperboloid_object))))
       return MOPS_OUTOFMEM; 
     memcpy(hyperboloid, source, sizeof(mops_hyperboloid_object)); 
     *dest = hyperboloid;
     break;
   case MOPS_OTPARABOLOID:
     if(!(paraboloid = calloc(1,sizeof(mops_paraboloid_object))))
       return MOPS_OUTOFMEM; 
     memcpy(paraboloid, source, sizeof(mops_paraboloid_object)); 
     *dest = paraboloid;
     break;
   }

 return mops_status;
}


int
mops_draw_sphere_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_sphere_object *sphere = NULL;
 GLdouble radius = 0.0;
 GLdouble w = 0.0;

  if(!o)
    return MOPS_OK;

  sphere = o->object;
  if(!sphere)
    return MOPS_OK;    

  radius = (GLdouble)(sphere->radius);
  w = (GLdouble)(sqrt(2.0)*0.5);

  /* draw */
  /* xy-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(radius*w,-radius*w,0.0);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(-radius*w,-radius*w,0.0);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(-radius*w,radius*w,0.0);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(radius*w,radius*w,0.0);
  glEnd();
  /* xz-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(radius*w,0.0,-radius*w);
  glVertex3d(0.0,0.0,-radius);
  glVertex3d(-radius*w,0.0,-radius*w);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(-radius*w,0.0,radius*w);
  glVertex3d(0.0,0.0,radius);
  glVertex3d(radius*w,0.0,radius*w);
  glEnd();
  /* yz-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(0.0,radius*w,-radius*w);
  glVertex3d(0.0,0.0,-radius);
  glVertex3d(0.0,-radius*w,-radius*w);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(0.0,-radius*w,radius*w);
  glVertex3d(0.0,0.0,radius);
  glVertex3d(0.0,radius*w,radius*w);
  glEnd();

 return mops_status;
}


int
mops_draw_sphere(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_sphere_object *sphere = NULL;
 GLdouble radius = 0.0;
 GLUquadricObj *qobj = NULL;

  if(!o)
    return MOPS_OK;

  sphere = o->object;
  if(!sphere)
    return MOPS_OK;

  if(!(qobj = gluNewQuadric()))
    return MOPS_OUTOFMEM;

  radius = (GLdouble)(sphere->radius);

  /* draw */
  gluSphere(qobj,radius,10,10);

  gluDeleteQuadric(qobj);

 return mops_status;
}


int
mops_draw_disk_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_disk_object *disk = NULL;
 GLdouble radius = 0.0;
 GLdouble w = 0.0;
 GLdouble height = 0.0;

  if(!o)
    return MOPS_OK;

  disk = o->object;
  if(!disk)
    return MOPS_OK;    

  radius = (GLdouble)(disk->radius);
  height = (GLdouble)(disk->height);
  w = (GLdouble)(sqrt(2.0)*0.5);

  /* draw */
  /* xy-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,height);
  glVertex3d(radius*w,-radius*w,height);
  glVertex3d(0.0,-radius,height);
  glVertex3d(-radius*w,-radius*w,height);
  glVertex3d(-radius,0.0,height);
  glVertex3d(-radius*w,radius*w,height);
  glVertex3d(0.0,radius,height);
  glVertex3d(radius*w,radius*w,height);
  glEnd();
  /* cross */
  glBegin(GL_LINES);
  glVertex3d(-radius,0.0,height);
  glVertex3d(radius,0.0,height);
  glVertex3d(0.0,-radius,height);
  glVertex3d(0.0,radius,height);
  glEnd();

 return mops_status;
}

int
mops_draw_disk(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_disk_object *disk = NULL;
 GLdouble radius = 0.0;
 GLUquadricObj *qobj = NULL;

  if(!o)
    return MOPS_OK;

  disk = o->object;
  if(!disk)
    return MOPS_OK;

  if(!(qobj = gluNewQuadric()))
    return MOPS_OUTOFMEM;

  radius = (GLdouble)(disk->radius);

  glTranslated(0.0,0.0,(GLdouble)(disk->height));
  /* draw */
  gluDisk(qobj,0.0,radius,10,10);

  gluDeleteQuadric(qobj);

 return mops_status;
}

int
mops_draw_cone_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cone_object *cone = NULL;
 GLdouble radius = 0.0;
 GLdouble w = 0.0;
 GLdouble height = 0.0;

  if(!o)
    return MOPS_OK;

  cone = o->object;
  if(!cone)
    return MOPS_OK;    

  radius = (GLdouble)(cone->radius);
  height = (GLdouble)(cone->height);
  w = (GLdouble)(sqrt(2.0)*0.5);

  /* draw */
  /* xy-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(radius*w,-radius*w,0.0);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(-radius*w,-radius*w,0.0);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(-radius*w,radius*w,0.0);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(radius*w,radius*w,0.0);
  glEnd();
  /* cone */
  glBegin(GL_LINES);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(0.0,0.0,height);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(0.0,0.0,height);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(0.0,0.0,height);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(0.0,0.0,height);
  glEnd();

 return mops_status;
}

int
mops_draw_cone(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cone_object *cone = NULL;
 GLdouble radius = 0.0;
 GLdouble height = 0.0;
 GLUquadricObj *qobj = NULL;

  if(!o)
    return MOPS_OK;

  cone = o->object;
  if(!cone)
    return MOPS_OK;

  if(!(qobj = gluNewQuadric()))
    return MOPS_OUTOFMEM;

  radius = (GLdouble)(cone->radius);
  height = (GLdouble)(cone->height);

  /* draw */
  gluCylinder(qobj,radius,0.0,height,10,10);

  gluDeleteQuadric(qobj);

 return mops_status;
}



int
mops_draw_cylinder_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cylinder_object *cylinder = NULL;
 GLdouble radius = 0.0;
 GLdouble w = 0.0;
 GLdouble zmin = 0.0, zmax = 0.0;

  if(!o)
    return MOPS_OK;

  cylinder = o->object;
  if(!cylinder)
    return MOPS_OK;    

  radius = (GLdouble)(cylinder->radius);
  zmin = (GLdouble)(cylinder->zmin);
  zmax = (GLdouble)(cylinder->zmax);
  w = (GLdouble)(sqrt(2.0)*0.5);

  /* draw */
  /* xy-plane-ring */
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,zmin);
  glVertex3d(radius*w,-radius*w,zmin);
  glVertex3d(0.0,-radius,zmin);
  glVertex3d(-radius*w,-radius*w,zmin);
  glVertex3d(-radius,0.0,zmin);
  glVertex3d(-radius*w,radius*w,zmin);
  glVertex3d(0.0,radius,zmin);
  glVertex3d(radius*w,radius*w,zmin);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3d(radius,0.0,zmax);
  glVertex3d(radius*w,-radius*w,zmax);
  glVertex3d(0.0,-radius,zmax);
  glVertex3d(-radius*w,-radius*w,zmax);
  glVertex3d(-radius,0.0,zmax);
  glVertex3d(-radius*w,radius*w,zmax);
  glVertex3d(0.0,radius,zmax);
  glVertex3d(radius*w,radius*w,zmax);
  glEnd();
  /* wall */
  glBegin(GL_LINES);
  glVertex3d(radius,0.0,zmin);
  glVertex3d(radius,0.0,zmax);
  glVertex3d(0.0,-radius,zmin);
  glVertex3d(0.0,-radius,zmax);
  glVertex3d(-radius,0.0,zmin);
  glVertex3d(-radius,0.0,zmax);
  glVertex3d(0.0,radius,zmin);
  glVertex3d(0.0,radius,zmax);
  glEnd();

 return mops_status;
}

int
mops_draw_cylinder(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cylinder_object *cylinder = NULL;
 GLdouble radius = 0.0;
 GLdouble height = 0.0;
 GLUquadricObj *qobj = NULL;

  if(!o)
    return MOPS_OK;

  cylinder = o->object;
  if(!cylinder)
    return MOPS_OK;

  if(!(qobj = gluNewQuadric()))
    return MOPS_OUTOFMEM;

  radius = (GLdouble)(cylinder->radius);
  height = (GLdouble)(cylinder->zmax - cylinder->zmin);

  /* draw */
  glTranslated(0.0,0.0,(GLdouble)(cylinder->zmin)); 
  gluCylinder(qobj,radius,radius,height,10,10);

  gluDeleteQuadric(qobj);

 return mops_status;
}


int
mops_draw_torus_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_torus_object *torus = NULL;
 GLdouble w = 0.0, rmin = 0.0, rmax = 0.0;

  if(!o)
    return MOPS_OK;

  torus = o->object;
  if(!torus)
    return MOPS_OK;    

 w = (GLdouble)(sqrt(2.0)*0.5);
 rmin = (GLdouble)(torus->minorrad);
 rmax = (GLdouble)(torus->majorrad);

 glBegin(GL_LINE_LOOP);
 glVertex3d(rmax,0.0,rmin);
 glVertex3d(rmax*w,-rmax*w,rmin);
 glVertex3d(0.0,-rmax,rmin);
 glVertex3d(-rmax*w,-rmax*w,rmin);
 glVertex3d(-rmax,0.0,rmin);
 glVertex3d(-rmax*w,rmax*w,rmin);
 glVertex3d(0.0,rmax,rmin);
 glVertex3d(rmax*w,rmax*w,rmin);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(rmax,0.0,-rmin);
 glVertex3d(rmax*w,-rmax*w,-rmin);
 glVertex3d(0.0,-rmax,-rmin);
 glVertex3d(-rmax*w,-rmax*w,-rmin);
 glVertex3d(-rmax,0.0,-rmin);
 glVertex3d(-rmax*w,rmax*w,-rmin);
 glVertex3d(0.0,rmax,-rmin);
 glVertex3d(rmax*w,rmax*w,-rmin);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(rmax-rmin,0.0,0.0);
 glVertex3d((rmax-rmin)*w,-(rmax-rmin)*w,0.0);
 glVertex3d(0.0,-(rmax-rmin),0.0);
 glVertex3d(-(rmax-rmin)*w,-(rmax-rmin)*w,0.0);
 glVertex3d(-(rmax-rmin),0.0,0.0);
 glVertex3d(-(rmax-rmin)*w,(rmax-rmin)*w,0.0);
 glVertex3d(0.0,(rmax-rmin),0.0);
 glVertex3d((rmax-rmin)*w,(rmax-rmin)*w,0.0);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(rmax+rmin,0.0,0.0);
 glVertex3d((rmax+rmin)*w,-(rmax+rmin)*w,0.0);
 glVertex3d(0.0,-(rmax+rmin),0.0);
 glVertex3d(-(rmax+rmin)*w,-(rmax+rmin)*w,0.0);
 glVertex3d(-(rmax+rmin),0.0,0.0);
 glVertex3d(-(rmax+rmin)*w,(rmax+rmin)*w,0.0);
 glVertex3d(0.0,(rmax+rmin),0.0);
 glVertex3d((rmax+rmin)*w,(rmax+rmin)*w,0.0);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(rmax,0.0,rmin);
 glVertex3d(rmax-rmin*w,0.0,rmin*w);
 glVertex3d(rmax-rmin,0.0,0.0);
 glVertex3d(rmax-rmin*w,0.0,-rmin*w);
 glVertex3d(rmax,0.0,-rmin);
 glVertex3d(rmax+rmin*w,0.0,-rmin*w);
 glVertex3d(rmax+rmin,0.0,0.0);
 glVertex3d(rmax+rmin*w,0.0,rmin*w);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(0.0,-rmax,rmin);
 glVertex3d(0.0,-rmax-rmin*w,rmin*w);
 glVertex3d(0.0,-(rmax+rmin),0.0);
 glVertex3d(0.0,-rmax-rmin*w,-rmin*w);
 glVertex3d(0.0,-rmax,-rmin);
 glVertex3d(0.0,-rmax+rmin*w,-rmin*w);
 glVertex3d(0.0,-(rmax-rmin),0.0);
 glVertex3d(0.0,-rmax+rmin*w,rmin*w);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(-rmax,0.0,rmin);
 glVertex3d(-rmax-rmin*w,0.0,rmin*w);
 glVertex3d(-(rmax+rmin),0.0,0.0);
 glVertex3d(-rmax-rmin*w,0.0,-rmin*w);
 glVertex3d(-rmax,0.0,-rmin);
 glVertex3d(-rmax+rmin*w,0.0,-rmin*w);
 glVertex3d(-(rmax-rmin),0.0,0.0);
 glVertex3d(-rmax+rmin*w,0.0,rmin*w);
 glEnd();

 glBegin(GL_LINE_LOOP);
 glVertex3d(0.0,rmax,rmin);
 glVertex3d(0.0,rmax+rmin*w,rmin*w);
 glVertex3d(0.0,(rmax+rmin),0.0);
 glVertex3d(0.0,rmax+rmin*w,-rmin*w);
 glVertex3d(0.0,rmax,-rmin);
 glVertex3d(0.0,rmax-rmin*w,-rmin*w);
 glVertex3d(0.0,(rmax-rmin),0.0);
 glVertex3d(0.0,rmax-rmin*w,rmin*w);
 glEnd();

 return mops_status;
}

int
mops_draw_torus(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_torus_object *torus = NULL;

  if(!o)
    return MOPS_OK;

  torus = o->object;
  if(!torus)
    return MOPS_OK;

 return mops_status;
}


int
mops_draw_paraboloid_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_paraboloid_object *paraboloid = NULL;
 double rmax = 0.0, zmax = 0.0, zmin = 0.0, rmin = 0.0, zh = 0.0, rh = 0.0;
 GLdouble w = 0.0;
 int i = 0;

 w = (GLdouble)(sqrt(2.0)*0.5);

  if(!o)
    return MOPS_OK;

  paraboloid = o->object;
  if(!paraboloid)
    return MOPS_OK;    

  rmax = paraboloid->rmax;
  zmax = paraboloid->zmax;
  zmin = paraboloid->zmin;

  /* xy-plane-ring at zmax */
  glBegin(GL_LINE_LOOP);
  glVertex3d(rmax,0.0,zmax);
  glVertex3d(rmax*w,-rmax*w,zmax);
  glVertex3d(0.0,-rmax,zmax);
  glVertex3d(-rmax*w,-rmax*w,zmax);
  glVertex3d(-rmax,0.0,zmax);
  glVertex3d(-rmax*w,rmax*w,zmax);
  glVertex3d(0.0,rmax,zmax);
  glVertex3d(rmax*w,rmax*w,zmax);
  glEnd();

  if(zmin > 0.0)
    {
      /* xy-plane-ring at zmin */
      rmin = sqrt(zmin);
      glBegin(GL_LINE_LOOP);
      glVertex3d(rmin,0.0,zmin);
      glVertex3d(rmin*w,-rmin*w,zmin);
      glVertex3d(0.0,-rmin,zmin);
      glVertex3d(-rmin*w,-rmin*w,zmin);
      glVertex3d(-rmin,0.0,zmin);
      glVertex3d(-rmin*w,rmin*w,zmin);
      glVertex3d(0.0,rmin,zmin);
      glVertex3d(rmin*w,rmin*w,zmin);
      glEnd(); 
    }

  zh = zmin+((zmax-zmin)/2.0);
  rh = sqrt(zh);
  /* xy-plane-ring at zmin */
  glBegin(GL_LINE_LOOP);
  glVertex3d(rh,0.0,zh);
  glVertex3d(rh*w,-rh*w,zh);
  glVertex3d(0.0,-rh,zh);
  glVertex3d(-rh*w,-rh*w,zh);
  glVertex3d(-rh,0.0,zh);
  glVertex3d(-rh*w,rh*w,zh);
  glVertex3d(0.0,rh,zh);
  glVertex3d(rh*w,rh*w,zh);
  glEnd(); 

  /* vertical lines */
  for(i=0;i<4;i++)
    {
      glBegin(GL_LINE_STRIP);
      glVertex3d(rmin,0.0,zmin);
      glVertex3d(rh,0.0,zh);
      glVertex3d(rmax,0.0,zmax);
      glEnd();
      glRotated(90,0.0,0.0,1.0);
    }

 return mops_status;
}

int
mops_draw_paraboloid(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_paraboloid_object *paraboloid = NULL;

  if(!o)
    return MOPS_OK;

  paraboloid = o->object;
  if(!paraboloid)
    return MOPS_OK;

 return mops_status;
}


int
mops_draw_hyperboloid_c(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_hyperboloid_object *hyperboloid = NULL;
 GLdouble p1x = 0.0, p1y = 0.0, p1z = 0.0, p2x = 0.0, p2y = 0.0, p2z = 0.0;
 int i = 0;

  if(!o)
    return MOPS_OK;

  hyperboloid = o->object;
  if(!hyperboloid)
    return MOPS_OK;    

  p1x = hyperboloid->p1x;
  p1y = hyperboloid->p1y;
  p1z = hyperboloid->p1z;

  p2x = hyperboloid->p2x;
  p2y = hyperboloid->p2y;
  p2z = hyperboloid->p2z;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  for(i=0;i<4;i++)
    {
      glBegin(GL_LINES);
      glVertex3d(p1x,p1y,p1z);
      glVertex3d(p2x,p2y,p2z);
      glEnd();
      glRotated(90.0,0.0,0.0,1.0);
    }
  glPopMatrix();
  
  glBegin(GL_LINE_LOOP);
  glVertex3d(p1x,p1y,p1z);
  glVertex3d(p1y,p1x,p1z);
  glVertex3d(-p1x,-p1y,p1z);
  glVertex3d(-p1y,-p1x,p1z);
  glEnd();

  glBegin(GL_LINE_LOOP);
  glVertex3d(p2x,p2y,p2z);
  glVertex3d(p2y,p2x,p2z);
  glVertex3d(-p2x,-p2y,p2z);
  glVertex3d(-p2y,-p2x,p2z);
  glEnd();

 return mops_status;
}

int
mops_draw_hyperboloid(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_hyperboloid_object *hyperboloid = NULL;

  if(!o)
    return MOPS_OK;

  hyperboloid = o->object;
  if(!hyperboloid)
    return MOPS_OK;


 return mops_status;
}


void
mops_wrib_solidcylinder(double radius, double zmin, double zmax)
{

  RiSolidBegin(RI_PRIMITIVE);

  RiCylinder(radius, zmin, zmax, 360.0, RI_NULL);
  RiDisk(zmax, radius, 360.0, RI_NULL);

  RiAttributeBegin();
   RiReverseOrientation();
   RiDisk(zmin, radius, 360.0, RI_NULL);
  RiAttributeEnd();
  RiSolidEnd();

 return;
}


void
mops_wrib_solidsphere(double radius, double zmin, double zmax)
{

  RiSolidBegin(RI_PRIMITIVE);

  RiSphere((RtFloat)radius, (RtFloat)zmin, (RtFloat)zmax, 360.0, RI_NULL);

  if(fabs(zmax) < radius)
    RiDisk((RtFloat)zmax, (RtFloat)sqrt(radius*radius - zmax*zmax),
	   360.0, RI_NULL);

  if(fabs(zmin) < radius)
    {
      RiAttributeBegin();
       RiReverseOrientation();
       RiDisk((RtFloat)zmin, (RtFloat)sqrt(radius*radius - zmin*zmin),
	      360.0, RI_NULL);
      RiAttributeEnd();
    }

  RiSolidEnd();

 return;
}

void
mops_wrib_solidhemisphere(double radius, double zmin, double zmax)
{

  RiSolidBegin(RI_INTERSECTION);

  mops_wrib_solidsphere(radius, zmin, zmax);

  RiRotate(90.0, 1.0, 0.0, 0.0);

  mops_wrib_solidcylinder(radius, 0.0, radius);

  RiSolidEnd();

 return;
}

int
mops_wrib_sphere(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_sphere_object *sphere = NULL;

  if(!o)
   return MOPS_OK;

  sphere = (mops_sphere_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  if(!sphere->closed)
  {
    RiSphere((RtFloat)sphere->radius,
	     (RtFloat)sphere->zmin,
	     (RtFloat)sphere->zmax,
	     (RtFloat)sphere->thetamax,
	     NULL);
  }
  else
  {
    if(sphere->thetamax == 180.0)
      {
	mops_wrib_solidhemisphere(sphere->radius,sphere->zmin,sphere->zmax);
      }
    else if(sphere->thetamax == 360.0)
      {
	mops_wrib_solidsphere(sphere->radius,sphere->zmin,sphere->zmax);
      }
    else if(sphere->thetamax < 180.0)
      {
	RiSolidBegin(RI_INTERSECTION);
	mops_wrib_solidhemisphere(sphere->radius,sphere->zmin,sphere->zmax);
	RiRotate(sphere->thetamax/* - 180.0*/, 0.0,0.0,1.0);
	mops_wrib_solidhemisphere(sphere->radius,sphere->zmin,sphere->zmax);
	RiSolidEnd();
      }
    else if(sphere->thetamax < 360.0)
      {
	RiTransformBegin();
	RiSolidBegin(RI_UNION);
	mops_wrib_solidhemisphere(sphere->radius,sphere->zmin,sphere->zmax);
	RiRotate(sphere->thetamax, 0.0,0.0,1.0);
	mops_wrib_solidhemisphere(sphere->radius,sphere->zmin,sphere->zmax);
	RiSolidEnd();
	RiTransformEnd();
      }
  }
  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

int
mops_wrib_disk(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_disk_object *disk = NULL;

  if(!o)
   return MOPS_OK;

  disk = (mops_disk_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  RiDisk((RtFloat)disk->height,
	 (RtFloat)disk->radius,
	 (RtFloat)disk->thetamax,
	 NULL);

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

void
mops_wrib_solidcone(double height, double radius)
{

  RiSolidBegin(RI_PRIMITIVE);
  RiCone((RtFloat)height, (RtFloat)radius, 360.0, RI_NULL);
  RiAttributeBegin();
   RiReverseOrientation();
   RiDisk(0.0, (RtFloat)radius, 360.0, RI_NULL);
  RiAttributeEnd();
  RiSolidEnd();

 return;
}


int
mops_wrib_cone(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cone_object *cone = NULL;
 RtPoint rect[4];

  if(!o)
   return MOPS_OK;

  cone = (mops_cone_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  if(!cone->closed)
    {
      RiCone((RtFloat)cone->height,
	     (RtFloat)cone->radius,
	     (RtFloat)cone->thetamax,
	     RI_NULL);
    }
  else
    {
    if(cone->thetamax == 180.0)
      {
	RiSolidBegin(RI_PRIMITIVE);

	RiCone((RtFloat)cone->height,
	       (RtFloat)cone->radius,
	       (RtFloat)cone->thetamax,
	       RI_NULL);

	rect[0][0] = (RtFloat)(cone->radius);
	rect[0][1] = 0.0;
	rect[0][2] = 0.0;

	rect[1][0] = (RtFloat)(cone->radius);
	rect[1][1] = 0.0;
	rect[1][2] = 0.0;

	rect[2][0] = 0.0;
	rect[2][1] = 0.0;
	rect[2][2] = (RtFloat)cone->height;

	rect[3][0] = 0.0;
	rect[3][1] = 0.0;
	rect[3][2] = (RtFloat)cone->height;

	RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect,RI_NULL);

	RiAttributeBegin();
	 RiReverseOrientation();
	 RiDisk(0.0, (RtFloat)cone->radius, (RtFloat)cone->thetamax, RI_NULL);
	RiAttributeEnd();

	RiSolidEnd();
      }
    else if(cone->thetamax == 360.0)
      {
	mops_wrib_solidcone(cone->height,cone->radius);
      }
    else
      {
	RiSolidBegin(RI_PRIMITIVE);

	RiCone((RtFloat)cone->height,
	       (RtFloat)cone->radius,
	       (RtFloat)cone->thetamax,
	       RI_NULL);

	rect[0][0] = (RtFloat)cone->radius;
	rect[0][1] = 0.0;
	rect[0][2] = 0.0;

	rect[1][0] = 0.0;
	rect[1][1] = 0.0;
	rect[1][2] = 0.0;

	rect[2][0] = 0.0;
	rect[2][1] = 0.0;
	rect[2][2] = (RtFloat)cone->height;

	rect[3][0] = 0.0;
	rect[3][1] = 0.0;
	rect[3][2] = (RtFloat)cone->height;

	RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);


	RiTransformBegin();
	 RiRotate(cone->thetamax, 0.0, 0.0, 1.0);
	 RiAttributeBegin();
	  RiReverseOrientation();
	  RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
	 RiAttributeEnd();
	 RiTransformEnd();

	 RiAttributeBegin();
	  RiReverseOrientation();
	  RiDisk(0.0, (RtFloat)cone->radius, (RtFloat)cone->thetamax, RI_NULL);
	 RiAttributeEnd();

	RiSolidEnd();
      }
    }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

int
mops_wrib_cylinder(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_cylinder_object *cylinder = NULL;
  RtPoint rect[4];

  if(!o)
   return MOPS_OK;

  cylinder = (mops_cylinder_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);


  if(!cylinder->closed)
    {
      RiCylinder((RtFloat)cylinder->radius,
		 (RtFloat)cylinder->zmin,
		 (RtFloat)cylinder->zmax,
		 (RtFloat)cylinder->thetamax,
		 RI_NULL);
    }
  else
    {
    if(cylinder->thetamax == 180.0)
      {
	RiSolidBegin(RI_PRIMITIVE);
	RiCylinder((RtFloat)cylinder->radius,
		   (RtFloat)cylinder->zmin,
		   (RtFloat)cylinder->zmax,
		   (RtFloat)cylinder->thetamax,
		   RI_NULL);

	rect[0][0] = (RtFloat)cylinder->radius;
	rect[0][1] = 0.0;
	rect[0][2] = (RtFloat)cylinder->zmin;

	rect[1][0] = (RtFloat)(-cylinder->radius);
	rect[1][1] = 0.0;
	rect[1][2] = (RtFloat)cylinder->zmin;

	rect[2][0] = (RtFloat)(cylinder->radius);
	rect[2][1] = 0.0;
	rect[2][2] = (RtFloat)cylinder->zmax;

	rect[3][0] = (RtFloat)(-cylinder->radius);
	rect[3][1] = 0.0;
	rect[3][2] = (RtFloat)cylinder->zmax;

	RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);

	RiAttributeBegin();
 	 RiReverseOrientation();
	 RiDisk((RtFloat)cylinder->zmin, (RtFloat)cylinder->radius,
		(RtFloat)cylinder->thetamax, RI_NULL);
	RiAttributeEnd();

	RiDisk((RtFloat)cylinder->zmax, (RtFloat)cylinder->radius,
	       (RtFloat)cylinder->thetamax, RI_NULL);

	RiSolidEnd();
      }
    else if(cylinder->thetamax == 360.0)
      {
	RiSolidBegin(RI_PRIMITIVE);
	RiCylinder((RtFloat)cylinder->radius,
		   (RtFloat)cylinder->zmin,
		   (RtFloat)cylinder->zmax,
		   (RtFloat)cylinder->thetamax,
		   RI_NULL);

	RiAttributeBegin();
 	 RiReverseOrientation();
	 RiDisk((RtFloat)cylinder->zmin, (RtFloat)cylinder->radius,
		(RtFloat)cylinder->thetamax, RI_NULL);
	RiAttributeEnd(); 

	RiDisk((RtFloat)cylinder->zmax, (RtFloat)cylinder->radius,
	       (RtFloat)cylinder->thetamax, RI_NULL);

	RiSolidEnd();
      }
    else
      {
	RiSolidBegin(RI_PRIMITIVE);
	RiCylinder((RtFloat)cylinder->radius,
		   (RtFloat)cylinder->zmin,
		   (RtFloat)cylinder->zmax,
		   (RtFloat)cylinder->thetamax,
		   RI_NULL);

	rect[0][0] = (RtFloat)cylinder->radius;
	rect[0][1] = 0.0;
	rect[0][2] = (RtFloat)cylinder->zmin;

	rect[1][0] = 0.0;
	rect[1][1] = 0.0;
	rect[1][2] = (RtFloat)cylinder->zmin;

	rect[2][0] = (RtFloat)(cylinder->radius);
	rect[2][1] = 0.0;
	rect[2][2] = (RtFloat)cylinder->zmax;

	rect[3][0] = 0.0;
	rect[3][1] = 0.0;
	rect[3][2] = (RtFloat)cylinder->zmax;


	RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);

	RiTransformBegin();
	RiRotate(cylinder->thetamax,0.0,0.0,1.0);
	RiAttributeBegin();
	 RiReverseOrientation();
	 RiPatch(RI_BILINEAR, RI_P, (RtPointer)rect, RI_NULL);
	RiAttributeEnd();
	RiTransformEnd();

	RiAttributeBegin();
	 RiReverseOrientation();
	RiDisk((RtFloat)cylinder->zmin, (RtFloat)cylinder->radius,
	       (RtFloat)cylinder->thetamax, RI_NULL);
	RiAttributeEnd();

	RiDisk((RtFloat)cylinder->zmax, (RtFloat)cylinder->radius,
	       (RtFloat)cylinder->thetamax, RI_NULL);

	RiSolidEnd();
      }
    }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

int
mops_wrib_torus(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_torus_object *torus = NULL;
  RtPoint p1, p2;

  if(!o)
   return MOPS_OK;

  torus = (mops_torus_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  if(!torus->closed)
    {
      RiTorus((RtFloat)torus->majorrad, (RtFloat)torus->minorrad,
	      (RtFloat)torus->phimin, (RtFloat)torus->phimax,
	      (RtFloat)torus->thetamax,
	      NULL);
    }
  else
    {
      RiSolidBegin(RI_UNION);

      RiTorus((RtFloat)torus->majorrad, (RtFloat)torus->minorrad,
	      (RtFloat)torus->phimin, (RtFloat)torus->phimax,
	      (RtFloat)torus->thetamax,
	      NULL);

      if(torus->thetamax != 360.0)
	{
	  RiTransformBegin();
	  RiTranslate((RtFloat)torus->majorrad, 0.0, 0.0);
	  RiRotate((RtFloat)torus->phimin, 0.0, 0.0, 1.0);
	  RiRotate(90.0, 1.0, 0.0, 0.0);

	  RiDisk(0.0, (RtFloat)torus->minorrad,
		 (RtFloat)(torus->phimax - torus->phimin), RI_NULL);
	  RiTransformEnd();

	  RiTransformBegin();
	  RiRotate((RtFloat)torus->thetamax, 0.0, 0.0, 1.0);
	  RiTranslate((RtFloat)torus->majorrad, 0.0, 0.0);
	  RiRotate((RtFloat)torus->phimin, 0.0, 0.0, 1.0);
	  RiRotate(90.0, 1.0, 0.0, 0.0);

	  RiDisk(0.0, (RtFloat)torus->minorrad,
		 (RtFloat)(torus->phimax - torus->phimin), RI_NULL);
	  RiTransformEnd();

	}

      if(torus->phimin != 0.0 || torus->phimax != 360.0)
	{
	  p1[0] = (RtFloat)torus->majorrad;
	  p1[1] = 0.0;
	  p1[2] = 0.0;

	  p2[0] = (RtFloat)(cos(torus->phimin*MOPS_PI/180.0)*torus->minorrad);
	  p2[1] = 0.0;
	  p2[2] = (RtFloat)(sin(torus->phimin*MOPS_PI/180.0)*torus->minorrad);
	  p2[0] += torus->majorrad;

	  RiHyperboloid(p1, p2, (RtFloat)torus->thetamax, RI_NULL);

	  p2[0] = (RtFloat)(cos(torus->phimax*MOPS_PI/180.0)*torus->minorrad);
	  p2[1] = 0.0;
	  p2[2] = (RtFloat)(sin(torus->phimax*MOPS_PI/180.0)*torus->minorrad);
	  p2[0] += torus->majorrad;

	  RiHyperboloid(p1, p2, (RtFloat)torus->thetamax, RI_NULL);

	}

      RiSolidEnd();
    }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

int
mops_wrib_hyperboloid(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_hyperboloid_object *hyperboloid = NULL;
  RtPoint p1, p2;
  RtPoint patch[4];
  RtFloat angle = 0.0, radius = 0.0;

  if(!o)
   return MOPS_OK;

  hyperboloid = (mops_hyperboloid_object*)(o->object);

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  p1[0] = (RtFloat)hyperboloid->p1x;
  p1[1] = (RtFloat)hyperboloid->p1y;
  p1[2] = (RtFloat)hyperboloid->p1z;
  p2[0] = (RtFloat)hyperboloid->p2x;
  p2[1] = (RtFloat)hyperboloid->p2y;
  p2[2] = (RtFloat)hyperboloid->p2z;

  if(!hyperboloid->closed)
    {
      RiHyperboloid(p1, p2,
		    (RtFloat)hyperboloid->thetamax,
		    NULL);
    }
  else
    {
      RiSolidBegin(RI_PRIMITIVE);
      RiHyperboloid(p1, p2,
		    (RtFloat)hyperboloid->thetamax,
		    NULL);
      if (p1[2] == p2[2])
	{
	  RiSolidEnd();
	  return mops_status;
	}

      radius = (RtFloat)sqrt(p1[0]*p1[0]+p1[1]*p1[1]);
      angle = (RtFloat)(180.0/MOPS_PI * acos(p1[0]/radius));
      if (p1[1] < 0.0)
	angle = -angle;

      RiTransformBegin();
       RiRotate(angle, 0.0, 0.0, 1.0);
       RiDisk(p1[2], radius, (RtFloat)hyperboloid->thetamax, RI_NULL);
      RiTransformEnd();     

      radius = (RtFloat)sqrt(p2[0]*p2[0] + p2[1]*p2[1]);
      angle = (RtFloat)(180.0/MOPS_PI * acos(p2[0]/radius));
      if (p2[1] < 0.0)
	angle = -angle;

      RiAttributeBegin();
       RiRotate(angle, 0.0, 0.0, 1.0);
       RiReverseOrientation();
       RiDisk(p2[2], radius, (RtFloat)hyperboloid->thetamax, RI_NULL);
      RiAttributeEnd();

      if(hyperboloid->thetamax != 360.0)
	{
	  patch[0][0] = p2[0]; patch[0][1] = p2[1]; patch[0][2] = p2[2];
	  patch[1][0] = 0.0;   patch[1][1] = 0.0;   patch[1][2] = p2[2];
	  patch[2][0] = p1[0]; patch[2][1] = p1[1]; patch[2][2] = p1[2];
	  patch[3][0] = 0.0;   patch[3][1] = 0.0;   patch[3][2] = p1[2];

	  RiPatch(RI_BILINEAR, RI_P, (RtPointer)patch, RI_NULL);
	  RiAttributeBegin();
	   RiReverseOrientation();
	   RiRotate((RtFloat)hyperboloid->thetamax, 0.0, 0.0, 1.0);
	   RiPatch(RI_BILINEAR, RI_P, (RtPointer)patch, RI_NULL);
	  RiAttributeEnd();
	}
      RiSolidEnd();
    }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}

int
mops_wrib_paraboloid(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_paraboloid_object *paraboloid = NULL;
  RtFloat rmin = 0.0;
  RtPoint patch[16];
  RtFloat  Sx, Sy, B0, B3, K;

  if(!o)
   return MOPS_OK;

  paraboloid = (mops_paraboloid_object*)(o->object);

  memset(patch, 0, sizeof(patch));

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  if(!paraboloid->closed)
    {
      RiParaboloid((RtFloat)paraboloid->rmax,
		   (RtFloat)paraboloid->zmin,
		   (RtFloat)paraboloid->zmax,
		   (RtFloat)paraboloid->thetamax,
		   NULL);
    }
  else
    {
      RiSolidBegin(RI_PRIMITIVE);

      RiParaboloid((RtFloat)paraboloid->rmax,
		   (RtFloat)paraboloid->zmin,
		   (RtFloat)paraboloid->zmax,
		   (RtFloat)paraboloid->thetamax,
		   NULL);
      RiDisk((RtFloat)paraboloid->zmax,(RtFloat)paraboloid->rmax,
	     (RtFloat)paraboloid->thetamax, RI_NULL);


      if(paraboloid->zmin > 0.0)
	{
	  rmin = (RtFloat)sqrt(paraboloid->zmin/paraboloid->zmax);
	  RiAttributeBegin();
 	   RiReverseOrientation();
	   RiDisk((RtFloat)paraboloid->zmin,
		 rmin,
		 (RtFloat)paraboloid->thetamax, RI_NULL);
	  RiAttributeEnd();
	}

      if(paraboloid->thetamax != 360.0)
	{
	  patch[0][2] = patch[1][2] = patch[2][2] = patch[3][2] =
	    (RtFloat)paraboloid->zmin;
	  patch[3][0] = (RtFloat) (paraboloid->zmin > 0.0 ? paraboloid->rmax *
				   rmin : 0.0);
	  patch[1][0] = (RtFloat)(patch[3][0]/3.0);
	  patch[2][0] = (RtFloat)(patch[1][0]*2.0);

	  patch[12][2] = patch[13][2] = patch[14][2] = patch[15][2] =
	    (RtFloat)paraboloid->zmax;
	  patch[15][0] = (RtFloat)paraboloid->rmax;
	  patch[13][0] = (RtFloat)(patch[15][0]/3.0);
	  patch[14][0] = (RtFloat)(patch[13][0]*2.0);

	  K = paraboloid->rmax*paraboloid->rmax;
	  B0 = patch[0][2]  - (2*paraboloid->zmin)/K;
	  B3 = patch[12][2] - 2*paraboloid->zmax;
	  K = (2*paraboloid->zmax)/K;
	  Sx = (B3 - B0)/( K*(rmin - paraboloid->rmax) );
	  Sy = K*rmin * Sx + B0;  
	  
	  patch[4][2] = patch[5][2] = patch[6][2] = patch[7][2] =
	    (RtFloat)((patch[0][2] + 2*Sy)/3.0);
	  patch[7][0] = (RtFloat)((rmin + 2*Sx)/3.0);
	  patch[5][0] = (RtFloat)patch[7][0]/3.0;
	  patch[6][0] = (RtFloat)(patch[5][0]*2.0);

	  patch[8][2] = patch[9][2] = patch[10][2] = patch[11][2] =
	    (RtFloat)((patch[12][2] + 2*Sy)/3.0);
	  patch[11][0] = (RtFloat)((paraboloid->rmax + 2*Sx)/3.0);
	  patch[9][0] = (RtFloat)(patch[11][0]/3.0);
	  patch[10][0] = (RtFloat)(patch[9][0]*2.0);

	  RiPatch(RI_BICUBIC, RI_P, (RtPointer)patch, RI_NULL);
	  RiAttributeBegin();
	   RiRotate((RtFloat)paraboloid->thetamax, 0.0, 0.0, 1.0);
	   RiReverseOrientation();
	   RiPatch(RI_BICUBIC, RI_P, (RtPointer)patch, RI_NULL);
	  RiAttributeEnd();
	}

      RiSolidEnd();
    }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}


int
mops_get_quadric_properties(Tcl_Interp *interp, int type)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  switch(type)
    {
    case MOPS_OTSPHERE:
      Tcl_SetVar(interp,"propertyList","RiSphereAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildSphereAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr sphere",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr sphere",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;

    case MOPS_OTCONE:
      Tcl_SetVar(interp,"propertyList","RiConeAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildConeAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr cone",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr cone",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_OTCYLINDER:
      Tcl_SetVar(interp,"propertyList","RiCylinderAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildCylinderAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr cylinder",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr cylinder",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_OTDISK:
      Tcl_SetVar(interp,"propertyList","RiDiskAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildDiskAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr disk",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr disk",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_OTHYPERBOLOID:
      Tcl_SetVar(interp,"propertyList","RiHyperbAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildHyperbAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr hyperboloid",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr hyperboloid",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_OTPARABOLOID:
      Tcl_SetVar(interp,"propertyList","RiParabAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildParabAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr paraboloid",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr paraboloid",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_OTTORUS:
      Tcl_SetVar(interp,"propertyList","RiTorusAttr", TCL_APPEND_VALUE |\
		 TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyProcList","buildTorusAttrGUI",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertyGetProcList","getQAttr torus",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      Tcl_SetVar(interp,"propertySetProcList","setQAttr torus",
		 TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
		 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    } /* switch */


  Tcl_SetVar(interp,"propertyDataList","QAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}



/* Tcl -> C! */
int
mops_set_qattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="QAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 mops_sphere_object *sphere = NULL;
 mops_cone_object *cone = NULL;
 mops_disk_object *disk = NULL;
 mops_cylinder_object *cylinder = NULL;
 mops_hyperboloid_object *hyperboloid = NULL;
 mops_paraboloid_object *paraboloid = NULL;
 mops_torus_object *torus = NULL;
 char fname[] = "set_qattr";
 int itemp = 0;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;
  
  if(argc<2)
    {
      mops_error(interp,MOPS_ARGS,fname,"type");
      return TCL_OK;
    }

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj(n1,-1);

  if(!(strcmp(argv[1],"sphere")))
    {
      sphere = (mops_sphere_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      sphere->closed = (char)itemp;

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &sphere->radius);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &sphere->zmin);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &sphere->zmax);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &sphere->thetamax);

    } else
  if(!(strcmp(argv[1],"cone")))
    {
      cone = (mops_cone_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      cone->closed = (char)itemp;

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cone->radius);

      Tcl_SetStringObj(ton,"Height",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cone->height);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cone->thetamax);

    } else
  if(!(strcmp(argv[1],"disk")))
    {
      disk = (mops_disk_object *)o->object;

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &disk->radius);

      Tcl_SetStringObj(ton,"Height",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &disk->height);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &disk->thetamax);
    } else
  if(!(strcmp(argv[1],"cylinder")))
    {
      cylinder = (mops_cylinder_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      cylinder->closed = (char)itemp;

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cylinder->radius);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cylinder->zmin);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cylinder->zmax);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &cylinder->thetamax);

    } else
  if(!(strcmp(argv[1],"hyperboloid")))
    {
      hyperboloid = (mops_hyperboloid_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      hyperboloid->closed = (char)itemp;

      Tcl_SetStringObj(ton,"P1_X",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p1x);
      Tcl_SetStringObj(ton,"P1_Y",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p1y);
      Tcl_SetStringObj(ton,"P1_Z",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p1z);

      Tcl_SetStringObj(ton,"P2_X",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p2x);
      Tcl_SetStringObj(ton,"P2_Y",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p2y);
      Tcl_SetStringObj(ton,"P2_Z",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->p2z);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &hyperboloid->thetamax);
    } else
  if(!(strcmp(argv[1],"paraboloid")))
    {
      paraboloid = (mops_paraboloid_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      paraboloid->closed = (char)itemp;

      Tcl_SetStringObj(ton,"RMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &paraboloid->rmax);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &paraboloid->zmin);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &paraboloid->zmax);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &paraboloid->thetamax);
    } else
  if(!(strcmp(argv[1],"torus")))
    {
      torus = (mops_torus_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetIntFromObj(interp,to, &itemp);
      torus->closed = (char)itemp;

      Tcl_SetStringObj(ton,"MajorRad",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &torus->majorrad);
      Tcl_SetStringObj(ton,"MinorRad",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &torus->minorrad);

      Tcl_SetStringObj(ton,"PhiMin",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &torus->phimin);

      Tcl_SetStringObj(ton,"PhiMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &torus->phimax);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_GetDoubleFromObj(interp,to, &torus->thetamax);
    }

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

/* C -> Tcl! */
int
mops_get_qattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="QAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 mops_sphere_object *sphere = NULL;
 mops_cone_object *cone = NULL;
 mops_disk_object *disk = NULL;
 mops_cylinder_object *cylinder = NULL;
 mops_hyperboloid_object *hyperboloid = NULL;
 mops_paraboloid_object *paraboloid = NULL;
 mops_torus_object *torus = NULL;
 char fname[] = "set_qattr";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(argc<2)
    {
      mops_error(interp,MOPS_ARGS,fname,"type");
      return TCL_OK;
    }

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj(n1,-1);

  if(!(strcmp(argv[1],"sphere")))
    {
      sphere = (mops_sphere_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(sphere->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_NewDoubleObj(sphere->radius);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_NewDoubleObj(sphere->zmin);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_NewDoubleObj(sphere->zmax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(sphere->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"cone")))
    {
      cone = (mops_cone_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(cone->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_NewDoubleObj(cone->radius);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"Height",-1);
      to = Tcl_NewDoubleObj(cone->height);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(cone->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"disk")))
    {
      disk = (mops_disk_object *)o->object;

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_NewDoubleObj(disk->radius);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"Height",-1);
      to = Tcl_NewDoubleObj(disk->height);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(disk->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"cylinder")))
    {
      cylinder = (mops_cylinder_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(cylinder->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"Radius",-1);
      to = Tcl_NewDoubleObj(cylinder->radius);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_NewDoubleObj(cylinder->zmin);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_NewDoubleObj(cylinder->zmax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(cylinder->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"hyperboloid")))
    {
      hyperboloid = (mops_hyperboloid_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(hyperboloid->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"P1_X",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p1x);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
      Tcl_SetStringObj(ton,"P1_Y",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p1y);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
      Tcl_SetStringObj(ton,"P1_Z",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p1z);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"P2_X",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p2x);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
      Tcl_SetStringObj(ton,"P2_Y",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p2y);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
      Tcl_SetStringObj(ton,"P2_Z",-1);
      to = Tcl_NewDoubleObj(hyperboloid->p2z);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(hyperboloid->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"paraboloid")))
    {
      paraboloid = (mops_paraboloid_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(paraboloid->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"RMax",-1);
      to = Tcl_NewDoubleObj(paraboloid->rmax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMin",-1);
      to = Tcl_NewDoubleObj(paraboloid->zmin);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ZMax",-1);
      to = Tcl_NewDoubleObj(paraboloid->zmax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(paraboloid->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    } else
    if(!(strcmp(argv[1],"torus")))
    {
      torus = (mops_torus_object *)o->object;

      Tcl_SetStringObj(ton,"Closed",-1);
      to = Tcl_NewIntObj(torus->closed);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"MajorRad",-1);
      to = Tcl_NewDoubleObj(torus->majorrad);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"MinorRad",-1);
      to = Tcl_NewDoubleObj(torus->minorrad);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"PhiMin",-1);
      to = Tcl_NewDoubleObj(torus->phimin);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"PhiMax",-1);
      to = Tcl_NewDoubleObj(torus->phimax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);

      Tcl_SetStringObj(ton,"ThetaMax",-1);
      to = Tcl_NewDoubleObj(torus->thetamax);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    }

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

