/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Read Scenes */

int
mops_read_string(FILE *fileptr, char **result)
{
 int mops_status = MOPS_OK;
 char readchar, *str;
 int read;
 Tcl_DString ds;

  Tcl_DStringInit(&ds);
  read = getc(fileptr);

  if(read == EOF)
    {Tcl_DStringFree(&ds); return MOPS_UEOF;}

  if((char)read == '\n')
    {Tcl_DStringFree(&ds); return MOPS_OK;}

  while((char)read != '\n')
    {
      readchar = (char)read;

      Tcl_DStringAppend(&ds, &readchar, 1);
      read = getc(fileptr);

      if(read == EOF)
	{Tcl_DStringFree(&ds); return MOPS_UEOF;}

    }

  Tcl_DStringAppend(&ds,"\0", 1);

  str = Tcl_DStringValue(&ds);

  if(str[strlen(str)-1] == '\r')
    {
      str[strlen(str)-1] = '\0';
    }

  if(!(*result = calloc(strlen(str)+1,sizeof(char))))
    {Tcl_DStringFree(&ds); return MOPS_OUTOFMEM;}

  strcpy(*result,str);

  Tcl_DStringFree(&ds);

 return mops_status;
}

/* skip to next entity, marked by a single Bell, or two Bell's! */
int
mops_read_skip(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 int read;

  read = getc(fileptr);

  if(read == EOF)
    {return MOPS_EOF;}

  if((char)read == '\a')
    {return MOPS_OK;}

  while((char)read != '\a')
    {
      read = getc(fileptr);

      if(read == EOF)
	return MOPS_EOF;
    }

 return mops_status;
}


int 
mops_read_header(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 char *version = NULL, *nbuffer = NULL /*nbuffer[7]*/;
 

 /*
   fgets(nbuffer,7,fileptr);
   if(strcmp(nbuffer,"Mops"))
   return MOPS_NOTAMOPS;
 */

 mops_read_string(fileptr, &nbuffer);
 if(!strstr(nbuffer,"Mops"))
   return MOPS_NOTAMOPS;

 mops_read_string(fileptr, &version);

 mops_read_version = 6;

 if(!strcmp(version,"0.42a"))
   mops_read_version = 0;
 else if(!strcmp(version,"0.42a2"))
   mops_read_version = 1;
 else if(!strcmp(version,"0.42a3"))
   mops_read_version = 2;
 else if(!strcmp(version,"0.42a4"))
   mops_read_version = 3;
 else if(!strcmp(version,"0.42a5"))
   mops_read_version = 4;
 else if(!strcmp(version,"0.42b"))
   mops_read_version = 5;
 else if(!strcmp(version,"0.42c"))
   mops_read_version = 6;
 else if(!strcmp(version,"0.42d"))
   mops_read_version = 7;

 free(version);

 return mops_status;
}


int 
mops_read_rioptions(FILE *fileptr, int insert)
{
 int mops_status = MOPS_OK;
 int itemp = 0;
 mops_riopt_object *riopt, otemp = {0};
 int read;

  if(insert || (! mops_prefs.readriopt))
    riopt = &otemp;
  else
    riopt = &mops_riopt;


  fscanf(fileptr,"%lg\n",&riopt->Variance);
  fscanf(fileptr,"%d\n",&riopt->Samples_X);
  fscanf(fileptr,"%d\n",&riopt->Samples_Y);
  fscanf(fileptr,"%d\n",&itemp);
  riopt->FilterFunc = (char)itemp;
  fscanf(fileptr,"%d\n",&riopt->FilterWidth);
  fscanf(fileptr,"%d\n",&riopt->FilterHeight);
  fscanf(fileptr,"%lg\n",&riopt->ExpGain);
  fscanf(fileptr,"%lg\n",&riopt->ExpGamma);
  fscanf(fileptr,"%lg\n",&riopt->RGBA_ONE);
  fscanf(fileptr,"%lg\n",&riopt->RGBA_MIN);
  fscanf(fileptr,"%lg\n",&riopt->RGBA_MAX);
  fscanf(fileptr,"%lg\n",&riopt->RGBA_Dither);
  fscanf(fileptr,"%d\n",&riopt->MinSamples);
  fscanf(fileptr,"%d\n",&riopt->MaxSamples);
  fscanf(fileptr,"%d\n",&riopt->MaxRayLevel);
  fscanf(fileptr,"%lg\n",&riopt->ShadowBias);
  fscanf(fileptr,"%d\n",&itemp);
  riopt->PRManSpec = itemp;
  fscanf(fileptr,"%d\n",&riopt->RadSteps);
  fscanf(fileptr,"%d",&riopt->PatchSamples);
  read = fgetc(fileptr);

  if(read == '\r')
    fgetc(fileptr);

  if(riopt->textures)
    {
      free(riopt->textures);
      riopt->textures = NULL;
    }
  mops_read_string(fileptr, &(riopt->textures));
  if(riopt->includes)
    {
      free(riopt->includes);
      riopt->includes = NULL;
    }
  mops_read_string(fileptr, &(riopt->includes));
  if(riopt->shaders)
    {
      free(riopt->shaders);
      riopt->shaders = NULL;
    }
  mops_read_string(fileptr, &(riopt->shaders));

  fscanf(fileptr,"%d\n",&riopt->texturemem);
  fscanf(fileptr,"%d\n",&riopt->geommem);

  if(mops_read_version > 4)
    {
      fscanf(fileptr,"%d\n",&riopt->width);
      fscanf(fileptr,"%d\n",&riopt->height);
    }

 return mops_status;
}


int 
mops_read_views(Tcl_Interp *interp, FILE *fileptr, int insert)
{
 int mops_status = MOPS_OK;
 int viewcount = 0, i;
 int width, height;
 char command[255] = {0}, update_cmd[] = "update";
 mops_object *root = mops_root, *o;
 mops_view_info vtemp = {0}, *v;

  /* skip to next entity */
  mops_status = mops_read_skip(fileptr);
  fscanf(fileptr,"%d\n", &viewcount);

  for(i=0;i<viewcount;i++)
    {
      /* skip to next entity */
      mops_status = mops_read_skip(fileptr);
      fscanf(fileptr,"%d\n", &width);
      fscanf(fileptr,"%d\n", &height);
      fscanf(fileptr,"%d\n", &vtemp.type);
      vtemp.up_x = 0.0;
      vtemp.up_y = 0.0;
      vtemp.up_z = 0.0;
      if(vtemp.type != MOPS_VTTOP)
	{
	  vtemp.up_y=1.0;
	}
      else
	{
	  vtemp.up_z=-1.0;
	}
      fscanf(fileptr,"%lg\n", &vtemp.grid);
      fscanf(fileptr,"%lg\n", &vtemp.bgr);
      fscanf(fileptr,"%lg\n", &vtemp.bgg);
      fscanf(fileptr,"%lg\n", &vtemp.bgb);
      fscanf(fileptr,"%lg\n", &vtemp.bga);
      fscanf(fileptr,"%d\n", &vtemp.drawwire);
      fscanf(fileptr,"%d\n", &vtemp.drawsurf);
      fscanf(fileptr,"%d\n", &vtemp.drawsel);
      fscanf(fileptr,"%d\n", &vtemp.redraw);
      fscanf(fileptr,"%d\n", &vtemp.drawgrid);
      fscanf(fileptr,"%lg\n", &vtemp.from_x);
      fscanf(fileptr,"%lg\n", &vtemp.from_y);
      fscanf(fileptr,"%lg\n", &vtemp.from_z);
      fscanf(fileptr,"%lg\n", &vtemp.to_x);
      fscanf(fileptr,"%lg\n", &vtemp.to_y);
      fscanf(fileptr,"%lg\n", &vtemp.to_z);
      fscanf(fileptr,"%lg\n", &vtemp.roll);
      fscanf(fileptr,"%lg\n", &vtemp.zoom);
      if(mops_read_version > 0)
	fscanf(fileptr,"%d\n", &vtemp.lighting);
      if(mops_read_version > 2)
	{
	  fscanf(fileptr,"%d\n", &vtemp.pos_x);
	  fscanf(fileptr,"%d\n", &vtemp.pos_y);

	}

      if(mops_read_version > 4)
	{
	  fscanf(fileptr,"%lg\n", &vtemp.rotx);
	  fscanf(fileptr,"%lg\n", &vtemp.roty);
	  fscanf(fileptr,"%lg\n", &vtemp.rotz);
	  fscanf(fileptr,"%lg\n", &vtemp.up_x);
	  fscanf(fileptr,"%lg\n", &vtemp.up_y);
	  fscanf(fileptr,"%lg\n", &vtemp.up_z);
	}
      else
	{
	  vtemp.rotx = 0.0;
	  vtemp.rotz = 0.0;
	  vtemp.roty = 180+MOPS_R2D(atan2((vtemp.to_x-vtemp.from_x),
					  (vtemp.to_z-vtemp.from_z)));

	}

      vtemp.drawhandles = MOPS_TRUE;

      /* open the view */
      sprintf(command,"newViewp %d %d\n",width,height);
      Tcl_Eval(interp, command);
      Tcl_Eval(interp, update_cmd);
      
      /* configure it */
      o = root->down;
      while(o->next)
	o = o->next;

      v = (mops_view_info*)o->object;
      vtemp.togl = v->togl;
      memcpy(v, &vtemp, sizeof(mops_view_info));
      reshape_cb(v->togl);


      /* set window title */
      if(vtemp.type == MOPS_VTFRONT)
	sprintf(command,
	 "global mops_views;viewTitlep [lindex $mops_views end] Front None");
      if(vtemp.type == MOPS_VTSIDE)
	sprintf(command,
	 "global mops_views;viewTitlep [lindex $mops_views end] Side None");
      if(vtemp.type == MOPS_VTTOP)
	sprintf(command,
	"global mops_views;viewTitlep [lindex $mops_views end] Top None");
      if(vtemp.type == MOPS_VTPERSP)
	sprintf(command,
         "global mops_views;viewTitlep [lindex $mops_views end] Persp None");
      if(vtemp.type == MOPS_VTTRIM)
	sprintf(command,
         "global mops_views;viewTitlep [lindex $mops_views end] Trim None");

      Tcl_Eval(interp, command);
      /* position window */
      sprintf(command,
"global mops_views;moveOrResize [lindex $mops_views end] \"+%d+%d\"\n",
	      vtemp.pos_x, vtemp.pos_y);

      Tcl_Eval(interp, command);
      Tcl_Eval(interp, update_cmd);

    }

 return mops_status;
}


int 
mops_read_root(Tcl_Interp *interp, FILE *fileptr, int insert)
{
 int mops_status = MOPS_OK;
 mops_object *root = mops_root;

 if(!root)
   return MOPS_OK;

 mops_status = mops_read_rioptions(fileptr, insert);
 if(mops_status)
   return mops_status;

 mops_status = mops_read_views(interp, fileptr, insert);
 if(mops_status)
   return mops_status;

 return mops_status;
}


int 
mops_read_level(FILE *fileptr, mops_object **optr)
{
 int mops_status = MOPS_OK;
 mops_level_object *level = NULL;
 int itemp = 0;
 mops_object *o = NULL;
 char read;

  o = *optr;

  if(!(level = calloc(1, sizeof(mops_level_object))))
    { return MOPS_OUTOFMEM; }

  o->object = level;

  if(mops_read_version > 0)
    {
      fscanf(fileptr,"%d",&itemp);
      read = (char)fgetc(fileptr);
      if(read == '\r')
	fgetc(fileptr);
    }
  else
    fscanf(fileptr,"%d\n",&itemp);

  level->type = (char)itemp;

  if(level->type == MOPS_LTENDLEVEL)
    {
      if(mops_read_version > 0)
	fscanf(fileptr,"\n");

      free(level);
      free(o);
      *optr = NULL;
      mops_del_level_object();
      mops_next_object = &(mops_current_level->object->next);
      mops_last_read_object = mops_current_level->object;
      mops_del_level_object();

    }
  else if(level->type == MOPS_LTDOWN)
    {
      if(mops_read_version > 0)
	fscanf(fileptr,"\n");
      free(level);
      free(o);
      *optr = NULL;
      mops_add_level_object(mops_last_read_object);
      mops_add_level_object(NULL);
      mops_next_object = &(mops_last_read_object->down);
    }
  else
    {
      /* get name of object */
      if(mops_read_version > 0)
	mops_read_string(fileptr, &(o->name));

      mops_status = mops_default_attributes(o);
      mops_status = mops_link_new_object(o);
      mops_add_level_object(o);
      mops_add_level_object(NULL);
      mops_next_object = &(o->down);
      *optr = NULL;
    }
 return mops_status;
}


int 
mops_read_attributes(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_last_read_object;

  if(!o)
    return MOPS_OK;

 fscanf(fileptr,"%lg\n",&o->movx);
 fscanf(fileptr,"%lg\n",&o->movy);
 fscanf(fileptr,"%lg\n",&o->movz);

 fscanf(fileptr,"%lg\n",&o->rotx);
 fscanf(fileptr,"%lg\n",&o->roty);
 fscanf(fileptr,"%lg\n",&o->rotz);

 fscanf(fileptr,"%g\n",&o->quat[0]);
 fscanf(fileptr,"%g\n",&o->quat[1]);
 fscanf(fileptr,"%g\n",&o->quat[2]);
 fscanf(fileptr,"%g\n",&o->quat[3]);

 fscanf(fileptr,"%lg\n",&o->scalx);
 fscanf(fileptr,"%lg\n",&o->scaly);
 fscanf(fileptr,"%lg\n",&o->scalz);
 
 fscanf(fileptr,"%d\n",&o->colr);
 fscanf(fileptr,"%d\n",&o->colg);
 fscanf(fileptr,"%d\n",&o->colb);
 fscanf(fileptr,"%d\n",&o->cola);

 fscanf(fileptr,"%d\n",&o->opr);
 fscanf(fileptr,"%d\n",&o->opg);
 fscanf(fileptr,"%d\n",&o->opb);
 fscanf(fileptr,"%d\n",&o->opa);

 fscanf(fileptr,"%lg\n",&o->shading_rate);
 fscanf(fileptr,"%d\n",&o->shading_interpolation);

 fscanf(fileptr,"%d\n",&o->cast_shadows);

 fscanf(fileptr,"%d\n",&o->true_displacement);
 fscanf(fileptr,"%d\n",&o->d_bound);
 fscanf(fileptr,"%lg\n",&o->d_bound_val);

 if(mops_read_version > 1)
   {

     fscanf(fileptr,"%d\n",&o->avr);
     fscanf(fileptr,"%d\n",&o->avg);
     fscanf(fileptr,"%d\n",&o->avb);
     fscanf(fileptr,"%d\n",&o->ava);

     fscanf(fileptr,"%d\n",&o->emr);
     fscanf(fileptr,"%d\n",&o->emg);
     fscanf(fileptr,"%d\n",&o->emb);
     fscanf(fileptr,"%d\n",&o->ema);

     fscanf(fileptr,"%d\n",&o->spr);
     fscanf(fileptr,"%d\n",&o->spg);
     fscanf(fileptr,"%d\n",&o->spb);
     fscanf(fileptr,"%d\n",&o->spa);

     fscanf(fileptr,"%lg\n",&o->patch_size);
     fscanf(fileptr,"%lg\n",&o->elem_size);
     fscanf(fileptr,"%lg\n",&o->min_size);

     fscanf(fileptr,"%d\n",&o->zonal);
     fscanf(fileptr,"%d\n",&o->has_caustics);

   }
 if(mops_read_version > 2)
   {
     fscanf(fileptr,"%d\n",&o->camera);
     fscanf(fileptr,"%d\n",&o->reflection);
     fscanf(fileptr,"%d\n",&o->shadow);
   }

 if(mops_read_version > 4)
   {
     fscanf(fileptr,"%d\n",&o->editor);
   }
     
 return mops_status;
}

int 
mops_read_trafos(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_last_read_object;

  if(!o)
    return MOPS_OK;

 fscanf(fileptr,"%lg\n",&o->movx);
 fscanf(fileptr,"%lg\n",&o->movy);
 fscanf(fileptr,"%lg\n",&o->movz);

 fscanf(fileptr,"%lg\n",&o->rotx);
 fscanf(fileptr,"%lg\n",&o->roty);
 fscanf(fileptr,"%lg\n",&o->rotz);

 fscanf(fileptr,"%g\n",&o->quat[0]);
 fscanf(fileptr,"%g\n",&o->quat[1]);
 fscanf(fileptr,"%g\n",&o->quat[2]);
 fscanf(fileptr,"%g\n",&o->quat[3]);

 fscanf(fileptr,"%lg\n",&o->scalx);
 fscanf(fileptr,"%lg\n",&o->scaly);
 fscanf(fileptr,"%lg\n",&o->scalz); 

 return mops_status;
}


/* XXXX leaks memory in low-mem situation */
int 
mops_read_shaders(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_last_read_object;
 int shcount = 0, argcount = 0, i = 0, j = 0;
 mops_shader *shader = NULL;
 mops_shader_arg *sarg = NULL, **prevargptr = NULL;
 int read;

 if(!o)
   return MOPS_OK;

 fscanf(fileptr,"%d\n",&shcount);

 for(i=0;i<shcount;i++)
   {
     shader = NULL;
     if(!(shader = calloc(1, sizeof(mops_shader))))
       return MOPS_OUTOFMEM;

     mops_read_string(fileptr,&(shader->name));

     fscanf(fileptr,"%d\n", &(shader->type));
     fscanf(fileptr,"%d\n", &(shader->subtype));
     fscanf(fileptr,"%d\n", &argcount);

     prevargptr = &(shader->arg);

     for(j=0; j < argcount; j++)
       {
	 if(!(sarg = calloc(1, sizeof(mops_shader_arg))))
	   return MOPS_OUTOFMEM; 

	 mops_read_string(fileptr, &(sarg->name));
	 fscanf(fileptr,"%d", &(sarg->type));
	 read = fgetc(fileptr);
	 switch(sarg->type)
	   {
	   case MOPS_SACOLOR:
	     fscanf(fileptr,"%g\n", &((sarg->val).color[0]));
	     fscanf(fileptr,"%g\n", &((sarg->val).color[1]));
	     fscanf(fileptr,"%g\n", &((sarg->val).color[2]));
	     fscanf(fileptr,"%g\n", &((sarg->val).color[3]));
	     break;
	   case MOPS_SAVECTOR:
	   case MOPS_SAPOINT:
	     fscanf(fileptr,"%g\n", &((sarg->val).point[0]));
	     fscanf(fileptr,"%g\n", &((sarg->val).point[1]));
	     fscanf(fileptr,"%g\n", &((sarg->val).point[2]));
	     break;
	   case MOPS_SASCALAR:
	     fscanf(fileptr,"%g\n", &((sarg->val).scalar));
	     break;
	   case MOPS_SASTRING:
	     mops_read_string(fileptr, &((sarg->val).string));
	     break;
	   } /* switch */
	  
	 *prevargptr = sarg;
	 prevargptr = &(sarg->next);

	} /* for */

     mops_status = mops_add_shader(o, shader);

   } /* for */

 return mops_status;
}

int 
mops_read_tags(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_last_read_object;
 mops_tag_object *tag = NULL;
 Tcl_HashEntry *entry = NULL;
 int tcount = 0, i = 0;

 if(!o)
   return MOPS_OK;

 fscanf(fileptr,"%d\n",&tcount);

 for(i=0;i<tcount;i++)
   {
     tag = NULL;
     if(!(tag = calloc(1, sizeof(mops_tag_object))))
       return MOPS_OUTOFMEM;

     mops_read_string(fileptr,&(tag->name));

     if(!(entry = Tcl_FindHashEntry(&mops_tagtypes_ht, tag->name)))
       {
	 /* 
	    return MOPS_HASH;
	    XXXX warn the user?
	 */

       }

     if(entry)
       tag->type = Tcl_GetHashValue(entry);

     mops_read_string(fileptr,&(tag->val));

     tag->next = o->tags;
     o->tags = tag;
   } /* for */

 return mops_status;
}


int 
mops_read_nurbcurve(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *curve = NULL;
 int i = 0, a = 0;

  if(!(curve = calloc(1, sizeof(mops_nurbcurve_object))))
    { return MOPS_OUTOFMEM; }

  o->object = curve;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n",&curve->length);
  fscanf(fileptr,"%d\n",&curve->order);
  fscanf(fileptr,"%d\n",&curve->knot_type);

  if(curve->knot_type == MOPS_KTCUSTOM)
    {
      if(!(curve->knotv =
	   calloc((curve->length + curve->order), sizeof(double))))
	{ free(curve); return MOPS_OUTOFMEM; }

      for(i=0; i<(curve->length + curve->order); i++)
	fscanf(fileptr,"%lg\n",&(curve->knotv[i]));
    }
  else
    {
      mops_status = mops_create_nurbcurve_knots(curve);
      if(mops_status)
	{ free(curve); return mops_status; }
    }

  fscanf(fileptr,"%d\n",&curve->has_weights);

  if(!(curve->controlv = calloc(curve->has_weights ?
				curve->length*4 : curve->length*3,
				sizeof(double))))
    { free(curve->knotv); free(curve); return MOPS_OUTOFMEM;}

  a = 0;
  for(i=0; i < curve->length; i++)
    {
      fscanf(fileptr,"%lg\n",&(curve->controlv[a]));
      a++;
      fscanf(fileptr,"%lg\n",&(curve->controlv[a]));
      a++;
      fscanf(fileptr,"%lg\n",&(curve->controlv[a]));
      a++;
      if(curve->has_weights)
	{
	  fscanf(fileptr,"%lg\n",&(curve->controlv[a]));
	  a++;
	}
    }

  fscanf(fileptr,"%d\n",&(curve->closed));
  fscanf(fileptr,"%d\n",&(curve->weight_handles));
  fscanf(fileptr,"%lg\n",&(curve->glu_sampling_tolerance));
  fscanf(fileptr,"%lg\n",&(curve->handle_size));

  mops_nurbcurve_recreatemp(curve);


 return mops_status;
}


int 
mops_read_nurbpatch(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *patch = NULL;
 int i = 0, j = 0, a = 0;

  if(!(patch = calloc(1, sizeof(mops_nurbpatch_object))))
    { return MOPS_OUTOFMEM; }

  o->object = patch;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n",&patch->width);
  fscanf(fileptr,"%d\n",&patch->height);
  fscanf(fileptr,"%d\n",&patch->uorder);
  fscanf(fileptr,"%d\n",&patch->vorder);
  fscanf(fileptr,"%d\n",&patch->uknot_type);
  fscanf(fileptr,"%d\n",&patch->vknot_type);
  
  mops_status = mops_create_nurbpatch_knots(patch);
  if(mops_status)
    { free(patch); return mops_status; }

  if(patch->uknot_type == MOPS_KTCUSTOM)
    {
      for(i=0; i<(patch->width + patch->uorder); i++)
	fscanf(fileptr,"%lg\n",&(patch->uknotv[i]));
    }

  if(patch->vknot_type == MOPS_KTCUSTOM)
    {
      for(i=0; i<(patch->height + patch->vorder); i++)
	fscanf(fileptr,"%lg\n",&(patch->vknotv[i]));
    }

  fscanf(fileptr,"%d\n",&patch->has_weights);

  if(!(patch->controlv = calloc(patch->has_weights ?
				patch->width*patch->height*4 :
				patch->width*patch->height*3,
				sizeof(double))))
    { free(patch->uknotv); free(patch->vknotv); free(patch);
      return MOPS_OUTOFMEM; }

  a = 0;
  for(i=0; i < patch->width; i++)
    for(j=0; j < patch->height; j++)
      {
	fscanf(fileptr,"%lg\n",&(patch->controlv[a]));
	a++;
	fscanf(fileptr,"%lg\n",&(patch->controlv[a]));
	a++;
	fscanf(fileptr,"%lg\n",&(patch->controlv[a]));
	a++;
	if(patch->has_weights)
	  {
	    fscanf(fileptr,"%lg\n",&(patch->controlv[a]));
	    a++;
	  }
      }

  fscanf(fileptr,"%d\n",&(patch->weight_handles));
  fscanf(fileptr,"%d\n",&(patch->glu_display_mode));
  fscanf(fileptr,"%lg\n",&(patch->glu_sampling_tolerance));
  fscanf(fileptr,"%lg\n",&(patch->handle_size));

 return mops_status;
}


int 
mops_read_box(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_box_object *box = NULL;

  if(!(box = calloc(1, sizeof(mops_box_object))))
    { return MOPS_OUTOFMEM; }

  o->object = box;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%lg\n",&box->width);
  fscanf(fileptr,"%lg\n",&box->length);
  fscanf(fileptr,"%lg\n",&box->height);

 return mops_status;
}


int 
mops_read_sphere(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_sphere_object *sphere = NULL;
 int itemp;

  if(!(sphere = calloc(1, sizeof(mops_sphere_object))))
    { return MOPS_OUTOFMEM; }

  o->object = sphere;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  sphere->closed = (char)itemp;
  fscanf(fileptr,"%lg\n",&sphere->radius);
  fscanf(fileptr,"%lg\n",&sphere->zmin);
  fscanf(fileptr,"%lg\n",&sphere->zmax);
  fscanf(fileptr,"%lg\n",&sphere->thetamax);

 return mops_status;
}


int 
mops_read_cone(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cone_object *cone = NULL;
 int itemp;

  if(!(cone = calloc(1, sizeof(mops_cone_object))))
    { return MOPS_OUTOFMEM; }

  o->object = cone;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  cone->closed = (char)itemp;
  fscanf(fileptr,"%lg\n",&cone->radius);
  fscanf(fileptr,"%lg\n",&cone->height);
  fscanf(fileptr,"%lg\n",&cone->thetamax);

 return mops_status;
}


int 
mops_read_disk(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_disk_object *disk = NULL;

  if(!(disk = calloc(1, sizeof(mops_disk_object))))
    { return MOPS_OUTOFMEM; }

  o->object = disk;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%lg\n",&disk->radius);
  fscanf(fileptr,"%lg\n",&disk->height);
  fscanf(fileptr,"%lg\n",&disk->thetamax);

 return mops_status;
}


int 
mops_read_cylinder(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cylinder_object *cylinder = NULL;
 int itemp;

  if(!(cylinder = calloc(1, sizeof(mops_cylinder_object))))
    { return MOPS_OUTOFMEM; }

  o->object = cylinder;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  cylinder->closed = (char)itemp;
  fscanf(fileptr,"%lg\n",&cylinder->radius);
  fscanf(fileptr,"%lg\n",&cylinder->zmin);
  fscanf(fileptr,"%lg\n",&cylinder->zmax);
  fscanf(fileptr,"%lg\n",&cylinder->thetamax);

 return mops_status;
}


int 
mops_read_hyperboloid(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_hyperboloid_object *hyperboloid = NULL;
 int itemp;

  if(!(hyperboloid = calloc(1, sizeof(mops_hyperboloid_object))))
    { return MOPS_OUTOFMEM; }

  o->object = hyperboloid;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  hyperboloid->closed = (char)itemp;

  fscanf(fileptr,"%lg\n",&hyperboloid->p1x);
  fscanf(fileptr,"%lg\n",&hyperboloid->p1y);
  fscanf(fileptr,"%lg\n",&hyperboloid->p1z);

  fscanf(fileptr,"%lg\n",&hyperboloid->p2x);
  fscanf(fileptr,"%lg\n",&hyperboloid->p2y);
  fscanf(fileptr,"%lg\n",&hyperboloid->p2z);

  fscanf(fileptr,"%lg\n",&hyperboloid->thetamax);

 return mops_status;
}


int 
mops_read_paraboloid(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_paraboloid_object *paraboloid = NULL;
 int itemp;

  if(!(paraboloid = calloc(1, sizeof(mops_paraboloid_object))))
    { return MOPS_OUTOFMEM; }

  o->object = paraboloid;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  paraboloid->closed = (char)itemp;
  fscanf(fileptr,"%lg\n",&paraboloid->rmax);
  fscanf(fileptr,"%lg\n",&paraboloid->zmin);
  fscanf(fileptr,"%lg\n",&paraboloid->zmax);
  fscanf(fileptr,"%lg\n",&paraboloid->thetamax);

 return mops_status;
}


int 
mops_read_torus(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_torus_object *torus = NULL;
 int itemp;

  if(!(torus = calloc(1, sizeof(mops_torus_object))))
    { return MOPS_OUTOFMEM; }

  o->object = torus;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n", &itemp);
  torus->closed = (char)itemp;
  fscanf(fileptr,"%lg\n",&torus->majorrad);
  fscanf(fileptr,"%lg\n",&torus->minorrad);
  fscanf(fileptr,"%lg\n",&torus->phimin);
  fscanf(fileptr,"%lg\n",&torus->phimax);
  fscanf(fileptr,"%lg\n",&torus->thetamax);

 return mops_status;
}


int 
mops_read_instance(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;

  o->object = NULL;

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

 return mops_status;
}


int 
mops_read_light(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_light_object *light = NULL;
 double dummy;

 if(!(light = calloc(1, sizeof(mops_light_object))))
    { return MOPS_OUTOFMEM; }

  /* get name of object */
  if(mops_read_version > 0)
    mops_read_string(fileptr, &(o->name));

  fscanf(fileptr,"%d\n",&light->shadows);

  if(mops_read_version > 0)
    fscanf(fileptr,"%d\n",&light->samples);

  /* default for very old scenes */
  light->on = MOPS_TRUE;

  if(mops_read_version > 3)
    {
      fscanf(fileptr,"%d\n",&light->type);
      fscanf(fileptr,"%d\n",&light->on);
      fscanf(fileptr,"%lg\n",&light->intensity);
      fscanf(fileptr,"%d\n",&light->colr);
      fscanf(fileptr,"%d\n",&light->colg);
      fscanf(fileptr,"%d\n",&light->colb);
      fscanf(fileptr,"%d\n",&light->cola);
      if(mops_read_version < 5)
	{
	  fscanf(fileptr,"%lg\n",&dummy);
	  fscanf(fileptr,"%lg\n",&dummy);
	  fscanf(fileptr,"%lg\n",&dummy);
	  fscanf(fileptr,"%lg\n",&dummy);
	  fscanf(fileptr,"%lg\n",&dummy);
	  fscanf(fileptr,"%lg\n",&dummy);
	}
      fscanf(fileptr,"%lg\n",&light->cone_angle);
      fscanf(fileptr,"%lg\n",&light->cone_delta_angle);
      fscanf(fileptr,"%lg\n",&light->beam_distribution);

    }

  o->object = light;

 return mops_status;
}


int 
mops_read_object(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 int type = 0;
 mops_object *o = NULL;

 if(feof(fileptr))
   return MOPS_EOF;

 /* skip to next entity */
 mops_status = mops_read_skip(fileptr);

 if(mops_status)
   return mops_status;

 if(!(o = calloc(1,sizeof(mops_object))))
   return MOPS_OUTOFMEM;

 mops_default_attributes(o);

 /* get type of object */
  if(mops_read_version > 0)
    {
      fscanf(fileptr,"%d",&type);
      fgetc(fileptr);
    }
  else
    fscanf(fileptr,"%d\n",&type);

  o->type = type;

 switch(type)
   {
   case MOPS_OTLEVEL:
     mops_status = mops_read_level(fileptr, &o);
     break;
   case MOPS_OTNURBCURVE:
     mops_status = mops_read_nurbcurve(fileptr, o);
     break;
   case MOPS_OTNURBPATCH:
     mops_status = mops_read_nurbpatch(fileptr, o);
     break;
   case MOPS_OTCUSTOM:
     mops_status = mops_read_custom(fileptr, o);
     break;
   case MOPS_OTLIGHT:
     mops_status = mops_read_light(fileptr, o);
     break;
   case MOPS_OTBOX:
     mops_status = mops_read_box(fileptr, o);
     break;
   case MOPS_OTSPHERE:
     mops_status = mops_read_sphere(fileptr, o);
     break;
   case MOPS_OTCONE:
     mops_status = mops_read_cone(fileptr, o);
     break;
   case MOPS_OTCYLINDER:
     mops_status = mops_read_cylinder(fileptr, o);
     break;
   case MOPS_OTDISK:
     mops_status = mops_read_disk(fileptr, o);
     break;
   case MOPS_OTTORUS:
     mops_status = mops_read_torus(fileptr, o);
     break;
   case MOPS_OTPARABOLOID:
     mops_status = mops_read_paraboloid(fileptr, o);
     break;
   case MOPS_OTHYPERBOLOID:
     mops_status = mops_read_hyperboloid(fileptr, o);
     break;
   case MOPS_OTINSTANCE:
     mops_status = mops_read_instance(fileptr, o);
     break;
   default:
     break;
   }

  if(mops_status)
   {
     free(o); o=NULL;
     return mops_status;
   }
 
 switch(type)
   {
   case MOPS_OTATTRIB:
     free(o); o=NULL;
     mops_status = mops_read_attributes(fileptr);
     break;
   case MOPS_OTTRAFO:
     free(o); o=NULL;
     mops_status = mops_read_trafos(fileptr);
     break;
   case MOPS_OTSHADER:
     free(o); o=NULL;
     mops_status = mops_read_shaders(fileptr);
     break;
   case MOPS_OTTAG:
     free(o); o=NULL;
     mops_status = mops_read_tags(fileptr);
     break;
   default:
     break;
   }

 if(o)
   {
     mops_status = mops_link_new_object(o);
     mops_last_read_object = o;
   }

 return mops_status;
}


int 
mops_read_scene(Tcl_Interp *interp, char *filename, int insert)
{
 int mops_status = MOPS_OK;
 FILE *fileptr = NULL;
 char *fname = "read_scene";
 mops_object *o = NULL;

 if(!(fileptr = fopen(filename,"rb")))
   return MOPS_OPENFILE;

 mops_status = mops_read_header(fileptr);

 if(mops_status)
   { fclose(fileptr); return mops_status; }

 mops_status = mops_read_root(interp, fileptr, insert);
 if(feof(fileptr))
   {fclose(fileptr); return MOPS_OK;}

 if(mops_status)
   {fclose(fileptr); return mops_status;}

 while(!mops_status)
   {
     mops_status = mops_read_object(fileptr);
     if(mops_status)
       if(mops_status != MOPS_EOF)
	 {
	   /* issue errmsg */
	   mops_error(interp, mops_status, fname, NULL);
	   if(mops_prefs.onerror)
	     {
	       mops_status = MOPS_OK;
	     }
	 }

   }

 if(mops_status == MOPS_EOF)
   mops_status = MOPS_OK;

 /* link instance objects to their originals */
 /* clear hashtable of oid ptrs */
 mops_create_oid_ht(NULL);
 /* build it */
 mops_create_readoid_ht(mops_root->next);
 /* link instance objects to their originals */
 mops_connect_instances(mops_root->next);
 /* force rebuild of all custom objects, that rely on childs */
 o = mops_root->next;
 while(o)
   {
     mops_force_notification(o);
     o = o->next;
   }

 return mops_status;
}


int
mops_new_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_root;
 mops_shader *shader = NULL;
 char fname[] = "new_scene";
 
  mops_status = mops_clear_scene();

  MOPS_CRERROR

  /* free shaders from root object */
  while(o->shaders)
    {
      shader = o->shaders;
      o->shaders = shader->next; 
      mops_status = mops_free_shader(shader);
    }

 return TCL_OK;
}


int
mops_replace_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char fname[] = "replace_scene";
 char done_cmd[] = "puts stdout \"Done replacing scene from: $filename\"";

 /* check args */
 if(argc != 2)
   {
     mops_error(interp, MOPS_ARGS, fname, "filename");
     return TCL_OK;
   }

 /* first, we completely clear the old scene */
 mops_status = mops_clear_scene();

 mops_last_read_object = mops_root;
 mops_free_selection();
 mops_del_all_level_objects();

 /* then, we read the new scene */
 mops_status = mops_read_scene(interp, argv[1], MOPS_FALSE);

 MOPS_CRERROR

 Tcl_Eval(interp, done_cmd);
    
 return TCL_OK;
}


int
mops_insert_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char fname[] = "insert_scene";
 char done_cmd[] = "puts stdout \"Done inserting scene from: $ifilename\"";

 /* check args */
 if(argc != 2)
   {
     mops_error(interp, MOPS_ARGS, fname, "filename");
     return TCL_OK;
   }

 mops_last_read_object = NULL;

 mops_status = mops_read_scene(interp, argv[1], MOPS_TRUE);

 MOPS_CRERROR

 Tcl_Eval(interp, done_cmd);

 return TCL_OK;
}

