/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

/*
 * revolve_rmdso.c - revolve RenderMan DynamicSharedObject
 */

#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <ri.h>

#ifdef __cplusplus
extern "C" RtPointer ConvertParameters(RtString paramstr);
extern "C" RtVoid Subdivide(RtPointer data, RtFloat detail);
extern "C" RtVoid Free(RtPointer data);
#endif

struct revolve_np {
  RtInt w, h;   /* width, height */
  RtInt uo, vo; /* uorder, vorder */
  RtFloat *cv;  /* control vector */
  RtFloat *ukv; /* u-knots */
  RtFloat *vkv; /* v-knots */
};

#define REVRMDSO_V3CROSS(r,v1,v2) {r[0]=(v1[1] * v2[2]) - (v1[2] * v2[1]);r[1]=(v1[2] * v2[0]) - (v1[0] * v2[2]);r[2]=(v1[0] * v2[1]) - (v1[1] * v2[0]);}

#define REVRMDSO_V3DOT(v1,v2) (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])

#define REVRMDSO_V3SUB(r,v1,v2) {r[0]=v1[0]-v2[0];r[1]=v1[1]-v2[1];r[2]=v1[2]-v2[2];}

#define REVRMDSO_D2R(x) ((x)*REVRMDSO_PI/180.0)

#define REVRMDSO_PI 3.1415926535897932384626433

/* was: mops_intersect_lines3D */
int
revolve_rmdso_il3D(double *p1, double *t1,
		   double *p2, double *t2,
		   double *p)
{
 double v[3] = {0}, px[3] = {0};
 double pd[3] = {0};
 double vw = 0.0, vw2 = 0.0;

  REVRMDSO_V3CROSS(px,t1,t2)
  REVRMDSO_V3CROSS(v,px,t1)
  vw = REVRMDSO_V3DOT(v,t2);

  if((vw*vw)<1e-7)
    return 0;

  REVRMDSO_V3SUB(pd,p1,p2)
  vw2 = REVRMDSO_V3DOT(v,pd);
  p[0] = p2[0]+(vw2/vw)*t2[0];
  p[1] = p2[1]+(vw2/vw)*t2[1];
  p[2] = p2[2]+(vw2/vw)*t2[2];

 return 1;
}


/* was: mops_nb_CreateNurbsCircle */
int
revolve_rmdso_crnc(double r, double ths, double the,
		   int *length, double **knotv, double **controlv)
{
 double theta = 0.0, dtheta = 0.0, angle = 0.0, w1 = 0.0;
 double *Pw = NULL, *U = NULL;
 int i, j, index = 0, narcs = 0, n = 0;
 double P0[4] = {0}, P1[4] = {0}, P2[4] = {0};
 double T0[4] = {0}, T2[4] = {0};

  if(the < ths)
    the = 360 - the;

  theta = the - ths;

  if(theta <= 90.0)
    narcs = 1;
  else
    if(theta <= 270.0)
      narcs = 2;
    else
      if(theta <= 180.0)
	narcs = 3;
      else
	narcs = 4;

  dtheta = theta/narcs;
  n = 2 * narcs;                  /* n+1 control points */
  w1 = cos(REVRMDSO_D2R(dtheta/2.0)); /* dtheta/2 == base angle */

  /* alloc mem for Pw and U */ 
  if(!(Pw = calloc((n+1)*4, sizeof(double))))
    return 1;

  if(!(U = calloc((n+1)+4, sizeof(double))))
    { free(Pw); return 1; }

 P0[0] = r * cos(REVRMDSO_D2R(ths));
 P0[1] = r * sin(REVRMDSO_D2R(ths));
 P0[3] = 1.0;
 T0[0] = -sin(REVRMDSO_D2R(ths));
 T0[1] = cos(REVRMDSO_D2R(ths));
 Pw[0] = P0[0];
 Pw[1] = P0[1];
 Pw[3] = 1.0;
 index=0; angle = ths;
 for(i=1; i<=narcs; i++)
   {
     angle += dtheta;
     P2[0] = r * cos(REVRMDSO_D2R(angle));
     P2[1] = r * sin(REVRMDSO_D2R(angle));
     P2[3] = 1.0;

     memcpy(&Pw[(index+2)*4],P2,4*sizeof(double));

     T2[0] = -sin(REVRMDSO_D2R(angle));
     T2[1] = cos(REVRMDSO_D2R(angle));

     revolve_rmdso_il3D(P0,T0,P2,T2,P1);

     Pw[(index+1)*4]   = w1*P1[0];
     Pw[((index+1)*4)+1] = w1*P1[1];
     Pw[((index+1)*4)+3] = w1;

     index += 2;
     if(i<index)
       {
	 memcpy(P0,P2,4*sizeof(double));
	 memcpy(T0,T2,2*sizeof(double));
       }
   }

 j = 2*narcs+1;
 for(i=0; i<3; i++)
   {
     U[i] = 0.0;
     U[i+j] = 1.0;
   }
 switch (narcs)
   {
   case 1: 
     break;
   case 2:
     U[3] = 0.5;
     U[4] = 0.5;
     break;
   case 3:
     U[3] = 1.0/3.0;
     U[4] = 1.0/3.0;
     U[5] = 2.0/3.0;
     U[6] = 2.0/3.0;
     break;
   case 4:
     U[3] = 0.25;
     U[4] = 0.25;
     U[5] = 0.5;
     U[6] = 0.5;
     U[7] = 0.75;
     U[8] = 0.75;
     break;
   }
 
 *knotv = U;
 *controlv = Pw;
 *length = n+1;

 return 0;
}


RtPointer
ConvertParameters(RtString paramstr)
{
 char *str = NULL;
 int i, j, n, o, h, has_weights, index, stride;
 double arc, radius, w, *cv, *kv, *tknotv = NULL, *tcontrolv = NULL;
 struct revolve_np *newp = NULL;

 str = (char *)paramstr;

 sscanf(str, "%lg", &arc);
 str = strchr(str, ',');
 if(arc >= 360.0 || arc < -360.0 || arc == 0.0)
   {
     arc = 360.0;
   }
 sscanf(str, ",%d", &n);
 str = strchr(str+1, ',');

 sscanf(str, ",%d", &o);
 str = strchr(str+1, ',');

 sscanf(str, ",%d", &stride);
 str = strchr(str+1, ',');

 sscanf(str, ",%d", &has_weights);
 str = strchr(str+1, ',');

 cv = calloc(n*stride, sizeof(double));
 for(i = 0; i < n; i++)
   {
     if(has_weights)
       sscanf(str, ",%lg %lg %lg", &(cv[i*stride]), &(cv[i*stride+1]),
	      &(cv[i*stride+2]));
     else
       sscanf(str, ",%lg %lg", &(cv[i*stride]), &(cv[i*stride+1]));

     str = strchr(str+1, ',');
   }

 kv = calloc(n+o, sizeof(double));
 for(i = 0; i < n+o; i++)
   {
     sscanf(str, ",%lg", &(kv[i]));
     str = strchr(str+1, ',');
   }
 
  newp = calloc(1, sizeof(struct revolve_np));

  newp->w = (RtInt)n;
  newp->uo = (RtInt)o;

  newp->ukv = calloc(n+o, sizeof(RtFloat));
  for(i = 0; i < (n+o); i++)
    {
      newp->ukv[i] = (RtFloat)kv[i];
    }

  for(i = 0; i < n; i++)
    {
      radius = cv[i*stride];

      if(has_weights)
	w = cv[(i*stride)+2];
      else
	w = 1.0;

      if(tknotv)
	free(tknotv);
      tknotv = NULL;
      if(tcontrolv)
	free(tcontrolv);
      tcontrolv = NULL;

      if(arc > 0.0)
	{
	 revolve_rmdso_crnc(radius, 0.0, arc, &h, &tknotv, &tcontrolv);
	}
      else
	{
	 revolve_rmdso_crnc(radius, arc, 0.0, &h, &tknotv, &tcontrolv);
	}

      if(!newp->cv)
	{
	  if(!(newp->cv = calloc(h*newp->w*4, sizeof(RtFloat))))
	    return NULL;
	}

      /* copy controlv */
      for(j = 0; j < h; j++)
	{
	  index = (i*h+j)*4;
	  newp->cv[index]  = (RtFloat)(tcontrolv[j*4]);
	  newp->cv[index+1] = (RtFloat)(cv[(i*stride)+1]*tcontrolv[j*4+3]);
	  newp->cv[index+2] = (RtFloat)(tcontrolv[j*4+1]);
	  newp->cv[index+3] = (RtFloat)(tcontrolv[j*4+3]*w);
	} /* for */
    }

  newp->vkv = calloc(h+4, sizeof(RtFloat));
  for(i = 0; i < h+3; i++)
    {
      newp->vkv[i] = (RtFloat)tknotv[i];
    }

  if(tknotv)
    free(tknotv);

  if(tcontrolv)
    free(tcontrolv);

  newp->h = (RtInt)h;
  newp->vo = (RtInt)3;


  /* return the blind data pointer */
  return (RtPointer)newp;
}

RtVoid
Subdivide(RtPointer data, RtFloat detail)
{
 struct revolve_np *p = NULL;

 p = (struct revolve_np *) data;

 
 RiNuPatch(p->h, p->vo, p->vkv, p->vkv[p->vo-1], p->vkv[p->h],
	   p->w, p->uo, p->ukv, p->ukv[p->uo-1], p->ukv[p->w],
	   "Pw", p->cv, NULL);

}


RtVoid
Free(RtPointer data)
{
  struct revolve_np *p = NULL;

  if(!data)
    return;

  p = (struct revolve_np *) data;

  if(p->ukv)
    free(p->ukv);

  if(p->vkv)
    free(p->vkv);
  
  if(p->cv)
    free(p->cv);

  free(data);

 return;
}
