/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

RtLightHandle mops_light_handle;

/* Shaders */

int
mops_scan_shader_tcmd(ClientData clientData, Tcl_Interp * interp,
		      int argc, char *argv[])
{
#ifndef MOPSNOSLCARGS
 char fname[] = "scan_shader";
 int i = 0, numargs = 0;
 SLC_VISSYMDEF *symbol = NULL;
 SLC_TYPE type;
 double deffltval;
 char *defstrval;
 int arraylen;
 Tcl_Obj *to = NULL, *tovname = NULL;

 if(argc<3)
   {
     mops_error(interp,MOPS_ARGS,fname,"shaderpath varname");
     return TCL_OK;
   }

 if((SLC_SetShader(argv[1])) == -1)
   {
     mops_error(interp,MOPS_ERROR,fname,"libslc cannot SetShader:");
     mops_error(interp,MOPS_ERROR,fname,argv[1]);
     return TCL_OK;
   }

 /* get name of shader */
 Tcl_SetVar(interp,argv[2],SLC_GetName(),TCL_APPEND_VALUE | 
	    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);

 /* get type of shader */
 type = SLC_GetType();
 switch (type)
   {
   case SLC_TYPE_SURFACE:
     Tcl_SetVar(interp,argv[2],"surface",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   case SLC_TYPE_DISPLACEMENT:
     Tcl_SetVar(interp,argv[2],"displacement",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   case SLC_TYPE_LIGHT:
     Tcl_SetVar(interp,argv[2],"light",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   case SLC_TYPE_VOLUME:
     Tcl_SetVar(interp,argv[2],"volume",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   case SLC_TYPE_IMAGER:
     Tcl_SetVar(interp,argv[2],"imager",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   case SLC_TYPE_TRANSFORMATION:
     Tcl_SetVar(interp,argv[2],"transformation",TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
     break;
   default:
     break;
   }

 tovname = Tcl_NewStringObj(argv[2], -1);

 /* get arguments of shader */
 numargs = SLC_GetNArgs();

 to = Tcl_NewIntObj(numargs);
 Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);

 for(i = 1; i <= numargs; i++)
   {
     symbol = NULL;
     symbol = SLC_GetArgById(i);

     if(!symbol)
       {
	 mops_error(interp,MOPS_ERROR,fname,"cannot get symbol from shader");
	 SLC_EndShader();
	 return TCL_OK;
       }

     Tcl_SetVar(interp,argv[2],symbol->svd_name,TCL_APPEND_VALUE | 
		TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);

     switch(symbol->svd_type)
       {
       case SLC_TYPE_POINT:
	 Tcl_SetVar(interp,argv[2],"point",TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	 break;
       case SLC_TYPE_COLOR:
	 Tcl_SetVar(interp,argv[2],"color",TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	 break;
       case SLC_TYPE_VECTOR:
	 Tcl_SetVar(interp,argv[2],"vector",TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	 break;
       case SLC_TYPE_SCALAR:
	 Tcl_SetVar(interp,argv[2],"float",TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	 break;
       case SLC_TYPE_STRING:
	 Tcl_SetVar(interp,argv[2],"string",TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	 break;
       default:
	 break;
       }

     arraylen = symbol->svd_arraylen;
     to = Tcl_NewIntObj(arraylen);
     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
		    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);

     if(arraylen < 1) /* XXXX we disregard default values for arrays! */
       {
	 switch(symbol->svd_type)
	   {
	   case SLC_TYPE_POINT:
	   case SLC_TYPE_COLOR:
	   case SLC_TYPE_VECTOR:
	     deffltval = (double)(symbol->svd_default).pointval->xval;;
	     to = Tcl_NewDoubleObj(deffltval);
	     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	     deffltval = (symbol->svd_default).pointval->yval;	 
	     to = Tcl_NewDoubleObj(deffltval);
	     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	     deffltval = (symbol->svd_default).pointval->zval;	 
	     to = Tcl_NewDoubleObj(deffltval);
	     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	     break;
	   case SLC_TYPE_SCALAR:
	     deffltval = *((symbol->svd_default).scalarval);	 
	     to = Tcl_NewDoubleObj(deffltval);
	     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	     break;
	   case SLC_TYPE_STRING:
	     defstrval = (symbol->svd_default).stringval;	 
	     to = Tcl_NewStringObj(defstrval,-1);
	     Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			    TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
	     break;
	   default:
	     break;
	   }
       }
     else
       {
	 to = Tcl_NewStringObj(NULL, -1);
	 Tcl_ObjSetVar2(interp,tovname,NULL,to,TCL_APPEND_VALUE | 
			TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);
       }
   }

 SLC_EndShader();

 Tcl_IncrRefCount(tovname);Tcl_DecrRefCount(tovname);
 
#endif
 return TCL_OK;
}



/********************************************/
/* mops_copy_shader_arg:
   copy a shader parameter                  */
/********************************************/
int
mops_copy_shader_arg(mops_shader_arg *source,
		       mops_shader_arg **dest)
{
 int mops_status = MOPS_OK;
 mops_shader_arg *newp = NULL;
 char *newval;

  if(!(newp = calloc(1,sizeof(mops_shader_arg))))
    return MOPS_OUTOFMEM; 

  memcpy(newp, source, sizeof(mops_shader_arg)); 
  /* danger! links point to original hierachy */

  /* copy name */
  if(!(newval = calloc(1, strlen(source->name)+1)))
    { free(newp); return MOPS_OUTOFMEM; }
  strcpy(newval, source->name);
  newp->name = newval;

  /* copy string */
  newval = NULL;
  if(source->type == MOPS_SASTRING)
    {
      if(!(newval = calloc(1, strlen(source->val.string)+1)))
	{ free(newp->name); free(newp); return MOPS_OUTOFMEM; }
      strcpy(newval, source->val.string);
      newp->val.string = newval;
    }

  *dest = newp;

 return mops_status;
}


/********************************************/
/* mops_copy_shader:
   copy a shader & associated parameters    */
/********************************************/
int
mops_copy_shader(mops_shader *source,
		 mops_shader **dest)
{
  int mops_status = MOPS_OK;
  mops_shader *news = NULL;
  mops_shader_arg *p = NULL, **newp;
  char *newval = NULL;

  if(!(news = calloc(1,sizeof(mops_shader))))
    return MOPS_OUTOFMEM; 

  memcpy(news, source, sizeof(mops_shader)); 
  /* danger! links point to original hierachy */

  /* copy name */
  if(!(newval = calloc(1, strlen(source->name)+1)))
    { free(news); return MOPS_OUTOFMEM; }
  strcpy(newval, source->name);
  news->name = newval;


  /* copy parameters */
  if(source->arg)
    {
      p = source->arg;
      newp = &(news->arg);
      while(p)
	{
	  mops_status = mops_copy_shader_arg(p, newp);
	  if(mops_status)
	    {
	      mops_free_shader(news);
	      return mops_status;
	    }
  
	  (*newp)->next = NULL;
	  newp = &((*newp)->next);
	  p = p->next;
	}
    }

  *dest = news;

 return mops_status;
}


/********************************************/
/* mops_add_shader:
   add shader to an object, overwriting 
   an eventually existing shader of the
   same type!                               */
/********************************************/
int
mops_add_shader(mops_object *o,
		mops_shader *new)
{
  int mops_status = MOPS_OK;
  mops_shader *old = NULL, *old2 = NULL;

  if(!o)
    return MOPS_OK;

  if(!o->shaders)
    o->shaders = new;
  else
    {
      old = o->shaders;
      if((old->type == new->type) && (old->subtype == new->subtype))
	{ /* delete old shader */
	  new->next = o->shaders->next;
	  o->shaders = new;
	  mops_status = mops_free_shader(old);
	  return mops_status;
	}

      while(old)
	{
	  if((old->type == new->type) && (old->subtype == new->subtype))
	    { /* delete old shader */
	      old2->next = new;
	      new->next = old->next;
	      mops_status = mops_free_shader(old);
	      return mops_status;
	    }
	  old2 = old;
	  old = old->next;
	} /* while */

      old2->next = new;

    } /* if */

  return mops_status;
}

/********************************************/
/* mops_free_shader:
   free a shader and associated arguments   */
/********************************************/
int
mops_free_shader(mops_shader *shader)
{
  int mops_status = MOPS_OK;
  mops_shader_arg *arg = NULL, *argtmp = NULL;

  if(!shader)
    return MOPS_OK;

  if(shader->arg)
    {
      /* delete arguments */
      arg = shader->arg;
      while(arg)
	{
	  argtmp = arg->next;

	  if(arg->type == MOPS_SASTRING)
	    free(arg->val.string);

	  free(arg);
	  arg = argtmp;
	}

    }


  /* finally, delete the shader*/
  free(shader);

 return mops_status;
}


int
mops_wrib_shaders(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_shader *shader = NULL;
 mops_shader_arg *sarg = NULL;
 RtToken *tokens = NULL;
 RtPointer *values = NULL;
 int count = 0;

  if(!o)
    return MOPS_OK;

  shader = o->shaders;
  while(shader)
    {
      if(shader->name)
	{
	    sarg = shader->arg;
	    while(sarg)
	      {
		count++;
		sarg = sarg->next;
	      }

	    if(count)
	      {
		if(!(tokens = calloc(count, sizeof(RtToken))))
		  return(MOPS_OUTOFMEM);
		if(!(values = calloc(count, sizeof(RtToken))))
		  { free(tokens); return(MOPS_OUTOFMEM); }
	      }

	    sarg = shader->arg;

	    count = 0;
	    while(sarg)
	      {
		switch(sarg->type)
		  {
		  case MOPS_SACOLOR:
		    RiDeclare(sarg->name,"color");
		    values[count] = (void*)(&sarg->val.color);
		    break;
		  case MOPS_SAVECTOR:
		  case MOPS_SAPOINT:
		    RiDeclare(sarg->name,"point");
		    values[count] = (void*)(&(sarg->val.point));
		    break;
		  case MOPS_SASCALAR:
		    RiDeclare(sarg->name,"float");
		    values[count] = (void*)(&(sarg->val.scalar));
		    break;
		  case MOPS_SASTRING:
		    RiDeclare(sarg->name,"string");
		    values[count] = (void*)(&(sarg->val.string));
		    break;
		  }

		tokens[count] = sarg->name;

		count++;
		sarg = sarg->next;
	      }

	    switch(shader->type)
	      {
	      case MOPS_STSURFACE:
		RiSurfaceV(shader->name,count,tokens,values);
		break;
	      case MOPS_STDISPLACEMENT:
		RiDisplacementV(shader->name,count,tokens,values);
		break;
	      case MOPS_STLIGHT:
		if(o->down)
		  { /* AreaLight */
		    mops_light_handle = RiAreaLightSourceV(shader->name,count,tokens,values);
		  }
		else
		  {
		    mops_light_handle = RiLightSourceV(shader->name,count,tokens,values);
		  }
		break;
	      case MOPS_STVOLUME:
		switch(shader->subtype)
		  {
		  case MOPS_STINTERIOR:
		    RiInteriorV(shader->name,count,tokens,values);
		    break;
		  case MOPS_STEXTERIOR:
		    RiExteriorV(shader->name,count,tokens,values);
		    break;
		  case MOPS_STATMOSPHERE:
		    RiAtmosphereV(shader->name,count,tokens,values);
		    break;
		  }
		break;
	      case MOPS_STIMAGER:
		RiImagerV(shader->name,count,tokens,values);
		break;
	      }

	    if(tokens)
	      free(tokens);
	    tokens = NULL;
	    if(values)
	      free(values);
	    values = NULL;

	  }

      shader = shader->next;
    }

  return mops_status;
}


/* C -> Tcl! */
int
mops_get_shaderprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_shader *shader = NULL;
 mops_shader_arg *arg = NULL;
 char hilf[255], *n1 = NULL, *sname = NULL/*, *sfile = NULL*/;
 char *typestr = NULL;
 int shadertype = 0, shadersubtype = 0, itemp = 0;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 char fname[] = "get_shaderprop";
 char update_cmd[] = "update";

  if(!sel) /* oops? */
    o = mops_root;
  else
    o = sel->object;

  /* parse args */
  if(argc < 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "shadertype arrayname subtype");
      return TCL_OK;
    }

  n1 = argv[2];
  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj(n1,-1);
  Tcl_GetInt(interp, argv[1], &shadertype);

  if(strstr(argv[1],"surface"))
    shadertype = MOPS_STSURFACE;
  else
  if(strstr(argv[1],"displacement"))
    shadertype = MOPS_STDISPLACEMENT;
  else
  if(strstr(argv[1],"volume"))
    {
      if(argc < 3)
	{
	  mops_error(interp, MOPS_ARGS, fname, "shadertype arrayname subtype");
	  return TCL_OK;
	}
      shadertype = MOPS_STVOLUME;
      if(strstr(argv[3],"interior"))
	shadersubtype = MOPS_STINTERIOR;
      else
      if(strstr(argv[3],"exterior"))
	shadersubtype = MOPS_STEXTERIOR;
      else
      if(strstr(argv[3],"atmosphere"))
	shadersubtype = MOPS_STATMOSPHERE;
    }
  else
  if(strstr(argv[1],"light"))
    shadertype = MOPS_STLIGHT;
  else
  if(strstr(argv[1],"imager"))
    shadertype = MOPS_STIMAGER;
  else
  if(strstr(argv[1],"transform"))
    shadertype = MOPS_STTRANSFORM;

  Tcl_SetVar2(interp,n1,"Name","",TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY);

  shader = o->shaders;
  while(shader)
    {

      if((shader->type == shadertype) && (shader->subtype == shadersubtype))
	{
	  arg = shader->arg;
	  sname = shader->name;
	  /*	  sfile = shader->loc;*/
	}
      shader = shader->next;
    }

  Tcl_SetVar2(interp,n1,"Name",sname,TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY);
  /*  Tcl_SetVar2(interp,n1,"File",sfile,TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY);*/
  Tcl_SetVar2(interp,n1,"ArgNames","",TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY);
  Tcl_SetVar2(interp,n1,"ArgTypes","",TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY);

  while(arg)
    {
      Tcl_SetVar2(interp,n1,"ArgNames",arg->name, TCL_LEAVE_ERR_MSG |
		  TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);     

      switch(arg->type)
	{
	case MOPS_SACOLOR:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "color", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_R", -1);

	  itemp = (int)(arg->val.color[0]*255);

	  to = Tcl_NewIntObj(itemp);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_G", -1);
	  itemp = (int)(arg->val.color[1]*255);
	  to = Tcl_NewIntObj(itemp);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_B", -1);
	  itemp = (int)(arg->val.color[2]*255);
	  to = Tcl_NewIntObj(itemp);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_A", -1);
	  itemp = (int)(arg->val.color[3]*255);
	  to = Tcl_NewIntObj(itemp);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  break;

	case MOPS_SAPOINT:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "point", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_X", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[0]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_Y", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[1]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_Z", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[2]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  break;

	case MOPS_SAVECTOR:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "vector", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_X", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[0]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_Y", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[1]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  Tcl_SetStringObj(ton,arg->name,-1);
	  Tcl_AppendToObj(ton, "_Z", -1);
	  to = Tcl_NewDoubleObj(arg->val.point[2]);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  break;

	case MOPS_SASCALAR:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "float", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	  Tcl_SetStringObj(ton,arg->name,-1);
	  to = Tcl_NewDoubleObj(arg->val.scalar);
	  Tcl_ObjSetVar2(interp,toa,ton,to, TCL_LEAVE_ERR_MSG |
			 TCL_GLOBAL_ONLY);
	  break;

	case MOPS_SASTRING:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "string", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);
	  Tcl_SetVar2(interp,n1, arg->name, arg->val.string,
		      TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	  break;

	default:
	  Tcl_SetVar2(interp,n1,"ArgTypes", "eimer", TCL_LEAVE_ERR_MSG |
		      TCL_GLOBAL_ONLY | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	  break;
	} /* switch */
      arg = arg->next;
    } /* while */

  /* we need to build an adapted buildShaderGUIproc ! */
  switch(shadertype)
    {
    case MOPS_STSURFACE:
      typestr = "surface";
      break;
    case MOPS_STDISPLACEMENT:
      typestr = "displacement";
      break;
    case MOPS_STLIGHT:
      typestr = "light";
      break;
    case MOPS_STVOLUME:
      typestr = "volume";
      break;
    case MOPS_STIMAGER:
      typestr = "imager";
      break;
    }

  Tcl_Eval(interp, update_cmd);
  sprintf(hilf,"buildNewSBuildGUIProcproc %s",typestr);
  Tcl_Eval(interp, hilf);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_shaderprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_shader *newshader = NULL;
 mops_shader_arg *newarg = NULL, **argnext = NULL;
 char *result, *n1=NULL;
 int sargnc = 0, sargtc = 0, i, shadertype = 0, shadersubtype = 0, argtype = 0;
 double dtemp = 0.0;
 char **sargnv, **sargtv;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 Tcl_DString ds;
 char fname[] = "set_shaderprop";

  if(!sel) /* oops? */
    o = mops_root;
  else
    o = sel->object; 

  /* parse args */
  if(argc < 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "shadertype arrayname subtype");
      return TCL_OK;
    }

  n1 = argv[2];
  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj(n1,-1);

  if(strstr(argv[1],"surface"))
    shadertype = MOPS_STSURFACE;
  else
  if(strstr(argv[1],"displacement"))
    shadertype = MOPS_STDISPLACEMENT;
  else
  if(strstr(argv[1],"volume"))
    {
      shadertype = MOPS_STVOLUME;
      if(strstr(argv[3],"interior"))
	shadersubtype = MOPS_STINTERIOR;
      else
      if(strstr(argv[3],"exterior"))
	shadersubtype = MOPS_STEXTERIOR;
      else
      if(strstr(argv[3],"atmosphere"))
	shadersubtype = MOPS_STATMOSPHERE;
    }
  else
  if(strstr(argv[1],"light"))
    shadertype = MOPS_STLIGHT;
  else
  if(strstr(argv[1],"imager"))
    shadertype = MOPS_STIMAGER;

  /* get shadername */
  result = 
    Tcl_GetVar2(interp, n1, "Name", TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  /* user specified no new shader -> delete old one! */
  if(!strcmp("",result))
    {
      Tcl_DStringInit(&ds);
      Tcl_DStringAppend(&ds, "delShader ", -1);
      Tcl_DStringAppend(&ds, argv[1], -1);
      if(strstr(argv[1],"volume"))
	{
	  Tcl_DStringAppend(&ds, " ", -1);
	  Tcl_DStringAppend(&ds, argv[3], -1);
	}

      Tcl_Eval(interp,Tcl_DStringValue(&ds));

      Tcl_DStringFree(&ds);

      return TCL_OK;
    }

  /* create shader-struct */
  if(!(newshader = calloc(1, sizeof(mops_shader))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
      Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);
      return TCL_OK;
    }
  if(!(newshader->name = calloc(strlen(result)+1, sizeof(char))))
    {
      free(newshader);
      Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
      Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }

  strcpy(newshader->name,result);

  newshader->type = shadertype;
  newshader->subtype = shadersubtype;

  /* decompose argument-list */
  Tcl_SplitList(interp,Tcl_GetVar2(interp, n1, "ArgNames",
				   TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY),
		&sargnc, &sargnv);
  Tcl_SplitList(interp,Tcl_GetVar2(interp, n1, "ArgTypes",
				   TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY),
		&sargtc, &sargtv);


  argnext = &(newshader->arg);
  for(i=0; i<sargnc; i++)
    {
      /* which argtype? */
      if(!strcmp(sargtv[i],"float"))
	argtype = MOPS_SASCALAR;
      else
      if(!strcmp(sargtv[i],"color"))
	argtype = MOPS_SACOLOR;
      else
      if(!strcmp(sargtv[i],"point"))
	argtype = MOPS_SAPOINT;
      else
      if(!strcmp(sargtv[i],"string"))
	argtype = MOPS_SASTRING;
      else
      if(!strcmp(sargtv[i],"vector"))
	argtype = MOPS_SAVECTOR;

      /* create new shader-arg-struct */
      if(!(newarg = calloc(1, sizeof(mops_shader_arg))))
	{
	  mops_free_shader(newshader);
	  Tcl_Free((char *) sargnv);
	  Tcl_Free((char *) sargtv);
	  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
	  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

	  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	  return TCL_OK;
	}
      *argnext = newarg;
      argnext = &(newarg->next);

      newarg->type = argtype;

      if(!(newarg->name = calloc(strlen(sargnv[i])+1, sizeof(char))))
	{
	  mops_free_shader(newshader);
	  Tcl_Free((char *) sargnv);
	  Tcl_Free((char *) sargtv);
	  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
	  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

	  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	  return TCL_OK;
	}
      strcpy(newarg->name, sargnv[i]);

      switch(argtype)
	{
	case MOPS_SACOLOR:
	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_R",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);

	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.color[0]=(float)(dtemp/255.0);

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_G",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.color[1]=(float)(dtemp/255.0);

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_B",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.color[2]=(float)(dtemp/255.0);

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_A",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.color[3]=(float)(dtemp/255.0);
	  break;
	case MOPS_SAPOINT:
	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_X",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[0] = (float)dtemp;

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_Y",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[1] = (float)dtemp;

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_Z",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[2] = (float)dtemp;
	  break;
	case MOPS_SAVECTOR:
	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_X",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[0] = (float)dtemp;

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_Y",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[1] = (float)dtemp;

	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  Tcl_AppendStringsToObj(ton,"_Z",NULL);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.point[2] = (float)dtemp;
	  break;
	case MOPS_SASCALAR:
	  Tcl_SetStringObj(ton,sargnv[i],-1);
	  to  = Tcl_ObjGetVar2(interp, toa, ton, TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  Tcl_GetDoubleFromObj(interp, to, &dtemp);
	  newarg->val.scalar = (float)dtemp;
	  break;
	case MOPS_SASTRING:
	  result = Tcl_GetVar2(interp, n1, sargnv[i], TCL_LEAVE_ERR_MSG |
			       TCL_GLOBAL_ONLY);
	  if(!(newarg->val.string = calloc(strlen(result)+1, sizeof(char))))
	  {
	    mops_free_shader(newshader);
	    Tcl_Free((char *) sargnv);
	    Tcl_Free((char *) sargtv);
	    mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	    return TCL_OK;
	  }
	  strcpy(newarg->val.string, result);
	  break;

	} /* switch */
      
    } /* for */

  /* add shader to object */
  mops_status = mops_add_shader(o, newshader);
  
  Tcl_Free((char *) sargnv);
  Tcl_Free((char *) sargtv);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


/* delete shader from selected object */
int
mops_del_shader_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_shader *shader = NULL, *nextshader = NULL;
 int shadertype = 0, shadersubtype = 0;
 char fname[] = "delShader";

  if(!sel) /* oops? */
    o = mops_root;
  else
    o = sel->object; 

  /* parse args */
  if(argc < 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "shadertype");
      return TCL_OK;
    }

  if(strstr(argv[1],"surface"))
    shadertype = MOPS_STSURFACE;
  else
  if(strstr(argv[1],"displacement"))
    shadertype = MOPS_STDISPLACEMENT;
  else
  if(strstr(argv[1],"interior"))
    {
      shadertype = MOPS_STVOLUME;
      shadersubtype = MOPS_STINTERIOR;
    }
  else 
  if(strstr(argv[1],"exterior"))
    {
      shadertype = MOPS_STVOLUME;
      shadersubtype = MOPS_STEXTERIOR;
    }
  else
  if(strstr(argv[1],"atmosphere"))
    {
      shadertype = MOPS_STVOLUME;
      shadersubtype = MOPS_STATMOSPHERE;
    }
  else
  if(strstr(argv[1],"light"))
    shadertype = MOPS_STLIGHT;
  else
  if(strstr(argv[1],"imager"))
    shadertype = MOPS_STIMAGER;
  else
    {
      mops_error(interp, MOPS_ERROR, fname,"unknown shader type");
      return TCL_OK;
    }

  while(o)
  {

   shader = o->shaders;
   if(shader)
     if(shader->type == shadertype)
       {
	 if(shadertype == MOPS_STVOLUME)
	   {	    
	     if(shader->subtype == shadersubtype)
	       {
		 o->shaders = shader->next;
		 mops_status = mops_free_shader(shader);
		 return TCL_OK;
	       }
	   }
	 else
	   {
	     o->shaders = shader->next;
	     mops_status = mops_free_shader(shader);
	     return TCL_OK;
	   }
       }

   while(shader)
     {
       nextshader = shader->next;
       if(nextshader)
	 if(nextshader->type == shadertype)
	   {
	     if(shadertype == MOPS_STVOLUME)
	       {	    
		 if(nextshader->subtype == shadersubtype)
		   {
		     shader->next = nextshader->next;
		     mops_status = mops_free_shader(nextshader);
		     return TCL_OK;
		   }
	       }
	     else
	       {
		 shader->next = nextshader->next;
		 mops_status = mops_free_shader(nextshader);
		 return TCL_OK;
	       }
	   }

       shader = nextshader;
     }

   if(sel)
     sel = sel->next;

   if(sel)
     o = sel->object;
   else
     o = NULL;

  } /* while */

 return TCL_OK;
}
