#ifndef SPEXPORT_H
#define SPEXPORT_H

#include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif

/*--------- Data and functions available to the plugin ---------*/

typedef long (SP_GET_NUM_PATCHES)(void *model, long layerID);

typedef struct {
	float	x, y, z;
} SP_Vector;

typedef struct {
	long	v1, v2, v3;
} SP_Triangle;

typedef struct {
	SP_Vector	ctrlPoints[4][4];

	long		numPoints, numTriangles;
	SP_Vector	*points, *normals;
	SP_Triangle	*triangles;
} SP_Patch;

/*------------------ Definition of the plugin ------------------*/
/*------------------- (Data and interfaces) --------------------*/

typedef BOOL (SP_BEGIN_EXPORT_PROC)(const char *pathname, SP_Vector *center, float radius);
typedef void (SP_BEGIN_LAYER_PROC)(long layerID, long numPatches);
typedef void (SP_EXPORT_PATCH_PROC)(SP_Patch *patch);
typedef BOOL (SP_END_LAYER_PROC)(long layerID);
typedef void (SP_END_EXPORT_PROC)();

typedef struct {
	/* The instance handle of the plugin. */
	HINSTANCE	thePlugin;
	/* The name of the export format to appear in the menu. */
	char		exportType[64];
	/* The export file type extension. */
	char		extn[8];
	/* Set the following to TRUE to ask user for # polygons/patch. */
	BOOL		askResolution;
	long		defaultResolution; /* 0, 1, or 2 */

	/* Called to begin the export process. */
	SP_BEGIN_EXPORT_PROC	*BeginExport;
	/* Called for each layer of the model. */
	SP_BEGIN_LAYER_PROC		*BeginLayer;
	SP_EXPORT_PATCH_PROC	*ExportPatch;
	SP_END_LAYER_PROC		*EndLayer;
	/* Called to finish up afetr all layers have been exported. */
	SP_END_EXPORT_PROC		*EndExport;
} SP_ExportPlugin;

/* Every plugin must have a function called SPDescribeMe that fits
the format below. When called, it must fill in all the data in the
SP_ExportPlugin structure. */
typedef void (SP_DESCRIBE_ME_PROC)(SP_ExportPlugin *plugin);

#ifdef __cplusplus
}
#endif

#endif
