#include "SPExport.h"
#include <stdio.h>

HINSTANCE	gMyself;

__declspec(dllexport) void SPDescribeMe(SP_ExportPlugin *plugin);
__declspec(dllexport) BOOL SPBeginExport(const char *pathname, SP_Vector *center, float radius);
__declspec(dllexport) void SPBeginLayer(long layerID, long numPatches);
__declspec(dllexport) void SPExportPatch(SP_Patch *patch);
__declspec(dllexport) BOOL SPEndLayer(long layerID);
__declspec(dllexport) void SPEndExport();

/* Standard DLL initialization. */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	gMyself = hinstDLL;
	return 1;
}

/* Export plugin initialization. */
void SPDescribeMe(SP_ExportPlugin *plugin)
{
  /* This function describes the export plugin to sPatch. */

  plugin->thePlugin = gMyself;

  /* The next line sets up the text that appears in the Export menu. */
  strcpy(plugin->exportType, "Mops");
  /* The next line sets the file type extension of your export format. */
  strcpy(plugin->extn, "mop");
  /* Set the next line to TRUE to ask the user for the polygon mesh resolution. */
  /* (Set to FALSE for patch export formats such as POV. */
  plugin->askResolution = FALSE;
  /* The next line sets the default polygon resolution.
     Can be 0, 1, or 2 - low, medium, high. */
  plugin->defaultResolution = 0;

  /* Leave the next 5 lines alone. */
  plugin->BeginExport = SPBeginExport;
  plugin->BeginLayer = SPBeginLayer;
  plugin->ExportPatch = SPExportPatch;
  plugin->EndLayer = SPEndLayer;
  plugin->EndExport = SPEndExport;
}

FILE	*gFile;

BOOL SPBeginExport(const char *pathname, SP_Vector *center, float radius)
{
  /* This function is called once to begin exporting. */
  /* Output any header information required for this file format. */
  /* The center and radius fields describe a sphere that bounds the sPatch model. */
  /* Return FALSE to cancel the export process. */

  gFile = fopen(pathname, "w");
  if (gFile == NULL)
    return FALSE;
  /* Mops-Header */
  fprintf(gFile,"Mops\n0.42d\n");
  /* Mops-RiOptions */
  fprintf(gFile,"0\n2\n2\n0\n2\n2\n1\n1\n255\n");
  fprintf(gFile,"0\n255\n0.5\n4\n64\n4\n0.01\n1\n0\n2\n\n\n\n");
  fprintf(gFile,"10000\n40000\n0\n0\n");
  /* new entity */
  putc('\a', gFile);
  /* no views */
  fprintf(gFile,"0\n");

	/* new entity */
  putc('\a', gFile);
  /* start level */
  fprintf(gFile,"%d\n",3 /*MOPS_OTLEVEL*/);
  fprintf(gFile,"%d\n",4 /*MOPS_LTLEVEL*/);
  fprintf(gFile,"%s\n","SPObject");
	
 return TRUE;
}

void SPBeginLayer(long layerID, long numPatches)
{
/* This function is called once for each layer. */
/* If your export format supports it, you may choose to 'group' the patches */
/* of each layer into one logical unit. numPatches tells how many patches will */
/* appear in this layer. (If you need the number of polygons in the layer, */
/* take the first patch passed into SPExportPatch and multiply numPatches by */
/* patch->numTriangles.) */


  /* new entity */
  putc('\a', gFile);
  /* start level */
  fprintf(gFile,"%d\n",3 /*MOPS_OTLEVEL*/);
  fprintf(gFile,"%d\n",4 /*MOPS_LTLEVEL*/);
  fprintf(gFile,"%s%d\n","Layer",layerID);

}

void SPExportPatch(SP_Patch *patch)
{
/* This function is called for each patch in a layer. */
/* Every patch is described as both a set of control points (for a cubic Bezier */
/* patch) and as a set of triangles - use whichever you need. */

 int u, v;

  /* new entity */
  putc('\a', gFile);
  /* write patch */
  fprintf(gFile,"%d\n",1 /*MOPS_OTNURBPATCH*/);
  
  fprintf(gFile,"\n"); /* name? */

  fprintf(gFile,"4\n4\n4\n4\n0\n0\n0\n");
  for (u = 0; u < 4; u++)
    {
      for (v = 0; v < 4; v++)
	{
	  fprintf(gFile,"%g\n%g\n%g\n",patch->ctrlPoints[u][v].x,
		  patch->ctrlPoints[u][v].y, patch->ctrlPoints[u][v].z);
	}
    }


  fprintf(gFile,"0\n0\n30\n6\n");

 return;
}

BOOL SPEndLayer(long layerID)
{
 /* Called after all the patches in a layer have been processed. */
 /* Perform any end-layer-specific operations here. */
 /* If you return FALSE, it tells sPatch that you aren't finished with */
 /* this layer, and sPatch will once again pass all the patches for that*/
 /* layer */
 /* through the SPExportPatch function. */

  /* new entity */
  putc('\a', gFile);
  /* end level */
  fprintf(gFile,"%d\n",3 /*MOPS_OTLEVEL*/);
  fprintf(gFile,"%d\n",10 /*MOPS_LTENDLEVEL*/);

 return TRUE;
}

void SPEndExport()
{
/* Called after all the model's layers have been exported. Do anything necessary */
/* to finish off the exported file here. */

  /* new entity */
  putc('\a', gFile);
  /* end level */
  fprintf(gFile,"%d\n",3 /*MOPS_OTLEVEL*/);
  fprintf(gFile,"%d\n",10 /*MOPS_LTENDLEVEL*/);

  fclose(gFile);

 return;
}
