/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Tags */

/********************************************/
/* mops_copy_tag:
   copy a tag */
/********************************************/
int
mops_copy_tag(mops_tag_object *source,
	      mops_tag_object **dest)
{
 int mops_status = MOPS_OK;
 mops_tag_object *new = NULL;

  if(!(new = calloc(1,sizeof(mops_tag_object))))
    return MOPS_OUTOFMEM; 

  memcpy(new, source, sizeof(mops_tag_object)); 
  /* danger! links point to original hierachy */

  new->next = NULL;

  /* copy name */
  if(!(new->name = calloc(1, strlen(source->name)+1)))
    { free(new); return MOPS_OUTOFMEM; }
  strcpy(new->name, source->name);

  /* copy val */
  if(!(new->val = calloc(1, strlen(source->val)+1)))
    { free(new); free(new->name); return MOPS_OUTOFMEM; }
  strcpy(new->val, source->val);

  *dest = new;

 return mops_status;
}



/********************************************/
/* mops_free_tag:
   free a tag */
/********************************************/
int
mops_free_tag(mops_tag_object *tag)
{
  int mops_status = MOPS_OK;

  if(!tag)
    return MOPS_OK;

  if(tag->name)
    free(tag->name);

  if(tag->val)
    free(tag->val);

  free(tag);

 return mops_status;
}

int
mops_register_tagtype(Tcl_Interp *interp, char *name, char **result)
{
 int new_item = 0;
 Tcl_HashEntry *entry = NULL;
 static char *tagcounter = (char *) 1; /* 0 - dummy */
 char  fname[] = "register_tagtype";

 /* check, if type is already registered */
 if((entry = Tcl_FindHashEntry(&mops_tagtypes_ht, name)))
   {
     mops_error(interp, MOPS_ERROR, fname, "tag type already registered");
     return TCL_OK;
   }

 tagcounter++;

 entry = Tcl_CreateHashEntry(&mops_tagtypes_ht, name, &new_item);
 Tcl_SetHashValue(entry, tagcounter);

 *result = tagcounter;

 return TCL_OK;
}


int
mops_set_tags_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_tag_object *new = NULL;
 mops_tag_object *tag = NULL, *nexttag = NULL;
 Tcl_HashEntry *entry = NULL;
 int index = 0;
 char fname[] = "setTags";

 /*
   if(argc )
   {
    mops_error(interp, MOPS_ARGS, fname, "[type value]");
    return TCL_OK;
   }
 */

 if(!sel)
   return TCL_OK;

 if(!sel->object)
   return TCL_OK;

 while(sel)
   {
     o = sel->object;

     /* delete old tags */
     tag = o->tags;
     while(tag)
       {
	 nexttag = tag->next;
	 mops_free_tag(tag);
	 tag = nexttag;
       }

     o->tags = NULL;

     for(index = 1; index < argc-1; index += 2)
       {
	 /* we first try to resolve the tag type */
	 if(!(entry = Tcl_FindHashEntry(&mops_tagtypes_ht, argv[index])))
	   {
	     /* 
	     XXXX Should we warn? 
	     mops_error(interp, MOPS_ERROR, fname, "tag type not registered");
	     return TCL_OK;
	     */
	   }
	 if(!(new = calloc(1,sizeof(mops_tag_object))))
	   {
	     mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	     return TCL_OK;
	   }
	 if(!(new->name = calloc(strlen(argv[index])+1, sizeof(char))))
	   {
	     mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	     return TCL_OK;
	   }
	 strcpy(new->name,argv[index]);

	 if(entry)
	   new->type = Tcl_GetHashValue(entry);

	 if(!(new->val = calloc(strlen(argv[index+1])+1, sizeof(char))))
	   {
	     mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	     return TCL_OK;
	   }
	 strcpy(new->val,argv[index+1]);

	 new->next = o->tags;
	 o->tags = new;
       } /* for */

     sel = sel->next;
   } /* while */
 
 return TCL_OK;
}

int
mops_get_tags_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 mops_tag_object *tag = NULL;
 char fname[] = "setTags";

  if(argc < 3)
    {
      mops_error(interp, MOPS_ARGS, fname, "varname varname2");
      return TCL_OK;
    }

 if(!sel)
   return TCL_OK;

 if(!sel->object)
   return TCL_OK;

 o = sel->object;
 if(o)
   {
     tag = o->tags;
     while(tag)
     {
       Tcl_SetVar(interp,argv[1],tag->name, TCL_APPEND_VALUE |\
		  TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
       Tcl_SetVar(interp,argv[2],tag->val, TCL_APPEND_VALUE |\
		  TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

       tag = tag->next;
     } /* while */
   } /* if */
 
 return TCL_OK;
}

