/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Simple SingleLevel Undo */

int
mops_clear_undo_info(void)
{
 int mops_status = MOPS_OK;
 mops_list_object *undl = mops_undo_list, *u;
 mops_object *undo = mops_undo_objects;
 mops_object *o = NULL;

   /* free old undo information */
  while(undo)
    {
      o = undo;
      undo = undo->next;
      mops_status = mops_delete_object(o);	  
    }
  mops_undo_objects = NULL;
  while(undl)
    {
      u=undl->next;
      free(undl);
      undl = u;
    }
  mops_undo_list = NULL;

 return MOPS_OK;
}


int
mops_save_undo_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_list_object *u;
 mops_object **nexto;
 mops_list_object *lastu = NULL;
 char fname[] = "saveUndo";

  /* free old undo information */
  mops_clear_undo_info();

  /* get pointers */
  while(sel)
    {
      if(!(u = calloc(1,sizeof(mops_list_object))))
	{
	  mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	  return TCL_OK;
	}

      if(mops_undo_list)
	{
	  lastu->next = u;
	}
      else
	mops_undo_list = u;

      lastu = u;

      u->object = sel->object;
     
      sel = sel->next;
    }
 
  /* copy all selected objects */
  nexto = &(mops_undo_objects);
  sel = mops_current_selection;
  while(sel)
    {
      mops_status = mops_copy_object(sel->object, nexto);
      if(mops_status)
	{
	  mops_error(interp,mops_status,fname,NULL);
	  return TCL_OK;
	}

      nexto = &((*nexto)->next);
      sel = sel->next;
    }
 
 return TCL_OK;
}


int
mops_undo_tcmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char *argv[])
{
 mops_object *undo = mops_undo_objects, temp, *old_orignext, *old_undonext;
 mops_list_object *undl = mops_undo_list;
 char fname[] = "undo";

 if(!undo || !undl)
   {
     mops_error(interp,MOPS_ERROR,fname,"no undo information available");
     return TCL_OK;
   }

 while(undl)
   {
     old_orignext = undl->object->next;
     old_undonext = undo->next;
     memcpy(&temp, undl->object, sizeof(mops_object));
     memcpy(undl->object, undo, sizeof(mops_object));
     undl->object->next = old_orignext;
     memcpy(undo, &temp, sizeof(mops_object));
     undo->next = old_undonext;
     undo = undo->next;
     undl = undl->next;
   }

 return TCL_OK;
}
