/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* RIB (RenderMan Interface Bytestream) Export */

/* functions local to this module */
void AimZ(RtPoint direction);
void PlaceCamera(RtPoint position, RtPoint direction, double roll);
int mops_wrib_rioptions(void);
void mops_wrib_rootsh(int imager);

/* cot() used by currently unused FrameCamera() */
/*
double
cot (double in)
{
if (sin (in) == 0)
 return (0);
else
 return (cos (in) / sin (in));
}
*/


void
AimZ(RtPoint direction)
{
 RtFloat xzlen = 0.0, yzlen = 0.0, xrot = 0.0, yrot = 0.0;
 RtFloat tmp;
 
  if((direction[0] == 0.0) && (direction[1] == 0.0) && (direction[2] == 0.0))
    return;

  xzlen = sqrt(direction[0]*direction[0] + direction[2]*direction[2]);
  if(xzlen == 0.0)
    yrot = (direction[1] < 0.0) ? 180.0 : 0.0;
  else
    {
      tmp = direction[2]/xzlen;

      if((tmp>1.0)||(tmp<-1.0))
	{
	  fprintf(stderr,"gotcha!\n");
	  fprintf(stderr,"%g %g\n",direction[0],direction[2]);
	}
      yrot = (RtFloat)MOPS_R2D(acos((fabs(tmp)<=1.0?tmp:(tmp<-1.0?-1.0:1.0))));
    }
 
  yzlen = sqrt(direction[1]*direction[1] + xzlen*xzlen);
  if(yzlen != 0.0)
    xrot = (RtFloat)MOPS_R2D(acos(xzlen/yzlen));
  else
    xrot = (RtFloat)0;

  if(direction[1] > 0.0)
    RiRotate(xrot, 1.0,0.0,0.0);
  else
    RiRotate(-xrot, 1.0,0.0,0.0);

  if(direction[0] < 0.0)
    RiRotate(yrot, 0.0,1.0,0.0);
  else
    RiRotate(-yrot, 0.0,1.0,0.0);

 return;
}

void
PlaceCamera(RtPoint position, RtPoint direction, double roll)
{
  /* XXXX RiIdentity(); */
  RiRotate(-roll, 0.0, 0.0, 1.0);
  AimZ(direction);
  RiTranslate(-position[0], -position[1],
	      -position[2]);

  return;
}


/* unused ruin */
/* XXXX Debug me! */
/*void
FrameCamera(double zoom, double framewidth, double frameheight)
{
  RtFloat fov;
  RtFloat focallength;

  focallength = (0.5 * (framewidth<frameheight)?framewidth:frameheight)*
    cot(45.0);
  
  fprintf(stderr,"focallength:%g\n",focallength);

  fov = 2.0 * atan((((framewidth<frameheight)?framewidth:frameheight)
		    * zoom * 0.5) / focallength) * 180.0 / MOPS_PI;

  fprintf(stderr,"fov:%g\n",fov);
  RiProjection("perspective", RI_FOV, (RtPointer) &fov, RI_NULL);

  RiFrameAspectRatio((RtFloat)(framewidth/frameheight));

 return;
}*/

int
mops_wrib_rioptions(void)
{
 int mops_status = MOPS_OK;
 RtInt fw = 0, fh = 0;
 RtFloat rtftemp = 0.0;
 RtInt rtitemp = 0;

  if(mops_riopt.Variance > 0.0)
    RiPixelVariance((RtFloat)mops_riopt.Variance);
  else
    RiPixelSamples((RtFloat)mops_riopt.Samples_X,
		   (RtFloat)mops_riopt.Samples_Y);

  if(mops_riopt.FilterWidth <= 0)
    fw = (RtInt)2;
  else
    fw = (RtInt)mops_riopt.FilterWidth;

  if(mops_riopt.FilterHeight <= 0)
    fh = (RtInt)2;
  else
    fh = (RtInt)mops_riopt.FilterHeight;

  switch(mops_riopt.FilterFunc)
    {
    case 1:
      RiPixelFilter(RiTriangleFilter,fw,fh);
      break;
    case 2:
      RiPixelFilter(RiCatmullRomFilter,fw,fh);
      break;
    case 3:
      RiPixelFilter(RiBoxFilter,fw,fh);
      break;
    case 4:
      RiPixelFilter(RiSincFilter,fw,fh);
      break;
    default:
      RiPixelFilter(RiGaussianFilter,fw,fh);
    }

  RiExposure((RtFloat)mops_riopt.ExpGain, (RtFloat)mops_riopt.ExpGamma);

  RiQuantize(RI_RGBA, mops_riopt.RGBA_ONE, mops_riopt.RGBA_MIN,
	     mops_riopt.RGBA_MAX, mops_riopt.RGBA_Dither);

  /* BMRT-Specific */
  if(!mops_prefs.ristandard)
  {
    rtitemp = mops_riopt.MinSamples;
    RiOption((RtToken)"render", (RtToken)"minsamples",
	     (RtPointer)(&rtitemp), RI_NULL);

    rtitemp = mops_riopt.MaxSamples;
    RiOption((RtToken)"render", (RtToken)"maxsamples",
	     (RtPointer)(&rtitemp), RI_NULL);
    
    rtitemp = mops_riopt.MaxRayLevel;
    RiOption((RtToken)"render", (RtToken)"max_raylevel",
	     (RtPointer)(&rtitemp), RI_NULL);

    rtftemp = mops_riopt.ShadowBias;
    RiOption((RtToken)"render", (RtToken)"minshadowbias",
	     (RtPointer)(&rtftemp), RI_NULL);

    rtitemp = mops_riopt.PRManSpec;
    RiOption((RtToken)"render", (RtToken)"prmanspecular",
	     (RtPointer)(&rtitemp), RI_NULL);

    rtitemp = mops_riopt.RadSteps;
    RiOption((RtToken)"radiosity", (RtToken)"steps",
	     (RtPointer)(&mops_riopt.RadSteps), RI_NULL);

    rtitemp = mops_riopt.PatchSamples;
    RiOption((RtToken)"radiosity", (RtToken)"minpatchsamples",
	     (RtPointer)(&mops_riopt.PatchSamples), RI_NULL);

    if(mops_riopt.textures)
      if((mops_riopt.textures)[0] != '\0')
	RiOption((RtToken)"searchpath", (RtToken)"texture",
		 (RtPointer)(&mops_riopt.textures), RI_NULL);

    if(mops_riopt.includes)
      if((mops_riopt.textures)[0] != '\0')
	RiOption((RtToken)"searchpath", (RtToken)"include",
		 (RtPointer)(&mops_riopt.includes), RI_NULL);

    if(mops_riopt.shaders)
      if((mops_riopt.textures)[0] != '\0')
	RiOption((RtToken)"searchpath", (RtToken)"shader",
		 (RtPointer)(&mops_riopt.shaders), RI_NULL);

    rtitemp = mops_riopt.texturemem;
    RiOption((RtToken)"limits", (RtToken)"texturememory",
	     &rtitemp, RI_NULL);

    rtitemp = mops_riopt.geommem;
    RiOption((RtToken)"limits", (RtToken)"geommemory",
	     &rtitemp, RI_NULL);
  }

 return mops_status;
}


int
mops_wrib_attributes(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_tag_object *tag = NULL;
 RtColor color = {0};
 char *camstr = "camera", *objstr = "object", *shdstr = "shader";
 char *str = NULL;
 char *tagvaltmp = NULL, *attrname = NULL, *parname = NULL, *partype = NULL,
   *parval = NULL;
 char tok[]=",";
 RtInt on = 1, itemp;
 char *nonstr = "none", *opstr = "opaque", *shadestr = "shade";
 char *zrstr = "zonal_receives", *zsstr = "zonal_shoots",
   *flstr = "fully_zonal";
 RtFloat m[4][4], dbound, ftemp;
 RtPoint ptemp={0};

  /* Color */
 if((o->colr != -1) && (o->colg != -1) && (o->colb != -1))
   {
     color[0] = (RtFloat)o->colr/255.0;
     color[1] = (RtFloat)o->colg/255.0;
     color[2] = (RtFloat)o->colb/255.0;
     RiColor(color);
   }

  /* Opacity */
 if((o->opr != -1) && (o->opg != -1) && (o->opb != -1))
   {
     color[0] = (RtFloat)o->opr/255.0;
     color[1] = (RtFloat)o->opg/255.0;
     color[2] = (RtFloat)o->opb/255.0;
     RiOpacity(color);
   }


  if(!o->refcount || mops_prefs.resolveinstances)
    {

      if(o->movx != 0.0 || o->movy != 0.0 || o->movz != 0.0)
	RiTranslate((RtFloat)o->movx, (RtFloat)o->movy, (RtFloat)o->movz);

      if(o->quat[0] != 0.0 || o->quat[1] != 0.0 || o->quat[2] != 0.0 ||
	 o->quat[3] != 1.0)
	{
	  mops_build_rotmatrix(m, o->quat);
	  RiConcatTransform(m);
	}

      if(o->scalx != 1.0 || o->scaly != 1.0 || o->scalz != 1.0)
	RiScale((RtFloat)o->scalx, (RtFloat)o->scaly, (RtFloat)o->scalz);
    }

  if(o->shading_rate != 1.0)
    RiShadingRate((RtFloat)o->shading_rate);

  if(o->shading_interpolation)
    RiShadingInterpolation(RI_SMOOTH);
  /*  else
      RiShadingInterpolation(RI_CONSTANT);*/

  if(!mops_prefs.ristandard)
    {
      if(o->true_displacement)
	{
	  RiAttribute("render","truedisplacement",&on,RI_NULL);
	  switch(o->d_bound)
	    {
	    case 0:
	      str = objstr;
	      break;
	    case 1:
	      str = shdstr;
	      break;
	    case 2:
	      str = camstr;
	      break;
	      
	    }
	  dbound = (RtFloat)o->d_bound_val;
	  RiAttribute("displacementbound", "coordinatesystem",
		      (RtPointer)&str, "sphere", &dbound, RI_NULL);
	} /* if */

      if(o->cast_shadows)
	{
	  if(o->cast_shadows == 1)
	    RiAttribute("render", "casts_shadows",
			(RtPointer)&nonstr, RI_NULL);
	  if(o->cast_shadows == 2)
	    RiAttribute("render", "casts_shadows",
			(RtPointer)&opstr, RI_NULL);
	  if(o->cast_shadows == 3)
	    RiAttribute("render", "casts_shadows",
			(RtPointer)&shadestr, RI_NULL);

	}

      /* Radiosity */
      if((o->avr != -1) && (o->avg != -1) && (o->avb != -1))
	{
	  color[0] = (RtFloat)o->avr/255.0;
	  color[1] = (RtFloat)o->avg/255.0;
	  color[2] = (RtFloat)o->avb/255.0;

	  RiAttribute("radiosity", "averagecolor", (RtPointer)&color, RI_NULL);
	}
      if((o->emr != -1) && (o->emg != -1) && (o->emb != -1))
	{
	  color[0] = (RtFloat)o->emr/255.0;
	  color[1] = (RtFloat)o->emg/255.0;
	  color[2] = (RtFloat)o->emb/255.0;

	  RiAttribute("radiosity", "emissioncolor",
		      (RtPointer)&color, RI_NULL);
	}
      if(o->patch_size != -1.0)
	{
	  ftemp = (RtFloat)o->patch_size;
	  RiAttribute("radiosity", "patchsize",(RtPointer)&ftemp, RI_NULL);
	}
      if(o->elem_size != -1.0)
	{
	  ftemp = (RtFloat)o->elem_size;
	  RiAttribute("radiosity", "elemsize",(RtPointer)&ftemp, RI_NULL);
	}
      if(o->min_size != -1.0)
	{
	  ftemp = (RtFloat)o->min_size;
	  RiAttribute("radiosity", "minsize",(RtPointer)&ftemp, RI_NULL);
	}
      if(o->zonal)
	{
	  if(o->zonal == 1)
	    RiAttribute("radiosity", "zonal", (RtPointer)&nonstr, RI_NULL);
	  if(o->zonal == 2)
	    RiAttribute("radiosity", "zonal", (RtPointer)&zrstr, RI_NULL);
	  if(o->zonal == 3)
	    RiAttribute("radiosity", "zonal", (RtPointer)&zsstr, RI_NULL);
	  if(o->zonal == 4)
	    RiAttribute("radiosity", "zonal", (RtPointer)&flstr, RI_NULL);
	}
      if((o->spr != -1) && (o->spg != -1) && (o->spb != -1))
	{
	  color[0] = (RtFloat)o->spr/255.0;
	  color[1] = (RtFloat)o->spg/255.0;
	  color[2] = (RtFloat)o->spb/255.0;

	  RiAttribute("radiosity", "specularcolor",
		      (RtPointer)&color, RI_NULL);
	}

      if(o->has_caustics)
	{
	  itemp = 1;
	  RiAttribute("radiosity", "has_caustics", (RtPointer)&itemp, RI_NULL);
	}

      /* visibility */
      if((!o->camera) || (!o->reflection) || (!o->shadow))
	{
	  itemp = 0;

	  if(o->camera)
	    itemp+=1;

	  if(o->reflection)
	    itemp+=2;

	  if(o->shadow)
	    itemp+=4;

	  RiAttribute("render", "visibility", (RtPointer)&itemp, RI_NULL);
	}
    } /* if !ristandard */

 /* process tags */
 tag = o->tags;
 while(tag)
   {
     if(tag->type == mops_riattr_tagtype)
       {
	 /* muto teleportilis shaelorasti */
	 if(tagvaltmp)
	   {
	     free(tagvaltmp);
	     tagvaltmp = NULL;
	   }

	 if(!(tagvaltmp = calloc(1, strlen(tag->val)+1)))
	   { return MOPS_OUTOFMEM; }
	 strcpy(tagvaltmp, tag->val);

	 /* get name of attribute */
	 attrname = strtok(tagvaltmp,tok);

	 /* get parameter(s) */
	 if(attrname)
	   {
	     parval = NULL;
	     partype = NULL;
	     parname = NULL;

	     /* get name */
	     parname = strtok(NULL,tok);
	     /* get type */
	     if(parname)
	       partype = strtok(NULL,tok);
	     else 
	       partype = NULL;

	     if(partype)
	       {
		 /* get value */
		 parval = strtok(NULL,tok);
		 if(parval)
		 {
		  if(*partype == 'i')
		   {
		     RiDeclare(parname, "int");
		     sscanf(parval, "%d", &itemp);
		     RiAttribute(attrname, parname,
				 (RtPointer)&itemp, RI_NULL);
		   } else {
		   if(*partype == 'f')
		    {
		      RiDeclare(parname, "float");
		      sscanf(parval, "%f", &ftemp);
		      RiAttribute(attrname, parname,
				  (RtPointer)&ftemp, RI_NULL);
		    } else {
		    if(*partype == 's')
		     {
		       RiDeclare(parname, "string");
		       RiAttribute(attrname, parname,
				   (RtPointer)&parval, RI_NULL);
		     } else {
		     if(*partype == 'p')
		     {
		       RiDeclare(parname, "point");
		       sscanf(parval, "%f,%f,%f", &ptemp[0], &ptemp[1],
			      &ptemp[2]);
		       RiAttribute(attrname, parname,
				   (RtPointer)&ptemp, RI_NULL);
		     } else {
		     if(*partype == 'c')
		     {
		       RiDeclare(parname, "color");
		       sscanf(parval, "%f,%f,%f", &color[0], &color[1],
			      &color[2]);
		       RiAttribute(attrname, parname,
				   (RtPointer)&color, RI_NULL);
		     }
		     }
		     }
		     }
		     }
		 }
	       }
	     

	   }
	 else
	   {
	     /* XXXX error */
	   }
       }

     tag = tag->next;
   }

 return mops_status;
}


int
mops_wrib_trafos(mops_object *o)
{
 int mops_status = MOPS_OK;
 RtFloat m[4][4];

  if(o->movx != 0.0 || o->movy != 0.0 || o->movz != 0.0)
    RiTranslate((RtFloat)o->movx, (RtFloat)o->movy, (RtFloat)o->movz);

  if(o->quat[0] != 0.0 || o->quat[1] != 0.0 || o->quat[2] != 0.0 ||
     o->quat[3] != 1.0)
     {
       mops_build_rotmatrix(m, o->quat);
       RiConcatTransform(m);
     }

  if(o->scalx != 1.0 || o->scaly != 1.0 || o->scalz != 1.0)
    RiScale((RtFloat)o->scalx, (RtFloat)o->scaly, (RtFloat)o->scalz);


 return mops_status;
}


int
mops_wrib_object(mops_object *o, char *file)
{
 int mops_status = MOPS_OK;
 mops_level_object *level = NULL;

  if(!o)
    return MOPS_OK;

  if(o->name)
    RiArchiveRecord(RI_COMMENT,o->name);

  if(!mops_prefs.resolveinstances && o->refcount)
    {
      mops_status = mops_wrib_instance(o,file);
      return mops_status; /* XXXX early exit! */
    }

  switch(o->type)
    {
    case MOPS_OTLEVEL:
      if(!o->down)
	return mops_status; /* XXXX early exit! */
      RiAttributeBegin();
      RiTransformBegin();
      mops_wrib_attributes(o);
      if(o->shaders)
	mops_status = mops_wrib_shaders(o);
      level = (mops_level_object*)o->object;
      switch(level->type)
	{
	case MOPS_LTUNION:
	  RiSolidBegin(RI_UNION);
	  break;
	case MOPS_LTINTERSECTION:
	  RiSolidBegin(RI_INTERSECTION);
	  break;
	case MOPS_LTDIFFERENCE:
	  RiSolidBegin(RI_DIFFERENCE);
	  break;
	case MOPS_LTPRIMITIVE:
	  RiSolidBegin(RI_PRIMITIVE);
	  break;
	}

      if(o->down)
	{
	  o = o->down;
	  mops_status = mops_wrib_object(o,file);
	  if(o->next)
	    {
	      o = o->next;
	      while(o)
		{
		  mops_status = mops_wrib_object(o,file);
		  o = o->next;
		}
	    }
	}
      switch(level->type)
	{
	case MOPS_LTUNION:
	case MOPS_LTINTERSECTION:
	case MOPS_LTDIFFERENCE:
	case MOPS_LTPRIMITIVE:
	  RiSolidEnd();
	  break;
	default:
	  break;
	}

      RiTransformEnd();
      RiAttributeEnd();
      break;
    case MOPS_OTNURBPATCH:
      RiAttributeBegin();
      RiTransformBegin();
      mops_wrib_attributes(o);
      mops_status = mops_wrib_trimcurves(o);
      mops_status = mops_wrib_npatch(o);
      RiTransformEnd();
      RiAttributeEnd();      
      break;
    case MOPS_OTCUSTOM:
      mops_status = mops_wrib_custom(o);
      break;

    case MOPS_OTSPHERE:
      mops_status = mops_wrib_sphere(o);
      break;
    case MOPS_OTDISK:
      mops_status = mops_wrib_disk(o);
      break;
    case MOPS_OTCONE:
      mops_status = mops_wrib_cone(o);
      break;
    case MOPS_OTCYLINDER:
      mops_status = mops_wrib_cylinder(o);
      break;
    case MOPS_OTTORUS:
      mops_status = mops_wrib_torus(o);
      break;
    case MOPS_OTPARABOLOID:
      mops_status = mops_wrib_paraboloid(o);
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mops_wrib_hyperboloid(o);
      break;
    case MOPS_OTBOX:
      mops_status = mops_wrib_box(o);
      break;
    case MOPS_OTINSTANCE:
      mops_status = mops_wrib_instance(o, file);
      break;

    }

 return mops_status;
}


int
mops_wrib_archive_object(mops_object *o, char *file)
{
 int mops_status = MOPS_OK;
 mops_level_object *level = NULL;

  if(!o)
    return MOPS_OK;

  switch(o->type)
    {
    case MOPS_OTLEVEL:
      RiAttributeBegin();
      RiTransformBegin();
      mops_wrib_attributes(o);
      if(o->shaders)
	mops_status = mops_wrib_shaders(o);
      level = (mops_level_object*)o->object;
      switch(level->type)
	{
	case MOPS_LTUNION:
	  RiSolidBegin(RI_UNION);
	  break;
	case MOPS_LTINTERSECTION:
	  RiSolidBegin(RI_INTERSECTION);
	  break;
	case MOPS_LTDIFFERENCE:
	  RiSolidBegin(RI_DIFFERENCE);
	  break;
	case MOPS_LTPRIMITIVE:
	  RiSolidBegin(RI_PRIMITIVE);
	  break;
	}

      if(o->down)
	{
	  o = o->down;
	  mops_status = mops_wrib_object(o,file);
	  if(o->next)
	    {
	      o = o->next;
	      while(o)
		{
		  mops_status = mops_wrib_object(o,file);
		  o = o->next;
		}
	    }
	}
      switch(level->type)
	{
	case MOPS_LTUNION:
	case MOPS_LTINTERSECTION:
	case MOPS_LTDIFFERENCE:
	case MOPS_LTPRIMITIVE:
	  RiSolidEnd();
	  break;
	default:
	  break;
	}

      RiTransformEnd();
      RiAttributeEnd();
      break;
    case MOPS_OTNURBPATCH:
      RiAttributeBegin();
      RiTransformBegin();
      mops_wrib_attributes(o);
      mops_status = mops_wrib_trimcurves(o);
      mops_status = mops_wrib_npatch(o);
      RiTransformEnd();
      RiAttributeEnd();      
      break;
    case MOPS_OTCUSTOM:
      mops_status = mops_wrib_custom(o);
      break;

    case MOPS_OTSPHERE:
      mops_status = mops_wrib_sphere(o);
      break;
    case MOPS_OTDISK:
      mops_status = mops_wrib_disk(o);
      break;
    case MOPS_OTCONE:
      mops_status = mops_wrib_cone(o);
      break;
    case MOPS_OTCYLINDER:
      mops_status = mops_wrib_cylinder(o);
      break;
    case MOPS_OTTORUS:
      mops_status = mops_wrib_torus(o);
      break;
    case MOPS_OTPARABOLOID:
      mops_status = mops_wrib_paraboloid(o);
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mops_wrib_hyperboloid(o);
      break;
    case MOPS_OTBOX:
      mops_status = mops_wrib_box(o);
      break;
    case MOPS_OTINSTANCE:
      mops_status = mops_wrib_instance(o,file);
      break;
    case MOPS_OTLIGHT:
      mops_status = mops_wrib_lights(o,file, MOPS_TRUE);
      break;
    }

 return mops_status;
}


int
mops_wrib_iarchives(mops_object *o, char *file)
{
 int mops_status = MOPS_OK;
 int found = MOPS_FALSE;
 mops_tag_object *tag = NULL;
 char *iafilename = NULL;

  while(o)
    {
      if(o->refcount)
	{
	  found = MOPS_FALSE;
	  tag = o->tags;
	  while(tag && !found)
	    {
	      if(tag->type == mops_oi_tagtype)
		{

		  /* create filename */
		  if(!file)
		    RiBegin(tag->val);
		  else
		    {
		      if(!(iafilename = calloc(1,
			   strlen(tag->val)+strlen(file)+2)))
			return MOPS_OUTOFMEM;
		      strcpy(iafilename,file);
		      iafilename[strlen(file)] = '-';
		      strcpy(&(iafilename[strlen(file)+1]),tag->val);
		      RiBegin(iafilename);
		      free(iafilename);
		    }

		  mops_status = mops_wrib_archive_object(o,file);

		  RiEnd();

		  found = MOPS_TRUE;
		}
	      tag = tag->next;
	    }

	  if(!found)
	    return MOPS_ERROR; /* This should never happen! */

	}

      if(o->down)
	mops_status = mops_wrib_iarchives(o->down, file);

      if(mops_status)
	return mops_status;

      o = o->next;
    }

 return mops_status;
}


void
mops_wrib_rootsh(int imager)
{
 mops_shader **before, *after,*after2, *sh;
 mops_object *o = mops_root;

 if(!imager)
   { /* write all shaders except the imager */
     before = &(o->shaders);
     sh = o->shaders;
     while(sh)
       {
	 after = sh->next;
	 if(sh->type == MOPS_STIMAGER)
	   {
	     /* break out the imager shader */
	     *before = after;
	     sh->next = NULL;
	     mops_wrib_shaders(mops_root);
	     *before = sh;
	     sh->next = after;
	     return; /* XXXX early exit */

	   }
	 before = &(sh->next);
	 sh = sh->next;
       }

     /* mops_root seems not to have an imager shader */
     mops_wrib_shaders(mops_root);

   }
 else
   { /* write just the imager shader */
     after = o->shaders;
     sh = o->shaders;
     while(sh)
       {
	 if(sh->type == MOPS_STIMAGER)
	   {
	     o->shaders = sh;
	     after2 = sh->next;
	     sh->next = NULL;
	     mops_wrib_shaders(mops_root);
	     o->shaders = after;
	     sh->next = after2;
	     return; /* XXXX early exit */

	   }
	 before = &(sh->next);
	 sh = sh->next;
       } /* while */

   } /* if */


 return;
}

int
mops_wrib_scene(char *file, char *image,
		double fromx, double fromy, double fromz,
		double tox, double toy, double toz,
		double roll, double zoom,
		int width, int height, int type)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_root;
 RtPoint from, to, direction;
 RtFloat aspect = 1.0, swleft, swright, swtop, swbot;

 if(!mops_prefs.resolveinstances)
  {
    /* reset oid counter */
    mops_create_oid(NULL);
    /* create OI tags for all original (referenced) objects */
    mops_status = mops_create_orig_ids(o);
    /* create OI tags for all instance (referencing) objects */
    mops_status = mops_create_instance_ids(o);
    /* write archive files for all original (referenced) objects */
    mops_status = mops_wrib_iarchives(o, file);

  }

  aspect = (RtFloat)(width/((double)height));
  /* assemble args */
  from[0] = (RtFloat) fromx;
  from[1] = (RtFloat) fromy;
  from[2] = (RtFloat) fromz;

  to[0] = (RtFloat) tox;
  to[1] = (RtFloat) toy;
  to[2] = (RtFloat) toz;

  direction[0] = (RtFloat)(tox - fromx);
  direction[1] = (RtFloat)(toy - fromy);
  direction[2] = (RtFloat)(toz - fromz);

  if(aspect < 1.0)
    {
      swleft = -1.0;
      swright = 1.0;
      swbot = -1.0/aspect;
      swtop = 1.0/aspect;
    }
  else
    {

      swleft = -aspect;
      swright =  aspect;
      swbot = -1.0;
      swtop = 1.0;
    }

  if(!file) /* dump .rib to stdout? */
    RiBegin(RI_NULL);
  else
    RiBegin(file);


  if(!image) /* render to image file or to frame buffer? */
    RiDisplay(RI_NULL, RI_FRAMEBUFFER, RI_RGBA, RI_NULL);
  else
    RiDisplay(image, RI_FILE, RI_RGBA, RI_NULL);

  /* write imager shader */
  if(mops_root->shaders)
    mops_wrib_rootsh(MOPS_TRUE);


  /* Camera! */
  RiFormat(width, height, -1.0);
  if(type != MOPS_VTPERSP)
    RiProjection("orthographic", RI_NULL);
  else
    RiProjection("perspective", RI_NULL);

  RiFrameAspectRatio((RtFloat)aspect);
  /*  RiScreenWindow(swright*zoom, swleft*zoom, swbot*zoom, swtop*zoom);*/
  RiScreenWindow((RtFloat)swleft*zoom, (RtFloat)swright*zoom,
		 (RtFloat)swbot*zoom, (RtFloat)swtop*zoom);

  RiIdentity();

  /* check for lights */
  if(mops_prefs.checklights)
    {
      if(!mops_check_lights(mops_root))
	{ /* no lights in scene -> add a distant headlight */
	  RiLightSource("distantlight",RI_NULL);
	}
    }

  /* convert rh to lh */
  RiScale(-1.0, 1.0, 1.0);
  PlaceCamera(from, direction, roll);

  mops_status = mops_wrib_rioptions();

  RiWorldBegin();

  /* Lights! */
  mops_status = mops_wrib_lights(mops_root->next, file, MOPS_FALSE);

  RiIdentity();

  /* Action! */

  /* write root shaders (!imager) */
  if(mops_root->shaders)
    mops_wrib_rootsh(MOPS_FALSE);

  o = mops_root->next;
  while(o)
    {
      mops_status = mops_wrib_object(o,file);
      o = o->next;
    }
  RiWorldEnd();

  /* Cut! */
  RiEnd();

 return mops_status;
}


int
mops_wrib_cb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 int width = Togl_Width (togl);
 int height = Togl_Height (togl);
 int mops_status;
 int i, temp = 0;
 char *file = NULL, *image = NULL;
 char fname[] = "write_rib";

  /* assemble args */
  i = 2;
  while(i+1<=argc)
    {
      if(!strcmp(argv[i],"-file"))
	file = argv[i+1];
      else
	if(!strcmp(argv[i],"-image"))
	  image = argv[i+1];
      else
	if(!strcmp(argv[i],"-temp"))
	  temp = 1;

      i+=2;
    }

  if(!temp)
    {
      /* override view width/height ? */
      if(mops_riopt.width > 0)
	width = mops_riopt.width;

      if(mops_riopt.height > 0)
	height = mops_riopt.height;
    }

  /* default */
  if(!file)
    file = "rendrib";

  /* adjust roll, if up vector points down */
  if(view->up_y < 0.0)
    {
      view->roll += 180.0;
    }

  mops_status = mops_wrib_scene(file, image,
				view->from_x, view->from_y, view->from_z,
				view->to_x, view->to_y, view->to_z, view->roll,
				view->zoom, width, height,
				view->type);

  if(view->up_y < 0.0)
    {
      view->roll -= 180.0;
    }

  MOPS_CRERROR

  if(!temp)
    Tcl_Eval(interp,
	     "puts stdout \"Done exporting scene to: $efilename\"");

 return TCL_OK;
}
