/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Write Scenes */

int
mops_write_header(FILE *fileptr)
{
 int mops_status = MOPS_OK;

 fprintf(fileptr,"Mops\n");
 fprintf(fileptr,"0.42d\n");

 return mops_status;
}

int
mops_write_rioptions(FILE *fileptr)
{
 int mops_status = MOPS_OK;

 fprintf(fileptr,"%g\n",mops_riopt.Variance);
 fprintf(fileptr,"%d\n",mops_riopt.Samples_X);
 fprintf(fileptr,"%d\n",mops_riopt.Samples_Y);
 fprintf(fileptr,"%d\n",(int)(mops_riopt.FilterFunc));
 fprintf(fileptr,"%d\n",mops_riopt.FilterWidth);
 fprintf(fileptr,"%d\n",mops_riopt.FilterHeight);
 fprintf(fileptr,"%g\n",mops_riopt.ExpGain);
 fprintf(fileptr,"%g\n",mops_riopt.ExpGamma);
 fprintf(fileptr,"%g\n",mops_riopt.RGBA_ONE);
 fprintf(fileptr,"%g\n",mops_riopt.RGBA_MIN);
 fprintf(fileptr,"%g\n",mops_riopt.RGBA_MAX);
 fprintf(fileptr,"%g\n",mops_riopt.RGBA_Dither);

 fprintf(fileptr,"%d\n",mops_riopt.MinSamples);
 fprintf(fileptr,"%d\n",mops_riopt.MaxSamples);
 fprintf(fileptr,"%d\n",mops_riopt.MaxRayLevel);
 fprintf(fileptr,"%g\n",mops_riopt.ShadowBias);
 fprintf(fileptr,"%d\n",(int)(mops_riopt.PRManSpec));
 fprintf(fileptr,"%d\n",mops_riopt.RadSteps);
 fprintf(fileptr,"%d\n",mops_riopt.PatchSamples);

 if(mops_riopt.textures)
   fprintf(fileptr,"%s\n",mops_riopt.textures);
 else
   fprintf(fileptr,"\n");

 if(mops_riopt.includes)
   fprintf(fileptr,"%s\n",mops_riopt.includes);
 else
   fprintf(fileptr,"\n");

 if(mops_riopt.shaders)
   fprintf(fileptr,"%s\n",mops_riopt.shaders);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr,"%d\n",mops_riopt.texturemem);
 fprintf(fileptr,"%d\n",mops_riopt.geommem);

 fprintf(fileptr,"%d\n",mops_riopt.width);
 fprintf(fileptr,"%d\n",mops_riopt.height);

 return mops_status;
}

int
mops_write_views(FILE *fileptr)
{
 int mops_status = MOPS_OK;
 mops_object *root = mops_root, *o;
 mops_view_info *v = NULL;
 struct Togl *togl = NULL;
 int vcount = 0,i;

 if(!mops_prefs.saveviews)
   {
     /* mark start of new entity */
     putc('\a', fileptr);
     /* no views */
     fprintf(fileptr,"0\n");
     return MOPS_OK;
   }

  o = root->down;
  while(o)
    {
      vcount++;
      o = o->next;
    }

  /* mark start of new entity */
  putc('\a', fileptr);
  fprintf(fileptr,"%d\n", vcount);
  o = root->down;
  for(i = 0; i < vcount; i++)
    {
      v = (mops_view_info*)o->object;
      togl = v->togl;
      /* mark start of new entity */
      putc('\a', fileptr);
      fprintf(fileptr,"%d\n",Togl_Width(togl));
      fprintf(fileptr,"%d\n",Togl_Height(togl));
      fprintf(fileptr,"%d\n",v->type);
      fprintf(fileptr,"%g\n",v->grid);
      fprintf(fileptr,"%g\n",v->bgr);
      fprintf(fileptr,"%g\n",v->bgg);
      fprintf(fileptr,"%g\n",v->bgb);
      fprintf(fileptr,"%g\n",v->bga);
      fprintf(fileptr,"%d\n",v->drawwire);
      fprintf(fileptr,"%d\n",v->drawsurf);
      fprintf(fileptr,"%d\n",v->drawsel);
      fprintf(fileptr,"%d\n",v->redraw);
      fprintf(fileptr,"%d\n",v->drawgrid);
      fprintf(fileptr,"%g\n",v->from_x);
      fprintf(fileptr,"%g\n",v->from_y);
      fprintf(fileptr,"%g\n",v->from_z);
      fprintf(fileptr,"%g\n",v->to_x);
      fprintf(fileptr,"%g\n",v->to_y);
      fprintf(fileptr,"%g\n",v->to_z);
      fprintf(fileptr,"%g\n",v->roll);
      fprintf(fileptr,"%g\n",v->zoom);
      fprintf(fileptr,"%d\n",v->lighting);

      fprintf(fileptr,"%d\n",v->pos_x);
      fprintf(fileptr,"%d\n",v->pos_y);

      fprintf(fileptr,"%g\n",v->rotx);
      fprintf(fileptr,"%g\n",v->roty);
      fprintf(fileptr,"%g\n",v->rotz);
      fprintf(fileptr,"%g\n",v->up_x);
      fprintf(fileptr,"%g\n",v->up_y);
      fprintf(fileptr,"%g\n",v->up_z);

      o = o->next;
    }

 return mops_status;
}


int 
mops_write_root(FILE *fileptr)
{
 int mops_status = MOPS_OK;

 mops_status = mops_write_rioptions(fileptr);

 mops_status = mops_write_views(fileptr);

 return mops_status;
}


int 
mops_write_attributes(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;

  if(!o)
    return MOPS_OK;

  /* mark start of new entity */
  putc('\a', fileptr);

  fprintf(fileptr,"%d\n", MOPS_OTATTRIB);

  fprintf(fileptr,"%g\n",o->movx);
  fprintf(fileptr,"%g\n",o->movy);
  fprintf(fileptr,"%g\n",o->movz);

  fprintf(fileptr,"%g\n",o->rotx);
  fprintf(fileptr,"%g\n",o->roty);
  fprintf(fileptr,"%g\n",o->rotz);

  fprintf(fileptr,"%g\n",o->quat[0]);
  fprintf(fileptr,"%g\n",o->quat[1]);
  fprintf(fileptr,"%g\n",o->quat[2]);
  fprintf(fileptr,"%g\n",o->quat[3]);

  fprintf(fileptr,"%g\n",o->scalx);
  fprintf(fileptr,"%g\n",o->scaly);
  fprintf(fileptr,"%g\n",o->scalz);

  fprintf(fileptr,"%d\n",o->colr);
  fprintf(fileptr,"%d\n",o->colg);
  fprintf(fileptr,"%d\n",o->colb);
  fprintf(fileptr,"%d\n",o->cola);
  
  fprintf(fileptr,"%d\n",o->opr);
  fprintf(fileptr,"%d\n",o->opg);
  fprintf(fileptr,"%d\n",o->opb);
  fprintf(fileptr,"%d\n",o->opa);

  fprintf(fileptr,"%g\n",o->shading_rate);
  fprintf(fileptr,"%d\n",o->shading_interpolation);

  fprintf(fileptr,"%d\n",o->cast_shadows);

  fprintf(fileptr,"%d\n",o->true_displacement);
  fprintf(fileptr,"%d\n",o->d_bound);
  fprintf(fileptr,"%g\n",o->d_bound_val);

  fprintf(fileptr,"%d\n",o->avr);
  fprintf(fileptr,"%d\n",o->avg);
  fprintf(fileptr,"%d\n",o->avb);
  fprintf(fileptr,"%d\n",o->ava);

  fprintf(fileptr,"%d\n",o->emr);
  fprintf(fileptr,"%d\n",o->emg);
  fprintf(fileptr,"%d\n",o->emb);
  fprintf(fileptr,"%d\n",o->ema);

  fprintf(fileptr,"%d\n",o->spr);
  fprintf(fileptr,"%d\n",o->spg);
  fprintf(fileptr,"%d\n",o->spb);
  fprintf(fileptr,"%d\n",o->spa);

  fprintf(fileptr,"%g\n",o->patch_size);
  fprintf(fileptr,"%g\n",o->elem_size);
  fprintf(fileptr,"%g\n",o->min_size);

  fprintf(fileptr,"%d\n",o->zonal);
  fprintf(fileptr,"%d\n",o->has_caustics);

  fprintf(fileptr,"%d\n",o->camera);
  fprintf(fileptr,"%d\n",o->reflection);
  fprintf(fileptr,"%d\n",o->shadow);
  fprintf(fileptr,"%d\n",o->editor);

 return mops_status;
}


int 
mops_write_trafos(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;

  if(!o)
    return MOPS_OK;

  /* mark start of new entity */
  putc('\a', fileptr);

  fprintf(fileptr,"%d\n", MOPS_OTTRAFO);

  fprintf(fileptr,"%g\n",o->movx);
  fprintf(fileptr,"%g\n",o->movy);
  fprintf(fileptr,"%g\n",o->movz);

  fprintf(fileptr,"%g\n",o->rotx);
  fprintf(fileptr,"%g\n",o->roty);
  fprintf(fileptr,"%g\n",o->rotz);

  fprintf(fileptr,"%g\n",o->quat[0]);
  fprintf(fileptr,"%g\n",o->quat[1]);
  fprintf(fileptr,"%g\n",o->quat[2]);
  fprintf(fileptr,"%g\n",o->quat[3]);

  fprintf(fileptr,"%g\n",o->scalx);
  fprintf(fileptr,"%g\n",o->scaly);
  fprintf(fileptr,"%g\n",o->scalz);

 return mops_status;
}

int 
mops_write_shaders(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int shcount = 0, argcount = 0;
 mops_shader *shader;
 mops_shader_arg *sarg;

  if((!o) || (!o->shaders))
    return MOPS_OK;

  /* count shaders */

  shader = o->shaders;
  while(shader)
    {
      shcount++;
      shader = shader->next;
    }

  /* mark start of new entity */
  putc('\a', fileptr);

  fprintf(fileptr,"%d\n",MOPS_OTSHADER);

  fprintf(fileptr,"%d\n",shcount);

  shader = o->shaders;
  while(shader)
    {
      fprintf(fileptr,"%s\n",shader->name);
      fprintf(fileptr,"%d\n",shader->type);
      fprintf(fileptr,"%d\n",shader->subtype);

      /* count args */
      argcount = 0;
      sarg = shader->arg;
      while(sarg)
	{
	  argcount++;
	  sarg = sarg->next;
	}
      sarg = shader->arg;

      fprintf(fileptr,"%d\n",argcount);

      while(sarg)
	{
	  fprintf(fileptr,"%s\n",sarg->name);
	  fprintf(fileptr,"%d\n",sarg->type);

	  switch(sarg->type)
	    {

	    case MOPS_SACOLOR:
	      fprintf(fileptr,"%g\n",sarg->val.color[0]);
	      fprintf(fileptr,"%g\n",sarg->val.color[1]);
	      fprintf(fileptr,"%g\n",sarg->val.color[2]);
	      fprintf(fileptr,"%g\n",sarg->val.color[3]);
	      break;
	    case MOPS_SAVECTOR:
	    case MOPS_SAPOINT:
	      fprintf(fileptr,"%g\n",sarg->val.point[0]);
	      fprintf(fileptr,"%g\n",sarg->val.point[1]);
	      fprintf(fileptr,"%g\n",sarg->val.point[2]);
	      break;
	    case MOPS_SASCALAR:
	      fprintf(fileptr,"%g\n",sarg->val.scalar);
	      break;
	    case MOPS_SASTRING:
	      if(sarg->val.string)
		fprintf(fileptr,"%s\n",sarg->val.string);
	      else
		fprintf(fileptr,"\n");
	      break;
	    }
	  sarg = sarg->next;
	}
      shader = shader->next;
    }
  
  return mops_status;
}

int 
mops_write_tags(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int tcount = 0;
 mops_tag_object *tag;


  if((!o) || (!o->tags))
    return MOPS_OK;

  /* count tags */

  tag = o->tags;
  while(tag)
    {
      tcount++;
      tag = tag->next;
    }

  /* mark start of new entity */
  putc('\a', fileptr);

  fprintf(fileptr,"%d\n",MOPS_OTTAG);

  fprintf(fileptr,"%d\n",tcount);

  tag = o->tags;
  while(tag)
    {
      fprintf(fileptr,"%s\n",tag->name);
      fprintf(fileptr,"%s\n",tag->val);

      tag = tag->next;
    }
  
  return mops_status;
}


int
mops_write_level(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_level_object *level = NULL;
 mops_object *otemp = o;

 if(o->object)
   level = (mops_level_object *) o->object;
 else
   return MOPS_OK;

 /* write level object */
 fprintf(fileptr,"%d\n",MOPS_OTLEVEL);
 fprintf(fileptr,"%d\n",level->type);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 /* write levels content */
 o = o->down;
 while(o)
   {
     mops_status = mops_write_object(fileptr, o);
     if(mops_status)
       return mops_status;

     o = o->next;
   }

 /* write endoflevel object */
 /* mark start of new entity */
 putc('\a', fileptr);
 fprintf(fileptr,"%d\n",otemp->type);
 fprintf(fileptr,"%d\n",MOPS_LTENDLEVEL);

 return mops_status;
}


int 
mops_write_view(FILE *fileptr, mops_view_info *view)
{
 int mops_status = MOPS_OK;

 if(!view)
   return MOPS_OK;

 fprintf(fileptr,"%d\n", MOPS_OTVIEW);

 fprintf(fileptr,"%d\n",view->type);
 fprintf(fileptr,"%g\n",view->grid);
 fprintf(fileptr,"%g\n",view->bgr);
 fprintf(fileptr,"%g\n",view->bgg);
 fprintf(fileptr,"%g\n",view->bgb);
 fprintf(fileptr,"%g\n",view->bga);

 fprintf(fileptr,"%d\n",view->drawwire);
 fprintf(fileptr,"%d\n",view->drawsurf);
 fprintf(fileptr,"%d\n",view->drawsel);
 fprintf(fileptr,"%d\n",view->redraw);
 fprintf(fileptr,"%d\n",view->drawgrid);
 fprintf(fileptr,"%d\n",view->lighting);

 fprintf(fileptr,"%g\n",view->from_x);
 fprintf(fileptr,"%g\n",view->from_y);
 fprintf(fileptr,"%g\n",view->from_z);

 fprintf(fileptr,"%g\n",view->to_x);
 fprintf(fileptr,"%g\n",view->to_y);
 fprintf(fileptr,"%g\n",view->to_z);

 fprintf(fileptr,"%g\n",view->roll);
 fprintf(fileptr,"%g\n",view->zoom);

 return mops_status;
}


int 
mops_write_nurbcurve(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int i = 0, a = 0;
 mops_nurbcurve_object *curve = NULL;

 if(!o)
   return MOPS_OK;

 curve = (mops_nurbcurve_object*)(o->object);
 
 if(!curve)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTNURBCURVE);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", curve->length);
 fprintf(fileptr, "%d\n", curve->order);

 fprintf(fileptr, "%d\n", curve->knot_type);

 if(curve->knot_type == MOPS_KTCUSTOM)
   {
     for(i=0;i<(curve->length+curve->order);i++)
       fprintf(fileptr,"%g\n",curve->knotv[i]);
   }

 fprintf(fileptr, "%d\n", curve->has_weights);
 a = 0;
 for(i=0;i<curve->length;i++)
 {
   fprintf(fileptr,"%g\n",curve->controlv[a]);
   a++;
   fprintf(fileptr,"%g\n",curve->controlv[a]);
   a++;
   fprintf(fileptr,"%g\n",curve->controlv[a]);
   a++;
   if(curve->has_weights)
     {
       fprintf(fileptr,"%g\n",curve->controlv[a]);
       a++;
     }
 }

 fprintf(fileptr, "%d\n", curve->closed);
 fprintf(fileptr, "%d\n", curve->weight_handles);
 fprintf(fileptr, "%g\n", curve->glu_sampling_tolerance);
 fprintf(fileptr, "%g\n", curve->handle_size);

 return mops_status;
}


int 
mops_write_nurbpatch(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int i = 0, j=0, a = 0;
 mops_nurbpatch_object *patch = NULL;

 if(!o)
   return MOPS_OK;

 patch = (mops_nurbpatch_object*)(o->object);
 
 if(!patch)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTNURBPATCH);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", patch->width);
 fprintf(fileptr, "%d\n", patch->height);
 fprintf(fileptr, "%d\n", patch->uorder);
 fprintf(fileptr, "%d\n", patch->vorder);

 fprintf(fileptr, "%d\n", patch->uknot_type);
 fprintf(fileptr, "%d\n", patch->vknot_type);

 if(patch->uknot_type == MOPS_KTCUSTOM)
   {
     for(i=0;i<(patch->width+patch->uorder);i++)
       fprintf(fileptr,"%g\n",patch->uknotv[i]);
   }

 if(patch->vknot_type == MOPS_KTCUSTOM)
   {
     for(i=0;i<(patch->height+patch->vorder);i++)
       fprintf(fileptr,"%g\n",patch->vknotv[i]);
   }

 fprintf(fileptr, "%d\n", patch->has_weights);
 a = 0;
 for(i=0;i<patch->width;i++)
 {
   for(j=0;j<patch->height;j++)
     {
       fprintf(fileptr,"%g\n",patch->controlv[a]);
       a++;
       fprintf(fileptr,"%g\n",patch->controlv[a]);
       a++;
       fprintf(fileptr,"%g\n",patch->controlv[a]);
       a++;
       if(patch->has_weights)
	 {
	   fprintf(fileptr,"%g\n",patch->controlv[a]);
	   a++;
	 }
     }
 }

 fprintf(fileptr, "%d\n", patch->weight_handles);
 fprintf(fileptr, "%d\n", patch->glu_display_mode);
 fprintf(fileptr, "%g\n", patch->glu_sampling_tolerance);
 fprintf(fileptr, "%g\n", patch->handle_size);

 return mops_status;
}

int 
mops_write_light(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_light_object *light = NULL;

 if(!o)
   return MOPS_OK;

 light = (mops_light_object*)(o->object);
 
 if(!light)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTLIGHT);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", light->shadows);

 fprintf(fileptr, "%d\n", light->samples);

 fprintf(fileptr, "%d\n", light->type);
 fprintf(fileptr, "%d\n", light->on);
 fprintf(fileptr, "%g\n", light->intensity);
 fprintf(fileptr, "%d\n", light->colr);
 fprintf(fileptr, "%d\n", light->colg);
 fprintf(fileptr, "%d\n", light->colb);
 fprintf(fileptr, "%d\n", light->cola);

 fprintf(fileptr, "%g\n", light->cone_angle);
 fprintf(fileptr, "%g\n", light->cone_delta_angle);
 fprintf(fileptr, "%g\n", light->beam_distribution);

 return mops_status;
}


int 
mops_write_box(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_box_object *box = NULL;

 if(!o)
   return MOPS_OK;

 box = (mops_box_object*)(o->object);
 
 if(!box)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTBOX);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%g\n", box->width);
 fprintf(fileptr, "%g\n", box->length);
 fprintf(fileptr, "%g\n", box->height);


 return mops_status;
}


int 
mops_write_sphere(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_sphere_object *sphere = NULL;

 if(!o)
   return MOPS_OK;

 sphere = (mops_sphere_object*)(o->object);
 
 if(!sphere)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTSPHERE);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", sphere->closed);
 fprintf(fileptr, "%g\n", sphere->radius);
 fprintf(fileptr, "%g\n", sphere->zmin);
 fprintf(fileptr, "%g\n", sphere->zmax);
 fprintf(fileptr, "%g\n", sphere->thetamax);

 return mops_status;
}


int 
mops_write_cone(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cone_object *cone = NULL;

 if(!o)
   return MOPS_OK;

 cone = (mops_cone_object*)(o->object);
 
 if(!cone)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTCONE);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", cone->closed);
 fprintf(fileptr, "%g\n", cone->radius);
 fprintf(fileptr, "%g\n", cone->height);
 fprintf(fileptr, "%g\n", cone->thetamax);

 return mops_status;
}


int 
mops_write_disk(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_disk_object *disk = NULL;

 if(!o)
   return MOPS_OK;

 disk = (mops_disk_object*)(o->object);
 
 if(!disk)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTDISK);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%g\n", disk->radius);
 fprintf(fileptr, "%g\n", disk->height);
 fprintf(fileptr, "%g\n", disk->thetamax);

 return mops_status;
}


int 
mops_write_cylinder(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_cylinder_object *cylinder = NULL;

 if(!o)
   return MOPS_OK;

 cylinder = (mops_cylinder_object*)(o->object);
 
 if(!cylinder)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTCYLINDER);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", cylinder->closed);
 fprintf(fileptr, "%g\n", cylinder->radius);
 fprintf(fileptr, "%g\n", cylinder->zmin);
 fprintf(fileptr, "%g\n", cylinder->zmax);
 fprintf(fileptr, "%g\n", cylinder->thetamax);

 return mops_status;
}

int 
mops_write_hyperboloid(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_hyperboloid_object *hyperboloid = NULL;

 if(!o)
   return MOPS_OK;

 hyperboloid = (mops_hyperboloid_object*)(o->object);
 
 if(!hyperboloid)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTHYPERBOLOID);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", hyperboloid->closed);
 fprintf(fileptr, "%g\n", hyperboloid->p1x);
 fprintf(fileptr, "%g\n", hyperboloid->p1y);
 fprintf(fileptr, "%g\n", hyperboloid->p1z);

 fprintf(fileptr, "%g\n", hyperboloid->p2x);
 fprintf(fileptr, "%g\n", hyperboloid->p2y);
 fprintf(fileptr, "%g\n", hyperboloid->p2z);

 fprintf(fileptr, "%g\n", hyperboloid->thetamax);

 return mops_status;
}


int 
mops_write_paraboloid(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_paraboloid_object *paraboloid = NULL;

 if(!o)
   return MOPS_OK;

 paraboloid = (mops_paraboloid_object*)(o->object);
 
 if(!paraboloid)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTPARABOLOID);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", paraboloid->closed);
 fprintf(fileptr, "%g\n", paraboloid->rmax);
 fprintf(fileptr, "%g\n", paraboloid->zmin);
 fprintf(fileptr, "%g\n", paraboloid->zmax);
 fprintf(fileptr, "%g\n", paraboloid->thetamax);

 return mops_status;
}


int 
mops_write_torus(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_torus_object *torus = NULL;

 if(!o)
   return MOPS_OK;

 torus = (mops_torus_object*)(o->object);
 
 if(!torus)
  return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTTORUS);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 fprintf(fileptr, "%d\n", torus->closed);
 fprintf(fileptr, "%g\n", torus->majorrad);
 fprintf(fileptr, "%g\n", torus->minorrad);
 fprintf(fileptr, "%g\n", torus->phimin);
 fprintf(fileptr, "%g\n", torus->phimax);
 fprintf(fileptr, "%g\n", torus->thetamax);

 return mops_status;
}

int 
mops_write_instance(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;


 if(!o)
   return MOPS_OK;

 fprintf(fileptr, "%d\n", MOPS_OTINSTANCE);

 if(o->name)
   fprintf(fileptr,"%s\n",o->name);
 else
   fprintf(fileptr,"\n");

 return mops_status;
}


int
mops_write_object(FILE *fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;

  if(!o)
    return MOPS_OK;

  /* mark start of new entity */
  putc('\a', fileptr);
  
  switch (o->type)
    {
    case MOPS_OTCUSTOM:
      mops_status = mops_write_custom(fileptr, o);
      break;
    case MOPS_OTLEVEL:
      mops_status = mops_write_level(fileptr, o);
      break;
    case MOPS_OTLIGHT:
      mops_status = mops_write_light(fileptr, o);
      break;
    case MOPS_OTNURBPATCH:
      mops_status = mops_write_nurbpatch(fileptr, o);
      break;
    case MOPS_OTNURBCURVE:
      mops_status = mops_write_nurbcurve(fileptr, o);
      break;
    case MOPS_OTBOX:
      mops_status = mops_write_box(fileptr, o);
      break;
    case MOPS_OTSPHERE:
      mops_status = mops_write_sphere(fileptr, o);
      break;
    case MOPS_OTCONE:
      mops_status = mops_write_cone(fileptr, o);
      break;
    case MOPS_OTCYLINDER:
      mops_status = mops_write_cylinder(fileptr, o);
      break;
    case MOPS_OTDISK:
      mops_status = mops_write_disk(fileptr, o);
      break;
    case MOPS_OTTORUS:
      mops_status = mops_write_torus(fileptr, o);
      break;
    case MOPS_OTPARABOLOID:
      mops_status = mops_write_paraboloid(fileptr, o);
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mops_write_hyperboloid(fileptr, o);
      break;
    case MOPS_OTINSTANCE:
      mops_status = mops_write_instance(fileptr, o);
      break;
    }

  if(o->type != MOPS_OTINSTANCE)
    mops_write_attributes(fileptr,o);
  else
    mops_write_trafos(fileptr,o);

  if(o->shaders)
    mops_write_shaders(fileptr,o);

  if(o->tags)
    mops_write_tags(fileptr,o);

  if((o->down) && (o->type != MOPS_OTLEVEL)) 
    {
      /* mark start of child objects */
      putc('\a', fileptr);
      fprintf(fileptr,"%d\n", MOPS_OTLEVEL);
      fprintf(fileptr,"%d\n", MOPS_LTDOWN);

      o = o->down;
      while(o)
	{
	  mops_status = mops_write_object(fileptr, o);
	  o = o->next;
	}

      /* write an EndOfLevel-Object, to mark end of child objects */
      putc('\a', fileptr);
      fprintf(fileptr,"%d\n", MOPS_OTLEVEL);
      fprintf(fileptr,"%d\n", MOPS_LTENDLEVEL);
    }


 return mops_status;
}


int
mops_write_scene(char *fname)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_root->next;
 FILE *fileptr = NULL;

  if(!fname)
    return MOPS_OK;

  if(!(fileptr = fopen(fname,"wb")))
    return MOPS_OPENFILE;

  /* reset oid counter for saving of Instance objects */
  mops_create_oid(NULL);
  mops_status = mops_create_orig_ids(o);
  mops_status = mops_create_instance_ids(o);

  mops_write_header(fileptr);

  mops_write_root(fileptr);

  if(mops_root->shaders)
    mops_write_shaders(fileptr,mops_root);

  if(mops_root->tags)
    mops_write_tags(fileptr,mops_root);

  while(o)
    {
      mops_status = mops_write_object(fileptr, o);
      if(mops_status)
	{fclose(fileptr); return mops_status;}
      o = o->next;
    }

  fclose(fileptr);

 return mops_status;
}


int
mops_save_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char fname[] = "save_scene";
 char done_cmd[] = "puts stdout \"Done writing scene to: $filename\"";

 /* check args */
 if(argc != 2)
   {
     mops_error(interp, MOPS_ARGS, fname, "filename");
     return TCL_OK;
   }

 mops_status = mops_write_scene(argv[1]);

 MOPS_CRERROR

 Tcl_Eval(interp, done_cmd);

 return TCL_OK;
}

